/*******************************************************************************
 * Copyright (c) 2014
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 * Marc R. Hoffmann - initial implementation for EclEmma plug-in
 * David Pfeffer	   - modification for use in Feature to Code mapping plug-in
 *
 ******************************************************************************/

/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/

package at.jku.ssw.ftcm;

import at.jku.ssw.marker.FeatureDrawingStrategy;
import org.eclipse.jface.text.source.AnnotationPainter;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.*;
import org.eclipse.ui.texteditor.AnnotationPreference;
import org.eclipse.ui.texteditor.ITextEditor;

import java.util.logging.Logger;

/**
 * The {@link EditorTracker} implements a mechanism to keep track of all opened
 * editors in a {@link IWorkbench}.
 *
 * @author David Pfeffer
 */
public class EditorTracker {

    // The workbench on which the tracking happens
    private final IWorkbench workbench;

    // The WindowListener adds or removes a PartListener on each opened or
    // closed window
    private IWindowListener windowListener = new IWindowListener() {

        @Override
        public void windowOpened(IWorkbenchWindow window) {
            window.getPartService().addPartListener(partListener);
        }

        @Override
        public void windowClosed(IWorkbenchWindow window) {
            window.getPartService().removePartListener(partListener);
        }

        @Override
        public void windowActivated(IWorkbenchWindow window) {
        }

        @Override
        public void windowDeactivated(IWorkbenchWindow window) {
        }
    };

    // The PartListener calls the modification or unmodification function for
    // each opened or closed {@link IWorkbenchPartReference}
    private IPartListener2 partListener = new IPartListener2() {

        @Override
        public void partOpened(IWorkbenchPartReference partref) {
            modifyAnnotaionPainterForEditor(partref);
        }

        @Override
        public void partClosed(IWorkbenchPartReference partref) {
            unmodifyAnnotaionPaitnerForEditor(partref);
        }

        @Override
        public void partActivated(IWorkbenchPartReference partref) {
        }

        @Override
        public void partBroughtToTop(IWorkbenchPartReference partref) {
        }

        @Override
        public void partVisible(IWorkbenchPartReference partref) {
        }

        @Override
        public void partInputChanged(IWorkbenchPartReference partref) {
        }

        @Override
        public void partDeactivated(IWorkbenchPartReference partref) {
        }

        @Override
        public void partHidden(IWorkbenchPartReference partref) {
        }
    };

    /**
     * Constructs a new {@link EditorTracker}.
     *
     * @param workbench The {@link IWorkbench} of which the {@link EditorTracker}
     *                  should keep track of.
     */
    public EditorTracker(IWorkbench workbench) {
        this.workbench = workbench;

        Logger.getLogger(Activator.PLUGIN_ID).info("Editor tracker started");

        // Add a PartListener for each workbench window which is already open
        for (final IWorkbenchWindow w : workbench.getWorkbenchWindows()) {
            w.getPartService().addPartListener(partListener);
        }

        // Add the WindowListener to the workbench to keep track of further
        // changes
        workbench.addWindowListener(windowListener);

        // Modify all editors which are already open
        modifyAllAnnotationPaintersForEditors();
    }

    /**
     * Cleans up the {@link EditorTracker} and removes all listeners.
     */
    public void dispose() {
        workbench.removeWindowListener(windowListener);
        for (final IWorkbenchWindow w : workbench.getWorkbenchWindows()) {
            w.getPartService().removePartListener(partListener);
        }
        Logger.getLogger(Activator.PLUGIN_ID).info("Editor tracker stopped");
    }

    /**
     * Iterates over all current open workbench windows and calls the
     * modification function for each found editor.
     */
    private void modifyAllAnnotationPaintersForEditors() {
        for (final IWorkbenchWindow w : workbench.getWorkbenchWindows()) {
            for (final IWorkbenchPage p : w.getPages()) {
                for (final IEditorReference e : p.getEditorReferences()) {
                    modifyAnnotaionPainterForEditor(e);
                }
            }
        }
    }

    /**
     * Modification function for a {@link IWorkbenchPartReference} object.
     *
     * @param partref The {@link IWorkbenchPartReference} object to modify.
     */
    private void modifyAnnotaionPainterForEditor(IWorkbenchPartReference partref) {
        IWorkbenchPart part = partref.getPart(false);

        // Only modify if it is a TextEditor
        if (part instanceof ITextEditor) {

            // Override the default BOX strategy with our implementation and
            // force a repaint
            final AnnotationPainter painter = EclipseHacks.getAnnotationPainter((ITextEditor) part);
            painter.addDrawingStrategy(AnnotationPreference.STYLE_BOX, new FeatureDrawingStrategy());
            Display.getDefault().asyncExec(new Runnable() {
                @Override
                public void run() {
                    painter.paint(AnnotationPainter.CONFIGURATION);
                }
            });
        }
    }

    /**
     * Unmodification function for a {@link IWorkbenchPartReference} object.
     *
     * @param partref The {@link IWorkbenchPartReference} object to unmodify.
     */
    private void unmodifyAnnotaionPaitnerForEditor(IWorkbenchPartReference partref) {
        IWorkbenchPart part = partref.getPart(false);

        // Only unmodify if it is a TextEditor
        if (part instanceof ITextEditor) {
            // It's not really possible to undo/cleanup anything for now
        }
    }
}
