/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/

package at.jku.ssw.ftcm;

import org.eclipse.jdt.core.*;

public class SignatureResolver {
    private static final String OBJECT = "Ljava/lang/Object;"; //$NON-NLS-1$
    private static final char SLASH = '/';

    public static String getResolvedSignature(final IMethod method) throws JavaModelException {
        if (method.isBinary()) {
            return method.getSignature();
        }
        final StringBuffer buffer = new StringBuffer("(");
        final String[] parameterTypes = method.getParameterTypes();
        for (final String t : parameterTypes) {
            resolveArrayParameterType(method, t, buffer);
        }
        buffer.append(")");

        final String returnType = method.getReturnType();
        resolveParameterType(method, returnType, buffer);

        return buffer.toString();
    }

    private static final void resolveArrayParameterType(final IMethod method, final String parameterType, final StringBuffer result) throws JavaModelException {
        final int arrayCount = Signature.getArrayCount(parameterType);
        for (int i = 0; i < arrayCount; i++) {
            result.append(Signature.C_ARRAY);
        }
        resolveParameterType(method, parameterType.substring(arrayCount), result);
    }

    private static final void resolveParameterType(final IMethod method, final String parameterType, final StringBuffer result) throws JavaModelException {
        final char kind = parameterType.charAt(0);
        switch (kind) {
            case Signature.C_UNRESOLVED:
                final String identifier = parameterType.substring(1, parameterType.length() - 1);
                if (resolveType(method.getDeclaringType(), identifier, result)) {
                    return;
                }
                if (resolveTypeParameter(method, identifier, result)) {
                    return;
                }
                break;
        }
        result.append(parameterType);
    }

    private static final boolean resolveType(final IType scope, final String identifier, final StringBuffer result) throws JavaModelException {
        final String[][] types = scope.resolveType(Signature.getTypeErasure(identifier));
        if (types == null || types.length != 1) {
            return false;
        }
        result.append(Signature.C_RESOLVED);
        final String qualifier = types[0][0];
        if (qualifier.length() > 0) {
            replace(qualifier, Signature.C_DOT, SLASH, result);
            result.append(SLASH);
        }
        replace(types[0][1], Signature.C_DOT, Signature.C_DOLLAR, result);
        result.append(Signature.C_SEMICOLON);
        return true;
    }

    private static final boolean resolveTypeParameter(final IMethod method, final String identifier, final StringBuffer result) throws JavaModelException {
        IType type = method.getDeclaringType();
        if (resolveTypeParameter(type, method.getTypeParameters(), identifier, result)) {
            return true;
        }
        while (type != null) {
            if (resolveTypeParameter(type, type.getTypeParameters(), identifier, result)) {
                return true;
            }
            type = type.getDeclaringType();
        }
        return false;
    }

    private static final boolean resolveTypeParameter(final IType context, final ITypeParameter[] typeParameters, final String identifier, final StringBuffer result)
        throws JavaModelException {
        for (final ITypeParameter p : typeParameters) {
            if (identifier.equals(p.getElementName())) {
                final String[] bounds = p.getBounds();
                if (bounds.length == 0) {
                    result.append(OBJECT);
                    return true;
                } else {
                    return resolveType(context, bounds[0], result);
                }
            }
        }
        return false;
    }

    private static final void replace(final String source, final char oldChar, final char newChar, final StringBuffer result) {
        final int len = source.length();
        for (int i = 0; i < len; i++) {
            final char c = source.charAt(i);
            result.append(c == oldChar ? newChar : c);
        }
    }

    private SignatureResolver() {
    }
}
