/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/

package at.jku.ssw.ftcm.preferences;

import at.jku.ssw.ftcm.Activator;
import at.jku.ssw.ftcm.builder.FeatureBuilder;
import org.eclipse.core.resources.*;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.preferences.IEclipsePreferences;
import org.eclipse.core.runtime.preferences.IScopeContext;

public class ProjectSettingsManager {

    public final static String FEATURE_LIST = "FEATURE_LIST";
    public final static String FEATURE_AUTO_CREATE = "FEATURE_AUTO_CREATE";

    // The feature list file name
    public final static String FEATURE_LIST_FILE_NAME = "features.lst"; //$NON-NLS-1$

    private IProject project;
    private IScopeContext projectScope;
    private IEclipsePreferences projectNode;

    private static Map<IProject, ProjectSettingsManager> map = new HashMap<>();

    private ProjectSettingsManager(IProject project) {
        this.project = project;
        projectScope = new ProjectScope(project);
        projectNode = projectScope.getNode(Activator.PLUGIN_ID);
    }

    public static ProjectSettingsManager getInstance(IProject project) {
        ProjectSettingsManager psm = map.get(project);
        if (psm == null) {
            psm = new ProjectSettingsManager(project);
            map.put(project, psm);
        }
        return psm;
    }

    public IProject getProject() {
        return project;
    }

    public IEclipsePreferences getProjectNode() {
        return projectNode;
    }

    public Feature getFeature(int id) {
        List<Feature> featureList = getFeatureList();
        for (Feature f : featureList) {
            if (f.getId() == id) {
                return f;
            }
        }
        return null;
    }

    public List<Feature> getFeatureList() {
        List<Feature> featureList = new ArrayList<>();
        String featureListString = projectNode.get(ProjectSettingsManager.FEATURE_LIST, "");

        try {
            StringTokenizer st1 = new StringTokenizer(featureListString, ";");
            while (st1.hasMoreTokens()) {
                StringTokenizer st2 = new StringTokenizer(st1.nextToken(), ",");
                featureList.add(new Feature(Integer.parseInt(st2.nextToken()),
                                            st2.nextToken(),
                                            Integer.parseInt(st2.nextToken()),
                                            Integer.parseInt(st2.nextToken()),
                                            Integer.parseInt(st2.nextToken())));
            }
        } catch (NoSuchElementException | NumberFormatException e) {
            // MessageDialog.openError(shell,
            // "Error while reading feature settings",
            // "There was an error while reading the projects feature setting file:\n\t -There is a format error in the saved feature
            // list.\n\nWhy did this
            // happen?\n\t- Most
            // likly the feature settings file got modified outside Eclipse and is now corrupted.\n\nWhat can I do now?\n\t- Delete the
            // corrupted settings
            // file, refresh your
            // project and try again.");
            featureList.clear();
        }
        return featureList;
    }

    public void setFeatureList(List<Feature> featureList) {

        StringBuffer sb = new StringBuffer();
        for (Feature feature : featureList) {
            sb.append(feature.getId())
              .append(",")
              .append(feature.getDescription())
              .append(",")
              .append(feature.getR())
              .append(",")
              .append(feature.getG())
              .append(",")
              .append(feature.getB())
              .append(";");
        }

        projectNode.put(ProjectSettingsManager.FEATURE_LIST, sb.toString());
        writeFeatureList(featureList);
    }

    public void addFeatureBuilder() {
        try {
            // check if builder is already registered
            if (isBuilderRegistered()) {
                return;
            }

            // get project description
            IProjectDescription description = project.getDescription();

            // add builder
            ICommand buildCommand = description.newCommand();
            buildCommand.setBuilderName(FeatureBuilder.BUILDER_ID);
            java.util.List<ICommand> commands = new ArrayList<ICommand>();
            commands.addAll(Arrays.asList(description.getBuildSpec()));
            commands.add(buildCommand);

            // update description
            description.setBuildSpec(commands.toArray(new ICommand[commands.size()]));
            project.setDescription(description, null);
        } catch (CoreException e) {
            e.printStackTrace();
        }
    }

    public void removeFeatureBuilder() {
        try {
            // check if builder is already removed
            if (!isBuilderRegistered()) {
                return;
            }

            // get project description
            final IProjectDescription description = project.getDescription();
            final java.util.List<ICommand> commands = new ArrayList<ICommand>();
            commands.addAll(Arrays.asList(description.getBuildSpec()));

            // look for builder and remove it
            for (ICommand buildSpec : description.getBuildSpec()) {
                if (FeatureBuilder.BUILDER_ID.equals(buildSpec.getBuilderName())) {
                    // remove builder
                    commands.remove(buildSpec);
                }
            }

            // update description
            description.setBuildSpec(commands.toArray(new ICommand[commands.size()]));
            project.setDescription(description, null);
        } catch (CoreException e) {
            e.printStackTrace();
        }
    }

    public boolean isBuilderRegistered() {
        try {
            for (ICommand buildSpec : project.getDescription().getBuildSpec()) {
                if (FeatureBuilder.BUILDER_ID.equals(buildSpec.getBuilderName())) {
                    return true;
                }
            }
        } catch (CoreException e) {
            e.printStackTrace();
        }

        return false;
    }

    public void writeFeatureList(List<Feature> featureList) {
        try (OutputStream outputStream = new FileOutputStream(getFeatureListFilePath()); BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(outputStream))) {
            for (Feature feature : featureList) {
                writer.write(feature.getId() + ";" + feature.getDescription() + ";" + feature.getR() + ";" + feature.getG() + ";" + feature.getB());
                writer.newLine();
            }
            outputStream.flush();
            this.getProject().refreshLocal(IResource.DEPTH_ONE, null);
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        } catch (CoreException e) {
            e.printStackTrace();
        }
    }

    private String getFeatureListFilePath() {
        IFile contributorsFile = this.getProject().getFile(FEATURE_LIST_FILE_NAME);
        IPath path = contributorsFile.getRawLocation();
        return path.toOSString();
    }
}
