/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.classification;

import at.jku.mevss.anttracks.classification.enumerations.ClassifierSourceCollection;
import at.jku.mevss.anttracks.classification.enumerations.ClassifierType;
import at.jku.mevss.anttracks.features.FeatureMap;
import at.jku.mevss.anttracks.heap.Closures;
import at.jku.mevss.anttracks.heap.IndexBasedHeap;
import at.jku.mevss.anttracks.heap.objects.ObjectInfo;
import at.jku.mevss.anttracks.heap.roots.RootPtr;
import at.jku.mevss.anttracks.heap.space.SpaceInfo;
import at.jku.mevss.anttracks.heap.symbols.*;
import at.jku.mevss.anttracks.parser.EventType;
import at.jku.mevss.anttracks.util.AnnotationHelper;
import at.jku.mevss.anttracks.util.ApplicationStatistics;
import at.jku.mevss.anttracks.util.ImagePack;
import javafx.scene.image.ImageView;

import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.function.Supplier;
import java.util.logging.Logger;

public abstract class Classifier<T> implements Comparable<Classifier<T>> {

    public enum ClassificationMode {
        MAP,
        LIST
    }

    public static final ClassificationMode CLASSIFICATION_MODE = ClassificationMode.LIST;

    private Supplier<Symbols> symbolsSupplier = null;
    private Supplier<IndexBasedHeap> fastHeapSupplier = null;

    protected boolean iconsLoaded = false;
    protected ImagePack[] icons = null;
    protected ImageView[] iconNodes = null;
    private boolean isCustom = false;
    private String sourceCode = null;

    protected Logger LOGGER = Logger.getLogger(this.getClass().getSimpleName());

    public void setup(Supplier<Symbols> symbolsSupplier) {
        this.symbolsSupplier = symbolsSupplier;
        this.fastHeapSupplier = null;
    }

    public void setup(Supplier<Symbols> symbolsSupplier, Supplier<IndexBasedHeap> fastHeapSupplier) {
        this.symbolsSupplier = symbolsSupplier;
        this.fastHeapSupplier = fastHeapSupplier;
    }

    @Override
    public String toString() {
        String s = getName();
        return s != null ? s : super.toString();
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((getName() == null) ? 0 : getName().hashCode());
        result = prime * result + ((getDesc() == null) ? 0 : getDesc().hashCode());
        result = prime * result + ((getExample() == null) ? 0 : getExample().hashCode());
        result = prime * result + ((getType() == null) ? 0 : getType().hashCode());
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        Classifier<?> other = (Classifier<?>) obj;
        return other.getName().equals(getName()) && other.getDesc().equals(getDesc()) && other.getExample().equals(getExample()) && other.getType().equals(getType());
    }

    protected IndexBasedHeap fastHeap() {
        if (fastHeapSupplier == null) {
            LOGGER.warning("FastHeap supplier is null! Probably setup() hasn't been called.");
            return null;
        }
        IndexBasedHeap h = fastHeapSupplier.get();
        if (h == null) {
            LOGGER.warning("FastHeap is null! Supplier returned null.");
            return null;
        }
        return h;
    }

    protected Symbols symbols() {
        if (symbolsSupplier == null) {
            LOGGER.warning("Symbols supplier is null! Probably setup() hasn't been called.");
            return null;
        }
        Symbols s = symbolsSupplier.get();
        if (s == null) {
            LOGGER.warning("Symobls is null! Supplier returned null.");
            return null;
        }
        return s;
    }

    protected FeatureMap features() {
        Symbols s = symbols();
        if (s != null) {
            return s.features;
        }
        return null;
    }

    protected AllocationSites allocationSites() {
        Symbols s = symbols();
        if (s != null) {
            return s.sites;
        }
        return null;
    }

    protected AllocatedTypes types() {
        Symbols s = symbols();
        if (s != null) {
            return s.types;
        }
        return null;
    }

    public ImagePack getIcon(Object key) {
        if (!iconsLoaded) {
            // Load classifierIcon only once
            icons = loadIcons();
            loadIconNodes();
            iconsLoaded = true;
        }

        if (icons != null && icons.length > 0) {
            return icons[0];
        } else {
            return null;
        }
    }

    /**
     * Method that is invoked once to load the classifierIcon.
     *
     * @return Returns the classifier's classifierIcon, or null if no classifierIcon is available
     */
    protected ImagePack[] loadIcons() { return null; }

    public ImageView getIconNode(Object key) {
        if (!iconsLoaded) {
            icons = loadIcons();
            loadIconNodes();
            iconsLoaded = true;
        }

        if (iconNodes != null && iconNodes.length > 0) {
            return iconNodes[0];
        } else {
            return null;
        }
    }

    protected void loadIconNodes() {
        if (icons != null) {
            iconNodes = new ImageView[icons.length];
            for (int i = 0; i < icons.length; i++) {
                ImagePack icon = icons[i];
                iconNodes[i] = icon.getAsNewNode();
            }
        }
    }

    public boolean isCustom() {
        return isCustom;
    }

    public void setIsCustom(boolean b) {
        isCustom = b;
    }

    public String getSourceCode() {
        return sourceCode;
    }

    public void setSourceCode(String sc) {
        sourceCode = sc;
    }

    public boolean isOnTheFlyCompilable() {
        return isCustom && sourceCode != null && !sourceCode.equals("");
    }

    protected ClassifierType type = null;

    /**
     * Gets the classifier's type
     *
     * @return The classifier's type
     */
    public ClassifierType getType() {
        if (type == null) {
            at.jku.mevss.anttracks.classification.annotations.Classifier annotation = AnnotationHelper.getAnnotation(this,
                                                                                                                     at.jku.mevss.anttracks.classification.annotations
                                                                                                                             .Classifier
                                                                                                                             .class);
            type = annotation != null ? annotation.type() : ClassifierType.ONE;
        }
        return type;
    }

    protected String desc = null;

    /**
     * Gets the classifier's description
     *
     * @return The classifier's description
     */
    public String getDesc() {
        if (desc == null) {
            at.jku.mevss.anttracks.classification.annotations.Classifier annotation = AnnotationHelper.getAnnotation(this,
                                                                                                                     at.jku.mevss.anttracks.classification.annotations
                                                                                                                             .Classifier
                                                                                                                             .class);
            desc = annotation != null ? annotation.desc() : "No description";
        }
        return desc;
    }

    protected String example = desc;

    /**
     * Gets the classifier's example
     *
     * @return The classifier's example
     */
    public String getExample() {
        if (example == null) {
            at.jku.mevss.anttracks.classification.annotations.Classifier annotation = AnnotationHelper.getAnnotation(this,
                                                                                                                     at.jku.mevss.anttracks.classification.annotations
                                                                                                                             .Classifier
                                                                                                                             .class);
            example = annotation != null ? annotation.example() : "No example";
        }
        return example;
    }

    protected String name = null;

    /**
     * Gets the classifier's name
     *
     * @return The classifier's name
     */
    public String getName() {
        if (name == null) {
            at.jku.mevss.anttracks.classification.annotations.Classifier annotation = AnnotationHelper.getAnnotation(this,
                                                                                                                     at.jku.mevss.anttracks.classification.annotations
                                                                                                                             .Classifier
                                                                                                                             .class);
            name = annotation != null ? annotation.name() : "No name";
        }
        return name;
    }

    protected ClassifierSourceCollection sourceCollection;

    /**
     * Gets the classifier's name
     *
     * @return The classifier's name
     */
    public ClassifierSourceCollection getSourceCollection() {
        if (sourceCollection == null) {
            at.jku.mevss.anttracks.classification.annotations.Classifier annotation = AnnotationHelper.getAnnotation(this,
                                                                                                                     at.jku.mevss.anttracks.classification.annotations
                                                                                                                             .Classifier
                                                                                                                             .class);
            sourceCollection = annotation != null ? annotation.collection() : ClassifierSourceCollection.ALL;
        }
        return sourceCollection;
    }

    @SuppressWarnings("rawtypes")
    public ClassifierProperty<?>[] configurableProperties() {
        List<Field> fields = new ArrayList<Field>();
        for (Class<?> c = getClass(); c != null; c = c.getSuperclass()) {
            fields.addAll(Arrays.asList(c.getDeclaredFields()));
        }
        return fields.stream()
                     .filter(field -> field.getAnnotation(at.jku.mevss.anttracks.classification.annotations.ClassifierProperty.class) != null)
                     .map(field -> new ClassifierProperty<>(this, field))
                     .toArray(ClassifierProperty[]::new);
    }

    @Override
    public int compareTo(Classifier<T> other) {
        return getName().compareTo(other.getName());
    }

    ThreadLocal<Integer> currentObjIndex = ThreadLocal.withInitial(() -> -1);

    ThreadLocal<Long> currentAddress = ThreadLocal.withInitial(() -> (long) -1);
    ThreadLocal<ObjectInfo> currentObjectInfo = ThreadLocal.withInitial(() -> null);
    ThreadLocal<SpaceInfo> currentSpace = ThreadLocal.withInitial(() -> null);
    ThreadLocal<AllocatedType> currentType = ThreadLocal.withInitial(() -> null);
    ThreadLocal<Integer> currentSize = ThreadLocal.withInitial(() -> -1);
    ThreadLocal<Boolean> currentIsArray = ThreadLocal.withInitial(() -> false);
    ThreadLocal<Integer> currentArrayLength = ThreadLocal.withInitial(() -> -1);
    ThreadLocal<AllocationSite> currentAllocationSite = ThreadLocal.withInitial(() -> null);
    ThreadLocal<long[]> currentPointedFrom = ThreadLocal.withInitial(() -> null);
    ThreadLocal<long[]> currentPointsTo = ThreadLocal.withInitial(() -> null);
    ThreadLocal<EventType> currentEventType = ThreadLocal.withInitial(() -> null);
    ThreadLocal<List<RootPtr>> currentRootPointers = ThreadLocal.withInitial(() -> null);

    private boolean fastHeapUsed() {
        return currentObjIndex.get() >= 0;
    }

    protected int index() { return currentObjIndex.get(); }

    protected long address() {
        return fastHeapUsed() ? fastHeap().getAddress(currentObjIndex.get()) : currentAddress.get();
    }

    protected ObjectInfo objectInfo() {
        return fastHeapUsed() ? fastHeap().getObjectInfo(currentObjIndex.get()) : currentObjectInfo.get();
    }

    protected SpaceInfo space() {
        return fastHeapUsed() ? fastHeap().getSpace(currentObjIndex.get()) : currentSpace.get();
    }

    protected AllocatedType type() {
        return fastHeapUsed() ? fastHeap().getType(currentObjIndex.get()) : currentType.get();
    }

    protected int size() {
        return fastHeapUsed() ? fastHeap().getSize(currentObjIndex.get()) : currentSize.get();
    }

    protected boolean isArray() {
        return fastHeapUsed() ? fastHeap().isArray(currentObjIndex.get()) : currentIsArray.get();
    }

    protected int arrayLength() {
        return fastHeapUsed() ? fastHeap().getArrayLength(currentObjIndex.get()) : currentArrayLength.get();
    }

    protected AllocationSite allocationSite() {
        return fastHeapUsed() ? fastHeap().getAllocationSite(currentObjIndex.get()) : currentAllocationSite.get();
    }

    protected long[] pointedFrom() throws ClassifierException {
        if (fastHeapUsed()) {
            throw new ClassifierException("Pointed from addresses can only be obtained from Heap iteration, not FastHeap");
        } else {
            return currentPointedFrom.get();
        }
    }

    protected long[] pointsTo() throws ClassifierException {
        if (fastHeapUsed()) {
            throw new ClassifierException("Points to addresses can only be obtained from Heap iteration, not FastHeap");
        } else {
            return currentPointsTo.get();
        }
    }

    protected int[] pointedFromIndices() throws ClassifierException {
        if (!fastHeapUsed()) {
            throw new ClassifierException("Pointed from indices can only be obtained from FastHeap iteration, not DetailedHeap");
        } else {
            return fastHeap().getFromPointers(currentObjIndex.get());
        }
    }

    protected int[] pointsToIndices() throws ClassifierException {
        if (!fastHeapUsed()) {
            throw new ClassifierException("Points to indices can only be obtained from FastHeap iteration, not DetailedHeap");
        } else {
            return fastHeap().getToPointers(currentObjIndex.get());
        }
    }

    protected EventType eventType() {
        return fastHeapUsed() ? fastHeap().getEventType(currentObjIndex.get()) : currentEventType.get();
    }

    protected RootPtr closestGCRoot(int maxDepth, boolean onlyVariables) throws ClassifierException {
        if (!fastHeapUsed()) {
            throw new ClassifierException("Root pointer information can only be obtained from FastHeap iteration, not DetailedHeap");
        } else {
            return fastHeap().findClosestRoot(currentObjIndex.get(), maxDepth, onlyVariables);
        }
    }

    protected List<RootPtr> indirectGCRoots(int maxDepth) throws ClassifierException {
        if (!fastHeapUsed()) {
            throw new ClassifierException("Root pointer information can only be obtained from FastHeap iteration, not DetailedHeap");
        } else {
            ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance().createMeasurement("indirectGCRoots");
            //Version 1:
            List<RootPtr> x = fastHeap().indirectGCRoots(currentObjIndex.get(), maxDepth);
            m.end();

            return x;

            // Version 2:
            /*
            ArrayList<RootPtr> rootPtrs = new ArrayList<>();
            BitSet fromClosure = fastHeap().fromClosure(currentObjIndex.get());
            for (int idx = fromClosure.nextSetBit(0); idx != -1; idx = fromClosure.nextSetBit(idx + 1)) {
                List<RootPtr> rootPtrAtAddress = fastHeap().getRootPointerBlocking().get(idx);
                if (rootPtrAtAddress != null) {
                    rootPtrs.addAll(rootPtrAtAddress);
                }
            }
            m.end();
            return rootPtrs;
            */

            // Version 3:
            /*
            int currentObjectIdx = currentObjIndex.get();
            List<RootPtr> rootPtrs = new ArrayList<>();
            fastHeap().getRootPointerBlocking().forEach((integer, rootPtrList) -> {
                if (rootPtrList.get(0).reachables.get(currentObjectIdx)) {
                    rootPtrs.addAll(rootPtrList);
                }
            });

            m.end();
            return rootPtrs;
            */

            // Version 4:
            /*
            int currentObjectIdx = currentObjIndex.get();
            RootPtr[] rootPtrs = fastHeap().getRootPtrListBlocking().parallelStream().filter(rootPtr -> rootPtr.reachables.get(currentObjectIdx)).toArray(RootPtr[]::new);

            m.end();
            return rootPtrs;
            */

            // Version 5:
            /*
            int currentObjectIdx = currentObjIndex.get();
            Stream<RootPtr> rootPtrs = fastHeap().getRootPtrListBlocking().parallelStream().filter(rootPtr -> rootPtr.reachables.get(currentObjectIdx));

            m.end();
            return rootPtrs;
            */
        }
    }

    protected List<RootPtr.RootInfo> traceIndirectRootPointers(int maxDepth, boolean onlyVariables) throws ClassifierException {
        if (!fastHeapUsed()) {
            throw new ClassifierException("Root pointer information can only be obtained from FastHeap iteration, not DetailedHeap");
        } else {
            return fastHeap().traceAllRootsDFS(currentObjIndex.get(), maxDepth, onlyVariables);
        }
    }

    protected List<RootPtr.RootInfo> traceClosestRoots(int maxDepth, boolean onlyVariables) throws ClassifierException {
        if (!fastHeapUsed()) {
            throw new ClassifierException("Root pointer information can only be obtained from FastHeap iteration, not DetailedHeap");
        } else {
            return fastHeap().traceClosestRootBFS(currentObjIndex.get(), maxDepth, onlyVariables);
        }
    }

    protected List<RootPtr> rootPointers() {
        return fastHeapUsed() ? fastHeap().getRoot(currentObjIndex.get()) : currentRootPointers.get();
    }

    protected Closures closures() throws ClassifierException {
        if (!fastHeapUsed()) {
            throw new ClassifierException("Closure information can only be obtained from FastHeap iteration, not DetailedHeap");
        } else {
            return fastHeap().getClosures(currentObjIndex.get());
        }
    }

    protected abstract T classify() throws Exception;

    public T classify(Integer objIndex) throws Exception {
        currentObjIndex.set(objIndex);
        return classify();
    }

    public T classify(long address,
                      ObjectInfo objectInfo,
                      SpaceInfo space,
                      AllocatedType type,
                      int size,
                      boolean isArray,
                      int arrayLength,
                      AllocationSite allocationSite,
                      long[] pointedFrom,
                      long[] pointsTo,
                      EventType eventType,
                      List<RootPtr> rootPointers) throws Exception {
        currentObjIndex.set(-1);
        currentObjectInfo.set(objectInfo);
        currentAddress.set(address);
        currentSpace.set(space);
        currentType.set(type);
        currentSize.set(size);
        currentIsArray.set(isArray);
        currentArrayLength.set(arrayLength);
        currentAllocationSite.set(allocationSite);
        currentPointedFrom.set(pointedFrom);
        currentPointsTo.set(pointsTo);
        currentEventType.set(eventType);
        currentRootPointers.set(rootPointers);
        return classify();
    }
}
