package at.jku.mevss.anttracks.classification;

import at.jku.mevss.anttracks.classification.enumerations.ClassifierSourceCollection;
import at.jku.mevss.anttracks.heap.IndexBasedHeap;
import at.jku.mevss.anttracks.heap.symbols.Symbols;
import at.jku.mevss.anttracks.util.ClassifierUtil;
import at.jku.mevss.anttracks.util.DirectoryClassLoader;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.ServiceConfigurationError;
import java.util.ServiceLoader;
import java.util.function.Supplier;
import java.util.logging.Logger;

public class ClassifierFactoryList extends ArrayList<ClassifierFactory> {

    private static final long serialVersionUID = 3157139101821000231L;

    protected final Logger LOGGER = Logger.getLogger(this.getClass().getSimpleName());
    private final Supplier<Symbols> symbolsSupplier;
    private final Supplier<IndexBasedHeap> fastHeapSupplier;

    public void remove(Classifier<?> filter) {
        ClassifierFactory factory = stream().filter(f -> f.getName().equals(filter.getName())).findFirst().orElse(null);
        if (factory != null) {
            this.remove(factory);
        }
    }

    @SuppressWarnings("rawtypes")
    public ClassifierFactoryList(Supplier<Symbols> symbolsSupplier, Supplier<IndexBasedHeap> fastHeapSupplier, String directory) {
        this(symbolsSupplier, fastHeapSupplier, directory, null);
    }

    @SuppressWarnings("rawtypes")
    public ClassifierFactoryList(Supplier<Symbols> symbolsSupplier,
                                 Supplier<IndexBasedHeap> fastHeapSupplier,
                                 String directory,
                                 ClassifierSourceCollection sourceCollection) {
        this.symbolsSupplier = symbolsSupplier;
        this.fastHeapSupplier = fastHeapSupplier;

        loadCompiledClassifiers(Thread.currentThread().getContextClassLoader(), sourceCollection);
        if (directory != null) {
            loadCompiledClassifiers(new DirectoryClassLoader(directory), sourceCollection);
            loadSourceCodeClassifiers(directory, sourceCollection);
        }
    }

    private void loadSourceCodeClassifiers(String directory, ClassifierSourceCollection sourceCollection) {
        try {
            Files.list(Paths.get(directory)).forEach(f -> {
                if (f.toString().endsWith(ClassifierUtil.CUSTOM_CLASSIFIER_FILE_EXTENSION)) {
                    ClassifierFactory classifierFactory = null;
                    try {
                        classifierFactory = ClassifierUtil.loadClassifier(f.toFile(), fastHeapSupplier, symbolsSupplier);
                        if (classifierFactory != null && (classifierFactory.getSourceCollection() == sourceCollection || classifierFactory.getSourceCollection() ==
                                ClassifierSourceCollection.ALL)) {
                            add(classifierFactory.create());
                        }
                    } catch (InstantiationException | IllegalAccessException | IOException e) {
                        e.printStackTrace();
                    }
                }
            });
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private void loadCompiledClassifiers(ClassLoader cl, ClassifierSourceCollection sourceCollection) {
        @SuppressWarnings("rawtypes")
        ServiceLoader<Classifier> serviceLoader = ServiceLoader.load(Classifier.class, cl);
        try {
            for (Classifier<?> oc : serviceLoader) {
                if (sourceCollection == null || oc.getSourceCollection() == sourceCollection || oc.getSourceCollection() == ClassifierSourceCollection.ALL) {
                    add(oc);
                }
            }
        } catch (ServiceConfigurationError sce) {
            LOGGER.warning("Loading object classifiers failed!\n" + sce);
        }
    }

    public Classifier<?> get(String name) {
        for (ClassifierFactory oc : this) {
            Classifier<?> classifier = oc.create();
            if (classifier.getName().equals(name)) {
                return classifier;
            }
        }
        return null;
    }

    public void add(Classifier<?> oc) {
        for (int i = size() - 1; i >= 0; i--) {
            if (get(i).getName().equals(oc.getName())) {
                remove(i);
            }
        }
        add(new ClassifierFactory(symbolsSupplier, fastHeapSupplier, oc));
    }

    public boolean contains(String name) {
        return get(name) != null;
    }
}

