package at.jku.mevss.anttracks.classification.trees;

import at.jku.mevss.anttracks.classification.ClassifierChain;
import at.jku.mevss.anttracks.classification.ConditionalClassifierList;
import at.jku.mevss.anttracks.classification.Filter;
import at.jku.mevss.anttracks.classification.nodes.GroupingException;
import at.jku.mevss.anttracks.classification.nodes.GroupingNode;
import at.jku.mevss.anttracks.util.ApplicationStatistics;
import javafx.concurrent.Task;

import javax.naming.OperationNotSupportedException;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.List;
import java.util.Stack;
import java.util.concurrent.atomic.AtomicBoolean;

public abstract class ClassificationTree {

    public Filter[] getFilters() {
        return filters;
    }

    public ConditionalClassifierList getConditionalClassifiers() {
        return conditionalClassifiers;
    }

    public interface ClosureCalcListener {
        public void update(double progress);
    }

    protected GroupingNode root;
    private final Filter[] filters;
    private final ClassifierChain classifiers;
    private final ConditionalClassifierList conditionalClassifiers;

    // metrics
    protected long objectSum = -1;  // sum of object counts of ALL nodes
    protected long nNodes = -1;
    protected long nDataNodes = -1;    // nodes with at least one data entry
    protected long nDataCollectionParts = -1;
    private double avgTreeNodeDataCollectionPartsPerNode = -1;
    private double avgObjectsPerNode = -1;
    private double avgTreeNodeDataCollectionPartsPerDataNode = -1;
    private double avgObjectsPerDataNode = -1;
    private double avgObjectsPerTreeNodeDataCollectionPart = -1;

    protected AtomicBoolean cancellationToken;

    public ClassificationTree(GroupingNode root, Filter[] filters, ClassifierChain classifiers, ConditionalClassifierList conditionalClassifiers) {
        this.root = root;
        this.filters = filters;
        this.classifiers = classifiers;
        this.conditionalClassifiers = conditionalClassifiers;
    }

    public void init() {
        // Try to save memory and time by calculating the whole tree's object and byte counts in one traversal
        calculateObjectAndByteCounts();
        getRoot().fillClosureSizes();

        // calculate additional metrics
        if (nNodes != -1 && nDataNodes != -1 && nDataCollectionParts != -1) {
            avgTreeNodeDataCollectionPartsPerNode = 1.0 * nDataCollectionParts / nNodes;
            avgObjectsPerNode = 1.0 * root.getObjectCount() / nNodes;
            avgTreeNodeDataCollectionPartsPerDataNode = 1.0 * nDataCollectionParts / nDataNodes;
            avgObjectsPerDataNode = 1.0 * root.getObjectCount() / nDataNodes;
            avgObjectsPerTreeNodeDataCollectionPart = 1.0 * root.getObjectCount() / nDataNodes;
        }
    }

    public Task<Void> initClosureTask() {
        cancellationToken = new AtomicBoolean(false);
        Task<Void> t = new Task<Void>() {
            @Override
            protected Void call() throws Exception {
                try {
                    updateTitle("Calculate closures");
                    updateMessage("Transitive closure, GC closure & DS closures");
                    calculateClosures(progress -> updateProgress(progress, 1.0));
                } catch (Exception e) {
                    e.printStackTrace();
                }
                return null;
            }

            @Override
            protected void cancelled() {
                super.cancelled();
                cancellationToken.set(true);
            }
        };

        return t;
    }

    public abstract Task<Void> writeTreeTask(String metaDir, String postfix) throws OperationNotSupportedException;

    protected void writeHeapMetrics(DataOutputStream stream) throws IOException {
        stream.writeLong(objectSum);
        stream.writeLong(nNodes);
        stream.writeLong(nDataNodes);
        stream.writeLong(nDataCollectionParts);

        stream.writeDouble(avgTreeNodeDataCollectionPartsPerNode);
        stream.writeDouble(avgObjectsPerNode);
        stream.writeDouble(avgTreeNodeDataCollectionPartsPerDataNode);
        stream.writeDouble(avgObjectsPerDataNode);
        stream.writeDouble(avgObjectsPerTreeNodeDataCollectionPart);
    }

    protected void readHeapMetrics(DataInputStream stream) throws IOException {
        objectSum = stream.readLong();
        nNodes = stream.readLong();
        nDataNodes = stream.readLong();
        nDataCollectionParts = stream.readLong();

        avgTreeNodeDataCollectionPartsPerNode = stream.readDouble();
        avgObjectsPerNode = stream.readDouble();
        avgTreeNodeDataCollectionPartsPerDataNode = stream.readDouble();
        avgObjectsPerDataNode = stream.readDouble();
        avgObjectsPerTreeNodeDataCollectionPart = stream.readDouble();
        ;
    }

    public abstract void calculateObjectAndByteCounts();

    public abstract void calculateClosures(ClosureCalcListener listener) throws OperationNotSupportedException;

    public void sampleTopDown() {
        root.sampleTopDown();
    }

    public List<GroupingException> getExceptions() {
        return root.getExceptions();
    }

    public GroupingNode getRoot() {
        return root;
    }

    public boolean closuresInitialized() {
        Stack<GroupingNode> toVisit = new Stack<>();
        toVisit.push(root);

        ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance().createMeasurement("isClosuresInitialized");
        while (!toVisit.empty()) {
            GroupingNode node = toVisit.pop();
            if (!node.isClosureSizeCalculated() || !node.isGCSizeCalculated()) {
                return false;
            }
            for (GroupingNode child : node.getChildren()) {
                toVisit.push(child);
            }
        }
        m.end();
        return true;
    }

    public long getnNodes() {
        return nNodes;
    }

    public long getnDataNodes() {
        return nDataNodes;
    }

    public long getnDataCollectionParts() {
        return nDataCollectionParts;
    }

    public double getAvgTreeNodeDataCollectionPartsPerNode() {
        return avgTreeNodeDataCollectionPartsPerNode;
    }

    public double getAvgObjectsPerNode() {
        return avgObjectsPerNode;
    }

    public double getAvgTreeNodeDataCollectionPartsPerDataNode() {
        return avgTreeNodeDataCollectionPartsPerDataNode;
    }

    public double getAvgObjectsPerDataNode() {
        return avgObjectsPerDataNode;
    }

    public double getAvgObjectsPerTreeNodeDataCollectionPart() {
        return avgObjectsPerTreeNodeDataCollectionPart;
    }

    public ClassifierChain getClassifiers() {
        return classifiers;
    }
}