package at.jku.mevss.anttracks.classification.trees;

import at.jku.mevss.anttracks.classification.Classifier;
import at.jku.mevss.anttracks.classification.ClassifierChain;
import at.jku.mevss.anttracks.classification.ConditionalClassifierList;
import at.jku.mevss.anttracks.classification.Filter;
import at.jku.mevss.anttracks.classification.nodes.GroupingNode;
import at.jku.mevss.anttracks.classification.nodes.IndexCollection;
import at.jku.mevss.anttracks.classification.nodes.ListGroupingNode;
import at.jku.mevss.anttracks.heap.Closures;
import at.jku.mevss.anttracks.util.ApplicationStatistics;
import at.jku.mevss.anttracks.util.Consts;
import at.jku.mevss.anttracks.util.Counter;
import at.jku.mevss.anttracks.util.ParallelizationUtil;
import javafx.concurrent.Task;

import java.io.*;
import java.util.BitSet;
import java.util.HashMap;
import java.util.List;

public class ListClassificationTree extends ClassificationTree {
    private static class ClosureInitInfo {
        public final IndexCollection data;
        public final Closures closures;

        public ClosureInitInfo(IndexCollection data, Closures closure) {
            this.data = data;
            this.closures = closure;
        }
    }

    private boolean closuresCalculated = false;
    private ClosureCalcListener progressListener;
    private Counter progressCounter;

    public ListClassificationTree(ListGroupingNode grouping, Filter[] filters, ClassifierChain classifiers, ConditionalClassifierList conditionalClassifiers) {
        super(grouping, filters, classifiers, conditionalClassifiers);
    }

    @Override
    public void calculateObjectAndByteCounts() {
        fillObjectAndByteCounts((ListGroupingNode) root);
    }

    @Override
    public void calculateClosures(ClosureCalcListener listener) {
        progressListener = listener;
        progressCounter = new Counter();
        closuresCalculated = fillClosures((ListGroupingNode) root, true) != null;
        progressListener = null;
        progressCounter = null;
    }

    private ClosureInitInfo fillClosures(ListGroupingNode node, boolean multiThread) {
        ClosureInitInfo ret;
        List<GroupingNode> childrenOnSameSubtreeLevel = node.getFirstChildrenOnSameSubtreeLevel();

        if (node.getChildren().isEmpty()) {
            ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance().createMeasurement("Fill Closures - Leaf");
            // Init children, no big deal
            Closures closures = node.getData().calculateClosures();

            ApplicationStatistics.Measurement m2 = ApplicationStatistics.getInstance().createMeasurement("Fill Closures - Leaf - Calc byte counts");
            node.setClosureSize(closures.getTransitiveClosureByteCount());
            node.setGCSize(closures.getGCClosureByteCount());
            node.setDataStructureSize(closures.getDataStructureClosureByteCount());
            node.setDeepDataStructureSize(closures.getDeepDataStructureClosureByteCount());
            m2.end();

            closureCalculationProgressed();
            m.end();
            ret = new ClosureInitInfo(node.getData(), closures);
        } else {
            // Return my data plus the data of all children
            if (cancellationToken != null && cancellationToken.get()) {
                // avoid all recursion to end algorithm as quickly as possible
                return new ClosureInitInfo((IndexCollection) node.getData(false), new Closures(null, new int[0], new BitSet(), new BitSet(), new BitSet(), new BitSet()));
            }
            childrenOnSameSubtreeLevel.sort((o1, o2) -> (int) (o2.getObjectCount() - o1.getObjectCount()));

            ClosureInitInfo singleChildClosuresInitInfo = null;
            BitSet combinedChildClosure = new BitSet();
            IndexCollection[] childData = new IndexCollection[childrenOnSameSubtreeLevel.size()];
            Counter sameSubtreeLevelChildId = new Counter();

            if (multiThread && childrenOnSameSubtreeLevel.size() > 1) {
                ParallelizationUtil.temporaryExecutorServiceBlocking((threadId, threadCount) -> {
                    for (int childId = threadId; childId < node.getChildren().size(); childId += threadCount) {
                        ClosureInitInfo childClosuresInfo = fillClosures((ListGroupingNode) node.getChildren().get(childId), false);
                        if (childrenOnSameSubtreeLevel.contains(node.getChildren().get(childId))) {
                            synchronized (combinedChildClosure) {
                                combinedChildClosure.or(childClosuresInfo.closures.getTransitiveClosure());
                            }
                            synchronized (childData) {
                                childData[(int) sameSubtreeLevelChildId.get()] = childClosuresInfo.data;
                                sameSubtreeLevelChildId.inc();
                            }
                        }
                    }
                });
            } else {
                for (int childId = 0; childId < node.getChildren().size(); childId++) {
                    ClosureInitInfo childClosuresInfo = fillClosures((ListGroupingNode) node.getChildren().get(childId), false);
                    if (childrenOnSameSubtreeLevel.contains(node.getChildren().get(childId))) {
                        combinedChildClosure.or(childClosuresInfo.closures.getTransitiveClosure());
                        childData[childId] = childClosuresInfo.data;
                        if (childrenOnSameSubtreeLevel.size() == 1) {
                            singleChildClosuresInitInfo = childClosuresInfo;
                        }
                    }
                }
            }

            ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance().createMeasurement("Fill Closures - Parent");

            Closures closures;
            IndexCollection data = ((IndexCollection) node.getData(false)).clone();
            if (data.getCount() == 0 && singleChildClosuresInitInfo != null) {
                // Special case: We can use the child's data directly
                data = singleChildClosuresInitInfo.data;
                closures = singleChildClosuresInitInfo.closures;
            } else {
                data.unionWith(childData);
                closures = data.calculateClosures(combinedChildClosure);
            }

            ApplicationStatistics.Measurement m2 = ApplicationStatistics.getInstance().createMeasurement("Fill Closures - Parent - Calc byte counts");
            node.setClosureSize(closures.getTransitiveClosureByteCount());
            node.setGCSize(closures.getGCClosureByteCount());
            node.setDataStructureSize(closures.getDataStructureClosureByteCount());
            node.setDeepDataStructureSize(closures.getDeepDataStructureClosureByteCount());
            m2.end();

            m.end();

            closureCalculationProgressed();

            ret = new ClosureInitInfo(data, closures);
        }

        return ret;
    }

    public synchronized void closureCalculationProgressed() {
        if (progressListener != null && progressCounter != null) {
            progressCounter.inc();
            progressListener.update((1.0 * progressCounter.get()) / nNodes);
        }
    }

    private IndexCollection fillObjectAndByteCounts(ListGroupingNode node) {
        nNodes++;
        nDataNodes += Math.min(1, node.getData().getObjectCount());
        nDataCollectionParts += node.getData().getObjectCount();

        if (node.getChildren().isEmpty()) {
            // Init children, no big deal
            objectSum += node.getData().getObjectCount();
            node.getByteCount();

            // Add child objects to accumulator
            return (IndexCollection) node.getData(false);

        } else {
            // Return my data plus the data of all children
            List<GroupingNode> childrenOnSameSubtreeLevel = node.getFirstChildrenOnSameSubtreeLevel();
            IndexCollection[] childData = new IndexCollection[childrenOnSameSubtreeLevel.size()];
            for (int i = 0; i < childrenOnSameSubtreeLevel.size(); i++) {
                childData[i] = fillObjectAndByteCounts((ListGroupingNode) childrenOnSameSubtreeLevel.get(i));
            }
            IndexCollection ret = ((IndexCollection) node.getData(false)).clone();
            ret.unionWith(childData);

            node.setObjectCount(ret.getObjectCount());
            objectSum += ret.getObjectCount();
            node.setByteCount(ret.getBytes());

            node.getChildren()
                .stream()
                .filter(child -> !childrenOnSameSubtreeLevel.contains(child))
                .map(child -> (ListGroupingNode) child)
                .forEach(this::fillObjectAndByteCounts);

            return ret;
        }
    }

    public boolean isClosuresCalculated() {
        return closuresCalculated;
    }

    @Override
    public Task<Void> writeTreeTask(String metaDir, String postfix) {
        Task<Void> task = new Task<Void>() {
            @Override
            protected Void call() throws Exception {
                updateTitle("Classification tree");
                updateMessage("Write to file");

                try {
                    File file = new File(metaDir + File.separator + getClassifiers().toString()
                                                                                    .replace(" ", "_")
                                                                                    .replace("(", "")
                                                                                    .replace(")", "")
                                                                                    .replace("[", "")
                                                                                    .replace("]", "")
                                                                                    .replace(">", "-")
                                                                                    .replace("<", "-") + "_" + postfix + Consts.LIST_TREE_EXTENSION);
                    if (!file.exists()) {
                        // System.out.println(file.getAbsolutePath());
                        file.createNewFile();
                    }

                    DataOutputStream stream = new DataOutputStream(new FileOutputStream(file, false));
                    ListGroupingNode.writeTree(stream, (ListGroupingNode) root, new Counter(), nNodes, progress -> updateProgress(progress, 1.0));
                    writeHeapMetrics(stream);
                } catch (IOException e) {
                    e.printStackTrace();
                }
                return null;
            }
        };
        return task;
    }

    public static ClassificationTree readTree(String metaDir,
                                              Filter[] filters,
                                              ClassifierChain classifiers,
                                              ConditionalClassifierList conditionalClassifiers,
                                              String postfix,
                                              HashMap<String, Classifier<?>> classifierMap) {
        try {
            File file = new File(metaDir + File.separator + classifiers.toString()
                                                                       .replace(" ", "_")
                                                                       .replace("(", "")
                                                                       .replace(")", "")
                                                                       .replace("[", "")
                                                                       .replace("]", "")
                                                                       .replace(">", "-")
                                                                       .replace("<", "-") + "_" + postfix + Consts.LIST_TREE_EXTENSION);
            if (file.exists() && file.isFile()) {
                DataInputStream stream = new DataInputStream(new FileInputStream(file));
                ListClassificationTree tree =
                        new ListClassificationTree(ListGroupingNode.readTree(stream, null, classifierMap), filters, classifiers, conditionalClassifiers);
                tree.readHeapMetrics(stream);
                return tree;
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
        return null;
    }
}
