/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.diff;

import at.jku.mevss.anttracks.diff.Interval.IntervalListener;
import at.jku.mevss.anttracks.heap.DetailedHeap;
import at.jku.mevss.anttracks.heap.objects.ObjectInfo;
import at.jku.mevss.anttracks.heap.symbols.Symbols;
import at.jku.mevss.anttracks.util.ApplicationStatistics;

import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.logging.Logger;

public class HeapDiff {
    // TODO: Remove, only use one interval
    private final List<Interval> intervals;
    private Interval currentInterval;
    private int currentIntervalId;

    private final int intervalsToFill;
    private long[] fromTime;
    private long[] toTime;

    private long lastMeasureTime = -1;

    //private Measurement moveMeasurement = null;

    private CopyOnWriteArrayList<IntervalListener> listener = new CopyOnWriteArrayList<>();

    private final Logger LOGGER = Logger.getLogger(HeapDiff.class.getSimpleName());

    public HeapDiff(long[] fromTime, long[] toTime) {
        assert fromTime.length == toTime.length : "Number of intervals can not be determined if array-length does not match";
        intervalsToFill = fromTime.length;
        this.fromTime = fromTime;
        this.toTime = toTime;

        intervals = new LinkedList<>();
        currentInterval = null;

        currentIntervalId = 0;
    }

    public HeapDiff(long fromTime, long toTime) {
        this(new long[]{fromTime}, new long[]{toTime});
    }

    public void add(DetailedHeap heap, long time, boolean gcStart) {
        // Check if all intervals filled
        if (handlingFinished()) {
            return;
        }

        // handle current interval
        if (isIntervalOpen()) {
            addToInterval(heap, time, gcStart);
        }

        // check if at end of interval
        if (time == toTime[currentIntervalId]) {
            endInterval(heap, time);
        }
        // check if at start of interval
        if (!handlingFinished() && !isIntervalOpen() && time == fromTime[currentIntervalId]) {
            startInterval(heap, time, gcStart);
        }
    }

    public void updateAddress(long fromAddr, long toAddr) {
        if (isIntervalOpen()) {
            currentInterval.updateAddress(fromAddr, toAddr);
        }
    }

    public void updateMultipleAddresses(long fromAddr, long toAddr, List<ObjectInfo> movedObjects) {
        if (isIntervalOpen()) {
            currentInterval.updateMultipleAddresses(fromAddr, toAddr, movedObjects);
        }
    }

    private void startInterval(DetailedHeap initial, long time, boolean gcStart) {
        assert currentInterval == null : "Can only start a new interval if interval not already existing";
        currentInterval = new Interval(initial, time, gcStart);
        intervals.add(currentInterval);
        if (gcStart) {
            listener.forEach(x -> x.gcStartHeapStateAdded(time,
                                                          currentInterval.getPermConsumption(),
                                                          currentInterval.getDiedConsumption(),
                                                          currentInterval.getBornConsumption(),
                                                          currentInterval.getTempConsumption()));
            //moveMeasurement = ApplicationStatistics.getInstance().createMeasurement("Diffing: Move", Thread.currentThread().toString());
            lastMeasureTime = System.nanoTime();
        } else {
            listener.forEach(x -> x.gcEndHeapStateAdded(time,
                                                        currentInterval.getPermConsumption(),
                                                        currentInterval.getDiedConsumption(),
                                                        currentInterval.getBornConsumption(),
                                                        currentInterval.getTempConsumption(),
                                                        currentInterval.info));
        }
    }

    private void addToInterval(DetailedHeap next, long time, boolean gcStart) {
        if (gcStart) {
            //moveMeasurement = ApplicationStatistics.getInstance().createMeasurement("Diffing: Move", Thread.currentThread().toString());
            lastMeasureTime = System.nanoTime();
        } else {
            ApplicationStatistics.getInstance().addCompletedMeasurement("Diffing: Move", System.nanoTime() - lastMeasureTime);
            lastMeasureTime = -1;
            //moveMeasurement.end();
            //moveMeasurement = null;
        }

        currentInterval.add(next, time, gcStart);
        if (gcStart) {
            listener.forEach(x -> x.gcStartHeapStateAdded(time,
                                                          currentInterval.getPermConsumption(),
                                                          currentInterval.getDiedConsumption(),
                                                          currentInterval.getBornConsumption(),
                                                          currentInterval.getTempConsumption()));
        } else {
            listener.forEach(x -> x.gcEndHeapStateAdded(time,
                                                        currentInterval.getPermConsumption(),
                                                        currentInterval.getDiedConsumption(),
                                                        currentInterval.getBornConsumption(),
                                                        currentInterval.getTempConsumption(),
                                                        currentInterval.info));
        }
    }

    private boolean isIntervalOpen() {
        return currentInterval != null;
    }

    private boolean handlingFinished() {
        return intervalsToFill == currentIntervalId;
    }

    private void endInterval(DetailedHeap last, long time) {
        //if (moveMeasurement != null) {
        //    moveMeasurement.end();
        //    moveMeasurement = null;
        //}

        if (lastMeasureTime >= 0) {
            ApplicationStatistics.getInstance().addCompletedMeasurement("Diffing: Move", System.nanoTime() - lastMeasureTime);
            lastMeasureTime = -1;
        }

        currentInterval.end(last);
        listener.forEach(x -> x.intervalEnded(time,
                                              last,
                                              currentInterval.getPermConsumption(),
                                              currentInterval.getDiedConsumption(),
                                              currentInterval.getBornConsumption(),
                                              currentInterval.getTempConsumption(),
                                              currentInterval.info));
        currentInterval = null;
        currentIntervalId++;
    }

    public void printIntervalState(Symbols symbols) {
        if (isIntervalOpen()) {
            System.out.println(currentInterval);
        } else {
            if (currentIntervalId > 0) {
                System.out.println("Interval currently closed, showing latest interval");
                System.out.println(intervals.get(currentIntervalId - 1));
            }
        }
    }

    public List<Interval> getIntervals() {
        return intervals;
    }

    public void close(DetailedHeap heap) {
        if (isIntervalOpen()) {
            // This can only happen when the diff ends on an "GC start event". The last "GC start event" is not read from the trace,
            // therefore we have to to artificially create it.
            LOGGER.info("Close on open interval received");
            // Create artificial GC start event
            addToInterval(heap, toTime[currentIntervalId], true);
            endInterval(heap, toTime[currentIntervalId]);
        }
    }

    public void addIntervalListener(IntervalListener listener) {
        this.listener.add(listener);
    }

    public void removeAllListener() {
        listener.clear();
    }

    public void clear() {
        intervals.clear();
        currentInterval = null;

        fromTime = null;
        toTime = null;

        //moveMeasurement = null;

        listener.clear();
        listener = null;
    }
}
