/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.diff;

import at.jku.mevss.anttracks.classification.Classifier;
import at.jku.mevss.anttracks.classification.ClassifierChain;
import at.jku.mevss.anttracks.compiler.OnTheFlyCompiler;
import at.jku.mevss.anttracks.heap.GarbageCollectionLookup;
import at.jku.mevss.anttracks.heap.io.MetaDataReaderConfig;
import at.jku.mevss.anttracks.heap.io.MetaDataWriterConfig;
import at.jku.mevss.anttracks.heap.labs.Lab;
import at.jku.mevss.anttracks.heap.symbols.Symbols;
import at.jku.mevss.anttracks.parser.EventType;
import at.jku.mevss.anttracks.parser.TraceParser;
import at.jku.mevss.anttracks.parser.symbols.SymbolsParser;
import at.jku.mevss.anttracks.util.Consts;
import at.jku.mevss.anttracks.util.GCReporter;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.List;

public class HeapDiffMain {
    @SuppressWarnings("unused")
    public static void main(String argv[]) throws IOException {
        List<File> symbols = new ArrayList<>();
        String metaPath = ".ant_tracks_trace_meta_data";
        File classifierPath = null;
        int metaDumpCount = MetaDataWriterConfig.DEFAULT_DUMP_COUNT;
        boolean metaMajorGCsOnly = MetaDataWriterConfig.DEFAULT_AT_MAJOR_GCS_ONLY;
        boolean check = false;

        String startTime = "";
        String endTime = "";

        for (int index = 0; index < argv.length; index++) {
            String arg = argv[index];
            if (arg.contains("=")) {
                int splitterIndex = arg.indexOf("=");
                String key = arg.substring(0, splitterIndex);
                String value = arg.substring(splitterIndex + 1);
                if (key.equals("ConsistencyTest")) {
                    check = Boolean.parseBoolean(value);
                } else if (key.equals("MetaDataPath")) {
                    metaPath = value;
                } else if (key.equals("ClassifierPath")) {
                    classifierPath = new File(value);
                } else if (key.equals("IntervalRange")) {
                    splitterIndex = value.indexOf("-");
                    startTime = value.substring(0, splitterIndex);
                    endTime = value.substring(splitterIndex + 1);
                } else {
                    System.err.println("Did not understand parameter '" + arg + "'");
                    System.exit(1);
                }
            } else if (arg.equals("GenerateMetaData")) {
                metaPath = Consts.ANT_META_DIRECTORY;
            } else {
                symbols.add(new File(arg));
            }
        }

        // affects performance only if LightweightStorage is enabled
        String addrGranularity = System.getProperty("EvalAddrGranularity");
        Lab.ADDR_GRANULARITY = addrGranularity != null ? Integer.parseInt(addrGranularity) : Lab.ADDR_GRANULARITY;
        // benchmark flag to compare parsing vs consuming of ptrs
        String parsePtrs = System.getProperty("ParsePtrs");
        SymbolsParser.parsePtrs = parsePtrs != null ? Boolean.parseBoolean(parsePtrs) : true;

        TraceParser.CONSISTENCY_CHECK = check;

        int errors = 0;
        GCReporter.getInstance();
        for (File file : symbols) {
            try {
                SymbolsParser symbolsParser = new SymbolsParser(file, null);
                Symbols sym = symbolsParser.parse();
                MetaDataReaderConfig config = metaPath != null ? new MetaDataReaderConfig(sym.root + File.separator + metaPath) : null;

                HeapDiffTraceParser parser = null;
                if ((startTime.toUpperCase().startsWith("S") || startTime.toUpperCase().startsWith("E")) && (endTime.toUpperCase()
                                                                                                                    .startsWith("S") || endTime.toUpperCase()
                                                                                                                                               .startsWith("E"))) {
                    GarbageCollectionLookup start;
                    GarbageCollectionLookup end;

                    if (startTime.toUpperCase().startsWith("S")) {
                        start = new GarbageCollectionLookup(EventType.GC_START, null, null, Integer.valueOf(startTime.substring(1)));
                    } else {
                        start = new GarbageCollectionLookup(EventType.GC_END, null, null, Integer.valueOf(startTime.substring(1)));
                    }

                    if (endTime.toUpperCase().startsWith("S")) {
                        end = new GarbageCollectionLookup(EventType.GC_START, null, null, Integer.valueOf(startTime.substring(1)));
                    } else {
                        end = new GarbageCollectionLookup(EventType.GC_END, null, null, Integer.valueOf(endTime.substring(1)));
                    }

                    parser = new HeapDiffTraceParser(sym, config, start, end);
                }
                HeapDiff diff = parser.parse();
                OnTheFlyCompiler compiler = new OnTheFlyCompiler();
                Class<?> classifierClazz = compiler.compile(Files.readAllLines(classifierPath.toPath()).stream().reduce((a, b) -> a + "\n" + b).orElse(""));
                System.out.println(diff.getIntervals().get(0).info.classify(new ClassifierChain((Classifier<?>) classifierClazz.newInstance()), null, null));
                System.gc();
            } catch (Throwable e) {
                e.printStackTrace(System.err);
                errors++;
            }
        }
        if (errors > 0) {
            System.err.println(errors + " errors found.");
        }
    }

}
