/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.diff;

import at.jku.mevss.anttracks.classification.ClassifierChain;
import at.jku.mevss.anttracks.classification.ConditionalClassifierList;
import at.jku.mevss.anttracks.classification.Filter;
import at.jku.mevss.anttracks.heap.DetailedHeap;
import at.jku.mevss.anttracks.heap.ObjectStream.ThreadVisitorGenerator;
import at.jku.mevss.anttracks.heap.ObjectVisitor;
import at.jku.mevss.anttracks.heap.objects.ObjectInfo;
import at.jku.mevss.anttracks.heap.roots.RootPtr;
import at.jku.mevss.anttracks.heap.space.SpaceInfo;
import at.jku.mevss.anttracks.heap.statistics.MemoryConsumption;
import at.jku.mevss.anttracks.heap.symbols.AllocatedType;
import at.jku.mevss.anttracks.heap.symbols.AllocationSite;
import at.jku.mevss.anttracks.parser.EventType;
import at.jku.mevss.anttracks.util.ApplicationStatistics;
import at.jku.mevss.anttracks.util.ApplicationStatistics.Measurement;
import at.jku.mevss.anttracks.util.Counter;
import at.jku.mevss.anttracks.util.MutableInt;

import javax.naming.OperationNotSupportedException;
import java.io.*;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Logger;

public class Interval {
    public static class IntervalInfo {
        public long time;
        public AddressCollection perm;
        public AddressCollection born;
        public MultipleObjectCollection died;
        public MultipleObjectCollection temp;

        private DetailedHeap lastHeap;

        private IntervalInfo() {
        }

        @Override
        public IntervalInfo clone() {
            IntervalInfo cloned = new IntervalInfo();
            cloned.time = time;
            cloned.perm = perm.clone();
            cloned.born = born.clone();
            cloned.died = died.clone();
            cloned.temp = temp.clone();

            // Heap does not get cloned
            cloned.lastHeap = lastHeap;
            return cloned;
        }

        public void setLastHeap(DetailedHeap heap) {
            lastHeap = heap;
            perm.setLastHeap(lastHeap);
            born.setLastHeap(lastHeap);
        }

        public DetailedHeap getLastHeap() {
            return lastHeap;
        }

        public DiffGrouping classify(ClassifierChain classifier, ConditionalClassifierList conditionalClassifier, List<Filter> filters) {
            DiffGrouping grouping = new DiffGrouping(perm.classify(classifier, conditionalClassifier, filters),
                                                     born.classify(classifier, conditionalClassifier, filters),
                                                     died.classify(classifier, conditionalClassifier, filters),
                                                     temp.classify(classifier, conditionalClassifier, filters));
            grouping.sampleTopDown();
            return grouping;
        }

        public void writeToMetaData(File metaFile) throws OperationNotSupportedException {
            try {
                DataOutputStream dos = new DataOutputStream(new FileOutputStream(metaFile));
                dos.writeLong(time);
                perm.writeToMetaData(dos);
                born.writeToMetaData(dos);
                died.writeToMetaData(dos);
                temp.writeToMetaData(dos);
            } catch (FileNotFoundException e) {
                e.printStackTrace();
            } catch (IOException e) {
                e.printStackTrace();
            }
        }

        public static IntervalInfo readFromMetaData(DetailedHeap heap, File metaFile) {
            try {
                DataInputStream dis = new DataInputStream(new FileInputStream(metaFile));
                IntervalInfo info = new IntervalInfo();
                info.time = dis.readLong();
                info.perm = SimpleAddressCollection.readFromMetaData(dis);
                info.born = SimpleAddressCollection.readFromMetaData(dis);
                info.died = MultipleObjectCollection.readFromMetaData(heap, dis);
                info.temp = MultipleObjectCollection.readFromMetaData(heap, dis);
                info.setLastHeap(heap);
                return info;
            } catch (FileNotFoundException e) {
                e.printStackTrace();
            } catch (IOException e) {
                e.printStackTrace();
            }
            return null;
        }
    }

    public interface IntervalListener {
        void gcStartHeapStateAdded(long time, MemoryConsumption perm, MemoryConsumption died, MemoryConsumption born, MemoryConsumption temp);

        void gcEndHeapStateAdded(long time, MemoryConsumption perm, MemoryConsumption died, MemoryConsumption born, MemoryConsumption temp, IntervalInfo intervalInfo);

        void intervalEnded(long time,
                           DetailedHeap heap,
                           MemoryConsumption perm,
                           MemoryConsumption died,
                           MemoryConsumption born,
                           MemoryConsumption temp,
                           IntervalInfo intervalInfo);
    }

    private enum IntervalState {
        NONE,
        MUTATOR,
        GC
    }

    private static final Logger LOGGER = Logger.getLogger(SimpleAddressCollection.class.getSimpleName());

    public IntervalInfo info;
    private final ConcurrentHashMap<Thread, AddressCollection> permMovedTo;
    private final ConcurrentHashMap<Thread, AddressCollection> bornMovedTo;
    private final ConcurrentHashMap<Thread, AddressCollection> movedFromAddresses;

    private AddressCollection backPerm;
    private AddressCollection backBorn;

    private IntervalState state = IntervalState.NONE;

    private long lastAddMillis = 0;

    // Constructor only for testing!
    public Interval(DetailedHeap initial) {
        this(initial, 0, false);
    }

    /**
     * Constructs a new interval, where every object is categorized as PERM.
     *
     * @param initial The objects initially in the heap
     * @param time    The elapsed time since application start
     * @param gcStart True if the application is at the start of a GC phase, false
     *                if the application is at the start of a mutator phase
     */
    public Interval(DetailedHeap initial, long time, boolean gcStart) {
        info = new IntervalInfo();
        info.time = time;

        // Everything alive starts in perm
        info.perm = createObjectCollectionFromHeap(initial, "Perm");

        // Nothing born yet
        info.born = new SimpleAddressCollection("Born");

        // Nothing died yet or was only temporary
        info.died = new MultipleObjectCollection("Died");
        info.temp = new MultipleObjectCollection("Temp");

        permMovedTo = new ConcurrentHashMap<>();
        bornMovedTo = new ConcurrentHashMap<>();
        movedFromAddresses = new ConcurrentHashMap<>();

        if (gcStart) {
            swapOnGCStart();
        }
        setState(gcStart);
        setLastHeap(initial);
        lastAddMillis = System.currentTimeMillis();
    }

    public MemoryConsumption getPermConsumption() {
        Measurement m = ApplicationStatistics.getInstance().createMeasurement("Interval: perm consumption");
        MemoryConsumption consumption;
        if (state == IntervalState.GC) {
            consumption = new MemoryConsumption(backPerm.getObjectCount(), calculateMemoryUsage(backPerm));
        } else {
            consumption = new MemoryConsumption(info.perm.getObjectCount(), calculateMemoryUsage(info.perm));
        }
        m.end();
        return consumption;
    }

    public MemoryConsumption getDiedConsumption() {
        Measurement m = ApplicationStatistics.getInstance().createMeasurement("Interval: died consumption");
        MemoryConsumption consumption = new MemoryConsumption(info.died.getObjectCount(), calculateMemoryUsage(info.died));
        m.end();
        return consumption;
    }

    public MemoryConsumption getBornConsumption() {
        Measurement m = ApplicationStatistics.getInstance().createMeasurement("Interval: born consumption");
        MemoryConsumption consumption;
        if (state == IntervalState.GC) {
            consumption = new MemoryConsumption(backBorn.getObjectCount(), calculateMemoryUsage(backBorn));
        } else {
            consumption = new MemoryConsumption(info.born.getObjectCount(), calculateMemoryUsage(info.born));
        }
        m.end();
        return consumption;
    }

    public MemoryConsumption getTempConsumption() {
        Measurement m = ApplicationStatistics.getInstance().createMeasurement("Interval: temp consumption");
        MemoryConsumption consumption = new MemoryConsumption(info.temp.getObjectCount(), calculateMemoryUsage(info.temp));
        m.end();
        return consumption;
    }

    private long calculateMemoryUsage(AddressCollection collection) {
        collection.setLastHeap(getLastHeap());
        return collection.getBytes();
    }

    private long calculateMemoryUsage(MultipleObjectCollection collection) {
        return collection.getBytes();
    }

    /**
     * Sets the intervals state if the application is currently in a GC or a
     * MUTATOR phase
     *
     * @param gcStart True if the application is at the start of a GC phase, false
     *                if the application is at the start of a mutator phase
     */
    private void setState(boolean gcStart) {
        state = gcStart ? IntervalState.GC : IntervalState.MUTATOR;
    }

    /**
     * Moves an object from its old address to its new one. May be called
     * multi-threaded (multiple GC threads move objects at the same time),
     * objects are stored thread-local.
     *
     * @param fromAddr The object's from address
     * @param toAddr   The objects to address
     */
    public void updateAddress(long fromAddr, long toAddr) {
        assert state == IntervalState.GC : "address updates may only happen during GC (= while back exists)";
        /*
         * ClassifyableAddressCollection containingFrontSet;
         *
         * if (backPerm.remove(fromAddr)) { containingFrontSet = info.perm; }
         * else { boolean check = backBorn.remove(fromAddr); assert check :
         * String.format(
         * "address %d must be contained in either BackPerm or BackBorn",
         * fromAddr); containingFrontSet = info.born; }
         *
         * boolean wasInBackLast = back.remove(fromAddr); assert wasInBackLast;
         */

        // assert !front.contains(toAddr) : "two moves to the same address in
        // the same GC --> impossible";
        // assert !containingFrontSet.contains(toAddr) : "two moves to the same
        // address in the same GC --> impossible";

        ConcurrentHashMap<Thread, AddressCollection> containingFrontMap;

        if (backPerm.contains(fromAddr)) {
            containingFrontMap = permMovedTo;
        } else {
            containingFrontMap = bornMovedTo;
        }

        AddressCollection containingFrontList = containingFrontMap.get(Thread.currentThread());
        if (containingFrontList == null) {
            containingFrontMap.put(Thread.currentThread(), new SimpleAddressCollection("default"));
            containingFrontList = containingFrontMap.get(Thread.currentThread());
        }
        containingFrontList.put(toAddr);

        AddressCollection containingMovedList = movedFromAddresses.get(Thread.currentThread());
        if (containingMovedList == null) {
            movedFromAddresses.put(Thread.currentThread(), new SimpleAddressCollection("default"));
            containingMovedList = movedFromAddresses.get(Thread.currentThread());
        }
        containingMovedList.put(fromAddr);
    }

    /**
     * Moves multiple address-consecutive objects from their old address to
     * their new one. May be called multi-threaded (multiple GC threads move
     * objects at the same time), objects are stored thread-local.
     *
     * @param fromAddr     The first object's from address
     * @param toAddr       The first object's to address
     * @param movedObjects A list of all moved objects
     */
    public void updateMultipleAddresses(long fromAddr, long toAddr, List<ObjectInfo> movedObjects) {
        for (ObjectInfo obj : movedObjects) {
            updateAddress(fromAddr, toAddr);
            fromAddr += obj.size;
            toAddr += obj.size;
        }
    }

    public void add(DetailedHeap next, long time, boolean gcStart) {
        Measurement m = ApplicationStatistics.getInstance().createMeasurement("Interval: add");
        info.time = time;
        LOGGER.finest("-----add(ObjectVisitable next, boolean gcStart)-----");
        long startMethodTime = System.currentTimeMillis();
        if (!gcStart) {
            LOGGER.finest("GC END:\tTime since last ADD (GC START) - Moves: " + (System.currentTimeMillis() - lastAddMillis) + "ms");
            // Make sure interval was in MUTATOR state
            assert state == IntervalState.GC : "A GC end can only occur after a GC phase";
            // After a GC, merge back into front
            long startMergeTime = System.currentTimeMillis();
            mergeBackToFrontOnGCEnd(next);
            LOGGER.finest("GC END:\tMerge back to front took " + (System.currentTimeMillis() - startMergeTime) + "ms");
        } else {
            LOGGER.finest("GC START:\tTime since last ADD (GC END) - Object allocations: " + (System.currentTimeMillis() - lastAddMillis) + "ms");
        }

        assert (state == IntervalState.GC && !gcStart) || (state == IntervalState.MUTATOR && gcStart) : "There has to be a GC phase switch";

        long startAddTime = System.currentTimeMillis();
        add(next);
        LOGGER.finest((gcStart ? "GC START" : "GC END") + ":\tAdding heap took " + (System.currentTimeMillis() - startAddTime) + "ms");

        if (gcStart) {
            // Make sure interval was in MUTATOR state
            assert state == IntervalState.MUTATOR : "A GC start can only occur after a MUTATOR phase";
            // Move all objects from front to back when GC starts
            swapOnGCStart();
        }

        LOGGER.finest((gcStart ? "GC START" : "GC END") + ":\tComplete method took " + (System.currentTimeMillis() - startMethodTime) + "ms");
        setState(gcStart);
        setLastHeap(next);
        lastAddMillis = System.currentTimeMillis();
        LOGGER.finest("----------------------------------------------------");
        m.end();
    }

    private void add(DetailedHeap heap) {
        // State here:
        // info contains either
        // (1) after GC -> State after all object moves (no knowledge about
        // objects created / died during GC (except overitten objs))
        // (2) after mutator phase -> Last heap state (no knowledge about
        // objects created during mutator phase)
        // Heap contains current heap

        SimpleAddressCollection heapAsAddressCollection = createObjectCollectionFromHeap(heap, "Default");

        // Objects could only die if the last phase was a GC phase, not in
        // mutator phase
        // IntervalState.NONE = GC without moves (only used in unit-tests)
        if (state == IntervalState.GC || state == IntervalState.NONE) {
            AddressCollection freedInPerm = info.perm.cloneWithout(heapAsAddressCollection);
            AddressCollection freedInBorn = info.born.cloneWithout(heapAsAddressCollection);

            // Perm
            info.perm.without(freedInPerm);

            info.born.without(freedInBorn);

            // Died
            info.died.unionWith(freedInPerm, heap);
            // Temp
            info.temp.unionWith(freedInBorn, heap);
        }

        heapAsAddressCollection.without(info.perm);
        info.born.unionWith(heapAsAddressCollection);

        // info now contains current state
    }

    private void mergeBackToFrontOnGCEnd(DetailedHeap next) {
        // We have the following state here:
        // backPerm / backBorn contains perm / born elements at the GC-Start
        // state (haven't been touched during Mutator phase)
        // permsFronts contains thread-local to-addresses of perm moves
        // bornFronts contains thread-local to-addresses of born moves
        // movedFromAddresses contains thread-local from-addresses of all moves
        // currentHeap contains the exact current heap
        info.perm.unionWith(permMovedTo.values().toArray(new AddressCollection[permMovedTo.size()]));
        permMovedTo.clear();

        info.born.unionWith(bornMovedTo.values().toArray(new AddressCollection[bornMovedTo.size()]));
        bornMovedTo.clear();

        backPerm.without(movedFromAddresses.values().toArray(new AddressCollection[movedFromAddresses.size()]));
        backBorn.without(movedFromAddresses.values().toArray(new AddressCollection[movedFromAddresses.size()]));
        movedFromAddresses.clear();

        // We have the following state here:
        // BackPerm / BackBorn contains perm / born elements that have not been
        // moved
        // Info.perm / Info.born contains all moved perm objects / born objects
        // permsFronts contains empty
        // bornFronts contains empty
        // movedFromAddresses empty
        // CurrentHeap contains the exact current heap
        AddressCollection overwrittenPerm = backPerm.cloneAndIntersectWith(info.born);
        overwrittenPerm.unionWith(backPerm.cloneAndIntersectWith(info.perm));

        AddressCollection overwrittenBorn = backBorn.cloneAndIntersectWith(info.born);
        overwrittenBorn.unionWith(backBorn.cloneAndIntersectWith(info.perm));

        info.died.unionWith(overwrittenPerm, next);
        info.temp.unionWith(overwrittenBorn, next);
        backPerm.without(overwrittenPerm);
        backBorn.without(overwrittenBorn);

        // We have the following state here:
        // BackPerm / BackBorn contains all / perm / born elements that have not
        // been moved & not overwritten by moves
        // Info.perm / Info.born contains all moved perm objects / born objects
        // Info.died contains overwritten perm objects
        // Info.temp contains overwritten temp objects
        // permsFronts contains empty
        // bornFronts contains empty
        // movedFromAddresses empty
        // CurrentHeap contains the exact current heap
        info.perm.unionWith(backPerm);
        info.born.unionWith(backBorn);

        backPerm.clear();
        backBorn.clear();
        backPerm = null;
        backBorn = null;

    }

    private void swapOnGCStart() {
        backPerm = info.perm;
        backBorn = info.born;

        info.perm = new SimpleAddressCollection("Perm");
        info.born = new SimpleAddressCollection("Born");
    }

    @Override
    public String toString() {
        StringBuilder string = new StringBuilder();
        string.append("Read from ");
        string.append(state == IntervalState.MUTATOR ? "FRONT:\n" : "BACK:\n");
        if (state == IntervalState.MUTATOR) {
            string.append("perm = ");
            string.append(info.perm.getObjectCount());
            string.append("\n");
            string.append("born = ");
            string.append(info.born.getObjectCount());
            string.append("\n");
        } else {
            string.append("perm = ");
            string.append(backPerm.getObjectCount());
            string.append("\n");
            string.append("born = ");
            string.append(backBorn.getObjectCount());
            string.append("\n");
        }

        string.append("died = ");
        string.append(info.died.getObjectCount());
        string.append("\n");
        string.append("temp = ");
        string.append(info.temp.getObjectCount());
        string.append("\n");

        return string.toString();
    }

    private SimpleAddressCollection createObjectCollectionFromHeap(DetailedHeap heap, String name) {
        class CounterVisitor extends Counter implements ObjectVisitor {

            @Override
            public void visit(long address,
                              ObjectInfo objectInfo,
                              SpaceInfo space,
                              AllocatedType type,
                              int size,
                              boolean isArray,
                              int arrayLength,
                              AllocationSite allocationSite,
                              long[] pointedFrom,
                              long[] pointsTo,
                              EventType eventType,
                              List<RootPtr> rootPtrs) {
                inc();
            }
        }
        ThreadVisitorGenerator<CounterVisitor> gen = () -> new CounterVisitor();

        Measurement m = ApplicationStatistics.getInstance().createMeasurement("Interval: collection from iterable", Thread.currentThread().getName());

        // Count heap size
        Counter sum = new Counter();
        heap.toObjectStream().forEachParallel(gen, new ObjectVisitor.Settings(false, true, false)).forEach(sum::add);

        final MutableInt i = new MutableInt();
        long[] addresses = new long[(int) sum.get()];
        sum.reset();
        heap.toObjectStream()
            .sorted()
            .forEach((address, objectInfo, space, type, size, isArray, arrayLength, allocationSite, pointedFrom, pointsTo, eventType, rootPtrs) -> {
                addresses[i.get()] = address;
                i.inc();
            }, ObjectVisitor.Settings.NO_INFOS);

        SimpleAddressCollection result = SimpleAddressCollection.from(addresses, name);

        m.end();
        LOGGER.finest(m.toString());

        return result;
    }

    public void setLastHeap(DetailedHeap lastHeap) {
        info.setLastHeap(lastHeap);
    }

    public DetailedHeap getLastHeap() {
        return info.lastHeap;
    }

    public void end(DetailedHeap last) {
        if (state == IntervalState.GC) {
            // We just received an artificial GC start event without moves. We
            // can close the interval now, we can just move back to front
            info.perm = backPerm;
            info.born = backBorn;
            backPerm = null;
            backBorn = null;
            setState(false);
        }

        setLastHeap(last);
    }
}
