/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.diff;

import at.jku.mevss.anttracks.classification.ClassifierChain;
import at.jku.mevss.anttracks.classification.ConditionalClassifierList;
import at.jku.mevss.anttracks.classification.Filter;
import at.jku.mevss.anttracks.classification.nodes.MapGroupingNode;
import at.jku.mevss.anttracks.heap.DetailedHeap;
import at.jku.mevss.anttracks.heap.ObjectVisitor;
import at.jku.mevss.anttracks.heap.objects.ArrayObjectInfo;
import at.jku.mevss.anttracks.heap.objects.InstanceMirrorObjectInfo;
import at.jku.mevss.anttracks.heap.objects.InstanceObjectInfo;
import at.jku.mevss.anttracks.heap.objects.ObjectInfo;
import at.jku.mevss.anttracks.util.ApplicationStatistics;
import at.jku.mevss.anttracks.util.ApplicationStatistics.Measurement;
import at.jku.mevss.anttracks.util.Counter;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.AbstractMap;
import java.util.HashMap;
import java.util.List;
import java.util.Set;
import java.util.logging.Logger;
import java.util.stream.Stream;

public class MultipleObjectCollection {
    private final HashMap<ObjectInfo, Counter> map = new HashMap<>();
    @SuppressWarnings("unused")
    private String name = null;

    private static final Logger LOGGER = Logger.getLogger(MultipleObjectCollection.class.getSimpleName());

    public MultipleObjectCollection() {
        this("Default");
    }

    public MultipleObjectCollection(String name) {
        this.name = name;
    }

    public void clear() {
        map.clear();
    }

    public void put(ObjectInfo object) {
        this.put(object, 1);
    }

    public void put(ObjectInfo object, long count) {
        Counter objectCounter = map.get(object);
        if (objectCounter == null) {
            map.put(object, new Counter(0));
            objectCounter = map.get(object);
        }
        objectCounter.add(count);
    }

    public boolean remove(ObjectInfo object) {
        boolean wasRemoved = remove(object, 1);
        return wasRemoved;
    }

    public boolean remove(ObjectInfo object, long count) {
        Counter objectCounter = map.get(object);
        if (objectCounter == null) {
            return false;
        }
        objectCounter.sub(count);
        return true;
    }

    public boolean contains(ObjectInfo object) {
        Counter objectCounter = map.get(object);
        return (objectCounter != null && objectCounter.get() > 0);
    }

    public long getObjectCount() {
        return map.values().stream().mapToLong(counterMapper -> counterMapper.get()).sum();
    }

    public long getObjectCount(ObjectInfo object) {
        return map.get(object).get();
    }
    // TODO: these getObject methods return only ObjectIDs, i.e. no info about
    // how many objects of the given ObjectID are present in the set.
    // This count is accessible via getObjectCount(ObjectID obj), but this needs
    // 2 additional hashmap-gets-->make better?
    // currently unused, if usage becomes relevant fix!

    public Set<ObjectInfo> getObjectInfos() {
        return map.keySet();
    }

    // public Set<ObjectInfo> getObjectsClassified(Object key) {
    // if(!isCorrectlyClassifiedOrClassifyable()) {
    // // TODO: What to do if classifier is not set?
    // return null;
    // }
    // if(isClassificationNeeded()) {
    // reclassify();
    // }
    // return classified.get(key).keySet();
    // }

    public Stream<ObjectInfo> stream() {
        return map.keySet().stream();
    }

    @Override
    public MultipleObjectCollection clone() {
        MultipleObjectCollection dolly = new MultipleObjectCollection();
        dolly.cloneFrom(this);
        return dolly;
    }

    private void cloneFrom(MultipleObjectCollection other) {
        map.clear();
        unionWith(other);
    }

    public void unionWith(MultipleObjectCollection other) {
        Measurement m = ApplicationStatistics.getInstance().createMeasurement("MultipleObjectCollection: union multi<->multi", Thread.currentThread().getName());
        other.stream().forEach(obj -> this.put(obj, other.getObjectCount(obj)));
        m.end();
        LOGGER.finest(m.toString());
    }

    public void unionWith(AddressCollection addresses, DetailedHeap next) {
        InstanceObjectInfo instancePrototype = new InstanceObjectInfo();
        InstanceMirrorObjectInfo instanceMirrorPrototype = new InstanceMirrorObjectInfo();
        ArrayObjectInfo arrayPrototype = new ArrayObjectInfo();

        Measurement m = ApplicationStatistics.getInstance().createMeasurement("MultipleObjectCollection: union multi<->address", Thread.currentThread().getName());
        long addrCount = addresses.getObjectCount();
        if (addrCount > 0) {
            // Single threaded
            AddressIterationHelper iterationHelper = new AddressIterationHelper(addresses);
            next.toObjectStream(false)
                .sorted()
                .forEach((address, objectInfo, space, type, size, isArray, arrayLength, allocationSite, pointedFrom, pointsTo, eventType, rootPtrs) -> {
                    if (iterationHelper.checkNextAddress(address)) {
                        put(objectInfo);
                    }
                }, ObjectVisitor.Settings.NO_INFOS);
        }
        m.end();
        LOGGER.finest(m.toString());
    }

    public MapGroupingNode classify(ClassifierChain classifier, ConditionalClassifierList conditionalClassifier, List<Filter> filters) {
        Measurement m = ApplicationStatistics.getInstance().createMeasurement("MultipleObjectCollection: Classify", Thread.currentThread().getName());
        MapGroupingNode group = new MapGroupingNode();
        stream().map(obj -> new AbstractMap.SimpleEntry<>(obj, map.get(obj).get()))
                .filter(objInfoPlusCount -> filters == null ? true : filters.stream().allMatch(filter -> {
                    try {
                        return filter.classify(-1,
                                               null,
                                               null,
                                               objInfoPlusCount.getKey().type,
                                               objInfoPlusCount.getKey().size,
                                               objInfoPlusCount.getKey() instanceof ArrayObjectInfo,
                                               objInfoPlusCount.getKey() instanceof ArrayObjectInfo ? ((ArrayObjectInfo) objInfoPlusCount.getKey()).getLength() : -1,
                                               objInfoPlusCount.getKey().allocationSite,
                                               null,
                                               null,
                                               objInfoPlusCount.getKey().eventType,
                                               null);
                    } catch (Exception e) {
                        e.printStackTrace();
                        return false;
                    }
                }))
                .forEach(objInfoPlusCount -> group.classify(-1,
                                                            null,
                                                            null,
                                                            objInfoPlusCount.getKey().type,
                                                            objInfoPlusCount.getKey().size,
                                                            objInfoPlusCount.getKey() instanceof ArrayObjectInfo,
                                                            objInfoPlusCount.getKey() instanceof ArrayObjectInfo ?
                                                            ((ArrayObjectInfo) objInfoPlusCount.getKey()).getLength() :
                                                            -1,
                                                            objInfoPlusCount.getKey().allocationSite,
                                                            null,
                                                            null,
                                                            objInfoPlusCount.getKey().eventType,
                                                            null,
                                                            classifier,
                                                            conditionalClassifier,
                                                            objInfoPlusCount.getValue()));
        m.end();
        return group;
    }

    public HashMap<ObjectInfo, Counter> getMap() {
        return map;
    }

    public long getBytes() {
        return stream().mapToLong(key -> map.get(key).get() * key.size).sum();
    }

    public void writeToMetaData(DataOutputStream dos) {
        try {
            dos.writeInt(map.size());
            map.forEach((objectInfo, count) -> {
                try {
                    dos.writeInt(objectInfo.hashCode());
                    dos.writeLong(count.get());
                } catch (IOException e) {
                    e.printStackTrace();
                }
            });
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public static MultipleObjectCollection readFromMetaData(DetailedHeap heap, DataInputStream dis) {
        MultipleObjectCollection moc = new MultipleObjectCollection();
        int count = 0;
        try {
            count = dis.readInt();
            for (int i = 0; i < count; i++) {
                int objectInfoHash = dis.readInt();
                long c = dis.readLong();
                moc.map.put(heap.getCache().getByHash(objectInfoHash), new Counter(c));
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
        return null;
    }
    /*
     * @Override public String toString() { StringBuilder sb = new StringBuilder(String.format("%s: %d", name, getObjectCount())); if
     * (!changeSinceLastClassification) { sb.append(" {"); boolean first = true; for (Object key : getKeys()) { long count =
     * getObjectCountClassified(key); if (count != 0) { if (first) { first = false; } else { sb.append(", "); } sb.append(key); sb.append(
     * " -> "); sb.append(count); } } sb.append("}"); } return sb.toString(); }
     */
}
