/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.heap;

import at.jku.mevss.anttracks.heap.iteration.FakeHeapSpliterator;
import at.jku.mevss.anttracks.heap.iteration.HeapSpliterator;
import at.jku.mevss.anttracks.heap.iteration.IteratingObjectStream;
import at.jku.mevss.anttracks.heap.iteration.SpliteratorObjectStream;
import at.jku.mevss.anttracks.heap.labs.Lab;
import at.jku.mevss.anttracks.heap.objects.ObjectInfo;
import at.jku.mevss.anttracks.heap.objects.ObjectInfoCache;
import at.jku.mevss.anttracks.heap.roots.RootPtr;
import at.jku.mevss.anttracks.heap.space.Space;
import at.jku.mevss.anttracks.heap.space.SpaceInfo;
import at.jku.mevss.anttracks.heap.space.SpaceRedefinition;
import at.jku.mevss.anttracks.heap.statistics.Statistics;
import at.jku.mevss.anttracks.heap.symbols.Symbols;
import at.jku.mevss.anttracks.parser.*;
import at.jku.mevss.anttracks.parser.heap.ThreadInfo;
import at.jku.mevss.anttracks.parser.heap.pointer.IncompletePointerInfo;
import at.jku.mevss.anttracks.parser.heap.pointer.PointerHandling;
import at.jku.mevss.anttracks.util.ApplicationStatistics;
import at.jku.mevss.anttracks.util.TraceException;

import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.logging.Logger;
import java.util.stream.Stream;
import java.util.stream.StreamSupport;

public class DetailedHeap implements Heap, FrontBackObjectAccess {

    private final Symbols symbols;

    private final ObjectInfoCache cache;
    private Space[] spaces;
    private final Map<Integer, int[]> collectedSpaces;
    private final Map<Integer, GarbageCollectionCause> causes;

    private List<ParserGCInfo> gcInfos;
    private GarbageCollectionType gc;
    private int gcNo;
    private long lastGCtime;
    private final List<HeapListener> listeners;

    public final ConcurrentHashMap<Long, List<RootPtr>> rootPtrs;
    public final ConcurrentHashMap<Long, ThreadInfo> threads;

    public final CopyOnWriteArrayList<IncompletePointerInfo> notYetMatchedMultiThreadedPtrEventsToHandleAtGCEnd;
    public final ConcurrentHashMap<Long, IncompletePointerInfo> multiThreadedPtrEventsToHandleAtGCEnd;
    public final CopyOnWriteArrayList<SpaceRedefinition> spaceRedefinitions = new CopyOnWriteArrayList<>();

    public final CopyOnWriteArrayList<Long> addressesThatMayHaveInvalidPointersDueToNullingOnNonDirtyObjects = new CopyOnWriteArrayList<>();

    // root pointers must be written before GC start thus we need to signal when the current root pointer DS should be cleared
    // if they would be written after GC start, inspecting the heap at a GC start would not yield up to date root info
    private boolean clearRootPointers = false;

    public Statistics statistics;

    private Logger LOGGER = Logger.getLogger(DetailedHeap.class.getSimpleName());
    private List<Tag> tags;

    public DetailedHeap(Symbols symbols) {
        this(symbols, new ObjectInfoCache(), new Space[0], 0, 0, new ConcurrentHashMap<>(), new ConcurrentHashMap<>());
    }

    public DetailedHeap(Symbols symbols,
                        ObjectInfoCache prototypes,
                        Space[] spaces,
                        int gcs,
                        long lastGCtime,
                        ConcurrentHashMap<Long, List<RootPtr>> rootPtrs,
                        ConcurrentHashMap<Long, ThreadInfo> threads) {
        this.symbols = symbols;
        cache = prototypes;
        this.spaces = spaces.clone();
        collectedSpaces = new HashMap<>(7, 0.8f);
        causes = new HashMap<>(7, 0.8f);
        gc = GarbageCollectionType.MUTATOR;
        gcNo = gcs;
        listeners = new CopyOnWriteArrayList<>();
        this.rootPtrs = rootPtrs;
        this.threads = threads;
        notYetMatchedMultiThreadedPtrEventsToHandleAtGCEnd = new CopyOnWriteArrayList<>();
        multiThreadedPtrEventsToHandleAtGCEnd = new ConcurrentHashMap<>();
        gcInfos = new ArrayList<>();
        this.lastGCtime = lastGCtime;
        tags = new ArrayList<>();
    }

    public Symbols getSymbols() {
        return symbols;
    }

    public void addListener(HeapListener listener) {
        listeners.add(listener);
    }

    public void removeListener(HeapListener listener) {
        boolean removed = listeners.remove(listener);
        if (removed) {
            listener.close(this);
        }
    }

    public void removeAllListeners() {
        for (HeapListener listener : listeners) {
            removeListener(listener);
        }
    }

    public ObjectInfoCache getCache() {
        return cache;
    }

    public void addSpace(int index, Space space) throws TraceException {
        if (index < spaces.length) {
            if (spaces[index] != null) {
                throw new TraceException("There is already a space with id " + index + "!");
            }
        } else {
            Space[] spaces = new Space[index + 1];
            System.arraycopy(this.spaces, 0, spaces, 0, this.spaces.length);
            this.spaces = spaces;
        }
        spaces[index] = space;
        space.setId(index);

        if (gc != GarbageCollectionType.MUTATOR) {
            // We are currently in a garbage collection phase
            space.startTransition(SpaceInfo.TransitionType.Accumulative);
        }
    }

    public void removeSpace(int index) {
        spaces[index] = null;
    }

    public GarbageCollectionType getGC() {
        return gc;
    }

    public int getGCNo() {
        return gcNo;
    }

    public void startGC(ParserGCInfo info, long beforeEventPosition, ParsingInfo parsingInfo) throws TraceException {
        if (info.type.isFull() && collectedSpaces.size() != 0) {
            throw new TraceException("Can't start a full GC if another GC is still active!");
        }
        if (collectedSpaces.put(info.id, new int[0]) != null) {
            throw new TraceException("GC already active");
        }

        this.statistics = Statistics.collect(this, info.type, info.cause, info.concurrent, false, EventType.GC_START, info.time);

        firePhaseChanging(gc,
                          info.type,
                          info.cause,
                          info.concurrent,
                          false,
                          info.time,
                          beforeEventPosition,
                          parsingInfo.getTraceLength(),
                          parsingInfo.isWithinParseTimeWindow(info.time));

        // Set all spaces to ReplaceAll if MAJOR_GC (every object that is not moved dies)
        // Set all spaces to Accumulative if MINOR_GC (keep non-moved objects if not overwritten
        // --> Following GC_INFO events sets collected spaces to ReplaceAll (i.e., spaces that must be evacuated) ...
        // See markSpaceForCollection()
        SpaceInfo.TransitionType type = info.type.isFull() ? SpaceInfo.TransitionType.ReplaceAll : SpaceInfo.TransitionType.Accumulative;
        for (Space space : spaces) {
            if (space != null) {
                if (space.getType() == null) {
                    // Empty region
                    space.startTransition(SpaceInfo.TransitionType.ReplaceAll);
                    space.commitTransition();
                }
                if (!space.isBeingCollected()) {
                    space.startTransition(type);
                }
            }
        }

        firePhaseChanged(gc,
                         info.type,
                         info.cause,
                         info.concurrent,
                         false,
                         info.time,
                         beforeEventPosition,
                         parsingInfo.getTraceLength(),
                         parsingInfo.isWithinParseTimeWindow(info.time));

        clearRootPointers = true;

        gcInfos.add(info);
        lastGCtime = info.time;
        gc = info.type;
        GarbageCollectionCause replaced = causes.put(info.id, info.cause);
        if (replaced != null) {
            throw new TraceException("Cause already registered under this GCID");
        }
    }

    public void markSpaceForCollection(int spaceID, int gcId) throws TraceException {
        errorOnAlreadyCollectedSpace(spaceID);
        int[] spaceIDs = collectedSpaces.get(gcId);
        if (spaceIDs == null) {
            throw new TraceException("Space=\"" + gcId + "\" can not be registered for GCID=\"" + gcId + "\", as the GCID is not registered! (gc_start missing?)");
        }
        int[] newSpaceIDs = new int[spaceIDs.length + 1];
        System.arraycopy(spaceIDs, 0, newSpaceIDs, 0, spaceIDs.length);
        newSpaceIDs[spaceIDs.length] = spaceID;
        collectedSpaces.put(gcId, newSpaceIDs);

        // This event tells us that a garbage collection happens in this space and that all living objects have to be moved
        // Objects that don't have get moved are assumed dead (i.e., Transition type is ReplaceAll)
        Space space = spaces[spaceID];
        if (space.getTransitionType() == SpaceInfo.TransitionType.Accumulative) {
            space.commitTransition();
        }
        space.startTransition(SpaceInfo.TransitionType.ReplaceAll);
    }

    public Space failGC(int spaceID) throws TraceException {
        LOGGER.info("Fail GC - " + spaces[spaceID].toShortString());
        Space space = spaces[spaceID];
        space.rollbackTransition();
        return space;
    }

    public void stopGC(ParserGCInfo info, boolean failed, long afterEventPosition, ParsingInfo parsingInfo) throws TraceException {
        ApplicationStatistics.Measurement mInner = ApplicationStatistics.getInstance().createMeasurement("handlePointers");
        PointerHandling.handlePtrsOnGCEnd(this, failed);
        mInner.end();

        if (gc == GarbageCollectionType.MUTATOR) {
            if (collectedSpaces.size() == 0) {
                throw new TraceException("no GC active");
            } else {
                throw new TraceException("internal state inconsistency");
            }
        }
        final int[] spaceIDs = collectedSpaces.remove(info.id);
        GarbageCollectionCause replaced = causes.remove(info.id);
        if (spaceIDs == null) {
            throw new TraceException(String.format("No GC active for ID=%d!", info.id));
        }

        if (replaced == null) {
            throw new TraceException(String.format("No cause found for GCID=%d!", info.id));
        }

        boolean anyGCActive = !collectedSpaces.isEmpty();
        GarbageCollectionType gcType = gc;
        gc = (anyGCActive) ? GarbageCollectionType.MINOR : GarbageCollectionType.MUTATOR;
        lastGCtime = info.time;

        firePhaseChanging(gcType,
                          gc,
                          info.cause,
                          info.concurrent,
                          failed,
                          info.time,
                          afterEventPosition,
                          parsingInfo.getTraceLength(),
                          parsingInfo.isWithinParseTimeWindow(info.time));
        // spaces handled by this GC
        for (int sID : spaceIDs) {
            Space space = spaces[sID];
            if (space != null) {
                if (space.getTransitionType() != SpaceInfo.TransitionType.None) {
                    space.commitTransition();
                    if (anyGCActive) {
                        space.startTransition(SpaceInfo.TransitionType.Accumulative);
                        // may be used in other GC
                    }
                }
                if (space.getType() == null) {
                    space.clear();
                }
            }
        }
        // spaces handled by other GCs (and current GC)
        if (!anyGCActive) {
            for (Space s : spaces) {
                if (s != null) {
                    if (s.getTransitionType() != SpaceInfo.TransitionType.None) {
                        s.commitTransition();
                    }
                    if (s.getType() == null) {
                        s.clear();
                    }
                }
            }
        } else {
            for (Space s : spaces) {
                if (s != null && s.isBeingCollected()) {
                    throw new TraceException("No concurrent GCs!");
                }
            }
        }

        this.statistics = Statistics.collect(this, info.type, info.cause, info.concurrent, failed, EventType.GC_END, info.time);

        // Pointers have to be validated before phase change.
        // Otherwise listerns could be notified with a heap that is in an inconsistant state
        if (TraceParser.CONSISTENCY_CHECK) {
            mInner = ApplicationStatistics.getInstance().createMeasurement("heapValidation");
            validate(symbols.isHeapFragmented);
            mInner.end();
        }

        mInner = ApplicationStatistics.getInstance().createMeasurement("pointerValidation");
        PointerHandling.validateAllPointers(this);
        mInner.end();

        firePhaseChanged(gcType,
                         gc,
                         info.cause,
                         info.concurrent,
                         failed,
                         info.time,
                         afterEventPosition,
                         parsingInfo.getTraceLength(),
                         parsingInfo.isWithinParseTimeWindow(info.time));
        gcNo++;

        // LOGGER.info(String.format("ObjectInfo cache size: %,d", cache.size()));
        if (gcType.isFull()) {
            shrinkObjectInfoCache();
        }

        if (TraceParser.CONSISTENCY_CHECK) {
            for (Map.Entry<Long, List<RootPtr>> rp : rootPtrs.entrySet()) {
                if (rp.getKey() >= 0) {
                    try {
                        // Throws an exception if object does not exist
                        this.getObject(rp.getKey());
                    } catch (Exception ex) {
                        /*
                        throw new TraceException(String.format("Root pointed address %,d without object @ GC End:\n%s",
                                                               rp.getKey(),
                                                               rp.getValue().stream().map(x -> x.toString()).collect(Collectors.joining())));
                    */
                    }
                }
            }
        }
    }

    public void shrinkObjectInfoCache() {
        cache.clear();
        toObjectStream().forEach(
                (address, objectInfo, space, type, size, isArray, arrayLength, allocationSite, pointedFrom, pointsTo, eventType, rootPtrs) -> cache.get(objectInfo),
                ObjectVisitor.Settings.NO_INFOS);
    }

    public void insertFillers(ThreadLocalHeap tlh) throws TraceException {
        for (Space space : tlh.retiredLabs.keySet()) {
            space.insertFillers(tlh.retiredLabs.get(space));
        }
    }

    public void assignLabsIncorrectlyAssumedToBeFillers(boolean allowUnusedFillers) throws TraceException {
        for (Space space : spaces) {
            if (space != null) {
                space.assignLabsIncorrectlyTreatedAsFillers(allowUnusedFillers);
            }
        }
    }

    @SuppressWarnings("unused")
    private void errorOnMajorGC(GarbageCollectionType type) throws TraceException {
        if (type == GarbageCollectionType.MAJOR || type == GarbageCollectionType.MAJOR_SYNC) {
            throw new TraceException("#" + type + ", GC MOVE Event to SURVIVOR_TO is only possible during a Minor GC. SURVIVOR_TO should be empty here. ");
        }

    }

    public boolean validate(boolean allowFragmentation) throws TraceException {
        for (Space space : spaces) {
            if (space != null) {
                space.validate(allowFragmentation);
            }
        }
        return true;
    }

    private void firePhaseChanging(GarbageCollectionType from,
                                   GarbageCollectionType to,
                                   GarbageCollectionCause cause,
                                   boolean concurrent,
                                   boolean failed,
                                   long time,
                                   long position,
                                   long length,
                                   boolean inSelectedTimeWindow) {
        listeners.forEach(l -> l.phaseChanging(this, from, to, cause, concurrent, failed, time, position, length, inSelectedTimeWindow));
    }

    private void firePhaseChanged(GarbageCollectionType from,
                                  GarbageCollectionType to,
                                  GarbageCollectionCause cause,
                                  boolean concurrent,
                                  boolean failed,
                                  long time,
                                  long position,
                                  long length,
                                  boolean inSelectedTimeWindow) {
        listeners.forEach(l -> l.phaseChanged(this, from, to, cause, concurrent, failed, time, position, length, inSelectedTimeWindow));
    }

    @Override
    public String toString() {
        StringBuilder result = new StringBuilder("Heap (" + Arrays.stream(spaces).filter(Objects::nonNull).count() + " spaces)\n");
        Arrays.stream(spaces).filter(Objects::nonNull).sorted(Comparator.comparingLong(Space::getAddress)).forEach(s -> result.append(s.toShortString() + "\n"));
        // Delete last new line
        result.deleteCharAt(result.length() - 1);
        return result.toString();
    }

    @Override
    public DetailedHeap clone() {
        return new DetailedHeap(symbols,
                                cache,
                                Stream.of(spaces).filter(Objects::nonNull).map(Space::clone).toArray(Space[]::new),
                                gcNo,
                                lastGCtime,
                                new ConcurrentHashMap<>(rootPtrs),
                                new ConcurrentHashMap<>(threads));
    }

    private void errorOnAlreadyCollectedSpace(int idx) throws TraceException {
        for (int[] arr : collectedSpaces.values()) {
            if (contains(arr, idx)) {
                throw new TraceException("Space#" + idx + "(" + spaces[idx].getType() + ") is already beeing collected!");
            }
        }
    }

    private static boolean contains(int[] array, int value) {
        for (int i : array) {
            if (i == value) {
                return true;
            }
        }
        return false;
    }

    /*
     * Returns a sequential stream
     */
    public Stream<HeapObject> stream() {
        return StreamSupport.stream(new HeapSpliterator(this, 0, 0, lastValidSpace(true) + 1, spaces[lastValidSpace(true)].getLabCount(), true, false), false);
    }

    /*
     * Returns a parallel stream
     */
    public Stream<HeapObject> parallelStream() {
        return StreamSupport.stream(new HeapSpliterator(this, 0, 0, lastValidSpace(true) + 1, spaces[lastValidSpace(true)].getLabCount(), true, false), true);
    }

    public GarbageCollectionCause getGarbageCollectionCause(int gcID) {
        return causes.get(gcID);
    }

    public long getLabCount() {
        long count = 0;
        for (Space s : spaces) {
            if (s != null) {
                count += s.getLabs().length;
            }
        }
        return count;
    }

    public long getObjectCount() {
        long count = 0;
        for (Space s : spaces) {
            if (s != null && s.getType() != null) {
                for (Lab l : s.getLabs()) {
                    if (l != null) {
                        count += l.getObjectCount();
                    }
                }
            }
        }
        return count;
    }

    public long getMinimumAddress() {
        long min = Long.MAX_VALUE;
        for (Space s : spaces) {
            if (s != null && s.getType() != null) {
                long addr = s.getAddress();
                if (addr < min) {
                    min = addr;
                }
            }
        }
        return min;
    }

    private int lastValidSpace(boolean current) {
        int lastValidSpace = spaces.length - 1;
        while (lastValidSpace > 0 && (spaces[lastValidSpace] == null || spaces[lastValidSpace].getLabCount(current) == 0)) {
            // Search for last valid space (not null and contains at least one lab)
            lastValidSpace--;
        }
        return lastValidSpace;
    }

    public ObjectStream toSpliteratorObjectStream(long multithreadingThreshold) {
        return toSpliteratorObjectStream(true, multithreadingThreshold);
    }

    public ObjectStream toSpliteratorObjectStream(boolean current, long multithreadingThreshold) {
        return new SpliteratorObjectStream(new FakeHeapSpliterator(this,
                                                                   0,
                                                                   0,
                                                                   lastValidSpace(current) + 1,
                                                                   spaces[lastValidSpace(current)].getLabCount(current),
                                                                   current,
                                                                   false),
                                           multithreadingThreshold);
    }

    public ObjectStream toObjectStream() {
        return toObjectStream(true);
    }

    public ObjectStream toObjectStream(boolean current) {
        return new IteratingObjectStream(this, current);
    }

    public void complete() throws TraceException {
        ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance().createMeasurement("completing heap");

        ApplicationStatistics.Measurement m1 = ApplicationStatistics.getInstance().createMeasurement("completing heap: setting from pointers");
        setFromPointers();
        m1.end();

        ApplicationStatistics.Measurement m2 = ApplicationStatistics.getInstance().createMeasurement("completing heap: reducing size");
        reduceSize();
        m2.end();

        ApplicationStatistics.Measurement m3 = ApplicationStatistics.getInstance().createMeasurement("completing heap: init markings");
        for (Space s : getSpacesCloned()) {
            if (s == null) {
                continue;
            }
            for (Lab l : s.getLabs()) {
                l.initMarkings();
            }
        }
        m3.end();

        ApplicationStatistics.Measurement m4 = ApplicationStatistics.getInstance().createMeasurement("completing heap: shrink roots");
        // shrink root pointer ds by removing unnecessary elements
        Iterator<Map.Entry<Long, List<RootPtr>>> entryIterator = rootPtrs.entrySet().iterator();
        while (entryIterator.hasNext()) {
            Map.Entry<Long, List<RootPtr>> entry = entryIterator.next();
            Iterator<RootPtr> rootIterator = entry.getValue().iterator();
            while (rootIterator.hasNext()) {
                RootPtr rootPtr = rootIterator.next();

                // remove dead strings
                if (rootPtr.getRootType() == RootPtr.RootType.INTERNED_STRING) {
                    // must exist
                    ObjectInfo oi = null;
                    try {
                        oi = this.getObject(rootPtr.getAddr());

                        if (oi == null || !oi.type.name.equals("Ljava/lang/String;")) {
                            rootIterator.remove();
                            ApplicationStatistics.getInstance().inc("Removing dead string roots");
                        }
                    } catch (TraceException e) {
                        rootIterator.remove();
                        ApplicationStatistics.getInstance().inc("Removing dead string roots");
                    }
                }

                // remove thread internal roots duplicated by thread local roots
                if (rootPtr.getRootType() == RootPtr.RootType.VM_INTERNAL_THREAD_DATA_ROOT && entry.getValue()
                                                                                                   .stream()
                                                                                                   .anyMatch(r -> r.getRootType() == RootPtr.RootType
                                                                                                           .LOCAL_VARIABLE_ROOT)) {
                    rootIterator.remove();
                }
            }

            // remove
            if (entry.getValue().isEmpty()) {
                entryIterator.remove();
            }
        }
        m4.end();

        ApplicationStatistics.Measurement m5 = ApplicationStatistics.getInstance().createMeasurement("completing heap: resolve roots");
        // resolve roots
        rootPtrs.values().stream().flatMap(Collection::stream).forEach(rootPtr -> {
            try {
                rootPtr.resolve(this);
            } catch (Exception e) {
                e.printStackTrace();
            }
        });
        m5.end();

        // resolve thread callstacks
        threads.values().forEach(t -> {
            try {
                t.resolveCallStack(this);
            } catch (Exception e) {
                e.printStackTrace();
            }
        });

        m.end();
    }

    private void reduceSize() {
        for (Space space : spaces) {
            if (space != null) {
                space.reduceSize();
            }
        }
    }

    private void setFromPointers() throws TraceException {
        if (symbols.expectPointers) {
            if (!multiThreadedPtrEventsToHandleAtGCEnd.isEmpty()) {
                throw new TraceException("All multithreaded pointer events must have been handled before from pointers get calculated!");
            }

            /*
            LOGGER.info("Set from-pointers");
            HashMap<Long, Counter> fromPointerCounters = new HashMap<>();

            LOGGER.info("Counting from-pointers");
            ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance().createMeasurement("Counting from-pointers");
            ObjectVisitor fromPointerCounterClosure =
                    (address, objectInfo, space, type, size, isArray, arrayLength, allocationSite, pointedFrom, pointsTo, eventType, rootPtrs) -> {
                        for (int i = 0; pointsTo != null && i < pointsTo.length; i++) {
                            if (pointsTo[i] < 0) {
                                // Null pointer
                                continue;
                            }
                            Counter ptrCounter = fromPointerCounters.get(pointsTo[i]);
                            if (ptrCounter == null) {
                                fromPointerCounters.put(pointsTo[i], new Counter());
                                ptrCounter = fromPointerCounters.get(pointsTo[i]);
                            }
                            ptrCounter.inc();
                        }
                    };
            toObjectStream().forEach(fromPointerCounterClosure);
            m.end();

            LOGGER.info("Creating from-pointer arrays");
            m = ApplicationStatistics.getInstance().createMeasurement("Creating from-pointer arrays");
            HashMap<Long, long[]> fromPointers = new HashMap<>();
            Iterator<Map.Entry<Long, Counter>> fromPointerCounterIterator = fromPointerCounters.entrySet().iterator();
            while (fromPointerCounterIterator.hasNext()) {
                Map.Entry<Long, Counter> entry = fromPointerCounterIterator.next();
                if (entry.getValue().get() > 0) {
                    long[] from = new long[(int) entry.getValue().get()];
                    fromPointers.put(entry.getKey(), from);
                }
                fromPointerCounterIterator.remove();
            }
            m.end();

            LOGGER.info("Creating from-pointer current index counter");
            m = ApplicationStatistics.getInstance().createMeasurement("Creating from-pointer current index counter");
            HashMap<Long, Counter> fromPointerCurCounters = new HashMap<>();
            fromPointers.keySet().forEach(addr -> fromPointerCurCounters.put(addr, new Counter()));
            m.end();

            LOGGER.info("Set from-pointer addresses in arrays");
            m = ApplicationStatistics.getInstance().createMeasurement("Set from-pointer addresses in arrays");
            ObjectVisitor fromPointerCombiner =
                    (address, objectInfo, space, type, size, isArray, arrayLength, allocationSite, pointedFrom, pointsTo, eventType, rootPtrs) -> {
                        for (int i = 0; pointsTo != null && i < pointsTo.length; i++) {
                            if (pointsTo[i] < 0) {
                                // Null pointer
                                continue;
                            }
                            long[] ptrs = fromPointers.get(pointsTo[i]);
                            Counter curIndex = fromPointerCurCounters.get(pointsTo[i]);
                            ptrs[(int) curIndex.get()] = address;
                            curIndex.inc();
                            if (curIndex.get() == ptrs.length) {
                                fromPointerCurCounters.remove(pointsTo[i]);
                            }
                        }
                    };
            toObjectStream().forEach(fromPointerCombiner);
            m.end();

            LOGGER.info("Set from-pointer array in object");
            m = ApplicationStatistics.getInstance().createMeasurement("Set from-pointer array in object");
            Iterator<Map.Entry<Long, long[]>> pointerIterator = fromPointers.entrySet().iterator();
            while (pointerIterator.hasNext()) {
                Map.Entry<Long, long[]> entry = pointerIterator.next();
                getLabInFront(entry.getKey()).setFromPointers(entry.getKey(), entry.getValue());
                pointerIterator.remove();
            }
            m.end();
            LOGGER.info("Calculating from-pointers finished");
            */
        }
    }

    public void interruptGC(int id, long address) throws TraceException {
        Space space = getSpace(address);
        assert space.isBeingCollected();

        boolean isValidInterrupt = false;
        int[] spaceIDs = collectedSpaces.get(id);

        for (int s : spaceIDs) {
            if (spaces[s].equals(space)) {
                isValidInterrupt = true;
            }
        }
        if (!isValidInterrupt) {
            throw new TraceException("collection of space is interrupted by different GC!");
        }

        space.commitTransition(address);
        space.startTransition(SpaceInfo.TransitionType.Accumulative);
        // check if other Space is beeing collected
        // changes on the code here should also be made to
        // Heap::continueGC(int,long)!
        long spCnt = Arrays.stream(collectedSpaces.get(id)).filter(s -> spaces[s].isBeingCollected()).count();

        if (0 != spCnt) {
            for (Space s : spaces) {
                if (s != null && s.isBeingCollected()) {
                    throw new TraceException(s.toShortString() + " is beeing collected. Forgot an interrupt?");
                }
            }
        }
    }

    public void continueGC(int id, long address) throws TraceException {
        Space space = getSpace(address);
        assert !space.isBeingCollected();

        boolean isValidContinue = false;
        int[] spaceIDs = collectedSpaces.get(id);

        for (int s : spaceIDs) {
            if (spaces[s].equals(space)) {
                isValidContinue = true;
            }
        }

        if (!isValidContinue) {
            throw new TraceException("collection of space is continued by different GC!");
        }

        // check if other Space is beeing collected
        // changes on the code here should also be made to
        // Heap::interruptGC(int,long)!
        long spCnt = Arrays.stream(collectedSpaces.get(id)).filter(s -> spaces[s].isBeingCollected()).count();

        if (0 != spCnt) {
            for (Space s : spaces) {
                if (s != null && s.isBeingCollected()) {
                    throw new TraceException(s.toShortString() + " is being collected. Forgot an interrupt?");
                }
            }
        }
        space.startTransition(SpaceInfo.TransitionType.ReplaceAll, address);
    }

    public String getLabSizes() {

        StringBuilder sb = new StringBuilder(";"); // to later insert avg at the beginning (easier viewing)
        long objCount = 0;
        int labCount = 0;

        for (Space s : getSpacesCloned()) {
            if (s == null) {
                continue;
            }
            labCount += s.getLabCount();
            for (Lab l : s.getLabs()) {
                sb.append(l.getObjectCount());
                sb.append(";");
                objCount += l.getObjectCount();
            }
        }

        // write avg obj count per lab
        sb.insert(0, (labCount > 0 ? objCount / labCount : 0));
        sb.append("\n");

        return sb.toString();
    }

    public void addRoot(RootPtr newRoot) throws TraceException {
        // TODO: Learn how interned strings exactly work and if they have to be considered as strong roots
        if (newRoot.getRootType() == RootPtr.RootType.INTERNED_STRING) {
            return;
        }
        // IMPORTANT: root pointers must not be parsed simultaneously because heap root ds is NOT threadsafe
        if (clearRootPointers) {
            rootPtrs.clear();
            threads.values().forEach(ThreadInfo::clearCallstack);
            clearRootPointers = false;
        }

        if (TraceParser.CONSISTENCY_CHECK) {
            try {
                try {
                    newRoot.resolve(this);
                } catch (Exception ex) {
                    ex.printStackTrace();
                }
                if (newRoot.getAddr() != -1) {
                    if (this.getObject(newRoot.getAddr()) == null) {
                        throw new TraceException(String.format("Root pointed address %,d without object @ GC Start:\n%s", newRoot.getAddr(), newRoot));
                    }
                }
            } catch (TraceException e) {
                throw new TraceException(String.format("Root pointed address %,d without object @ GC Start:\n%s", newRoot.getAddr(), newRoot));
            }
        }

        List<RootPtr> existingRoots = rootPtrs.get(newRoot.getAddr());

        if (existingRoots != null) {
            if (!existingRoots.contains(newRoot)) {
                existingRoots.add(newRoot);
                existingRoots.sort(Comparator.comparing(RootPtr::getRootType));
            }
        } else {
            existingRoots = new ArrayList<>();
            existingRoots.add(newRoot);
            rootPtrs.put(newRoot.getAddr(), existingRoots);
        }
    }

    // -------------------------------------
    // ------ Space access -----------------
    // -------------------------------------

    public Space[] getSpacesUncloned() {
        return spaces;
    }

    public Space[] getSpacesCloned() {
        return spaces.clone();
    }

    public Space getSpace(int id) {
        return 0 <= id && id < spaces.length ? spaces[id] : null;
    }

    /**
     * Used to find the corresponding space to a certain address.
     * Uses binary search instead of linear search for speed-up
     *
     * @param addr The address for which the corresponding space should be found
     * @return The found space, or null if not found
     */
    public Space getSpace(long addr) {
        Space result = getSpaceBinarySearch(addr);
        assert result != null || getSpaceLinearSearch(addr) == null : "binary search unable to find space, linear search is. what is " + "going on?";
        return result;
    }

    /**
     * Used to find the corresponding space to a certain address.
     * Uses binary search instead of linear search for speed-up
     *
     * @param addr The address for which the corresponding space should be found
     * @return The found space, or null if not found
     */
    private Space getSpaceBinarySearch(long addr) {
        int lo = 0, hi = spaces.length - 1;
        while (lo <= hi) {
            int mi = (lo + hi) / 2;
            Space space = getSpace(mi);

            while (space == null && mi < hi) {
                space = getSpace(++mi);
            }
            while (space == null && mi > lo) {
                space = getSpace(--mi);
            }
            if (space == null) {
                break;
            }

            if (space.contains(addr)) {
                return space;
            } else if (addr < space.getAddress()) {
                hi = mi - 1;
            } else if (space.getAddress() + space.getLength() <= addr) {
                lo = mi + 1;
            }
        }
        return null;
    }

    /**
     * DO NOT USE ANYMORE!
     * It is only used in assertion mode if the binary search does not find a certain space to check if also the linear search does not find
     * it, i.e., to verfy that the binary search is working correct.
     *
     * @param addr The address for which the corresponding space should be found
     * @return The found space, or null if not found
     */
    @Deprecated
    private Space getSpaceLinearSearch(long addr) {
        for (Space space : spaces) {
            if (space != null && space.contains(addr)) {
                return space;
            }
        }
        return null;
    }

    // -------------------------------------
    // ------ Lab access -------------------
    // -------------------------------------

    public Lab getLab(long objAddr, boolean inFront) throws TraceException {
        return inFront ? getLabInFront(objAddr) : getLabInBack(objAddr);
    }

    public Lab getCurrentLab(long objAddr) throws TraceException {
        Space space = getSpace(objAddr);
        if (space == null) {
            throw new TraceException(String.format("No space found containing address %d", objAddr));
        }
        Lab lab = space.getCurrentLab(objAddr);
        if (lab == null) {
            throw new TraceException(String.format("No lab found containing address %,d in space %s", objAddr, space));
        }
        return lab;
    }

    public Lab getLabInFront(long objAddr) throws TraceException {
        Space space = getSpace(objAddr);
        if (space == null) {
            throw new TraceException(String.format("No space found containing address %d", objAddr));
        }
        Lab lab = space.getLab(objAddr, true);
        if (lab == null) {
            throw new TraceException(String.format("No lab found containing address %,d in space %s", objAddr, space));
        }
        return lab;
    }

    public Lab getLabInBack(long objAddr) throws TraceException {
        Space space = getSpace(objAddr);
        if (space == null) {
            throw new TraceException(String.format("No space found containing address %d", objAddr));
        }
        Lab lab = space.getLab(objAddr, false);
        if (lab == null) {
            throw new TraceException(String.format("No lab found containing address %,d in space %s", objAddr, space));
        }
        return lab;
    }

    // ----------------------------------------------
    // ------ Object access -------------------------
    // ----------------------------------------------

    @Override
    public void setObjectPointers(long objAddr, long[] ptrs) throws TraceException {
        getCurrentLab(objAddr).setObjectPointers(objAddr, ptrs);
    }

    @Override
    public void setObjectForwardingAddr(long objAddr, long forwAddr) throws TraceException {
        getCurrentLab(objAddr).setObjectForwardingAddr(objAddr, forwAddr);
    }

    @Override
    public void setObjectForwardingAddrAndPtrs(long objAddr, long forwAddr, long[] ptrs) throws TraceException {
        getCurrentLab(objAddr).setObjectForwardingAddrAndPtrs(objAddr, forwAddr, ptrs);
    }

    @Override
    public void setFromPointerCount(long objAddr, int n) throws TraceException {
        getCurrentLab(objAddr).setFromPointerCount(objAddr, n);
    }

    @Override
    public ObjectInfo getObject(long objAddr) throws TraceException {
        return getCurrentLab(objAddr).getObject(objAddr);
    }

    @Override
    public long[] getObjectPointers(long objAddr) throws TraceException {
        return getCurrentLab(objAddr).getObjectPointers(objAddr);
    }

    @Override
    public long[] getObjectFromPointers(long objAddr) throws TraceException {
        return getCurrentLab(objAddr).getObjectFromPointers(objAddr);
    }

    @Override
    public long getObjectForwardingAddressUnescaped(long objAddr) throws TraceException {
        return getCurrentLab(objAddr).getObjectForwardingAddressUnescaped(objAddr);
    }

    @Override
    public long getObjectForwardingAddressEscaped(long objAddr) throws TraceException {
        return getCurrentLab(objAddr).getObjectForwardingAddressEscaped(objAddr);
    }

    @Override
    public void mark(long objAddr) throws TraceException {
        getCurrentLab(objAddr).mark(objAddr);
    }

    @Override
    public boolean isMarked(long objAddr) throws TraceException {
        return getCurrentLab(objAddr).isMarked(objAddr);
    }

    @Override
    public void setFromPointers(long objAddr, long[] fromPtrs) throws TraceException {
        getCurrentLab(objAddr).setFromPointers(objAddr, fromPtrs);
    }

    @Override
    public boolean ptrNeedAdjustment(long objAddr) throws TraceException {
        return getCurrentLab(objAddr).ptrNeedAdjustment(objAddr);
    }

    @Override
    public void setPtrNeedAdjustment(long objAddr, boolean needAdjustment) throws TraceException {
        getCurrentLab(objAddr).setPtrNeedAdjustment(objAddr, needAdjustment);
    }

    @Override
    public void setObjectPointersInFront(long objAddr, long[] ptrs) throws TraceException {
        Lab lab = getLabInFront(objAddr);
        lab.setObjectPointers(objAddr, ptrs);
    }

    @Override
    public void setObjectForwardingAddrInFront(long objAddr, long forwAddr) throws TraceException {
        Lab lab = getLabInFront(objAddr);
        lab.setObjectForwardingAddr(objAddr, forwAddr);
    }

    @Override
    public void setObjectForwardingAddrAndPtrsInFront(long objAddr, long forwAddr, long[] ptrs) throws TraceException {
        Lab lab = getLabInFront(objAddr);
        lab.setObjectForwardingAddrAndPtrs(objAddr, forwAddr, ptrs);
    }

    @Override
    public void setFromPointerCountInFront(long objAddr, int n) throws TraceException {
        Lab lab = getLabInFront(objAddr);
        lab.setFromPointerCount(objAddr, n);
    }

    @Override
    public ObjectInfo getObjectInFront(long objAddr) throws TraceException {
        Lab lab = getLabInFront(objAddr);
        return lab.getObject(objAddr);
    }

    @Override
    public long[] getObjectPointersInFront(long objAddr) throws TraceException {
        Lab lab = getLabInFront(objAddr);
        return lab.getObjectPointers(objAddr);
    }

    @Override
    public long[] getObjectFromPointersInFront(long objAddr) throws TraceException {
        Lab lab = getLabInFront(objAddr);
        return lab.getObjectFromPointers(objAddr);
    }

    @Override
    public long getObjectForwardingAddressUnescapedInFront(long objAddr) throws TraceException {
        Lab lab = getLabInFront(objAddr);
        return lab.getObjectForwardingAddressUnescaped(objAddr);
    }

    @Override
    public long getObjectForwardingAddressEscapedInFront(long objAddr) throws TraceException {
        Lab lab = getLabInFront(objAddr);
        return lab.getObjectForwardingAddressEscaped(objAddr);
    }

    @Override
    public void markInFront(long objAddr) throws TraceException {
        Lab lab = getLabInFront(objAddr);
        lab.mark(objAddr);
    }

    @Override
    public boolean isMarkedInFront(long objAddr) throws TraceException {
        Lab lab = getLabInFront(objAddr);
        return lab.isMarked(objAddr);
    }

    @Override
    public void setFromPointersInFront(long objAddr, long[] fromPtrs) throws TraceException {
        Lab lab = getLabInFront(objAddr);
        lab.setFromPointers(objAddr, fromPtrs);
    }

    @Override
    public boolean ptrNeedAdjustmentInFront(long objAddr) throws TraceException {
        Lab lab = getLabInFront(objAddr);
        return lab.ptrNeedAdjustment(objAddr);
    }

    @Override
    public void setPtrNeedAdjustmentInFront(long objAddr, boolean needAdjustment) throws TraceException {
        Lab lab = getLabInFront(objAddr);
        lab.setPtrNeedAdjustment(objAddr, needAdjustment);
    }

    @Override
    public void setObjectPointersInBack(long objAddr, long[] ptrs) throws TraceException {
        Lab lab = getLabInBack(objAddr);
        lab.setObjectPointers(objAddr, ptrs);
    }

    @Override
    public void setObjectForwardingAddrInBack(long objAddr, long forwAddr) throws TraceException {
        Lab lab = getLabInBack(objAddr);
        lab.setObjectForwardingAddr(objAddr, forwAddr);
    }

    @Override
    public void setObjectForwardingAddrAndPtrsInBack(long objAddr, long forwAddr, long[] ptrs) throws TraceException {
        Lab lab = getLabInBack(objAddr);
        lab.setObjectForwardingAddrAndPtrs(objAddr, forwAddr, ptrs);
    }

    @Override
    public void setFromPointerCountInBack(long objAddr, int n) throws TraceException {
        Lab lab = getLabInBack(objAddr);
        lab.setFromPointerCount(objAddr, n);
    }

    @Override
    public ObjectInfo getObjectInBack(long objAddr) throws TraceException {
        Lab lab = getLabInBack(objAddr);
        return lab.getObject(objAddr);
    }

    @Override
    public long[] getObjectPointersInBack(long objAddr) throws TraceException {
        Lab lab = getLabInBack(objAddr);
        return lab.getObjectPointers(objAddr);
    }

    @Override
    public long[] getObjectFromPointersInBack(long objAddr) throws TraceException {
        Lab lab = getLabInBack(objAddr);
        return lab.getObjectFromPointers(objAddr);
    }

    @Override
    public long getObjectForwardingAddressUnescapedInBack(long objAddr) throws TraceException {
        Lab lab = getLabInBack(objAddr);
        return lab.getObjectForwardingAddressUnescaped(objAddr);
    }

    @Override
    public long getObjectForwardingAddressEscapedInBack(long objAddr) throws TraceException {
        Lab lab = getLabInBack(objAddr);
        return lab.getObjectForwardingAddressEscaped(objAddr);
    }

    @Override
    public void markInBack(long objAddr) throws TraceException {
        Lab lab = getLabInBack(objAddr);
        lab.mark(objAddr);
    }

    @Override
    public boolean isMarkedInBack(long objAddr) throws TraceException {
        Lab lab = getLabInBack(objAddr);
        return lab.isMarked(objAddr);
    }

    @Override
    public void setFromPointersInBack(long objAddr, long[] fromPtrs) throws TraceException {
        Lab lab = getLabInBack(objAddr);
        lab.setFromPointers(objAddr, fromPtrs);
    }

    @Override
    public boolean ptrNeedAdjustmentInBack(long objAddr) throws TraceException {
        Lab lab = getLabInBack(objAddr);
        return lab.ptrNeedAdjustment(objAddr);
    }

    @Override
    public void setPtrNeedAdjustmentInBack(long objAddr, boolean needAdjustment) throws TraceException {
        Lab lab = getLabInBack(objAddr);
        lab.setPtrNeedAdjustment(objAddr, needAdjustment);
    }

    public IndexBasedHeap toIndexBasedHeap() {
        ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance().createMeasurement("build fast heap");
        LOGGER.info("create fast heap");
        // build fast heap
        IndexBasedHeap fastHeap = IndexBasedHeapFactory.INSTANCE.create(this);
        m.end();
        LOGGER.info("fast heap created");
        return fastHeap;
    }

    public void addTag(String tagText) {
        this.tags.add(new Tag(gcInfos.get(gcInfos.size() - 1), new GarbageCollectionLookup(null, gc, getGarbageCollectionCause(gcNo), gcNo), tagText));
    }

    public long getLastGCTime() {
        return lastGCtime;
    }

    public List<Tag> getTags() {
        return tags;
    }
}