package at.jku.mevss.anttracks.heap;

import at.jku.mevss.anttracks.classification.ClassifierChain;
import at.jku.mevss.anttracks.classification.ConditionalClassifierList;
import at.jku.mevss.anttracks.classification.Filter;
import at.jku.mevss.anttracks.classification.nodes.ListGroupingNode;
import at.jku.mevss.anttracks.classification.trees.ListClassificationTree;
import at.jku.mevss.anttracks.heap.datastructures.DataStructure;
import at.jku.mevss.anttracks.heap.objects.ArrayObjectInfo;
import at.jku.mevss.anttracks.heap.objects.ObjectInfo;
import at.jku.mevss.anttracks.heap.objects.ObjectInfoCache;
import at.jku.mevss.anttracks.heap.roots.RootPtr;
import at.jku.mevss.anttracks.heap.space.Space;
import at.jku.mevss.anttracks.heap.space.SpaceInfo;
import at.jku.mevss.anttracks.heap.symbols.AllocatedType;
import at.jku.mevss.anttracks.heap.symbols.AllocationSite;
import at.jku.mevss.anttracks.heap.symbols.Symbols;
import at.jku.mevss.anttracks.parser.EventType;
import at.jku.mevss.anttracks.parser.heap.ThreadInfo;
import at.jku.mevss.anttracks.util.*;
import javafx.beans.property.BooleanProperty;
import org.jgrapht.DirectedGraph;
import org.jgrapht.graph.DefaultDirectedGraph;

import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.BiConsumer;
import java.util.function.Supplier;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import static at.jku.mevss.anttracks.parser.TraceSlaveParser.NULL_PTR;

public abstract class IndexBasedHeap implements Heap {
    /**
     * This flag is used to turn off unnecessary operations that may distort measurements regarding heap state creating time.
     * This is currently used by BF's bachelor thesis and probably may be removed afterwards
     */
    public static boolean REDUCE_TIMING_MEASUREMENT_DISTORTION = false;
    public static boolean CALCULATE_DATA_STRUCTURES = true;

    public static final boolean USE_ANT_TRACKS_BIT_SET = false;
    public static final Supplier<BitSet> bitSetSupplier = () -> USE_ANT_TRACKS_BIT_SET ? new AntTracksBitSet() : new BitSet();

    public final ObjectInfoCache objectInfoCache;

    public static final int NULL_INDEX = NULL_PTR;
    public static final int SUPER_ROOT_INDEX = -2;

    protected final int objectCount;
    protected int byteCount;

    protected final SpaceInfo[] spaceInfos;
    protected final long[] spaceStartAdresses;

    protected final List<RootPtr> rootPtrList;
    protected final HashMap<Integer, List<RootPtr>> rootPtrs;
    protected boolean rootPtrsReachablesBeingCalculated = false;
    protected final BitSet reachableFromRoot;

    protected final ThreadInfo[] threads;
    protected final Symbols symbols;

    protected Dominators<Integer, Tuple<Integer, Integer>> dominators;

    protected DataStructure[] dataStructures;
    protected int[][] dataStructureComposition;
    protected final int gcNo;

    public IndexBasedHeap(DetailedHeap heap) {
        objectInfoCache = heap.getCache().clone();
        gcNo = heap.getGCNo();

        this.objectCount = Math.toIntExact(heap.getObjectCount()); // safe conversion
        long[] addressesArray = new long[objectCount];
        ObjectInfo[] objectInfos = new ObjectInfo[objectCount];
        this.reachableFromRoot = bitSetSupplier.get();
        this.rootPtrs = new HashMap<>();
        rootPtrList = new ArrayList<>();
        Space[] spaces = Arrays.stream(heap.getSpacesUncloned())
                               .filter(space -> space != null && space.getType() != null && space.getMode() != null)
                               .sorted(Comparator.comparingLong(Space::getAddress))
                               .toArray(Space[]::new);
        this.spaceInfos = new SpaceInfo[spaces.length];
        this.spaceStartAdresses = new long[spaces.length];

        this.threads = heap.threads.values()
                                   .stream()
                                   .toArray(ThreadInfo[]::new);
        this.symbols = heap.getSymbols();

        long[][] addressToPointers = new long[objectCount][];

        AtomicInteger objIndex = new AtomicInteger(0);  // only for sequential stream!
        for (int i = 0; i < spaces.length; i++) {
            spaceInfos[i] = spaces[i].getInfo();
            spaceStartAdresses[i] = spaceInfos[i].getAddress();

            // copy heap objects
            spaces[i].iterate(heap, (address, objectInfo, space, type, size, isArray, arrayLength, allocationSite, pointedFrom, pointsTo, eventType, rootPtrs) -> {
                int i1 = objIndex.get();
                addressesArray[i1] = address;
                addressToPointers[i1] = pointsTo;
                objectInfos[i1] = objectInfo;
                objIndex.incrementAndGet();

                byteCount += size;
            }, new ObjectVisitor.Settings(false, true, false));
        }
        assert objIndex.get() == objectCount;

        // translate long pointers to indices
        int[][] toPointersArray = new int[objectCount][];

        Counter[] fromPtrCounter = new Counter[objectCount];
        int[][] fromPointersArray = new int[objectCount][];

        for (int i = 0; i < addressToPointers.length; i++) {
            if (addressToPointers[i] != null) {
                toPointersArray[i] = new int[addressToPointers[i].length];
                for (int j = 0; j < addressToPointers[i].length; j++) {
                    toPointersArray[i][j] = addressToPointers[i][j] != NULL_PTR ? toIndex(addressToPointers[i][j], addressesArray) : NULL_INDEX;
                }
            }
        }

        //LOGGER.info("Count from pointers");
        for (int idx = 0; idx < toPointersArray.length; idx++) {
            int[] ptrs = toPointersArray[idx];
            if (ptrs != null) {
                for (int ptrI = 0; ptrI < ptrs.length; ptrI++) {
                    int ptr = ptrs[ptrI];
                    if (valid(ptr)) {
                        if (fromPtrCounter[ptr] == null) {
                            fromPtrCounter[ptr] = new Counter();
                        }
                        fromPtrCounter[ptr].inc();
                    }
                }
            }
        }
        //LOGGER.info("Set from pointer arrays");
        for (int idx = 0; idx < objectCount; idx++) {
            if (fromPtrCounter[idx] != null) {
                fromPointersArray[idx] = new int[(int) fromPtrCounter[idx].get()];
                fromPtrCounter[idx].reset();
            } else {
                // No from pointers
                fromPointersArray[idx] = new int[0];
            }
        }
        for (int idx = 0; idx < objectCount; idx++) {
            int[] ptrs = toPointersArray[idx];
            if (ptrs != null) {
                for (int ptrI = 0; ptrI < ptrs.length; ptrI++) {
                    int ptr = ptrs[ptrI];
                    if (valid(ptr)) {
                        fromPointersArray[ptr][(int) fromPtrCounter[ptr].get()] = idx;
                        fromPtrCounter[ptr].inc();
                    }
                }
            }
        }

        // This call ensures that everything is set up right in the child classes
        setup(objectInfos, addressesArray, toPointersArray, fromPointersArray);

        // translate root pointers to indices
        heap.rootPtrs.forEach((adress, rootPointerList) -> {
            int rootIndex = adress != NULL_INDEX ? toIndex(adress, addressesArray) : NULL_INDEX;
            rootPtrList.addAll(rootPointerList);
            rootPtrs.put(rootIndex, rootPointerList);
            rootPointerList.forEach(rp -> rp.setIdx(rootIndex));
        });

        if (!REDUCE_TIMING_MEASUREMENT_DISTORTION) {
            // mark everything that is reachable from roots
            for (int rootIdx : rootPtrs.keySet()) {
                markReachable(rootIdx);
            }

            //        Counter c = new Counter();
            //        ExecutorServiceUtil.temporaryExecutorServiceAsync(
            //                rootPtrs.entrySet().iterator(),
            //                rootPtrEntry -> {
            //                    BitSet closure = closure(rootPtrEntry.getKey());
            //                    rootPtrEntry.getValue().forEach(rootPtr -> rootPtr.reachables = closure);
            //                    c.inc();
            //                    return null;
            //                },
            //                () -> rootPtrsReachablesBeingCalculated = false);

            // detect datastructures in heap
            initDataStructures();
        }
    }

    public void initDataStructures() {
        ApplicationStatistics.Measurement mCreateDataStructures = ApplicationStatistics.getInstance().createMeasurement("Create data structures");
        List<DataStructure> ds = new ArrayList<>();
        if (CALCULATE_DATA_STRUCTURES) {
            stream().forEach(idx -> {
                DataStructure dataStructure = DataStructure.createDataStructure(idx, this);
                if (dataStructure != null) {
                    ds.add(dataStructure);
                }
            });
        }
        dataStructures = new DataStructure[ds.size()];
        ds.toArray(dataStructures);
        mCreateDataStructures.end();

        ApplicationStatistics.Measurement mBuildDataStructureComposition = ApplicationStatistics.getInstance().createMeasurement("Build data structure composition");
        dataStructureComposition = new int[objectCount][];
        // build data structure composition array...
        for (int i = 0; i < dataStructures.length; i++) {
            final int dataStructureIndex = i;

            dataStructures[dataStructureIndex].visitAllObjects(idx -> {
                int[] dataStructuresOfObjIndex = dataStructureComposition[idx];

                // update data structures for object at given index...
                if (dataStructuresOfObjIndex == null) {
                    // create new array with first entry
                    dataStructuresOfObjIndex = new int[]{dataStructureIndex};
                } else {
                    // append to existing array
                    ApplicationStatistics.Measurement mBuildDataStructureCompositionArraycopy = ApplicationStatistics.getInstance()
                                                                                                                     .createMeasurement("Copy composition array");
                    dataStructuresOfObjIndex = Arrays.copyOf(dataStructuresOfObjIndex, dataStructuresOfObjIndex.length + 1);
                    mBuildDataStructureCompositionArraycopy.end();
                    dataStructuresOfObjIndex[dataStructuresOfObjIndex.length - 1] = dataStructureIndex;
                }

                // update data structures array of object
                dataStructureComposition[idx] = dataStructuresOfObjIndex;

            }, this);
        }
        mBuildDataStructureComposition.end();
    }

    protected abstract void setup(ObjectInfo[] objectInfos, long[] addressesArray, int[][] toPointersArray, int[][] fromPointersArray);

    protected void markReachable(int objectIndex) {
        if (objectIndex < 0 || isRootReachable(objectIndex)) {
            return;
        }

        reachableFromRoot.set(objectIndex);

        Stack<Integer> stack = new Stack<>();
        stack.push(objectIndex);

        while (!stack.isEmpty()) {
            int idx = stack.pop();
            int[] ptrs = getToPointers(idx);
            if (ptrs != null) {
                for (int i = 0; i < ptrs.length; i++) {
                    int ptrIdx = ptrs[i];
                    if (ptrIdx >= 0 && !isRootReachable(ptrIdx)) {
                        stack.push(ptrIdx);
                        reachableFromRoot.set(ptrIdx);
                    }
                }
            }
        }
    }

    //================================================================================
    // utilities
    //================================================================================

    public boolean valid(int objIndex) {
        return objIndex >= 0 && objIndex < objectCount;
    }

    private int toIndex(long address, long[] addresses) {
        int objIndex = Arrays.binarySearch(addresses, address);
        assert objIndex < objectCount : "Index must not be out of bounds!";
        if (!valid(objIndex)) {
            // although it is not a big deal for this method, it shouldn't happen considering current usage of this method
            //assert false;
            return NULL_INDEX;
        }
        return objIndex;
    }

    abstract public void clear();

    //================================================================================
    // iteration
    //================================================================================
    public IntStream stream() {
        return IntStream.rangeClosed(0, objectCount - 1);
    }

    public ListClassificationTree groupListParallel(Filter[] filters,
                                                    ClassifierChain classifiers,
                                                    ConditionalClassifierList conditionalClassifiers,
                                                    ObjectStream.IterationListener listener,
                                                    BooleanProperty cancellationToken) {
        ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance()
                                                                   .createMeasurement("fast heap: group list");
        if (filters != null) {
            Arrays.stream(filters).filter(f -> f != null).forEach(f -> f.setup(() -> symbols, () -> this));
        }
        if (classifiers != null) {
            classifiers.getList().stream().filter(c -> c != null).forEach(c -> c.setup(() -> symbols, () -> this));
        }
        if (conditionalClassifiers != null) {
            conditionalClassifiers.stream()
                                  .filter(cc -> cc != null)
                                  .forEach(cc -> cc.getClassifiers()
                                                   .getList()
                                                   .stream()
                                                   .filter(c -> c != null)
                                                   .forEach(c -> c.setup(() -> symbols, () -> this)));
        }

        ListGroupingNode combinedTree = new ListGroupingNode(this);
        ParallelizationUtil.temporaryExecutorServiceBlocking((threadId, threadCount) -> {
            ListGroupingNode threadLocalTree = new ListGroupingNode(this);
            Counter threadLocalCounter = new Counter();
            for (int i = threadId; i < objectCount; i += threadCount) {
                boolean classify = true;
                if (filters != null) {
                    for (Filter filter : filters) {
                        try {
                            if (!filter.classify(i)) {
                                classify = false;
                                break;
                            }
                        } catch (Exception e) {
                            e.printStackTrace();
                            classify = false;
                            break;
                        }
                    }
                }
                if (classify && (cancellationToken == null || !cancellationToken.get())) {
                    threadLocalTree.classify(this, i, classifiers, conditionalClassifiers);
                }

                if (listener != null) {
                    // progress indication...
                    threadLocalCounter.inc();
                    if (threadLocalCounter.get() % 10_000 == 0) {
                        listener.objectsIterated(threadLocalCounter.get());
                        threadLocalCounter.reset();
                    }
                }
            }
            if (listener != null) {
                listener.objectsIterated(threadLocalCounter.get());
            }
            synchronized (combinedTree) {
                combinedTree.merge(threadLocalTree);
            }
        });
        m.end();
        return new ListClassificationTree(combinedTree, filters, classifiers, conditionalClassifiers);

    }

    //================================================================================
    // Getters
    //================================================================================
    public int getObjectCount() {
        return objectCount;
    }

    public int getReachableFromRootObjectCount() {
        return reachableFromRoot.cardinality();
    }

    public long getByteCount() {
        return byteCount;
    }

    public AllocatedType getType(int objIndex) {
        return getObjectInfo(objIndex).type;
    }

    public int getSize(int objIndex) {
        assert objIndex < objectCount : "Object index may not be out of range";
        return objIndex == NULL_INDEX ? 0 : getObjectInfo(objIndex).size;
    }

    public boolean isArray(int objIndex) {
        return getObjectInfo(objIndex) instanceof ArrayObjectInfo;
    }

    public int getArrayLength(int objIndex) {
        return isArray(objIndex) ? ((ArrayObjectInfo) getObjectInfo(objIndex)).getLength() : -1;
    }

    public AllocationSite getAllocationSite(int objIndex) {
        if (!valid(objIndex)) {
            return null;
        }

        return getObjectInfo(objIndex).allocationSite;
    }

    public EventType getEventType(int objIndex) {
        if (!valid(objIndex)) {
            return null;
        }

        return getObjectInfo(objIndex).eventType;
    }

    public SpaceInfo getSpace(int objIndex) {
        long address = getAddress(objIndex);
        int spaceIndex = Arrays.binarySearch(spaceStartAdresses, address);
        if (spaceIndex < 0) {
            // address is not the start of a space
            spaceIndex = Math.abs(spaceIndex);
            assert spaceIndex >= 2;   // otherwise address < spaceStartAddresses[0]
            spaceIndex = spaceIndex - 2;
        }

        return spaceInfos[spaceIndex];
    }

    public List<RootPtr> getRoot(int objIndex) {
        return rootPtrs.get(objIndex);
    }

    public Symbols getSymbols() {
        return symbols;
    }

    public List<RootPtr> getRootPtrListBlocking() {
        while (rootPtrsReachablesBeingCalculated) {
            // Busy waiting while root pointer reachables are still being calculated
            try {
                Thread.sleep(10);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }
        return rootPtrList;
    }

    public HashMap<Integer, List<RootPtr>> getRootPointerBlocking() {
        while (rootPtrsReachablesBeingCalculated) {
            // Busy waiting while root pointer reachables are still being calculated
            try {
                Thread.sleep(10);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }
        return rootPtrs;
    }

    public boolean isRootReachable(int objIndex) {
        return reachableFromRoot.get(objIndex);
    }

    public DataStructure[] getDataStructures() {
        return dataStructures;
    }

    public DataStructure[] getDataStructures(int objIndex, boolean includeIndirectDataStructures, boolean reduceToTopLevelDataStructures) {
        // TODO reduce flag is dependent on include flag which makes it weird to use...

        if (!valid(objIndex)) {
            return null;
        }

        int[] dataStructureIndices = dataStructureComposition[objIndex];

        if (dataStructureIndices != null && dataStructureIndices.length > 0) {
            Set<DataStructure> closedSet = new HashSet<>();
            for (int i = 0; i < dataStructureIndices.length; i++) {
                closedSet.add(dataStructures[dataStructureIndices[i]]);
            }

            Set<DataStructure> topLevelDSs = null;
            if (reduceToTopLevelDataStructures) {
                topLevelDSs = new HashSet<>();
                // only add those ds to top-level set that are not dominated by any other ds in the closed set
                closedSet.stream()
                         .filter(ds -> closedSet.stream()
                                                .filter(ds2 -> ds2 != ds)
                                                .noneMatch(ds2 -> dominates(ds2.headIdx, ds.headIdx)))
                         .forEach(topLevelDSs::add);
            }
            if (includeIndirectDataStructures) {
                // follow data structure heads recursively to collect also data structures that indirectly reference the object at the given index
                Stack<DataStructure> stack = new Stack<>();
                stack.addAll(closedSet);
                while (!stack.isEmpty()) {
                    DataStructure cur = stack.pop();
                    int[] curContainingDSIndices = dataStructureComposition[cur.headIdx];
                    for (int i = 0; i < curContainingDSIndices.length; i++) {
                        DataStructure dsContainingCur = dataStructures[curContainingDSIndices[i]];
                        if (closedSet.add(dsContainingCur)) {
                            stack.push(dsContainingCur);

                            if (reduceToTopLevelDataStructures) {
                                topLevelDSs.add(dsContainingCur);
                                if (dsContainingCur != cur && dominates(dsContainingCur.headIdx, cur.headIdx)) {   // TODO do really only head indices matter?
                                    topLevelDSs.remove(cur);
                                }
                            }
                        }
                    }
                }
            }

            return reduceToTopLevelDataStructures ? topLevelDSs.toArray(new DataStructure[topLevelDSs.size()]) : closedSet.toArray(new DataStructure[closedSet.size()]);
        }

        return null;
    }

    /**
     * Returns the DataStructure of which the object at the given index is the head
     *
     * @param headIndex
     * @return the DataStructure object or null if the given index is not a head index
     */
    public DataStructure getHeadedDataStructure(int headIndex) {
        if (!valid(headIndex)) {
            return null;
        }

        int[] dataStructureIndices = dataStructureComposition[headIndex];
        for (int i = 0; i < dataStructureIndices.length; i++) {
            DataStructure ds = dataStructures[dataStructureIndices[i]];
            if (ds.headIdx == headIndex) {
                return ds;
            }
        }

        return null;
    }

    /**
     * This is a faster variant of DataStructure::isDataStructureHead because it operates on the already recognized data structures of this heap
     */
    public boolean isDataStructureHead(int objIndex) {
        if (!valid(objIndex)) {
            return false;
        }

        int[] dataStructureIndices = dataStructureComposition[objIndex];
        if (dataStructureIndices == null) {
            return false;
        }

        for (int i = 0; i < dataStructureIndices.length; i++) {
            DataStructure dataStructure = dataStructures[dataStructureIndices[i]];
            if (dataStructure.headIdx == objIndex) {
                return true;
            }
        }

        return false;
    }

    //getters abstract
    public abstract int toIndex(long address);

    public abstract long getAddress(int objIndex);

    public abstract int[] getToPointers(int objIndex);

    public abstract int[] getFromPointers(int objIndex);

    public abstract ObjectInfo getObjectInfo(int objIndex);

    //================================================================================
    // Closures
    //================================================================================
    public Closures getClosures(int objIndex) {
        return getClosures(new int[]{objIndex}, null);
    }

    public Closures getClosures(int[] objIndices) {
        return getClosures(objIndices, null);
    }

    public Closures getClosures(int[] objIndices, BitSet assumeChildClosure) {
        ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance()
                                                                   .createMeasurement("Calculate closure Alt");
        BitSet transitiveClosure = transitiveClosure(objIndices, assumeChildClosure);
        m.end();

        m = ApplicationStatistics.getInstance().createMeasurement("Calculate GC closure Alt");
        BitSet gcClosure = gcClosure(objIndices, transitiveClosure);
        m.end();

        m = ApplicationStatistics.getInstance().createMeasurement("Calculate DS closure");
        BitSet dsClosure = getDataStructureClosure(objIndices);
        m.end();

        m = ApplicationStatistics.getInstance().createMeasurement("Calculate deep DS closure");
        BitSet deepDSClosure = getDeepDataStructureClosure(objIndices);
        m.end();

        return new Closures(this, objIndices, transitiveClosure, gcClosure, dsClosure, deepDSClosure);
    }

    //================================================================================
    // Closure size
    //================================================================================
    public BitSet transitiveClosure(int[] objIndices, BitSet assumeClosure) {
        BitSet closure = assumeClosure != null ? assumeClosure : bitSetSupplier.get();

        BitSet toProcess = bitSetSupplier.get();

        // build closure of objIndices
        for (int i = 0; i < objIndices.length; i++) {
            if (valid(objIndices[i])) {
                toProcess.set(objIndices[i]);
                closure.set(objIndices[i]);
            }
        }

        while (!toProcess.isEmpty()) {
            for (int idx = toProcess.nextSetBit(0); idx != -1; idx = toProcess.nextSetBit(idx + 1)) {
                toProcess.clear(idx);
                int[] pointers = getToPointers(idx);
                if (pointers != null) {
                    for (int j = 0; j < pointers.length; j++) {
                        int ptr = pointers[j];
                        if (valid(ptr) && !closure.get(ptr)) {
                            closure.set(ptr);
                            toProcess.set(ptr);
                        }
                    }
                }
            }
        }

        return closure;
    }

    //================================================================================
    // GC size
    //================================================================================
    public BitSet gcClosure(int[] objIndices, BitSet transitiveClosure) {
        // everything reachable from given indices (= upper limit of gc size)
        BitSet gcClosure = (BitSet) transitiveClosure.clone();
        BitSet objIndicesSet = bitSetSupplier.get();
        for (int i = 0; i < objIndices.length; i++) {
            if (valid(objIndices[i])) {
                objIndicesSet.set(objIndices[i]);
            }
        }

        BitSet toRemove = bitSetSupplier.get();

        // find all objects in transitiveClosure that are referenced from the outside (except transitiveClosure root objects)
        for (int idx = transitiveClosure.nextSetBit(0); idx != -1; idx = transitiveClosure.nextSetBit(idx + 1)) {
            if (objIndicesSet.get(idx)) {
                continue;
            }

            // check direct root pointer for idx
            if (getRoot(idx) != null) {
                toRemove.set(idx);
            } else {
                // object itself is not directly root pointed, check if any referee is 1) outside of the transitiveClosure and 2) reachable from root
                int[] fromPointers = getFromPointers(idx);
                if (fromPointers != null) {
                    for (int i = 0; i < fromPointers.length; i++) {
                        int fromPtr = fromPointers[i];

                        if (!transitiveClosure.get(fromPtr) && isRootReachable(fromPtr)) {
                            // Object is kept alive from outside the transitiveClosure: Remove it and all its children from the transitiveClosure
                            toRemove.set(idx);
                            break;
                        }
                    }
                }
            }
        }

        while (!toRemove.isEmpty()) {
            for (int idx = toRemove.nextSetBit(0); idx != -1; idx = toRemove.nextSetBit(idx + 1)) {
                toRemove.clear(idx);
                gcClosure.clear(idx);
                int[] pointers = getToPointers(idx);
                if (pointers != null) {
                    for (int j = 0; j < pointers.length; j++) {
                        int ptr = pointers[j];
                        if (valid(ptr) && !objIndicesSet.get(ptr) && gcClosure.get(ptr)) {
                            toRemove.set(ptr);
                            gcClosure.clear(idx);
                        }
                    }
                }
            }
        }

        return gcClosure;
    }

    //================================================================================
    // Data structure size
    //================================================================================
    public BitSet getDataStructureClosure(int headIndex) {
        return getDataStructureClosure(new int[]{headIndex});
    }

    public BitSet getDataStructureClosure(int[] headIndices) {
        BitSet closure = bitSetSupplier.get();

        for (int i = 0; i < headIndices.length; i++) {
            if (valid(headIndices[i]) && dataStructureComposition[headIndices[i]] != null) {
                for (int j = 0; j < dataStructureComposition[headIndices[i]].length; j++) {
                    DataStructure dataStructure = dataStructures[dataStructureComposition[headIndices[i]][j]];
                    if (dataStructure.headIdx == headIndices[i]) {
                        // ...headIndices[i] is the head of dataStructure (and not just contained)
                        // ==> mark all objects in dataStructure in the closure
                        dataStructure.visitAllObjects(closure::set, this);
                    }
                }
            }
        }

        return closure;
    }

    private BitSet getDeepDataStructureClosure(int headIndex) {
        return getDeepDataStructureClosure(new int[]{headIndex});
    }

    private BitSet getDeepDataStructureClosure(int[] headIndices) {
        // this set serves as a queue of head indices still to handle
        BitSet dsHeadsToVisit = bitSetSupplier.get();
        // ...initialized with the given set of head indices
        for (int i = 0; i < headIndices.length; i++) {
            if (valid(headIndices[i]) && isDataStructureHead(headIndices[i])) {
                dsHeadsToVisit.set(headIndices[i]);
            }
        }

        // the final closure; also serves as closed set for already visited head indices
        BitSet deepDataStructureClosure = bitSetSupplier.get();
        // the to visit queue is iterated from front to back until it is empty (new head indices might be set before the one that is currently handled)
        while (!dsHeadsToVisit.isEmpty()) {
            for (int headIndex = dsHeadsToVisit.nextSetBit(0); headIndex != -1; headIndex = dsHeadsToVisit.nextSetBit(headIndex + 1)) {
                // visiting the head index of a data structure:
                // 1) remove it from the to visit queue
                dsHeadsToVisit.clear(headIndex);
                // 2) if it was not yet handled (i.e. not yet in the final closure) ...
                if (!deepDataStructureClosure.get(headIndex)) {
                    // ... add it
                    deepDataStructureClosure.set(headIndex);
                    // 3) retrieve the data structure object of the head index
                    DataStructure ds = getHeadedDataStructure(headIndex);
                    if (ds != null) {
                        // 4) visit all objects part of the data structure...
                        ds.visitAllObjects(objIndex -> {
                            if (isDataStructureHead(objIndex)) {
                                // ... data structure heads encountered in the process are not added to final closure yet but only put in the to visit queue
                                dsHeadsToVisit.set(objIndex);
                            } else {
                                deepDataStructureClosure.set(objIndex);
                            }
                        }, this);
                    }
                }
            }
        }

        return deepDataStructureClosure;
    }

    //================================================================================
    // Root pointers
    //================================================================================
    public List<RootPtr.RootInfo> traceClosestRootBFS(int objIndex, int maxDepth, boolean onlyVariables) {

        if (!valid(objIndex)) {
            return new ArrayList<>();
        }

        Set<Integer> closedSet = new HashSet<>();
        ArrayDeque<int[]> fifo = new ArrayDeque<>();
        List<RootPtr.RootInfo> ret = new ArrayList<>();

        int[] initialPath = new int[]{objIndex};
        fifo.add(initialPath);
        closedSet.add(objIndex);

        while (!fifo.isEmpty()) {
            int[] curPath = fifo.poll();
            int curObjIndex = curPath[curPath.length - 1];

            // Continue to check all from-pointers on the same level on which the first root-pointer was found.
            // We always take the one which's typeId is lower
            if (!ret.isEmpty() && curPath.length > ret.get(0).path.length) {
                return ret;
            }

            // is there a corresponding root ptr?
            List<RootPtr> matchingRoots = rootPtrs.get(curObjIndex);
            if (matchingRoots != null && onlyVariables) {
                matchingRoots = matchingRoots.stream()
                                             .filter(r -> r.getRootType().isVariable)
                                             .collect(Collectors.toList());
            }
            if (matchingRoots != null && matchingRoots.size() != 0) {
                // If we find multiple root pointers on the same depth, take the one with the lower type id
                if (ret.isEmpty() || ret.get(0).ptrs[0].getRootType().byteVal > matchingRoots.get(0)
                                                                                             .getRootType().byteVal) {
                    ret.clear();
                    for (RootPtr root : matchingRoots) {
                        ret.add(new RootPtr.RootInfo(new RootPtr[]{root}, curPath));
                    }
                }
            }

            // ... and keep searching
            int[] fromPointers = getFromPointers(curObjIndex);
            if (fromPointers != null && curPath.length < maxDepth) {
                for (int i = 0; i < fromPointers.length; i++) {
                    // skip already handled addresses
                    if (closedSet.add(fromPointers[i])) {
                        // extend path by next step
                        int[] newPath = Arrays.copyOf(curPath, curPath.length + 1);
                        newPath[newPath.length - 1] = fromPointers[i];
                        fifo.offer(newPath);
                    }
                }
            }
        }

        return ret;
    }

    public List<RootPtr.RootInfo> traceAllRootsDFS(int objIndex, int maxDepth, boolean onlyVariables) {
        return traceAllRootsDFS(new int[]{objIndex}, maxDepth, onlyVariables);
    }

    public List<RootPtr.RootInfo> traceAllRootsDFS(int[] objIndices, int maxDepth, boolean onlyVariables) {
        // the keys of this map are the object indices of all objects that have already been identified as part of a path to a root
        // the value of each key is an array of RootInfos i.e. the results (paths) that contain the key
        Map<Integer, List<RootPtr.RootInfo>> foundPathObjects = new HashMap<>();
        List<RootPtr.RootInfo> ret = new ArrayList<>();

        for (int objIndex : objIndices) {
            if (!valid(objIndex)) {
                return new ArrayList<>();
            }

            Stack<int[]> stack = new Stack<>();
            Set<Integer> closedSet = new HashSet<>();

            int[] initialPath = new int[]{objIndex};
            stack.add(initialPath);
            closedSet.add(objIndex);

            while (!stack.isEmpty()) {
                int[] curPath = stack.pop();
                int curObjIndex = curPath[curPath.length - 1];

                if (foundPathObjects.containsKey(curObjIndex)) {
                    // we reached an object index that was already identified as part of a path to a root
                    // retrieve this already existing result
                    List<RootPtr.RootInfo> existingPaths = foundPathObjects.get(curObjIndex);
                    // the current path can be completed with each of the already existing paths
                    for (RootPtr.RootInfo existingPath : existingPaths) {
                        List<Integer> completedPath = Arrays.stream(existingPath.path)
                                                            .boxed()
                                                            .collect(Collectors.toCollection(ArrayList::new));
                        completedPath = completedPath.subList(completedPath.indexOf(curObjIndex) + 1, completedPath.size());
                        for (int i = curPath.length - 1; i >= 0; i--) {
                            completedPath.add(0, curPath[i]);
                        }
                        // add the resulting path as a new result
                        RootPtr.RootInfo newResult = new RootPtr.RootInfo(existingPath.ptrs,
                                                                          completedPath.stream()
                                                                                       .mapToInt(i -> i)
                                                                                       .toArray());
                        ret.add(newResult);
                        for (int i = 0; i < curPath.length - 1; i++) {
                            List<RootPtr.RootInfo> rootInfos = foundPathObjects.get(curPath[i]);
                            if (rootInfos == null) {
                                rootInfos = new ArrayList<>();
                            }
                            rootInfos.add(newResult);
                            foundPathObjects.put(curPath[i], rootInfos);
                        }
                    }

                    // possible root paths from this node have already been discovered
                    continue;
                }

                // are there corresponding root ptrs?
                List<RootPtr> matchingRoots = rootPtrs.get(curObjIndex);
                if (matchingRoots != null) {
                    RootPtr.RootInfo newResult = new RootPtr.RootInfo(matchingRoots.stream()
                                                                                   .filter(root -> !onlyVariables || root.getRootType().isVariable)
                                                                                   .toArray(RootPtr[]::new), curPath);
                    ret.add(newResult);
                    for (Integer curPathObj : curPath) {
                        List<RootPtr.RootInfo> rootInfos = foundPathObjects.get(curPathObj);
                        if (rootInfos == null) {
                            rootInfos = new ArrayList<>();
                        }
                        rootInfos.add(newResult);
                        foundPathObjects.put(curPathObj, rootInfos);
                    }
                }

                // ... and keep searching
                int[] fromPointers = getFromPointers(curObjIndex);
                if (fromPointers != null && curPath.length < maxDepth) {
                    for (int i = 0; i < fromPointers.length; i++) {
                        // skip already handled addresses
                        if (closedSet.add(fromPointers[i])) {
                            // extend path by next step
                            int[] newPath = Arrays.copyOf(curPath, curPath.length + 1);
                            newPath[newPath.length - 1] = fromPointers[i];
                            stack.push(newPath);
                        }
                    }
                }
            }
        }

        ret.sort(Comparator.comparingInt(rootInfo -> rootInfo.path.length));
        return ret;
    }

    public RootPtr findClosestRoot(int objIndex, int maxDepth, boolean onlyVariables) {
        //ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance().createMeasurement("find closest Root BFS");
        if (!valid(objIndex)) {
            return null;
        }

        BitSet visited = bitSetSupplier.get();
        BitSet currentlyToVisit = bitSetSupplier.get();
        BitSet nextVisitRound = bitSetSupplier.get();

        RootPtr ret = null;

        nextVisitRound.set(objIndex);
        visited.set(objIndex);

        for (int currentDepth = 0; currentDepth < maxDepth && ret == null; currentDepth++) {
            currentlyToVisit.clear();
            currentlyToVisit.or(nextVisitRound);
            nextVisitRound.clear();
            for (int idx = currentlyToVisit.nextSetBit(0); idx != -1; idx = currentlyToVisit.nextSetBit(idx + 1)) {
                // is there a corresponding root ptr?
                List<RootPtr> matchingRoots = rootPtrs.get(idx);
                if (onlyVariables && matchingRoots != null) {
                    matchingRoots = matchingRoots.stream()
                                                 .filter(r -> r.getRootType().isVariable)
                                                 .collect(Collectors.toList());
                }
                if (matchingRoots != null && matchingRoots.size() != 0) {
                    // If we find multiple root pointers on the same depth, take the one with the lower type id
                    if (ret == null || ret.getRootType().byteVal > matchingRoots.get(0)
                                                                                .getRootType().byteVal) {
                        ret = matchingRoots.get(0);
                    }
                }
                if (ret == null) {
                    // ... and keep searching
                    int[] fromPointers = getFromPointers(idx);
                    if (fromPointers != null) {
                        for (int i = 0; i < fromPointers.length; i++) {
                            if (!visited.get(fromPointers[i])) {
                                visited.set(fromPointers[i]);
                                // extend path by next step
                                nextVisitRound.set(fromPointers[i]);
                            }
                        }
                    }
                }
            }
        }
        //m.end();
        return ret;
    }

    public List<RootPtr> indirectGCRoots(int[] objIndices, int maxDepth) {
        return indirectGCRoots(objIndices, maxDepth, null);
    }

    public List<RootPtr> indirectGCRoots(int[] objIndices, int maxDepth, BiConsumer<Integer, Integer> visitor) {
        List<RootPtr> ret = new ArrayList<>();

        BitSet visited = bitSetSupplier.get();
        BitSet currentlyToVisit = bitSetSupplier.get();
        BitSet nextVisitRound = bitSetSupplier.get();

        for (int i = 0; i < objIndices.length; i++) {
            if (valid(objIndices[i])) {
                nextVisitRound.set(objIndices[i]);
                visited.set(objIndices[i]);
            }
        }

        for (int currentDepth = 0; currentDepth < maxDepth && !nextVisitRound.isEmpty(); currentDepth++) {
            currentlyToVisit.clear();
            currentlyToVisit.or(nextVisitRound);
            nextVisitRound.clear();
            for (int idx = currentlyToVisit.nextSetBit(0); idx != -1; idx = currentlyToVisit.nextSetBit(idx + 1)) {

                // is there a corresponding root ptr?
                List<RootPtr> matchingRoots = rootPtrs.get(idx);
                if (matchingRoots != null && matchingRoots.size() != 0) {
                    ret.addAll(matchingRoots);
                }

                // ... and keep searching
                int[] fromPointers = getFromPointers(idx);
                if (fromPointers != null) {
                    for (int i = 0; i < fromPointers.length; i++) {
                        if (!visited.get(fromPointers[i])) {
                            visited.set(fromPointers[i]);
                            // extend path by next step
                            nextVisitRound.set(fromPointers[i]);

                            if (visitor != null) {
                                visitor.accept(idx, fromPointers[i]);
                            }
                        }
                    }
                }
            }
        }

        return ret;
    }

    public List<RootPtr> indirectGCRoots(int objIndex, int maxDepth) {
        return indirectGCRoots(new int[]{objIndex}, maxDepth);
    }

    public BitSet fromClosure(int objIndex) {
        ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance()
                                                                   .createMeasurement("From Closure");
        if (!valid(objIndex)) {
            return bitSetSupplier.get();
        }

        BitSet fromClosure = bitSetSupplier.get();

        int[] stack = new int[]{objIndex};
        int stackIndex = 0;
        int stackTop = 1;

        while (stackIndex < stackTop) {
            int curObjIndex = stack[stackIndex];
            stackIndex++;
            // Add current object to from-closure
            fromClosure.set(curObjIndex);

            // ... and keep following from-pointers
            int[] fromPointers = getFromPointers(curObjIndex);
            if (fromPointers != null) {
                for (int i = 0; i < fromPointers.length; i++) {
                    int pointedFromPtr = fromPointers[i];
                    if (!fromClosure.get(pointedFromPtr)) {
                        fromClosure.set(pointedFromPtr);
                        // extend path by next step
                        if (stackTop == stack.length) {
                            stack = Arrays.copyOf(stack, stack.length * 2);
                        }
                        stack[stackTop] = pointedFromPtr;
                        stackTop++;
                    }
                }
            }
        }
        m.end();
        return fromClosure;
    }

    //================================================================================
    // Dominators
    //================================================================================
    public boolean dominates(int dominatorCandidateIdx, int dominatedIdx) {
        return Arrays.stream(getDominators(dominatedIdx))
                     .anyMatch(dominator -> dominator == dominatorCandidateIdx);
    }

    public int[] getDominators(int objIndex) {
        if (dominators == null) {
            calculateDominators();
            if (dominators == null) {
                throw new IllegalStateException("Could not calculate dominators!");
            }
        }

        List<Integer> dominators = new ArrayList<>();
        for (int idom = getImmediateDominator(objIndex); idom != -1; idom = getImmediateDominator(idom)) {
            dominators.add(idom);
        }

        return dominators.stream()
                         .mapToInt(i -> i)
                         .toArray();
    }

    public int getImmediateDominator(int objIndex) {
        // returns the index of the object that immediately dominates the object at the given objIndex
        if (dominators == null) {
            calculateDominators();
            if (dominators == null) {
                throw new IllegalStateException("Could not calculate dominators!");
            }
        }

        Integer idom = dominators.getIDom(objIndex + 1); // might return null, so don't auto(un)box...

        return getOriginalObjectIndex(idom);
    }

    public int[] getDominatedObjects(int objIndex) {
        return getDominatedObjects(objIndex, true);
    }

    public int[] getImmediatelyDominatedObjects(int objIndex) {
        return getDominatedObjects(objIndex, false);
    }

    private int[] getDominatedObjects(int objIndex, boolean recursive) {
        // returns the indices of all objects that are immediately dominated by the object at the given objIndex
        // if the recursive flag is set, also the indices of non-immediately dominated children are included
        if (dominators == null) {
            calculateDominators();
            if (dominators == null) {
                throw new IllegalStateException("Could not calculate dominators!");
            }
        }

        Set<Integer> dominatedObjects = new HashSet<>();
        dominators.getNodesWithIDom(objIndex)
                  .forEach(dominatedObjects::add);

        if (recursive) {
            Stack<Integer> stack = new Stack<>();
            dominatedObjects.forEach(stack::push);
            while (!stack.isEmpty()) {
                int dom = stack.pop();
                for (int domDom : dominators.getNodesWithIDom(dom)) {
                    if (dominatedObjects.add(domDom)) {
                        stack.push(domDom);
                    }
                }
            }
        }

        return dominatedObjects.stream()
                               .mapToInt(this::getOriginalObjectIndex)
                               .toArray();
    }

    public void calculateDominators(List<RootPtr.RootInfo> pathsToAllRoots) {
        ApplicationStatistics.Measurement m_graph = ApplicationStatistics.getInstance()
                                                                         .createMeasurement("Building object graph for object subset");
        // build object graph for given object subset
        // TODO test this...
        DirectedGraph<Integer, Tuple<Integer, Integer>> objectGraph = new DefaultDirectedGraph<>(Tuple::new);
        objectGraph.addVertex(0);
        pathsToAllRoots.stream()
                       .map(rootInfo -> rootInfo.path)
                       .forEach(pathToRoot -> {
                           objectGraph.addVertex(pathToRoot[0] + 1);
                           for (int i = 0; i < pathToRoot.length - 1; i++) {
                               objectGraph.addVertex(pathToRoot[i + 1] + 1);
                               objectGraph.addEdge(pathToRoot[i + 1] + 1, pathToRoot[i] + 1);  // reverse order corresponds to object graph
                           }
                           objectGraph.addEdge(0, pathToRoot[pathToRoot.length - 1] + 1);
                       });

        m_graph.end();

        calculateDominators(objectGraph);
    }

    private void calculateDominators(DirectedGraph<Integer, Tuple<Integer, Integer>> objectGraph) {
        ApplicationStatistics.Measurement m_dom = ApplicationStatistics.getInstance()
                                                                       .createMeasurement("Computing dominators");
        // get dominator tree of the object graph
        dominators = Dominators.compute(objectGraph, 0);
        m_dom.end();
    }

    public void calculateDominators() {
        ApplicationStatistics.Measurement m_graph = ApplicationStatistics.getInstance()
                                                                         .createMeasurement("Building object graph");
        // build object graph of this heap
        DirectedGraph<Integer, Tuple<Integer, Integer>> objectGraph = new DefaultDirectedGraph<>(Tuple::new);
        objectGraph.addVertex(0);
        stream().forEach(objIndex -> objectGraph.addVertex(objIndex + 1));

        rootPtrs.keySet()
                .stream()
                .distinct()
                .filter(toPointer -> toPointer != NULL_INDEX)
                .forEach(toPointer -> objectGraph.addEdge(0, toPointer + 1));

        for (int i = 0; i < objectCount; i++) {
            int[] ptrs = getToPointers(i);
            if (ptrs != null) {
                for (int j = 0; j < ptrs.length; j++) {
                    if (ptrs[j] != NULL_INDEX) {
                        objectGraph.addEdge(i + 1, ptrs[j] + 1);
                    }
                }
            }
        }
        m_graph.end();

        calculateDominators(objectGraph);
    }

    private int getOriginalObjectIndex(Integer domTreeNodeIdx) {
        // object indices used for dominator calculation are all +1 (to introduce super root node at idx 0)
        // now we have to subtract 1 to get the original object index
        if (domTreeNodeIdx == null) {
            // the dominator of the super root node
            return NULL_INDEX;          // ...-1
        } else if (domTreeNodeIdx == 0) {
            // super root
            return SUPER_ROOT_INDEX;    // ...-2
        } else {
            return domTreeNodeIdx - 1;            // ...[0, objectCount[
        }
    }
}
