package at.jku.mevss.anttracks.heap.datastructures;

import at.jku.mevss.anttracks.heap.IndexBasedHeap;
import at.jku.mevss.anttracks.heap.symbols.AllocatedType;

import java.util.BitSet;
import java.util.Stack;
import java.util.function.Consumer;

import static at.jku.mevss.anttracks.heap.IndexBasedHeap.NULL_INDEX;
import static at.jku.mevss.anttracks.heap.IndexBasedHeap.bitSetSupplier;

public class DataStructure {

    public final int headIdx;
    /**
     * All objects that are part of this data structure according to its DSL definition, NOT the number of data elements in the data structure
     */
    private int objectCount;

    private DataStructure(int headIdx) {
        this.headIdx = headIdx;
        this.objectCount = 0;
    }

    /**
     * Creates {@link DataStructure} if given index is a data structure head. Do not call twice for the same objIndex
     *
     * @param objIndex
     * @param heap
     * @return
     */
    public static DataStructure createDataStructure(int objIndex, IndexBasedHeap heap) {
        if (isDataStructureHead(objIndex, heap)) {
            return new DataStructure(objIndex);
        }

        return null;
    }

    public static boolean isDataStructureHead(int objIndex, IndexBasedHeap heap) {
        DataStructureDefinition dataStructureDefinition = getDataStructureDefinition(objIndex, heap);
        return dataStructureDefinition.isHead;
    }

    public static DataStructureDefinition getDataStructureDefinition(int objIndex, IndexBasedHeap heap) {
        AllocatedType type = heap.getType(objIndex);
        return type.dataStructureDefinition;
    }

    /*
    void visit(current_object) {
        add current_object to current data structure
        mark current_object as visited
        for each not visited ptr_object of current_object {
            if(type of ptr_object is leaf type in current_object's type's data structure definiton) {
                // Just visit ptr_object, but do not continue recursive descent
                add ptr to current data structure
            } else if (type of ptr_object is fllow type in current_object's type's data structure definiton) {
                if(dsDefinition(ptr).isHead) {
                    // handle other data structe heads as leafs of current data structure
                    // Just visit ptr_object, but do not continue recursive descent
                    add ptr to current data structure
                } else {
                    // recursive descent
                    visit(ptr)
                }
            }
        }
    }
    */

    /**
     * Will call the accept method of the given visitor, passing the object indices of all objects that form this data structure.
     * Objects pointed by the data elements in the data structure are not visited.
     * Null pointers will not be visited.
     * Duplicate data elements will not be visited twice.
     * No order of visit guaranteed.
     *
     * @param visitor the int consumer whose accept method is called for each visited object
     * @param heap    fast heap needed for getType and getToPointers calls
     */
    public void visitAllObjects(Consumer<Integer> visitor, IndexBasedHeap heap) {
        int newObjectCount = 0;
        BitSet visited = bitSetSupplier.get();
        Stack<Integer> stack = new Stack<>();
        stack.push(headIdx);

        // visit head
        visitor.accept(headIdx);
        visited.set(headIdx);

        // visit the rest
        while (!stack.isEmpty()) {
            int idx = stack.pop();
            AllocatedType type = heap.getType(idx);
            DataStructureDefinition dsDefinition = type.dataStructureDefinition;

            int[] ptrs = heap.getToPointers(idx);
            if (ptrs != null) {
                for (int i = 0; i < ptrs.length; i++) {
                    int ptr = ptrs[i];
                    if (ptr != NULL_INDEX && !visited.get(ptr)) {
                        // ptr may be a part of the data structure...

                        AllocatedType ptrType = heap.getType(ptr);
                        DataStructureDefinition ptrDsDefinition = ptrType.dataStructureDefinition;
                        if (dsDefinition.hasFollower(ptrType)) {
                            // The ptr is covered by a data structure definition that should be followed => visit ptr
                            visitor.accept(ptr);
                            visited.set(ptr);
                            newObjectCount++;

                            if (!ptrDsDefinition.isHead) {
                                // do recursive descend
                                stack.push(ptr);
                            }
                        } else if (dsDefinition.containsFlat(ptrType)) {
                            // 3) the ptr is ONLY covered by a data structure definition that should not be followed => visit ptr and stop
                            visitor.accept(ptr);
                            visited.set(ptr);
                            newObjectCount++;
                        }

                        // 4) the pointer is not covered by any data structure definition => ignore
                    }
                }
            }

            // finally update the object count (avoid inconsistent object count of data structure during visit process)
            objectCount = newObjectCount;
        }
    }

    public int getObjectCount() {
        return objectCount;
    }
}
