/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.heap.io;

import at.jku.mevss.anttracks.heap.DetailedHeap;
import at.jku.mevss.anttracks.heap.GarbageCollectionCause;
import at.jku.mevss.anttracks.heap.GarbageCollectionLookup;
import at.jku.mevss.anttracks.heap.GarbageCollectionType;
import at.jku.mevss.anttracks.heap.symbols.Symbols;
import at.jku.mevss.anttracks.parser.EventType;
import at.jku.mevss.anttracks.parser.io.BaseFile;
import at.jku.mevss.anttracks.util.Consts;

import java.io.*;
import java.util.ArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;

public class HeapIndexReader implements AutoCloseable {
    private final DataInputStream in;
    private final String path;
    private final Symbols symbols;
    private final Logger logger = Logger.getLogger(this.getClass().getSimpleName());

    public HeapIndexReader(String heapReaderPath, Symbols symbols) throws IOException {
        this.symbols = symbols;
        path = heapReaderPath;
        in = new DataInputStream(new BufferedInputStream(BaseFile.openR(heapReaderPath + File.separator + Consts.HEAP_INDEX_META_FILE)));
    }

    private ArrayList<HeapIndexEntry> read() throws IOException {
        ArrayList<HeapIndexEntry> heapIndex = new ArrayList<>();
        int magic = in.readInt();
        assert (magic == Consts.HEAP_FILES_MAGIC_PREFIX) : "Index file must start with correct magic number";

        try {
            while (true) {
                long position = in.readLong();
                GarbageCollectionType gcType = GarbageCollectionType.fromID(in.readInt());
                EventType gcMeta = EventType.parse(in.readInt());
                GarbageCollectionCause gcCause = symbols.causes.get(in.readInt());
                long time = in.readLong();
                int gcCount = in.readInt();
                boolean fileExists = in.readBoolean();
                heapIndex.add(new HeapIndexEntry(position, gcType, gcMeta, gcCause, time, gcCount, fileExists));
            }
        } catch (EOFException e) {
            in.close();
        }

        return heapIndex;
    }

    public HeapPosition getRangeFromLastHeapDumpToGivenTime(long time) throws IOException {
        ArrayList<HeapIndexEntry> heapIndex = read();
        boolean found = false;
        HeapIndexEntry index = null;
        long toPosition = 0;
        int i = 0;
        int size = heapIndex.size();

        while (!found && i < size) {
            index = heapIndex.get(i);
            if (index.time == time) {
                found = true;
                toPosition = index.position;
            }

            i++;
        }

        if (!found) {
            logger.log(Level.WARNING, "Tried to obtain an entry from the GC index file which does not exist!");
            assert found : "GC must be found at given time";
        }
        found = false;
        do {
            i--;
            index = heapIndex.get(i);
            if (index.heapDumpFileExists) {
                found = true;
            }
        } while (!found && i > 0);

        HeapPosition position;
        if (!found) {
            position = new HeapPosition(-1, -1, toPosition);
        } else {
            // fromPosition = latests GC dump
            // toPosition = how far to parse (either _before_ GCstart or _after_ GCend)
            // For more information see call hierarchy of HeapIndexWriter#write()
            position = new HeapPosition(index.fileName, index.position, toPosition);
        }

        return position;
    }

    public DetailedHeap getHeap(int fileName, Symbols symbols) throws IOException {
        DetailedHeap heap;
        try (HeapReader heapReader = new HeapReader(path, fileName, symbols)) {
            heap = heapReader.read();
        }
        return heap;
    }

    @Override
    public void close() throws IOException {
        in.close();
    }

    public long getGCTime(GarbageCollectionLookup gc) throws IOException {
        ArrayList<HeapIndexEntry> heapIndex = read();
        int id = 0; // Is only incremented if matching GC type is found
        for (HeapIndexEntry entry : heapIndex) {
            if (!entry.matches(gc)) {
                continue;
            }
            if (id != gc.id) {
                id++;
                continue;
            }
            return entry.time;
        }
        return -1;
    }
}
