/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.heap.io;

import at.jku.mevss.anttracks.heap.DetailedHeap;
import at.jku.mevss.anttracks.heap.labs.Lab;
import at.jku.mevss.anttracks.heap.labs.MultiObjectLab;
import at.jku.mevss.anttracks.heap.labs.SingleObjectLab;
import at.jku.mevss.anttracks.heap.objects.*;
import at.jku.mevss.anttracks.heap.roots.RootPtr;
import at.jku.mevss.anttracks.heap.space.Space;
import at.jku.mevss.anttracks.heap.space.SpaceMode;
import at.jku.mevss.anttracks.heap.space.SpaceType;
import at.jku.mevss.anttracks.heap.symbols.Symbols;
import at.jku.mevss.anttracks.parser.EventType;
import at.jku.mevss.anttracks.parser.heap.ThreadInfo;
import at.jku.mevss.anttracks.parser.io.BaseFile;
import at.jku.mevss.anttracks.util.Consts;
import at.jku.mevss.anttracks.util.TraceException;

import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.File;
import java.io.IOException;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;

public class HeapReader implements AutoCloseable {

    private static final int VERSION = 0;

    private final DataInputStream in;
    private final Symbols symbols;

    public HeapReader(String path, Symbols symbols) throws IOException {
        in = new DataInputStream(new BufferedInputStream(BaseFile.openR(path)));
        this.symbols = symbols;
    }

    public HeapReader(String path, int fileName, Symbols symbols) throws IOException {
        in = new DataInputStream(new BufferedInputStream(BaseFile.openR(path + File.separator + fileName)));
        this.symbols = symbols;
    }

    public DetailedHeap read() throws IOException {
        if (in.readInt() != Consts.HEAP_FILES_MAGIC_PREFIX) {
            throw new IllegalArgumentException("Expected magic prefix");
        }
        if (in.readInt() != VERSION) {
            throw new IllegalArgumentException("Unsupported version");
        }

        int gcs = in.readInt();
        long lastGCTime = in.readLong();
        // CopyOnWriteArrayList<Long> ageRanges = readAgeRanges();

        ObjectInfoCache cache = new ObjectInfoCache();
        Map<Integer, ObjectInfo> prototypes = readPrototypes(cache);

        Space[] spaces = new Space[in.readInt()];
        for (int i = 0; i < spaces.length; i++) {
            spaces[i] = readSpace(prototypes);
            if (spaces[i] != null) {
                spaces[i].setId(i);
            }
        }

        ConcurrentHashMap<Long, List<RootPtr>> rootPtrs = new ConcurrentHashMap<>();
        int noOfRoots;
        if (symbols.expectPointers) {
            noOfRoots = in.readInt();
            for (int i = 0; i < noOfRoots; i++) {
                RootPtr root = null;
                try {
                    root = RootPtr.fromMetadata(in);
                } catch (Exception e) {
                    e.printStackTrace();
                }

                List<RootPtr> existingRoots = rootPtrs.get(root.getAddr());
                if (existingRoots != null) {
                    existingRoots.add(root);
                    existingRoots.sort(Comparator.comparing(RootPtr::getRootType));
                } else {
                    existingRoots = new ArrayList<>();
                    existingRoots.add(root);
                    rootPtrs.put(root.getAddr(), existingRoots);
                }
            }
        }

        ConcurrentHashMap<Long, ThreadInfo> threads = new ConcurrentHashMap<>();
        int noOfThreads = in.readInt();
        for (int i = 0; i < noOfThreads; i++) {
            ThreadInfo t = new ThreadInfo(in.readLong(), in.readUTF(), in.readBoolean());
            threads.put(t.threadId, t);

            // rebuild callstack
            int callStackSize = in.readInt();
            for (int j = 0; j < callStackSize; j++) {
                t.addStackframe(in.readInt(), in.readInt());
            }
        }

        DetailedHeap heap = new DetailedHeap(symbols, cache, spaces, gcs, lastGCTime, rootPtrs, threads);
        // heap.getOld().setAgeRanges(ageRanges);
        return heap;
    }

    /*
     * private CopyOnWriteArrayList<Long> readAgeRanges() throws IOException { int size = in.readInt(); CopyOnWriteArrayList<Long> ageRanges
     * = new CopyOnWriteArrayList<Long>(); for (int i = 0; i < size; i++) { ageRanges.add(in.readLong()); }
     *
     * return ageRanges; }
     */

    private Map<Integer, ObjectInfo> readPrototypes(ObjectInfoCache cache) throws IOException {
        int count = in.readInt();
        Map<Integer, ObjectInfo> prototypes = new HashMap<>();
        for (int i = 0; i < count; i++) {
            int id = in.readInt();
            ObjectInfo prototype = readPrototype(cache);
            prototypes.put(id, prototype);
        }
        return prototypes;
    }

    private ObjectInfo readPrototype(ObjectInfoCache cache) throws IOException {
        int allocatedTypeId = in.readInt();
        int allocationSiteId = in.readInt();
        EventType type = EventType.parse(in.readInt());
        ObjectInfo prototype;
        switch (in.readInt()) {
            case 0:
                int size = in.readInt();
                prototype = InstanceMirrorObjectInfo.get(cache, type, symbols.sites.getById(allocationSiteId), symbols.types.get(allocatedTypeId), size, symbols);
                break;
            case 1:
                prototype = InstanceObjectInfo.get(cache, type, symbols.sites.getById(allocationSiteId), symbols.types.get(allocatedTypeId), symbols);
                break;
            case 2:
                int length = in.readInt();
                prototype = ArrayObjectInfo.get(cache, type, symbols.sites.getById(allocationSiteId), symbols.types.get(allocatedTypeId), length, symbols);
                break;
            default:
                throw new IOException("Unknown object prototype");
        }
        return prototype;
    }

    private Space readSpace(Map<Integer, ObjectInfo> prototypes) throws IOException {
        if (in.readByte() == 0) {
            return null;
        }

        String name = in.readUTF();
        Space space = new Space(name);
        long addr = in.readLong();
        long length = in.readLong();
        space.resetAddressAndLength(addr, length);
        int spaceType = in.readInt();
        if (spaceType >= 0) {
            space.setType(SpaceType.values()[spaceType]);
        }
        int spaceMode = in.readInt();
        if (spaceMode >= 0) {
            space.setMode(SpaceMode.values()[spaceMode]);
        }
        int count = in.readInt();
        for (int i = 0; i < count; i++) {
            Lab lab = readLab(prototypes);
            try {
                space.assignLab(lab);
            } catch (TraceException e) {
                throw new IOException(e);
            }
        }
        return space;
    }

    private Lab readLab(Map<Integer, ObjectInfo> prototypes) throws IOException {
        String thread = in.readUTF();
        int kind = in.readInt();
        long addr = in.readLong();
        long capacity = in.readLong();
        int objectCount = in.readInt();
        if (objectCount == 1) {
            int id = in.readInt();
            ObjectInfo obj = prototypes.get(id);
            SingleObjectLab lab = new SingleObjectLab(thread, kind, addr, obj);
            if (symbols.expectPointers) {
                int pointerCount = in.readInt();
                long[] ptrs = new long[pointerCount];
                for (int ptrIdx = 0; ptrIdx < pointerCount; ptrIdx++) {
                    ptrs[ptrIdx] = in.readLong();
                }
                try {
                    lab.setObjectPointersAtIndex(0, ptrs);
                } catch (TraceException e) {
                    throw new IOException(e);
                }
            }
            return lab;
        } else {
            Lab lab = new MultiObjectLab(thread, kind, addr, capacity, symbols.expectPointers);
            for (int i = 0; i < objectCount; i++) {
                int id = in.readInt();
                ObjectInfo obj = prototypes.get(id);
                // add feature to obj?
                try {
                    long assignedAddr = lab.tryAllocate(obj, Consts.UNDEFINED_ADDR);
                    assert assignedAddr != Lab.OBJECT_NOT_ASSIGNED;
                    if (symbols.expectPointers) {
                        int pointerCount = in.readInt();
                        long[] ptrs = new long[pointerCount];
                        for (int ptrIdx = 0; ptrIdx < pointerCount; ptrIdx++) {
                            ptrs[ptrIdx] = in.readLong();
                        }
                        lab.setObjectPointersAtIndex(i, ptrs);
                    }
                } catch (TraceException e) {
                    throw new IOException(e);
                }
            }
            return lab;
        }
    }

    @Override
    public void close() throws IOException {
        in.close();
    }

}
