/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.heap.io;

import at.jku.mevss.anttracks.heap.DetailedHeap;
import at.jku.mevss.anttracks.heap.GarbageCollectionType;
import at.jku.mevss.anttracks.heap.ObjectVisitor;
import at.jku.mevss.anttracks.heap.labs.Lab;
import at.jku.mevss.anttracks.heap.objects.ArrayObjectInfo;
import at.jku.mevss.anttracks.heap.objects.InstanceMirrorObjectInfo;
import at.jku.mevss.anttracks.heap.objects.InstanceObjectInfo;
import at.jku.mevss.anttracks.heap.objects.ObjectInfo;
import at.jku.mevss.anttracks.heap.roots.RootPtr;
import at.jku.mevss.anttracks.heap.space.Space;
import at.jku.mevss.anttracks.heap.symbols.AllocationSites;
import at.jku.mevss.anttracks.parser.heap.ThreadInfo;
import at.jku.mevss.anttracks.parser.io.BaseFile;
import at.jku.mevss.anttracks.util.Consts;

import java.io.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.stream.Collectors;

public class HeapWriter implements AutoCloseable {

    public static final int VERSION = 0;

    private final DataOutputStream out;

    public HeapWriter(String path) throws IOException {
        out = new DataOutputStream(new BufferedOutputStream(BaseFile.openW(path)));
    }

    public HeapWriter(String path, int fileName) throws IOException {
        out = new DataOutputStream(new BufferedOutputStream(BaseFile.openW(path + File.separator + fileName)));
    }

    public void write(DetailedHeap heap, ExecutorService executor) throws IOException {
        if (heap == null) {
            throw new NullPointerException();
        }
        if (heap.getGC() != GarbageCollectionType.MUTATOR) {
            throw new IllegalArgumentException("Cannot write heap when a GC is in progress");
        }
        out.writeInt(Consts.HEAP_FILES_MAGIC_PREFIX);
        out.writeInt(VERSION);
        out.writeInt(heap.getGCNo());
        out.writeLong(heap.getLastGCTime());

        Map<ObjectInfo, Integer> prototypes = buildAndWrite(out, heap);

        boolean pointers = heap.getSymbols().expectPointers;

        Space[] spaces = heap.getSpacesCloned();
        out.writeInt(spaces.length);
        for (Space space : spaces) {
            writeSpace(heap, space, prototypes, pointers, executor);
        }

        if (pointers) {
            List<RootPtr> roots = heap.rootPtrs.values().stream().flatMap(x -> x.stream()).collect(Collectors.toList());
            out.writeInt(roots.size());
            for (RootPtr root : roots) {
                byte[] metadata = root.getMetadata();
                out.write(metadata);
            }
        }

        Collection<ThreadInfo> threads = heap.threads.values();
        out.writeInt(threads.size());
        for (ThreadInfo t : threads) {
            out.writeLong(t.threadId);
            out.writeUTF(t.threadName);
            out.writeBoolean(t.isAlive());

            // write callstack
            out.writeInt(t.getStackDepth());
            out.write(t.getCallstackMetadata());
        }
    }

    private Map<ObjectInfo, Integer> buildAndWrite(DataOutputStream out, DetailedHeap heap) throws IOException {
        Set<ObjectInfo> objects = new HashSet<>();
        heap.toObjectStream()
            .forEach(
                    (address, objectInfo, space, type, size, isArray, arrayLength, allocationSite, pointedFrom, pointsTo, eventType, rootPtrs) -> objects.add(objectInfo),
                    ObjectVisitor.Settings.NO_INFOS);

        Map<ObjectInfo, Integer> map = new HashMap<>();
        int size = objects.size();
        out.writeInt(size);

        int id = 0;
        for (ObjectInfo obj : objects) {
            map.put(obj, id);
            out.writeInt(id++);
            write(obj, heap.getSymbols().sites);
        }

        return map;
    }

    private void write(ObjectInfo prototype, AllocationSites sites) throws IOException {
        out.writeInt(prototype.type.id);
        out.writeInt(sites.getById(prototype.allocationSite.id).getOriginalID());
        out.writeInt(prototype.eventType.id);
        if (prototype instanceof InstanceMirrorObjectInfo) {
            out.writeInt(0);
            out.writeInt(prototype.size);
        } else if (prototype instanceof InstanceObjectInfo) {
            out.writeInt(1);
        } else if (prototype instanceof ArrayObjectInfo) {
            out.writeInt(2);
            out.writeInt(((ArrayObjectInfo) prototype).getLength());
        } else {
            assert false;
            return;
        }
    }

    private void writeSpace(DetailedHeap heap, Space space, Map<ObjectInfo, Integer> prototypes, boolean pointers, ExecutorService executor) throws IOException {

        if (space == null) {
            out.writeByte(0);
            return;
        }

        out.writeByte(1);
        out.writeUTF(space.getName());
        out.writeLong(space.getAddress());
        out.writeLong(space.getLength());
        out.writeInt(space.getType() != null ? space.getType().ordinal() : -1);
        out.writeInt(space.getMode() != null ? space.getMode().ordinal() : -1);
        Lab[] labs = space.getLabs();
        out.writeInt(labs.length);
        BlockingQueue<Future<ByteArrayOutputStream>> queue = new LinkedBlockingQueue<>();

        for (Lab lab : labs) {
            queue.add(executor.submit(() -> writeLab(heap, lab, prototypes, pointers)));
        }

        Future<ByteArrayOutputStream> f = null;
        try {
            while (!queue.isEmpty()) {
                f = queue.poll();
                if (f.isDone()) {
                    out.write(f.get().toByteArray());
                } else {
                    queue.add(f);
                }
            }
        } catch (InterruptedException | ExecutionException e) {
            throw new IOException(e);
        }
    }

    @SuppressWarnings("serial")
    private static class IORuntimeException extends RuntimeException {
        public IORuntimeException(IOException e) {
            super(e);
        }
    }

    private ByteArrayOutputStream writeLab(DetailedHeap heap, Lab lab, final Map<ObjectInfo, Integer> prototypes, boolean pointers) throws IOException {
        ByteArrayOutputStream bias = new ByteArrayOutputStream();
        DataOutputStream out = new DataOutputStream(bias);
        out.writeUTF(lab.thread);
        out.writeInt(lab.kind);
        out.writeLong(lab.addr);
        out.writeLong(lab.capacity());
        out.writeInt(lab.getObjectCount());
        try {
            lab.iterate(heap, null, null, (address, objectInfo, space, type, size, isArray, arrayLength, allocationSite, pointedFrom, pointsTo, eventType, rootPtrs) -> {
                try {
                    out.writeInt(prototypes.get(objectInfo));
                    if (pointers) {
                        if (pointsTo == null) {
                            out.writeInt(0);
                        } else {
                            out.writeInt(pointsTo.length);
                            for (int i = 0; i < pointsTo.length; i++) {
                                out.writeLong(pointsTo[i]);
                            }
                        }
                    }
                } catch (IOException e) {
                    throw new IORuntimeException(e);
                }
            }, new ObjectVisitor.Settings(false, true, false));
        } catch (IORuntimeException iore) {
            throw new IOException(iore);
        }

        return bias;
    }

    @Override
    public void close() throws IOException {
        out.close();
    }

}
