/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.heap.io;

import at.jku.mevss.anttracks.heap.DetailedHeap;
import at.jku.mevss.anttracks.heap.GarbageCollectionCause;
import at.jku.mevss.anttracks.heap.GarbageCollectionType;
import at.jku.mevss.anttracks.heap.HeapListener;
import at.jku.mevss.anttracks.heap.symbols.Symbols;
import at.jku.mevss.anttracks.parser.EventType;
import at.jku.mevss.anttracks.parser.io.BaseFile;
import at.jku.mevss.anttracks.util.Consts;
import at.jku.mevss.anttracks.util.FileUtil;
import at.jku.mevss.anttracks.util.StreamUtil;
import at.jku.mevss.anttracks.util.ZipFileUtil;

import java.io.*;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.logging.Level;
import java.util.logging.Logger;

public final class MetaDataWriterListener implements HeapListener {

    private final Logger LOGGER = Logger.getLogger(MetaDataWriterListener.class.getName());

    private final MetaDataWriterConfig config;
    private final Symbols symbols;

    private final HeapIndexWriter indexWriter;
    private final StatisticsWriter statisticsWriter;
    private final ExecutorService executors;

    private int dumps;

    public MetaDataWriterListener(MetaDataWriterConfig config, Symbols symbols) throws IOException {
        this.config = config;
        this.symbols = symbols;
        {
            File file = new File(config.path);
            if (file.exists()) {
                if (!file.isDirectory()) {
                    throw new IOException("Given path must be a directory! (" + config.path + ")");
                }
            } else {
                file.mkdir();
            }
        }

        indexWriter = new HeapIndexWriter(config.path);
        statisticsWriter = new StatisticsWriter(config.path);
        executors = Executors.newFixedThreadPool(Consts.AVAILABLE_PROCESSORS());

        if (symbols.featureFile != null) {
            try (InputStream in = FileUtil.openR(new File(symbols.featureFile))) {
                String path = config.path + File.separator + Consts.FEATURES_META_FILE;
                try (OutputStream out = ZipFileUtil.isZipFilePath(path) ? ZipFileUtil.openW(new File(path)) : FileUtil.openW(new File(path))) {
                    StreamUtil.copy(in, out);
                }
            }
        }

        dumps = 1;
    }

    /**
     * Event handling before the GC starts.
     */
    @Override
    public void phaseChanging(DetailedHeap heap,
                              GarbageCollectionType from,
                              GarbageCollectionType to,
                              GarbageCollectionCause cause,
                              boolean concurrent,
                              boolean failed,
                              long time,
                              long position,
                              long length,
                              boolean withinSelectedTimeWindow) {
        if (from == GarbageCollectionType.MUTATOR) {
            // Switching away from mutator = Switching into GC phase
            int id = heap.getGCNo() * 2;
            EventType event = EventType.GC_START;
            boolean writeHeap = false;
            handleHeapEvent(id, time, position, to, cause, event, concurrent, failed, heap, writeHeap);
        }
    }

    /**
     * Event handling after the GC ended.
     */
    @Override
    public void phaseChanged(DetailedHeap heap,
                             GarbageCollectionType from,
                             GarbageCollectionType to,
                             GarbageCollectionCause cause,
                             boolean concurrent,
                             boolean failed,
                             long time,
                             long position,
                             long length,
                             boolean withinSelectedTimeWindow) {
        if (to == GarbageCollectionType.MUTATOR) {
            // Switching into mutator phase
            int id = heap.getGCNo() * 2 + 1;
            EventType event = EventType.GC_END;
            double progress = 1.0 * position / length;
            boolean writeHeap = (!config.atMajorGCsOnly || from.isFull()) && dumps < (config.dumps * progress);
            handleHeapEvent(id, time, position, from, cause, event, concurrent, failed, heap, writeHeap);
        }
    }

    private void handleHeapEvent(int id,
                                 long time,
                                 long position,
                                 GarbageCollectionType gcIdentifier,
                                 GarbageCollectionCause cause,
                                 EventType event,
                                 boolean concurrent,
                                 boolean failed,
                                 DetailedHeap heap,
                                 boolean writeHeap) {
        try {
            indexWriter.write(position, gcIdentifier, event, cause, time, id, writeHeap);
            statisticsWriter.write(heap.statistics);
            if (writeHeap) {
                try (HeapWriter heapWriter = new HeapWriter(config.path, id)) {
                    heapWriter.write(heap, executors);
                }
                dumps++;
            }
            //if (writeHeap || heap.getGCNo() == 0) {
            // Fix AT-60: To reconstruct a heap state, we rely on the index entry to already be written
            // Flush every GC
            statisticsWriter.flush();
            indexWriter.flush();
            //}
        } catch (IOException e) {
            LOGGER.log(Level.SEVERE, "Unexpected error @ " + position, e);
        }
    }

    @Override
    public void close(DetailedHeap heap) {
        executors.shutdown();
        try {
            statisticsWriter.close();
        } catch (IOException e) {
            LOGGER.log(Level.SEVERE, "Unexpected error", e);
        }
        try {
            indexWriter.close();
        } catch (IOException e) {
            LOGGER.log(Level.SEVERE, "Unexpected error", e);
        }
        try (DataOutputStream out = new DataOutputStream(BaseFile.openW(config.path + File.separator + Consts.HEADERS_META_FILE))) {
            out.writeInt(symbols.header.length);
            for (int i = 0; i < symbols.header.length; i++) {
                out.writeInt(symbols.header[i]);
            }
        } catch (IOException e) {
            LOGGER.log(Level.SEVERE, "Unexpected error", e);
        }
    }
}