/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.heap.labs;

import at.jku.mevss.anttracks.classification.Filter;
import at.jku.mevss.anttracks.heap.DetailedHeap;
import at.jku.mevss.anttracks.heap.ObjectAccess;
import at.jku.mevss.anttracks.heap.ObjectVisitor;
import at.jku.mevss.anttracks.heap.objects.ObjectInfo;
import at.jku.mevss.anttracks.heap.space.Space;
import at.jku.mevss.anttracks.heap.space.SpaceInfo;
import at.jku.mevss.anttracks.parser.heap.pointer.PtrUpdateVisitor;
import at.jku.mevss.anttracks.util.TraceException;

import java.util.List;

public abstract class Lab implements Comparable<Lab>, ObjectAccess {

    public static int ADDR_GRANULARITY = 32; // TODO remove
    public static final int KIND_TLAB = 0;
    public static final int KIND_PLAB = 1;
    public static final int KIND_VIRTUAL = 2;
    public static final int KIND_REGION_VIRTUAL = 3;
    public static final int UNKNOWN_CAPACITY = -1;
    public static final int OBJECT_NOT_ASSIGNED = -1;
    public static final int OBJECT_NOT_FOUND = -1;
    public static final int OBJECT_PTR_DELETED = -2;
    public static final int UNSET_FORWARDING_ADDR = -2;
    public final String thread;
    public final int kind;
    public final long addr;

    protected Lab(String thread, int kind, long addr) {
        this.thread = thread;
        this.kind = kind;
        this.addr = addr;
    }

    @Override
    public abstract Lab clone();

    public long bottom() {
        return addr;
    }

    public long top() {
        return bottom() + position();
    }

    public long end() {
        return bottom() + capacity();
    }

    public abstract long capacity();

    public abstract long position();

    public abstract int getObjectCount();

    public abstract void iterate(DetailedHeap heap,
                                 SpaceInfo space,
                                 List<Filter> filter,
                                 ObjectVisitor visitor,
                                 ObjectVisitor.Settings visitorSettings);

    public abstract long tryAllocate(ObjectInfo obj, long objAddr) throws TraceException;

    public abstract void resetForwardingAddresses();

    public abstract void reduceSize();

    public abstract boolean mayContainPointers();

    public abstract ObjectInfo getObjectAtIndex(int index) throws TraceException;

    public abstract void setObjectPointersAtIndex(int index, long[] ptrs) throws TraceException;

    public abstract long[] getObjectPointersAtIndex(int index) throws TraceException;

    public abstract long[] getObjectFromPointersAtIndex(int index) throws TraceException;

    public abstract long getObjectForwardingAddrUnescapedAtIndex(int index) throws TraceException;

    public abstract long getObjectForwardingAddrEscapedAtIndex(int index) throws TraceException;

    public abstract long getObjectAddressAtIndex(int index) throws TraceException;

    public abstract int getAddressIndex(long addr) throws TraceException;

    public abstract void mark(int index) throws TraceException;

    public abstract boolean isMarked(int index) throws TraceException;

    public abstract void clearMarkings();

    public abstract void iterateUpdatePointer(PtrUpdateVisitor iterator, Space space);

    public void initMarkings() {}

    public void resetCapacity() {}

    public boolean isFull() {
        return position() == capacity();
    }

    public boolean isExtendable() {
        return false;
    }

    public Lab sublab(long bottom, long end) throws TraceException {
        if (bottom == end) {
            return null;
        }
        if (bottom() == bottom && end() == end) {
            return this;
        }
        if (!(bottom() <= bottom && bottom < end())) {
            throw new IllegalArgumentException(String.format("%,d e [%,d, %,d[", bottom, bottom(), end()));
        }
        if (!(bottom() <= end && end <= end())) {
            throw new IllegalArgumentException(String.format("%,d e [%,d, %,d]", end, bottom(), end()));
        }
        if (!(bottom < end)) {
            throw new IllegalArgumentException(String.format("%,d < %,d", bottom, end));
        }
        end = Math.min(end, top()); // necessary for non-full labs
        Lab sublab = new MultiObjectLab(thread, kind, bottom, end - bottom, mayContainPointers());
        long top = bottom;
        while (top < end) {
            ObjectInfo obj = getObject(top);
            long assigned = sublab.tryAllocate(obj, top);
            assert assigned != Lab.OBJECT_NOT_ASSIGNED;
            top += obj.size;
        }
        assert sublab.isFull();
        assert top == end;
        return sublab;
    }

    @Override
    public int compareTo(Lab lab) {
        long result = (addr - lab.addr);
        if (result < 0) {
            return -1;
        } else if (result > 0) {
            return 1;
        } else {
            return 0;
        }
    }

    @Override
    public String toString() {
        final StringBuilder string = new StringBuilder();
        string.append(String.format("LAB (%s) @ %,d - %,d (%,d / %,d)\n", kind, addr, addr + capacity(), position(), capacity()));

        iterate(null,
                null,
                null,
                (address, objectInfo, space, type, size, isArray, arrayLength, allocationSite, pointedFrom, pointsTo, eventType, rootPtrs) -> string.append(String.format(
                        " OBJ @ %,d - %,d (%,d)\n",
                        address,
                        address + size,
                        size)), ObjectVisitor.Settings.NO_INFOS);
        return string.toString();
    }

    public abstract void variableCapacity();
}