/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.heap.labs;

import at.jku.mevss.anttracks.classification.Filter;
import at.jku.mevss.anttracks.heap.DetailedHeap;
import at.jku.mevss.anttracks.heap.ObjectAccess;
import at.jku.mevss.anttracks.heap.ObjectVisitor;
import at.jku.mevss.anttracks.heap.objects.ArrayObjectInfo;
import at.jku.mevss.anttracks.heap.objects.ObjectInfo;
import at.jku.mevss.anttracks.heap.space.Space;
import at.jku.mevss.anttracks.heap.space.SpaceInfo;
import at.jku.mevss.anttracks.parser.heap.pointer.PtrUpdateVisitor;
import at.jku.mevss.anttracks.util.Consts;
import at.jku.mevss.anttracks.util.TraceException;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.logging.Logger;

public final class MultiObjectLab extends Lab {

    private static final Logger LOGGER = Logger.getLogger(MultiObjectLab.class.getName());

    private long[] addrIndex;
    private int lastIndex;

    private final ArrayList<ObjectInfo> objectsList;
    /**
     * This is not updated during heap parsing, but only set once at the end of heap parsing
     */
    private long[][] pointedFromList = null;
    private final ArrayList<long[]> pointersList;
    //private final int[] ageList;
    private long[] forwardedAddrList;

    private long capacity;
    private long position;

    private long[] markings;

    public MultiObjectLab(String thread, int kind, long addr, long capacity, boolean expectPointers) {
        this(thread, kind, addr, capacity, new long[1], new ArrayList<>((int) (capacity / Consts.AVERAGE_OBJECT_SIZE)), 0, 0, expectPointers);
    }

    public MultiObjectLab(String thread,
                          int kind,
                          long addr,
                          long capacity,
                          long[] addrIndex,
                          ArrayList<ObjectInfo> objectsList,
                          int lastIndex,
                          long position,
                          boolean expectPointers) {
        super(thread, kind, addr);
        this.capacity = capacity;

        if (expectPointers) {
            pointersList = new ArrayList<>((int) (capacity / Consts.AVERAGE_OBJECT_SIZE));
            forwardedAddrList = new long[1];
        } else {
            pointersList = null;
            forwardedAddrList = null;
        }
        this.objectsList = objectsList;
        //this.ageList = ageList;
        this.addrIndex = addrIndex;
        this.lastIndex = lastIndex;
        this.position = position;
    }

    @Override
    public long capacity() {
        return capacity != Lab.UNKNOWN_CAPACITY ? capacity : position;
    }

    @Override
    public long position() {
        return position;
    }

    @Override
    public boolean isFull() {
        return capacity != Lab.UNKNOWN_CAPACITY ? super.isFull() : false;
    }

    @Override
    public boolean isExtendable() {
        return capacity == Lab.UNKNOWN_CAPACITY;
    }

    @Override
    public void setFromPointers(long address, long[] ptrs) throws TraceException {
        ensureFromPointersInitialized();
        int index = getAddressIndex(address);
        if (index != Lab.OBJECT_NOT_FOUND) {
            if (pointedFromList != null) {
                pointedFromList[index] = ptrs;
            }
        } else {
            throw new TraceException(String.format("Could not set from-pointers, object not found at address: %,d", address));
        }
    }

    @Override
    public boolean ptrNeedAdjustment(long objAddr) throws TraceException {
        int index = getAddressIndex(objAddr);
        if (index != Lab.OBJECT_NOT_FOUND) {
            return (forwardedAddrList[index] & ObjectAccess.PTR_ADJUSTMENT_NEEDED_MASK) == 1 ? true : false;
        } else {
            throw new TraceException("Could not retrieve pointer-adjustment status, object not found at address: " + objAddr);
        }
    }

    @Override
    public void setPtrNeedAdjustment(long objAddr, boolean needAdjustment) throws TraceException {
        int index = getAddressIndex(objAddr);
        if (index != Lab.OBJECT_NOT_FOUND) {
            if (needAdjustment) {
                forwardedAddrList[index] = (forwardedAddrList[index] | ObjectAccess.PTR_ADJUSTMENT_NEEDED_MASK);
            } else {
                forwardedAddrList[index] = (forwardedAddrList[index] & ObjectAccess.PTR_ADJUSTMENT_NEEDED_RESET_MASK);
            }
        } else {
            throw new TraceException("Could not set pointer-adjustment status, object not found at address: " + objAddr);
        }
    }

    @Override
    public int getObjectCount() {
        return objectsList.size();
    }

    @Override
    public void iterate(DetailedHeap heap,
                        SpaceInfo space,
                        List<Filter> filter,
                        ObjectVisitor visitor,
                        ObjectVisitor.Settings visitorSettings) {
        long addr = this.addr;
        for (int i = 0; i < objectsList.size(); i++) {
            ObjectInfo obj = objectsList.get(i);
            long[] pointedFrom = null;
            long[] pointsTo = null;
            if (visitorSettings.pointerInfoNeeded) {
                try {
                    if (pointersList != null) {
                        pointsTo = pointersList.get(i);
                    }
                } catch (Exception ex) {
                    LOGGER.warning("Could not find pointers for object @ " + addr + "\n" + ex);
                }
                try {
                    if (pointedFromList != null) {
                        pointedFrom = pointedFromList[i];
                    }
                } catch (Exception ex) {
                    LOGGER.warning("Could not find from-pointers for object @ " + addr + "\n" + ex);
                }
            }
            boolean accept = true;
            if (filter != null) {
                for (Filter f : filter) {
                    try {
                        if (!f.classify(addr,
                                        obj,
                                        space,
                                        obj.type,
                                        obj.size,
                                        visitorSettings.arrayInfoNeeded ? obj instanceof ArrayObjectInfo : false,
                                        visitorSettings.arrayInfoNeeded ? (obj instanceof ArrayObjectInfo ? ((ArrayObjectInfo) obj).getLength() : -1) : -1,
                                        obj.allocationSite,
                                        pointedFrom,
                                        pointsTo,
                                        obj.eventType,
                                        visitorSettings.rootPointerInfoNeeded ? heap.rootPtrs.get(addr) : null)) {
                            accept = false;
                            break;
                        }
                    } catch (Exception e) {
                        e.printStackTrace();
                        accept = false;
                        break;
                    }
                }
            }
            if (accept) {
                visitor.visit(addr,
                              obj,
                              space,
                              obj.type,
                              obj.size,
                              obj instanceof ArrayObjectInfo,
                              (obj instanceof ArrayObjectInfo) ? ((ArrayObjectInfo) obj).getLength() : -1,
                              obj.allocationSite,
                              pointedFrom,
                              pointsTo,
                              obj.eventType,
                              heap == null ? null : heap.rootPtrs.get(addr));
            }
            addr += obj.size;
        }

    }

    @Override
    public long tryAllocate(ObjectInfo obj, long objAddr) throws TraceException {
        long assignedAddr = OBJECT_NOT_ASSIGNED;

        if (objAddr == Consts.UNDEFINED_ADDR) {
            assignedAddr = allocate(obj) + addr;
        } else if (isInLab(objAddr)) {
            errorOnWrongAddr(thread, objAddr, obj);
            assignedAddr = allocate(obj) + addr;
        }
        return assignedAddr;
    }

    private boolean isInLab(long objAddr) {
        if (capacity == Lab.UNKNOWN_CAPACITY) {
            return (objAddr == top());
        } else {
            return ((objAddr >= addr) && objAddr < (addr + capacity));
        }
    }

    private void trim() {
        addrIndex = Arrays.copyOf(addrIndex, lastIndex + 1);
    }

    private int searchTrueObjectIndex(long objAddr, int index) throws TraceException {

        ObjectInfo objectInfo;
        index = (index * -1) - 2; // insertion point

        if (index == addrIndex.length) {
            index = addrIndex.length - 1;
        }

        if (index == -1) {
            return -1;
        }

        long cursor = addrIndex[index];
        int trueIndex = index * ADDR_GRANULARITY;

        while (cursor < objAddr && trueIndex < objectsList.size()) {
            objectInfo = objectsList.get(trueIndex);
            cursor += objectInfo.size;
            trueIndex++;
        }

        if (cursor != objAddr) {
            trueIndex = Lab.OBJECT_NOT_FOUND;
            // might be filler
        }
        if (trueIndex >= (objectsList.size())) {
            // throw new TraceException("Object not in this lab, but the first
            // of the next one");
            trueIndex = Lab.OBJECT_NOT_FOUND;
        }

        return trueIndex;
    }

    private long allocate(ObjectInfo obj) throws TraceException {
        long assignedAddr = position;
        objectsList.add(obj);
        if (pointersList != null) {
            if (objectsList.size() >= forwardedAddrList.length) {
                int newSize = objectsList.size() + (objectsList.size() >> 1);
                if (newSize == objectsList.size()) {
                    newSize++;
                }
                forwardedAddrList = Arrays.copyOf(forwardedAddrList, newSize);
            }
            forwardedAddrList[objectsList.size() - 1] = UNSET_FORWARDING_ADDR;
            pointersList.add(null);
        }
        if ((objectsList.size() - 1) % ADDR_GRANULARITY == 0) {
            putPositionToAddrIndex();
        }
        errorOnNegativeObjectSize(obj.size);
        position += obj.size;
        errorOnLabOverflow(assignedAddr + addr, obj);

        return assignedAddr;
    }

    private void putPositionToAddrIndex() {
        int index = (objectsList.size() - 1) / ADDR_GRANULARITY;
        {
            int newSize = addrIndex.length;
            while (index >= newSize) {
                newSize *= 2;
            }
            if (newSize > addrIndex.length) {
                addrIndex = Arrays.copyOf(addrIndex, newSize);
            }
        }
        addrIndex[index] = position;
        lastIndex = index;
    }

    @Override
    public void resetCapacity() {
        assert (capacity == Lab.UNKNOWN_CAPACITY) : "Lab Capacity must not be set if capacity should be reset";
        capacity = position;
    }

    private void errorOnNegativeObjectSize(int size) throws TraceException {
        if (size <= 0) {
            throw new TraceException("#CalculateAddrByTLAB size of object must be positive but is " + size);
        }
    }

    private void errorOnLabOverflow(long assignedAddr, ObjectInfo obj) throws TraceException {
        if (capacity != Lab.UNKNOWN_CAPACITY && position > capacity) {
            throw new TraceException(String.format("LAB CAPACITY EXCEEDED\n" + "Object added to lab of kind %s of thread %s (Range = %,d - %,d, Size = %,d) which " +
                                                           "exceeded the " +
                                                           "LAB's size:\n" + "obj addr = %,d (Range = %,d - %,d, Size = %,d, Object = %s)\n" + "capacity = %,d, " +
                                                           "position now = %,d",
                                                   kind,
                                                   thread,
                                                   bottom(),
                                                   end(),
                                                   capacity(),
                                                   assignedAddr,
                                                   assignedAddr,
                                                   (assignedAddr + obj.size),
                                                   obj.size,
                                                   obj,
                                                   capacity,
                                                   position));
        }
    }

    private void errorOnWrongAddr(String thread, long objAddr, ObjectInfo objInfo) throws TraceException {
        if (objAddr != top()) {
            long expectedAddr = top();
            long offset = objAddr - expectedAddr;
            throw new TraceException(String.format("#CalculateAddrByTLAB, Lab of kind %s of thread %s (Range = %,d - %,d, Size = %,d) is not consecutive:\n" + "obj " +
                                                           "actual addr =" +
                                                           " %,d (Range = %,d - %,d, Size = %,d, Object = %s),\n" + "expected addr = %,d + %,d = %,d\n" + "offset = %,d",
                                                   kind,
                                                   thread,
                                                   bottom(),
                                                   end(),
                                                   capacity(),
                                                   objAddr,
                                                   objAddr,
                                                   (objAddr + objInfo.size),
                                                   objInfo.size,
                                                   objInfo,
                                                   addr,
                                                   position,
                                                   expectedAddr,
                                                   offset));
        }
    }

    @Override
    public void resetForwardingAddresses() {
        for (int i = 0; i < forwardedAddrList.length; i++) {
            forwardedAddrList[i] = UNSET_FORWARDING_ADDR;
        }

    }

    @SuppressWarnings("unchecked")
    @Override
    public Lab clone() {
        return new MultiObjectLab(thread,
                                  kind,
                                  addr,
                                  capacity,
                                  addrIndex.clone(),
                                  (ArrayList<ObjectInfo>) objectsList.clone(),
                                  lastIndex,
                                  position,
                                  pointersList != null);
    }

    @Override
    public void reduceSize() {
        objectsList.trimToSize();
        if (pointersList != null) {
            pointersList.trimToSize();
            assert objectsList.size() == pointersList.size() : "All lists must be the same size";
        }
        if (forwardedAddrList != null) {
            // Forward address array no longer needed
            forwardedAddrList = new long[0];
        }
    }

    @Override
    public boolean mayContainPointers() {
        return pointersList != null;
    }

    @Override
    public ObjectInfo getObject(long objAddr) throws TraceException {
        long relAddr = objAddr - addr;

        ObjectInfo objectInfo = null;
        if (lastIndex != addrIndex.length - 1) {
            trim();
        }

        int index = Arrays.binarySearch(addrIndex, relAddr);
        int trueIndex = index * ADDR_GRANULARITY;
        if (index < 0) {
            trueIndex = searchTrueObjectIndex(relAddr, index);
        }

        if (trueIndex != Lab.OBJECT_NOT_FOUND) {
            objectInfo = getObjectAtIndex(trueIndex);
        }
        return objectInfo;
    }

    @Override
    public long getObjectAddressAtIndex(int index) throws TraceException {
        long objAddr = addr;
        for (int i = 0; i < index; i++) {
            objAddr += getObjectAtIndex(i).size;
        }
        return objAddr;
    }

    @Override
    public ObjectInfo getObjectAtIndex(int index) {
        return objectsList.get(index);
    }

    @Override
    public void setObjectPointersAtIndex(int index, long[] ptrs) throws TraceException {
        if (index >= 0 && index < getObjectCount()) {
            if (pointersList != null) {
                pointersList.set(index, ptrs);
            }
        } else {
            throw new TraceException(String.format("Could not set new pointers, index not found in MultiObjectLab: %,d", index));
        }
    }

	/*
    @Override
	public ObjectInfoAge getObjectAge(long objAddr) throws TraceException {
		return getObjectAgeAtIndex(getAddressIndex(objAddr));
	}

	@Override
	public ObjectInfoAge getObjectAgeAtIndex(int index) throws TraceException {
		ObjectInfo obj = getObjectAtIndex(index);
		return new ObjectInfoAge(obj, ageList[index]);
	}
	*/

    @Override
    public long[] getObjectPointers(long objAddr) throws TraceException {
        int index;
        index = getAddressIndex(objAddr);
        return getObjectPointersAtIndex(index);
    }

    @Override
    public long[] getObjectPointersAtIndex(int index) throws TraceException {
        if (pointersList == null || index == Lab.OBJECT_NOT_FOUND) {
            return null;
        } else {
            return pointersList.get(index);
        }
    }

    @Override
    public long[] getObjectFromPointers(long objAddr) throws TraceException {
        int index;
        index = getAddressIndex(objAddr);
        return getObjectFromPointersAtIndex(index);
    }

    @Override
    public long[] getObjectFromPointersAtIndex(int index) throws TraceException {
        if (pointedFromList == null || index == Lab.OBJECT_NOT_FOUND) {
            return null;
        } else {
            return pointedFromList[index];
        }
    }

    @Override
    public long getObjectForwardingAddressUnescaped(long objAddr) throws TraceException {
        int index = getAddressIndex(objAddr);
        if (index != Lab.OBJECT_NOT_FOUND) {
            return getObjectForwardingAddrUnescapedAtIndex(index);
        } else {
            throw new TraceException(String.format("Cannot find object at address: %,d", objAddr));
        }
    }

    @Override
    public long getObjectForwardingAddressEscaped(long objAddr) throws TraceException {
        int index = getAddressIndex(objAddr);
        if (index != Lab.OBJECT_NOT_FOUND) {
            return getObjectForwardingAddrEscapedAtIndex(index);
        } else {
            throw new TraceException(String.format("Cannot find object at address: %,d", objAddr));
        }
    }

    @Override
    public long getObjectForwardingAddrUnescapedAtIndex(int index) throws TraceException {
        if (index != Lab.OBJECT_NOT_FOUND) {
            return forwardedAddrList[index];
        } else {
            throw new TraceException(String.format("Invalid index for forwarding address access: %,d", index));
        }
    }

    @Override
    public long getObjectForwardingAddrEscapedAtIndex(int index) throws TraceException {
        if (index != Lab.OBJECT_NOT_FOUND) {
            return forwardedAddrList[index] & ObjectAccess.PTR_ADJUSTMENT_NEEDED_RESET_MASK;
        } else {
            throw new TraceException(String.format("Invalid index for forwarding address access: %,d", index));
        }
    }

    @Override
    public int getAddressIndex(long objAddr) throws TraceException {
        long relAddr = objAddr - addr;
        if (lastIndex != addrIndex.length - 1) {
            trim();
        }
        int index = Arrays.binarySearch(addrIndex, relAddr);
        int trueIndex = index * ADDR_GRANULARITY;
        if (index < 0) {
            trueIndex = searchTrueObjectIndex(relAddr, index);
        }
        return trueIndex;
    }

    @Override
    public void setObjectPointers(long objAddr, long[] ptrs) throws TraceException {
        int index = getAddressIndex(objAddr);
        if (index != Lab.OBJECT_NOT_FOUND) {
            if (pointersList != null) {
                pointersList.set(index, ptrs);
            }
        } else {
            throw new TraceException(String.format("Could not set new Pointers, object not found at address: %,d", objAddr));
        }
    }

    @Override
    public void setObjectForwardingAddr(long objAddr, long forwAddr) throws TraceException {
        int index = getAddressIndex(objAddr);
        if (index != Lab.OBJECT_NOT_FOUND) {
            forwardedAddrList[index] = forwAddr;
        } else {
            throw new TraceException(String.format("Could not set new Pointer, object not found at address: %,d", objAddr));
        }
    }

    @Override
    public void setObjectForwardingAddrAndPtrs(long objAddr, long forwAddr, long[] ptrs) throws TraceException {
        int index = getAddressIndex(objAddr);
        if (index != Lab.OBJECT_NOT_FOUND) {
            pointersList.set(index, ptrs);
            forwardedAddrList[index] = forwAddr;
        } else {
            throw new TraceException(String.format("Could not set pointers / forwarding addr, object not found at address: %,d", objAddr));
        }
    }

	/*
    @Override
	public void setAge(long objAddr, int age) throws TraceException {
		int index = getAddressIndex(objAddr);
		if (index != Lab.OBJECT_NOT_FOUND) {
			ageList[index] = age + 1;
		} else {
			assert (false);
		}
	}
	*/

    @Override
    public void setFromPointerCount(long objAddr, int n) throws TraceException {
        ensureFromPointersInitialized();
        int index = getAddressIndex(objAddr);
        if (index != Lab.OBJECT_NOT_FOUND) {
            if (pointedFromList != null) {
                long[] fromPointers = new long[n];
                Arrays.fill(fromPointers, -1);
                pointedFromList[index] = fromPointers;
            }
        } else {
            throw new TraceException(String.format("Could not init from-pointers, object not found at address: %,d", objAddr));
        }
    }

    private void ensureFromPointersInitialized() {
        if (pointedFromList == null) {
            pointedFromList = new long[objectsList.size()][];
        }
    }

    @Override
    public void initMarkings() {
        // need one long per 64 objects
        markings = new long[getObjectCount() / Long.SIZE + 1];
    }

    @Override
    public void mark(int index) throws TraceException {
        if (index >= getObjectCount() || index < 0) {
            throw new TraceException("Given index exceeds object count of lab");
        }
        // find long containing object bit and set it to 1
        markings[index / Long.SIZE] |= 1L << index % Long.SIZE;
    }

    @Override
    public void mark(long objAddr) throws TraceException {
        int index = getAddressIndex(objAddr);
        if (index != Lab.OBJECT_NOT_FOUND) {
            mark(index);
        } else {
            throw new TraceException("Object not found at address: " + objAddr);
        }
    }

    @Override
    public boolean isMarked(int index) throws TraceException {
        if (index >= getObjectCount() || index < 0) {
            throw new TraceException("Given index exceeds object count of lab");
        }
        // find long containing bit and check whether it is set to 1
        return (markings[index / Long.SIZE] & (1L << index % Long.SIZE)) != 0;
    }

    @Override
    public boolean isMarked(long objAddr) throws TraceException {
        int index = getAddressIndex(objAddr);
        if (index != Lab.OBJECT_NOT_FOUND) {
            return isMarked(index);
        } else {
            throw new TraceException("Object not found at address: " + objAddr);
        }
    }

    @Override
    public void clearMarkings() {
        for (int i = 0; i < markings.length; i++) {
            markings[i] = 0L;
        }
    }

    @Override
    public void iterateUpdatePointer(PtrUpdateVisitor iterator, Space space) {
        long addr = this.addr;
        for (int i = 0; i < objectsList.size(); i++) {
            ObjectInfo obj = objectsList.get(i);
            long[] pointsTo = null;
            try {
                if (pointersList != null) {
                    pointsTo = pointersList.get(i);
                }
            } catch (Exception ex) {
                LOGGER.warning("Could not find pointers for object @ " + addr + "\n" + ex);
            }
            iterator.visit(space, this, addr, pointsTo);
            iterator.visit(space, this, addr, pointsTo);
            addr += obj.size;
        }
    }

    public void variableCapacity() {
        capacity = Lab.UNKNOWN_CAPACITY;
    }
}
