package at.jku.mevss.anttracks.heap.roots;

import at.jku.mevss.anttracks.heap.DetailedHeap;
import at.jku.mevss.anttracks.heap.objects.ObjectInfo;
import at.jku.mevss.anttracks.heap.symbols.AllocatedType;
import at.jku.mevss.anttracks.parser.heap.ThreadInfo;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class LocalVariableRoot extends RootPtr {
    public final long threadId;
    public final int classId;
    public final int methodId;
    public final int slot;
    private ThreadInfo thread;
    private AllocatedType klass;
    private AllocatedType.MethodInfo method;
    private AllocatedType localType;
    private String localName;

    private int callStackIndex;

    public LocalVariableRoot(long addr, long threadId, int classId, int methodId, int slot) {
        super(addr);
        this.threadId = threadId;
        this.classId = classId;
        this.methodId = methodId;
        this.slot = slot;
    }

    public LocalVariableRoot(long addr, long threadId, int classId, int methodId, int slot, int callStackIndex) {
        this(addr, threadId, classId, methodId, slot);

        this.callStackIndex = callStackIndex;
    }

    public String toString() {
        StringBuilder b = new StringBuilder("Local variable:\n");
        b.append("- ").append(thread != null ? thread.threadName : "???").append(" (").append(threadId).append(")\n");
        b.append("-- ").append(klass != null ? klass.name : "???").append(" (").append(classId).append(")\n");
        b.append("--- ").append(method != null ? method.name : "???").append(" (").append(methodId).append(")\n");
        b.append("---- ")
         .append(localType != null ? localType.getExternalName(true, false) : "???")
         .append(localName != null ? localName : "???")
         .append(" (#")
         .append(slot)
         .append(")\n");

        return b.toString();
    }

    @Override
    public String toShortString() {
        return getClass().getSimpleName() + " " +
                (localName != null ? localName : "??? (#" + slot + ")") +
                " in " + (klass != null ? klass.getExternalName(true, false) : "???") +
                "::" + (method != null ? method.name : "???") +
                " running in " + (thread != null ? thread.threadName : "???");
    }

    @Override
    public byte[] getMetadata() {
        return ByteBuffer.allocate(1 + 2 * Long.BYTES + 3 * Integer.BYTES + 1 * Integer.BYTES)
                         .put(RootType.LOCAL_VARIABLE_ROOT.byteVal)
                         .putLong(addr)
                         .putLong(threadId)
                         .putInt(classId)
                         .putInt(methodId)
                         .putInt(slot)

                         .putInt(callStackIndex)
                         .array();
    }

    @Override
    public RootType getRootType() {
        return RootType.LOCAL_VARIABLE_ROOT;
    }

    public void setCallStackIndex(int callStackIndex) {
        this.callStackIndex = callStackIndex;
    }

    @Override
    public void resolve(DetailedHeap heap) throws Exception {
        if (addr != -1) {
            ObjectInfo obj = heap.getObject(addr);
            if (obj != null) {
                localType = obj.type;
            }
        }
        thread = heap.threads.get(threadId);
        if (thread != null) {
            klass = heap.getSymbols().types.stream().filter(c -> c.id == this.classId).findFirst().orElse(null);
            if (klass != null && klass.methodInfos != null) {
                method = Arrays.stream(klass.methodInfos).filter(mi -> mi.idnum == this.methodId).findFirst().orElse(null);
                if (method != null) {
                    if (method.locals.containsKey(slot)) {
                        localName = method.locals.get(slot);
                    }
                    // it's fine if slot can't be resolved (data only available for user defined methods)
                    resolved = true;
                    return;
                }
            }
        }

        //throw new Exception("Could not resolve method local root: " + toString());
    }

    @Override
    public String[] hierarchy() {
        return hierarchy(false);
    }

    public String[] hierarchy(boolean includeCallStack) {
        String[] ret;
        String[] callStack;

        if (includeCallStack) {
            callStack = callStack();

            ret = new String[3 + callStack.length];

            for (int i = 0; i < callStack.length; i++) {
                ret[i + 2] = callStack[i];
            }
        } else {
            ret = new String[4];
            ret[2] = "Method: " + (klass != null ? klass.getExternalName(false, false) : "???") + " :: " + (method != null ? method.name : "???");
        }

        ret[0] = "Root: Local variable";
        ret[1] = "Thread: " + (thread != null ? thread.threadName : "???");
        ret[ret.length - 1] = "Variable: " + (localType != null ? localType.getExternalName(true, false) : "???") + " " + (localName != null ?
                                                                                                                           localName :
                                                                                                                           "???") + " (Slot #" + slot + ")";

        return ret;
    }

    private String[] callStack() {
        // returns an array of method names representing the stack frame starting at the given root
        List<String> frames = new ArrayList<>();
        for (int i = callStackIndex; i >= 0 && i < thread.getCallstack().size(); i = i + 1) {
            frames.add(0, "Method: " + (thread.getCallstack().get(i).type.getExternalName(false, false)) + " :: " + (thread.getCallstack().get(i).name));
        }
        return frames.toArray(new String[frames.size()]);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }

        if (!LocalVariableRoot.class.isAssignableFrom(obj.getClass())) {
            return false;
        }

        final LocalVariableRoot other = (LocalVariableRoot) obj;

        return this.addr == other.addr && this.threadId == other.threadId && this.classId == other.classId && this.methodId == other.methodId && this.slot == other.slot;
    }

    @Override
    public boolean isInternal() {
        if (!resolved) {
            throw new IllegalStateException("Must resolve root ptr before calling isInternal()!");
        }

        if (thread == null) {
            throw new IllegalStateException("Root ptr has unknown thread!");
        }

        return thread.threadName.equals("Reference Handler") || thread.threadName.equals("Finalizer");
    }
}
