package at.jku.mevss.anttracks.heap.roots;

import at.jku.mevss.anttracks.heap.DetailedHeap;
import at.jku.mevss.anttracks.heap.symbols.AllocatedType;

import java.nio.ByteBuffer;
import java.util.Arrays;

public class StaticFieldRoot extends RootPtr {
    public final int classId;
    public final int offset;
    private AllocatedType klass;
    private AllocatedType type;
    private String name;

    public StaticFieldRoot(long addr, int classId, int offset) {
        super(addr);
        this.classId = classId;
        this.offset = offset;
    }

    public String toString() {
        StringBuilder b = new StringBuilder("Static field:\n");
        b.append("- ").append(klass != null ? klass.name : "???").append(" (").append(classId).append(")\n");
        b.append("-- ").append(type != null ? type.name : "???").append(" ").append(name != null ? name : "???").append(" (").append(offset).append(")\n");

        return b.toString();
    }

    @Override
    public String toShortString() {
        return getClass().getSimpleName() + " " + (name != null ? name : "???") + " in " + (klass != null ? klass.getExternalName(true, false) : "???");
    }

    @Override
    public byte[] getMetadata() {
        return ByteBuffer.allocate(1 + Long.BYTES + 2 * Integer.BYTES).put(RootType.STATIC_FIELD_ROOT.byteVal).putLong(addr).putInt(classId).putInt(offset).array();
    }

    @Override
    public RootType getRootType() {
        return RootType.STATIC_FIELD_ROOT;
    }

    @Override
    public void resolve(DetailedHeap heap) throws Exception {
        klass = heap.getSymbols().types.stream().filter(c -> c.id == this.classId).findFirst().orElse(null);
        if (klass != null) {
            AllocatedType.FieldInfo field = Arrays.stream(klass.getAllFieldInfos()).filter(f -> f.offset == this.offset).findFirst().orElse(null);
            if (field != null) {
                type = heap.getSymbols().types.stream().filter(c -> c.id == field.getTypeId()).findFirst().orElse(null);
                name = field.name;
                resolved = true;
                return;
            }
        }

        //throw new Exception("Could not resolve static root: " + toString());
    }

    @Override
    public String[] hierarchy() {
        String[] ret = {"Root: Static field",
                        "Class: " + (klass != null ? klass.getExternalName(false, false) : "???"),
                        "Field: " + (type != null ? type.getExternalName(false, false) : "???") + " " + (name != null ? name : "???")};
        return ret;
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }

        if (!StaticFieldRoot.class.isAssignableFrom(obj.getClass())) {
            return false;
        }

        final StaticFieldRoot other = (StaticFieldRoot) obj;

        return this.addr == other.addr && this.classId == other.classId && this.offset == other.offset;
    }

    @Override
    public boolean isInternal() {
        if (!resolved) {
            throw new IllegalStateException("Must resolve root ptr before calling isInternal()!");
        }

        if (klass == null) {
            throw new IllegalStateException("Root ptr has unknown class!");
        }

        String klassName = klass.getExternalName(false, false);
        return klassName.contains("java.lang.ClassLoader") || klassName.contains("sun.launcher.LauncherHelper") || klassName.contains("sun.misc.Launcher") || klass
                .isSubtypeOf(
                        "java.lang.ClassLoader");
    }
}
