/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.heap.symbols;

import java.util.*;
import java.util.logging.Logger;
import java.util.stream.Stream;

public class AllocatedTypes implements Iterable<AllocatedType> {

    private static final Logger LOGGER = Logger.getLogger(AllocatedTypes.class.getSimpleName());

    public static final String MIRROR_CLASS_NAME = "Ljava/lang/Class;";
    public static final String MIRROR_CLASS_EXTERNAL_NAME = "java.lang.Class";

    private HashMap<Integer, AllocatedType> allocatedTypes;
    // Starts at 1 since "UNKNOWN" always exists
    private int count = 1;
    private AllocatedType classKlass = null;

    public AllocatedTypes() {
        allocatedTypes = new HashMap<>();
        allocatedTypes.put(AllocatedType.ALLOCATED_TYPE_IDENTIFIER_UNKNOWN,
                           new AllocatedType(AllocatedType.ALLOCATED_TYPE_IDENTIFIER_UNKNOWN,
                                             AllocatedType.ALLOCATED_TYPE_IDENTIFIER_UNKNOWN,
                                             String.valueOf(AllocatedType.ALLOCATED_TYPE_IDENTIFIER_UNKNOWN_SIG),
                                             0));
    }

    public AllocatedType add(int id, AllocatedType allocatedType) throws SymbolsFileException {
        if (allocatedTypes.get(id) == null) {
            count++;
            allocatedTypes.put(id, allocatedType);
            if (classKlass == null && allocatedType.name.equals(MIRROR_CLASS_NAME)) {
                // Remember Class<> class
                classKlass = allocatedType;
            }
        } else {
            throw new SymbolsFileException(String.format("Duplicate entry in symbols file for type ID %d, Old Type: %s, Duplicate Type: %s",
                                                         id,
                                                         allocatedType,
                                                         allocatedTypes.get(id)));
        }
        return allocatedType;
    }

    public void complete(Symbols symbols) {
        assert classKlass != null : "Class class has to exist after symbols parsing";
        // Convert class type into mirror class
        classKlass = new AllocatedType.MirrorAllocatedType(classKlass, classKlass);
        allocatedTypes.put(classKlass.id, classKlass);
        classKlass.complete(symbols);
        ((AllocatedType.MirrorAllocatedType) classKlass).mirrorKlass.complete(symbols);

        allocatedTypes.put(AllocatedType.ALLOCATED_TYPE_IDENTIFIER_UNKNOWN_MIRROR, new AllocatedType.MirrorAllocatedType(classKlass, null));

        HashMap<Integer, AllocatedType> mirrorTypes = new HashMap<>();
        for (Map.Entry<Integer, AllocatedType> entry : allocatedTypes.entrySet()) {
            entry.getValue().complete(symbols);

            if (entry.getKey() != AllocatedType.ALLOCATED_TYPE_IDENTIFIER_UNKNOWN && entry.getValue() != classKlass) {
                AllocatedType.MirrorAllocatedType mirrorType = new AllocatedType.MirrorAllocatedType(classKlass, entry.getValue());
                mirrorTypes.put(-entry.getKey(), mirrorType);
            }
        }

        for (Map.Entry<Integer, AllocatedType> entry : mirrorTypes.entrySet()) {
            allocatedTypes.put(entry.getKey(), entry.getValue());
        }

        fixFields(symbols);

        // resolve data structure types
        symbols.getDataStructureDefinitions().resolveDataStructureParts(this);
    }

    private void fixFields(Symbols symbols) {
        for (AllocatedType type : symbols.types) {
            for (AllocatedType.FieldInfo field : type.getAllFieldInfos()) {
                assert field != null : "Field info may not be null";
            }
        }

        List<String> nonFoundTypeMessages = new ArrayList<>();

        for (AllocatedType type : symbols.types) {
            for (AllocatedType.FieldInfo field : type.getAllFieldInfos()) {
                int typeByName = symbols.types.stream().filter(t -> t.name.equals(field.signature)).map(t -> t.id).findFirst().orElse(-1);
                if (typeByName < 0) {
                    nonFoundTypeMessages.add("Found field of type '" + field.signature + "' but no matching symbols entry. This may be because this type has " +
                                                     "never been loaded (i.e., not " + "instantiated once)");
                }
                field.setTypeId(typeByName);
            }
        }

        nonFoundTypeMessages.stream().distinct().forEach(LOGGER::info);
    }

    public AllocatedType get(int id) {
        return allocatedTypes.get(id);
    }

    public int getLength() {
        return count;
    }

    @Override
    public Iterator<AllocatedType> iterator() {
        return allocatedTypes.values().iterator();
    }

    public Stream<AllocatedType> stream() {
        return allocatedTypes.values().stream();
    }
}
