/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.parser;

import at.jku.mevss.anttracks.util.TraceException;

import java.io.Closeable;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.ArrayList;
import java.util.List;

public abstract class Scanner implements Closeable {
    private final static char[] signatures = {'Z', 'B', 'C', 'S', 'I', 'J', 'F', 'D', 'L', 'V', '['};
    public final static int META_BUFFER_INFO = 12;

    private final long globalOffset;

    protected Scanner(long globalOffset) {
        this.globalOffset = globalOffset;
    }

    public long getGlobalPosition() {
        return globalOffset + getPosition();
    }

    public abstract long getPosition();

    public abstract void skip(int length) throws IOException;

    public abstract boolean isAvailable(int length);

    public abstract long getLong() throws IOException;

    public abstract int getInt() throws IOException;

    public abstract short getShort() throws IOException;

    public abstract byte getByte() throws IOException;

    public abstract byte[] get(int length) throws IOException;

    public int getWord() throws IOException {
        return getInt();
    }

    public char getChar() throws IOException {
        return (char) getByte();
    }

    /*
     * Methods used by TraceFile and SymbolFile
     */

    public int[] getHeader() throws IOException {
        int magic = getInt();
        if (magic != 0xC0FFEE) {
            throw new IOException("Cannot find magic sequence!");
        }
        int[] header = new int[getInt()];
        for (int i = 0; i < header.length; i++) {
            header[i] = getInt();
        }
        return header;
    }

    /*
     * Methods used by TraceFile only
     */

    public String getThread(int heapWordSize) throws IOException {
        assert heapWordSize == 8 : "Adjust buffer (heapWordSize != 8)";

        if (isAvailable(8)) {
            return "0x" + Long.toHexString(getLong());
        } else {
            return null;
        }
    }

    public ByteBuffer getBuffer(int length) throws IOException, TraceException {
        byte[] arr = get(length);
        assert arr.length == length;
        ByteBuffer buffer = ByteBuffer.wrap(arr);
        assert buffer.limit() == arr.length;
        buffer.order(ByteOrder.LITTLE_ENDIAN);

        return buffer;
    }

    /*
     * Methods used by SymbolFile Only
     */

    public int getMagicByte() throws IOException {
        if (isAvailable(1)) {
            return getByte();
        } else {
            return 0;
        }
    }

    public String getNullTerminatedString(boolean removeSeperator) throws IOException, TraceException {
        ArrayList<Byte> signature = new ArrayList<Byte>();
        byte currentByte = getByte();

        addNextByteToSignature(' ', signature, currentByte);
        if (removeSeperator) {
            signature.remove(signature.size() - 1);
        }
        return byteListToString(signature);
    }

    public String getString(char separator) throws IOException, TraceException {
        ArrayList<Byte> signature = new ArrayList<Byte>();
        byte currentByte = getByte();

        if (separator == ')') {
            addNextByteToSignature(separator, signature, currentByte);
            currentByte = getByte();
        }

        reconstructType(signature, currentByte);
        return byteListToString(signature);
    }

    public String getStringZeroTerminated() throws IOException {
        StringBuilder string = new StringBuilder();
        for (char c = getChar(); c != '\0'; c = getChar()) {
            string.append(c);
        }
        return string.toString();
    }

    private boolean signaturesContain(byte currentByte) {
        for (int i = 0; i < signatures.length; i++) {
            if (currentByte == signatures[i]) {
                return true;
            }
        }
        return false;
    }

    private void reconstructType(ArrayList<Byte> signature, byte currentByte) throws IOException, TraceException {
        if (signaturesContain(currentByte)) {
            if (currentByte == '[') {
                signature.add(currentByte); // add [
                currentByte = getByte();
                reconstructType(signature, currentByte);
            } else if (currentByte == 'L') {
                addNextByteToSignature(';', signature, currentByte);
            } else {
                signature.add(currentByte); // add elementary type
            }
        } else {
            throw new TraceException("Invalid symbol file.");
        }

    }

    private void addNextByteToSignature(char separator, ArrayList<Byte> word, byte currentByte) throws IOException {
        do {
            word.add(currentByte);
            currentByte = getByte();
        } while (currentByte != separator);

        word.add(currentByte); // add separator
    }

    public String byteListToString(List<Byte> list) {
        if (list == null) {
            return null;
        }
        byte[] array = new byte[list.size()];
        int i = 0;
        for (Byte current : list) {
            array[i] = current;
            i++;
        }
        return new String(array);
    }

}
