/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.parser;

import at.jku.mevss.anttracks.heap.DetailedHeap;
import at.jku.mevss.anttracks.heap.labs.Lab;
import at.jku.mevss.anttracks.heap.objects.ArrayObjectInfo;
import at.jku.mevss.anttracks.heap.objects.InstanceMirrorObjectInfo;
import at.jku.mevss.anttracks.heap.objects.InstanceObjectInfo;
import at.jku.mevss.anttracks.heap.space.Space;
import at.jku.mevss.anttracks.heap.space.SpaceType;
import at.jku.mevss.anttracks.heap.symbols.AllocatedType;
import at.jku.mevss.anttracks.heap.symbols.AllocationSite;
import at.jku.mevss.anttracks.parser.heap.pointer.IncompletePointerInfo;
import at.jku.mevss.anttracks.parser.heap.pointer.PointerHandling;
import at.jku.mevss.anttracks.util.ApplicationStatistics;
import at.jku.mevss.anttracks.util.RingBuffer;
import at.jku.mevss.anttracks.util.TraceException;

import java.util.*;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

public class ThreadLocalHeap {

    public static final int STATE_PARKED = 0;
    public static final int STATE_IN_QUEUE = 1;
    public static final int STATE_IN_PROCESS = 2;
    public static final int STATE_IN_QUEUE_FOR_CLEAN_UP = 3;

    private static final int NUM_LAST_ALLOCATIONS = 15;

    public final BlockingQueue<QueueEntry> queue;
    public int state = -1;
    public final Map<Space, List<Lab>> retiredLabs;
    public final Map<SpaceType, Lab> currentLabs;
    public final String thread;

    public final RingBuffer<AllocationSite> lastAllocations;
    public long objectsAllocated;
    public long extendedStackFramesStatic;
    public long extendedStackFramesDynamic;

    // prototyping for performance
    public InstanceObjectInfo instanceInfo;
    public InstanceMirrorObjectInfo instanceMirrorInfo;
    public ArrayObjectInfo arrayInfo;

    // Ptr handling happens in two steps:
    // The first PTR event that is received for an object is stored in the fields currentPtrEvent, currentPtrFromAddr, currentPtrToAddr, currentPtrPointers and the
    // currentObjectPtrEvents list is null.
    // As soon as at least a second PTR event is received for the same object, the currentObjectPtrEvents list gets used to store pointer information.
    private EventType currentPtrEvent = EventType.NOP;
    private long currentPtrFromAddr = -1;
    public long currentPtrToAddr = -1;
    private long[] currentPtrPointers = null;
    private boolean currentPtrIsFiller;
    private int currentPtrTop = 0;

    private ArrayList<IncompletePointerInfo> notYetMatchedMultiThreadedPtrEventsToHandleAtGCEnd;

    public ThreadLocalHeap(String thread, int state) {
        this.thread = thread;
        queue = new LinkedBlockingQueue<>();
        this.state = state;
        retiredLabs = new HashMap<>();
        currentLabs = new HashMap<>();
        lastAllocations = new RingBuffer<>(NUM_LAST_ALLOCATIONS);
        instanceInfo = new InstanceObjectInfo();
        instanceMirrorInfo = new InstanceMirrorObjectInfo();
        arrayInfo = new ArrayObjectInfo();
        notYetMatchedMultiThreadedPtrEventsToHandleAtGCEnd = new ArrayList<>();
    }

    public void retireCurrentLabs(DetailedHeap heap) {
        for (SpaceType type : currentLabs.keySet().toArray(new SpaceType[currentLabs.size()])) {
            retireCurrentLab(heap, type, true);
        }
    }

    private void retireCurrentLab(DetailedHeap heap, SpaceType type, boolean clear) {
        Lab current = currentLabs.get(type);
        if (current == null) {
            return;
        }
        Space space = heap.getSpace(current.addr);
        List<Lab> retireds = retiredLabs.get(space);
        if (retireds == null) {
            retireds = new ArrayList<>(SpaceType.values().length);
            if (space != null) {
                retiredLabs.put(space, retireds);
            }
        }
        retireds.add(current);
        if (clear && current.isFull()) {
            currentLabs.remove(type);
        }
    }

    public void retireCurrentLab(DetailedHeap heap, SpaceType type) {
        retireCurrentLab(heap, type, true);
    }

    /**
     * Record an allocation.
     *
     * @param site               The allocation site.
     * @param dynamicCallContext {@code true} to add the allocation to {@link #lastAllocations} and count dynamic stack frames, {@code
     *                           false} to only count
     *                           the allocation and static stack frames.
     */
    public void recordAllocation(AllocationSite site, boolean dynamicCallContext) {
        if (dynamicCallContext) {
            lastAllocations.add(site);
            extendedStackFramesDynamic += site.callSites.length - site.stackSizeExtendedStatic;
        }
        objectsAllocated++;
        extendedStackFramesStatic += site.stackSizeExtendedStatic - site.stackSizeOriginal;
    }

    public void finishCurrentObjectPointers(DetailedHeap heap) throws TraceException {
        if (currentPtrEvent != EventType.NOP) {
            // Objects must be filled completely (except they have been handled in a multi-threaded way by the VM
            if (currentPtrTop == currentPtrPointers.length) {
                PointerHandling.handleMovedObjectWithPointers(heap, currentPtrEvent, currentPtrFromAddr, currentPtrToAddr, currentPtrPointers, currentPtrIsFiller);
            } else {
                // TODO: Find out why some class objects receive too few pointers, the second condition in the if should not be needed if pointers
                // got sent correctly, at least as far as we think...
                if (heap.getObjectInBack(currentPtrFromAddr).type.isKnownToHaveWrongPointerNumber(heap.getSymbols()) ||
                        heap.getObjectInBack(currentPtrFromAddr).type instanceof AllocatedType.MirrorAllocatedType) {
                    // weak references, etc.
                    PointerHandling.handleMovedObjectWithPointers(heap,
                                                                  currentPtrEvent,
                                                                  currentPtrFromAddr,
                                                                  currentPtrToAddr,
                                                                  Arrays.copyOf(currentPtrPointers, currentPtrTop),
                                                                  currentPtrIsFiller);
                } else {
                    // multi-threaded
                    IncompletePointerInfo info = new IncompletePointerInfo(currentPtrFromAddr, currentPtrToAddr, currentPtrPointers, currentPtrTop);
                    IncompletePointerInfo alreadyExisting = heap.multiThreadedPtrEventsToHandleAtGCEnd.put(currentPtrToAddr, info);
                    assert alreadyExisting == null : "There must not have been an already existing incomplete pointer info";
                }
            }
        }
        currentPtrEvent = EventType.NOP;
        currentPtrFromAddr = -1;
        currentPtrToAddr = -1;
        currentPtrPointers = null;
        currentPtrTop = 0;
    }

    public boolean isMatchingPtrEvent(long toAddr) {
        return currentPtrToAddr == toAddr;
    }

    public void startNewCurrentObjectPointers(EventType event, long from, long to, long[] ptrs, int top, boolean isFiller) {
        assert currentPtrTop == 0 : "Cannot start new object pointer handling if previous top is still set";
        assert currentPtrEvent == EventType.NOP : "Cannot start new object pointer handling if there are events currently set";
        assert currentPtrFromAddr == -1 : "Cannot start new object pointer handling if there are events currently set";
        assert currentPtrToAddr == -1 : "Cannot start new object pointer handling if there are events currently set";
        assert currentPtrPointers == null : "Cannot start new object pointer handling if there are events currently set";

        currentPtrEvent = event;
        currentPtrFromAddr = from;
        currentPtrToAddr = to;
        currentPtrPointers = ptrs;
        currentPtrTop = top;
        currentPtrIsFiller = isFiller;
    }

    public void addPtrToCurrentObject(long toAddr, long[] addPtrs) {
        assert toAddr == currentPtrToAddr : "All subsequent events must handle the same object";
        assert currentPtrTop + addPtrs.length <= currentPtrPointers.length : "There must be enough space to hold the additional pointers";
        System.arraycopy(addPtrs, 0, currentPtrPointers, currentPtrTop, addPtrs.length);
        currentPtrTop += addPtrs.length;
    }

    public void addMultithreadedPtrEvent(DetailedHeap heap, long toAddr, long[] ptrs) {
        IncompletePointerInfo info = heap.multiThreadedPtrEventsToHandleAtGCEnd.get(toAddr);
        if (info != null) {
            info.addPointers(ptrs);
            ApplicationStatistics.getInstance().inc("Matched Multi-threaded Pointer Event");
        } else {
            info = new IncompletePointerInfo(-1, toAddr, ptrs, ptrs.length);
            notYetMatchedMultiThreadedPtrEventsToHandleAtGCEnd.add(info);
            ApplicationStatistics.getInstance().inc("Non-matched Multi-threaded Pointer Event");
        }
    }

    public void copyAndClearMultiThreadedEvents(List<IncompletePointerInfo> dest) {
        dest.addAll(notYetMatchedMultiThreadedPtrEventsToHandleAtGCEnd);
        notYetMatchedMultiThreadedPtrEventsToHandleAtGCEnd.clear();
    }
}
