/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.parser.heap;

import at.jku.mevss.anttracks.heap.*;
import at.jku.mevss.anttracks.heap.io.*;
import at.jku.mevss.anttracks.heap.symbols.Symbols;
import at.jku.mevss.anttracks.parser.ParsingInfo;
import at.jku.mevss.anttracks.parser.TraceParser;
import at.jku.mevss.anttracks.parser.TraceParsingEventHandler;
import at.jku.mevss.anttracks.parser.TraceScannerFactory;
import at.jku.mevss.anttracks.util.TraceException;
import javafx.beans.property.BooleanProperty;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.logging.Level;

public class HeapTraceParser extends TraceParser<DetailedHeap> {

    private final List<HeapListener> heapListeners = new CopyOnWriteArrayList<>();

    public HeapTraceParser(Symbols symbols, MetaDataReaderConfig readerConfig, long toTime) throws Exception {
        super(symbols, readerConfig, toTime, toTime);
    }

    public HeapTraceParser(Symbols symbols, MetaDataReaderConfig readerConfig, long fromTime, long toTime) throws Exception {
        super(symbols, readerConfig, fromTime, toTime);
    }

    public HeapTraceParser(Symbols symbols, MetaDataReaderConfig readerConfig, GarbageCollectionLookup to) throws Exception {
        super(symbols, readerConfig, to, to);
    }

    public HeapTraceParser(Symbols symbols, MetaDataWriterConfig writerConfig) throws Exception {
        super(symbols, writerConfig);
    }

    // -------------------------
    // ------- Methods ---------
    // -------------------------

    public void addHeapListener(HeapListener l) {
        heapListeners.add(l);
    }

    public void removeHeapListener(HeapListener l) {
        heapListeners.remove(l);
    }

    // -------------------------
    // ------- Overrides -------
    // -------------------------

    @Override
    protected DetailedHeap generatePlainWorkspace(TraceScannerFactory factory) throws IOException {
        DetailedHeap heap = HeapBuilder.constructHeap(symbols);
        if (writerConfig != null) {
            heap.addListener(new MetaDataWriterListener(writerConfig, symbols));
        }
        for (HeapListener hl : heapListeners) {
            heap.addListener(hl);
        }
        return heap;
    }

    @Override
    protected DetailedHeap generateWorkspaceFromMetaData(HeapIndexReader heapIndexReader, HeapPosition heapPosition)
            throws FileNotFoundException,
            IOException {
        DetailedHeap heap;
        if (heapPosition.fileName != -1) {
            heap = heapIndexReader.getHeap(heapPosition.fileName, symbols);
        } else {
            heap = HeapBuilder.constructHeap(symbols);
        }
        for (HeapListener hl : heapListeners) {
            heap.addListener(hl);
        }

        // This is the case if the first selected GC point has exaclty been reconstructed from a heap dump file.
        // The GC-end event at which the heap dump file got created will not be read from the trace anymore, therefore
        // we have to start the diffing here manually.
        if (heapPosition.fromPosition == heapPosition.toPosition) {
            // We only dump heaps on GC-end, so this cannot be a GC start
            // Notify all listeners that we are at a GC end
            // TODO: Currently we say that the last GC was MINOR, but it could also have been a MAJOR
            // TODO: Also length is not correct but just set to 1
            heapListeners.stream()
                         .forEach(l -> l.phaseChanging(heap,
                                                       GarbageCollectionType.MUTATOR,
                                                       GarbageCollectionType.MINOR,
                                                       new GarbageCollectionCause(-1, "AntTracks Artification", false),
                                                       false,
                                                       false,
                                                       fromTime,
                                                       0,
                                                       1,
                                                       true));
        }

        return heap;
    }

    @Override
    protected void doWorkspaceCompletion(DetailedHeap workspace) {
        logger.log(Level.INFO, "bringing heap into consistent state");
        try {
            workspace.complete();
        } catch (TraceException e) {
            e.printStackTrace();
        }
    }

    @Override
    protected void doRemoveListenersOnCompletion(DetailedHeap workspace) {
        workspace.removeAllListeners();
        heapListeners.clear();
    }

    @Override
    protected void doParseCleanupAfterSuccessfulParse(DetailedHeap workspace) throws TraceException {
        workspace.assignLabsIncorrectlyAssumedToBeFillers(false);
    }

    protected TraceParsingEventHandler createMainEventHandler(ParsingInfo parsingInfo) {
        return new HeapEventHandler(workspace, symbols, parsingInfo);
    }

    public IndexBasedHeap parseFastHeap(BooleanProperty cancellationToken) throws InterruptedException, TraceException, IOException {
        return parse(cancellationToken).toIndexBasedHeap();
    }
}
