package at.jku.mevss.anttracks.parser.heap.pointer;

import at.jku.mevss.anttracks.heap.DetailedHeap;
import at.jku.mevss.anttracks.heap.ObjectVisitor;
import at.jku.mevss.anttracks.heap.labs.Lab;
import at.jku.mevss.anttracks.heap.objects.InstanceMirrorObjectInfo;
import at.jku.mevss.anttracks.heap.objects.InstanceObjectInfo;
import at.jku.mevss.anttracks.heap.objects.ObjectInfo;
import at.jku.mevss.anttracks.heap.roots.RootPtr;
import at.jku.mevss.anttracks.heap.space.Space;
import at.jku.mevss.anttracks.heap.space.SpaceInfo;
import at.jku.mevss.anttracks.heap.symbols.AllocatedType;
import at.jku.mevss.anttracks.heap.symbols.AllocationSite;
import at.jku.mevss.anttracks.parser.EventType;
import at.jku.mevss.anttracks.parser.TraceParser;
import at.jku.mevss.anttracks.util.CollectionsUtil;
import at.jku.mevss.anttracks.util.ParallelizationUtil;
import at.jku.mevss.anttracks.util.TraceException;

import java.util.*;
import java.util.stream.Collectors;

import static at.jku.mevss.anttracks.parser.EventType.isDedicatedPtrEvent;

public class PointerHandling {
    /*
    static BufferedWriter bw;

    static {
        try {
            bw = new BufferedWriter(new FileWriter("C:\\repos\\m3\\test.txt"));
        } catch (IOException e) {
            e.printStackTrace();
        }
    }*/
    public static void handleMovedObjectWithPointers(DetailedHeap heap, EventType event, long fromAddr, long toAddr, long[] pointers, boolean isFiller)
            throws TraceException {
        //ApplicationStatistics.getInstance().inc("Single pointer event");
        if (isFiller) {
            // Just ignore filler objects, what else should we do?
            //ApplicationStatistics.getInstance().inc("Single pointer event for filler object");
            return;
        }

        boolean isMove = !isDedicatedPtrEvent(event);
        boolean front = true;

        if (isMove) {
            heap.setObjectForwardingAddrInBack(fromAddr, toAddr);

            if (pointers == null || mayBeNonDirtyObject(heap, pointers, fromAddr)) {
                // Ptrs not set in move event
                // i.e., pointers did not change since last GC, copy pointers from from-address

                // Copy pointers from back to keep them
                assert heap.getObjectInBack(fromAddr) == heap.getObjectInFront(toAddr);
                pointers = heap.getObjectPointersInBack(fromAddr);
                // TODO: Includce again in more intelligent way, CopyOnWriteArrayList needs too much memory
                //heap.addressesThatMayHaveInvalidPointersDueToNullingOnNonDirtyObjects.add(toAddr);

                // Remember the address from where we copied the pointers since we probably have to reset them if we receive multi-threaded pointer events.
                // In the case that multi-threaded pointer events are received but pointers have been copied from the back object
                // These copied pointers then have to be overwritten by the multi-threaded pointers.
                /*
                synchronized (bw) {
                    try {
                        bw.write(String.format("Remember pointer copy from %,d (%s) to %,d (%s)\n", fromAddr, heap.getSpace(fromAddr).toShortString(), toAddr,
                                               heap.getSpace(toAddr).toShortString()));
                        bw.flush();
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                }*/
                heap.setObjectForwardingAddrInFront(toAddr, fromAddr);
                // Since objects may be moved during GC, mark the object for pointer update
                heap.setPtrNeedAdjustmentInFront(toAddr, true);
            } else {
                // Major GCs always send the referenced object's target address (address of the object after its possible move)
                // No pointer adjustment needed at the end of Major GC since pointers have been sent adjusted by the VM
                // Minor GCs always send the referenced object's base (from) address (address of an object before its possible move)
                // Pointer adjustment needed at the end of the Minor GC since objects may move change during GC
                heap.setPtrNeedAdjustmentInFront(toAddr, !heap.getGC().isFull());
            }
            heap.setObjectPointersInFront(toAddr, pointers);
        } else {
            // In this case, we got a PTR_UPDATE_PREMOVE or PTR_UPDATE_POSTMOVE event
            // These events are only allowed for non-moved objects, therefore we have to update the pointers in BACK, since those objects get merged back into FRONT later
            if (pointers == null || mayBeNonDirtyObject(heap, pointers, toAddr)) {
                // Keep pointers for object in back, but mark them that they may have changed and need adjustment
                pointers = heap.getObjectPointersInBack(toAddr);
                heap.setPtrNeedAdjustmentInBack(toAddr, true);
                // Fields may have been set to null
                // TODO: Includce again in more intelligent way, CopyOnWriteArrayList needs too much memory
                //heap.addressesThatMayHaveInvalidPointersDueToNullingOnNonDirtyObjects.add(toAddr);
            } else {
                heap.setPtrNeedAdjustmentInBack(toAddr, event == EventType.GC_PTR_UPDATE_PREMOVE);
                heap.setObjectPointersInBack(toAddr, pointers);
            }
            front = false;
        }

        if (TraceParser.CONSISTENCY_CHECK) {
            consistencyCheck(heap, event, () -> event, toAddr, isMove, pointers, pointers, front);
        }
    }

    private static void consistencyCheck(DetailedHeap heap,
                                         EventType moveEventType,
                                         CollectionsUtil.Factory<EventType> pointerUpdateEventType,
                                         long toAddr,
                                         boolean isMoved,
                                         long[] movePtrs,
                                         long[] ptrs,
                                         boolean front)
            throws TraceException {
        /*
        if (!hasCorrectPointerCount(heap, heap.getLab(toAddr, front), toAddr, ptrs)) {
            System.out.println(toAddr + " " + (moveEventType != null ? moveEventType : "NoMove-MultiEvent") + " " + Arrays.toString(ptrs));
        }
        */
        assert hasCorrectPointerCount(heap, heap.getLab(toAddr, front), toAddr, ptrs) :
                ("There must not be more pointers than non-static reference fields." +
                         "\nObject type: " + heap.getObject(toAddr, front).type.name +
                         "\nAssumed pointers: " + heap.getObject(toAddr, front).type.nonStaticReferenceFieldsCount() +
                         "\nReceived pointers: " + (ptrs == null ? "null" : ptrs.length));

        if (ptrs != null) {
            if (isMoved) {
                if (movePtrs == null) {
                    // Consistency check: All pointers must be found in BACK
                    // Cannot check pointers here since we could have accidentally copied pointers from back even though
                    // multi-threaded ptr events will follow that will fix the pointers
                    /*
                    for (long p : ptrs) {
                        if (p >= 0) {
                            if (heap.getObjectInBack(p) == null) {
                                throw new TraceException("Consistency check: Could not find referenced object in back after Minor GC " +
                                                                 "move + pointer dirty event handling");
                            }
                        }

                    }*/
                } else {
                    if (heap.getGC().full) {
                        // Consistency check: All pointers must be found in FRONT as soon as all moves are processed
                        // Therefore, we cannot check anything now
                        /*
                        for (long p : ptrs) {
                            if (p >= 0) {
                                if (heap.getObjectInFront(p) == null) {
                                    throw new TraceException(
                                            "Consistency check: Could not find referenced object in front after Major GC" +
                                                    " move + pointer event handling");
                                }
                            }
                        }*/

                    } else {
                        // Consistency check: All pointers must be found in BACK (or FRONT if GC failed)
                        for (long p : ptrs) {
                            if (p >= 0) {
                                Space s = heap.getSpace(p);
                                // Use getObject, not getObjectInBack because GC could have failed
                                if (s.getObject(p) == null) {
                                    throw new TraceException("Consistency check: Could not find referenced object in back after Minor GC " + "move + pointer event " +
                                                                     "handling");
                                }
                            }
                        }
                    }
                }
            } else {
                if (pointerUpdateEventType.create() == EventType.GC_PTR_UPDATE_PREMOVE) {
                    // Consistency check: All pointers must be found in BACK (or FRONT if GC failed)
                    for (long p : ptrs) {
                        if (p >= 0) {
                            try {
                                heap.getObjectInBack(p);
                            } catch (Exception ignored) {
                                throw new TraceException(String.format("Consistency check: Could not find referenced object in back after " + "PTR_UPDATE_PREMOVE " +
                                                                               "event" +
                                                                               " " +
                                                                               "handling\nObject=%,d\nPointer=%,d",
                                                                       toAddr,
                                                                       p));
                            }
                        }
                    }
                } else {
                    /* can't assert anything as objects have not been forwarded to their new addresses
                    for (long p : ptrs) {
                        if (p >= 0) {
                            Space toSpace = heap.getSpace(p);
                            ObjectInfo o = null;
                            switch(toSpace.getTransitionType()) {
                                case Accumulative:
                                    // if we point into an accumulative space, pointee may be in back
                                    try {
                                        o = heap.getLabInBack(p).getObject(p);
                                    } catch (TraceException ignored) { }
                                    if (o != null) break;
                                    // intended fallthrough
                                case ReplaceAll:
                                case None:
                                    try {
                                        o = heap.getLabInFront(p).getObject(p);
                                    } catch (TraceException ignored) { }
                            }
                            if (o == null) {
                                throw new TraceException(
                                        "Consistency check: Could not find referenced object in heap after Major GC" +
                                                " move + pointer event handling");
                            }
                        }
                    } */
                }
            }
        }
    }

    // manipulated by multiple threads during GC, so we need to wait until GC end, i.e., Obj ptr array & filler.
    public static void handleMultiThreadedPtrEvent(DetailedHeap heap, long fromAddr, long toAddr, long[] ptrs) throws TraceException {
        heap.setObjectPointersInFront(toAddr, ptrs);
        heap.setPtrNeedAdjustmentInFront(toAddr, true);
        heap.setObjectForwardingAddrInBack(fromAddr, toAddr);
        //heap.setPtrNeedAdjustmentInFront(toAddr, !heap.getGC().isFull());
    }

    public static void updateRefsNew(DetailedHeap heap, Space space, Lab lab, long address, long[] pointers) throws TraceException {
        if (pointers == null) {
            // Objects that don't have pointers don't have to be handled
            return;
        }

        // Information about current pointer that has to be updated (collected in back)
        long ptrAddress = -1;
        Space ptrSpace;
        long ptrForwardingAddress;

        // Update all BACK objects of accumulative spaces (objects that have not been moved but may need ptr adjustment)
        // For FRONT objects, only update pointers that have not already been sent updated by the VM
        if (lab.ptrNeedAdjustment(address)) {
            for (int i = 0; i < pointers.length; i++) {
                ptrAddress = pointers[i];
                if (ptrAddress >= 0) {
                    ptrSpace = heap.getSpace(ptrAddress);
                    if (ptrSpace.isBeingCollected()) {
                        Lab l = ptrSpace.getLabInBack(ptrAddress);
                        ptrForwardingAddress = l.getObjectForwardingAddressEscaped(ptrAddress);
                        if (ptrForwardingAddress < 0) {
                            ObjectInfo objInfo = lab.getObject(address);
                            throw new TraceException(String.format("Pointing object: %s @ %,d (Space: %s)" +
                                                                           "\nPointers: %s" +
                                                                           "\nForwarding address not set for object at " + "%,d (%s) in BACK (Space: %s) during GC " +
                                                                           "pointer update.\n",
                                                                   objInfo.type.name,
                                                                   address,
                                                                   heap.getSpace(address).toShortString(),
                                                                   Arrays.toString(pointers),
                                                                   ptrAddress,
                                                                   l.getObject(ptrAddress).type.name,
                                                                   ptrSpace.toShortString()));
                        }
                        pointers[i] = ptrForwardingAddress;
                    } else if (TraceParser.CONSISTENCY_CHECK) {
                        // throws an exception if object could not be found
                        if (ptrSpace.getObjectInBack(ptrAddress) == null) {
                            throw new TraceException("Expected to find referenced object in non-moved space's back but didn't find it!");
                        }
                    }
                }
            }
        }
        lab.setPtrNeedAdjustment(address, false);
    }

    public static void updateRefsFailedGC(DetailedHeap heap, Space space, Lab lab, long address, long[] pointers) throws TraceException {
        assert !heap.getGC().isFull() : "Major GC cannot fail";

        if (pointers == null) {
            // Objects that don't have pointers don't have to be handled
            return;
        }

        if (lab.ptrNeedAdjustment(address)) {
            for (int i = 0; i < pointers.length; i++) {
                long curPtrAddr = pointers[i];
                Space curPtrSpace = heap.getSpace(curPtrAddr);
                if (curPtrSpace != null) {
                    Lab curPtrLab = null;
                    if (curPtrSpace.getTransitionType() == SpaceInfo.TransitionType.None) {
                        // The object's space that the pointer points to has no transition == has failed during GC.
                        // Objects that have been moved yet have a forwarding address set, objects that have not been moved yet have no
                        // forwarding address.
                        curPtrLab = curPtrSpace.getLabForAddr(curPtrAddr, true);
                        if (curPtrLab == null) {
                            throw new TraceException(String.format("Pointing object: %s @ %,d (Space: %s)\n" + "Pointers: %s\n" + "Did not find lab for object at " +
                                                                           "%,d " +
                                                                           "in FRONT of" +
                                                                           " failed Space (Space: %s) during GC pointer update (Failed GC).\n",
                                                                   lab.getObject(address).type.name,
                                                                   address,
                                                                   space.toShortString(),
                                                                   Arrays.toString(pointers),
                                                                   curPtrAddr,
                                                                   curPtrSpace.toShortString()));
                        }
                        long forwardedAddr = curPtrLab.getObjectForwardingAddressEscaped(curPtrAddr);
                        // if the forwardedAddr >= 0, the object has been moved before the collection failed-->it exists twice. In this case,
                        // assume that the moved object will survive the following major collection and update the pointer accordingly.
                        // if the forwardedAddr < 0, (following sentence is POGC only) the object did not move: it either lives in Eden/SurvFrom and the gc failed
                        // before
                        // it could be moved or it lives in old and would not move at all. In either case, the pointer remains unchanged.
                        if (forwardedAddr >= 0) {
                            pointers[i] = forwardedAddr;
                        }
                    } else if (curPtrSpace.getTransitionType() == SpaceInfo.TransitionType.Accumulative) {
                        // The object's space that the pointer points to has not been collected at all, it cannot have been moved at all.
                        if (TraceParser.CONSISTENCY_CHECK) {
                            // throws an exception if object could not be found
                            if (curPtrSpace.getObjectInBack(curPtrAddr) == null) {
                                throw new TraceException("Expected to find referenced object in non-moved space's back but didn't find it!");
                            }
                        }
                        continue;
                    } else if (curPtrSpace.getTransitionType() == SpaceInfo.TransitionType.ReplaceAll) {
                        // The object's space that the pointer points to has been collected.
                        // This means the the object must have been moved, since it is pointed to and may not have been collected.
                        curPtrLab = curPtrSpace.getLabForAddr(curPtrAddr, false);
                        long forwardedAddr = curPtrLab.getObjectForwardingAddressEscaped(curPtrAddr);
                        if (forwardedAddr < 0) {
                            throw new TraceException(String.format("Pointing object: %s @ %,d (Space: %s)\n" + "Pointers: %s\n" + "Forwarding address not set for " +
                                                                           "object at %,d " +
                                                                           "(%s) in BACK (Space: %s) during GC pointer update.\n",
                                                                   lab.getObject(address).type.name,
                                                                   address,
                                                                   space.toShortString(),
                                                                   Arrays.toString(pointers),
                                                                   curPtrAddr,
                                                                   curPtrLab.getObject(curPtrAddr).type.name,
                                                                   curPtrSpace.toShortString()));
                        }
                        pointers[i] = forwardedAddr;
                    } else {
                        assert false : "Here be dragons. No transition type beside None, Accumulative and ReplaceAll";
                    }

                }
            }
        }

        lab.setPtrNeedAdjustment(address, false);
    }

    private static boolean hasCorrectPointerCount(DetailedHeap heap, Lab lab, long address, long[] ptrs) {
        int ptrCount = ptrs == null ? 0 : ptrs.length;

        ObjectInfo obj;
        AllocatedType type;
        try {
            obj = lab.getObject(address);
            type = obj.type;
        } catch (TraceException ex) {
            return false;
        }
        // Type info only stored for instance types, InstanceMirrorObjectInfo (java.lang.Class) and arrays should be ignored at the moment
        if (obj instanceof InstanceObjectInfo && !(obj instanceof InstanceMirrorObjectInfo) && !type.isKnownToHaveWrongPointerNumber(heap.getSymbols())) {
            return type.nonStaticReferenceFieldsCount() == ptrCount;
        }

        // Non-checkable type
        return true;
    }

    public static void handlePtrsOnGCEnd(DetailedHeap heap, boolean failed) throws TraceException {
        if (heap.getSymbols().expectPointers) {
            for (IncompletePointerInfo info : heap.notYetMatchedMultiThreadedPtrEventsToHandleAtGCEnd) {
                heap.multiThreadedPtrEventsToHandleAtGCEnd.get(info.toAddr).addPointers(info.ptrs);
            }
            heap.notYetMatchedMultiThreadedPtrEventsToHandleAtGCEnd.clear();

            for (Map.Entry<Long, IncompletePointerInfo> entry : heap.multiThreadedPtrEventsToHandleAtGCEnd.entrySet()) {
                if (!entry.getValue().isComplete()) {
                    assert entry.getValue().isComplete() : "Incomplete pointer info must have been completed by multi-threaded events but this is not the case here!";
                }
                PointerHandling.handleMultiThreadedPtrEvent(heap, entry.getValue().fromAddr, entry.getValue().toAddr, entry.getValue().ptrs);
            }
            heap.multiThreadedPtrEventsToHandleAtGCEnd.clear();

            ParallelizationUtil.temporaryExecutorServiceBlocking(Arrays.asList(heap.getSpacesCloned()).iterator(), (Space space) -> {
                try {
                    if (space != null) {
                        switch (space.getTransitionType()) {
                            case Accumulative:
                                space.iterateUpdatePointer(new PtrUpdateVisitor(true, heap, failed));
                                space.iterateUpdatePointer(new PtrUpdateVisitor(false, heap, failed));
                                break;
                            case ReplaceAll:
                                space.iterateUpdatePointer(new PtrUpdateVisitor(true, heap, failed));
                                break;
                            case None:
                                if (!failed) {
                                    throw new TraceException("Space needs to be in transition if gc did not fail");
                                }
                                space.iterateUpdatePointer(new PtrUpdateVisitor(true, heap, failed));
                                break;
                            default:
                                assert false : "here be dragons";
                        }
                    }
                } catch (TraceException e) {
                    e.printStackTrace();
                }
            });

            if (failed) {
                Arrays.stream(heap.getSpacesCloned())
                      .filter(s -> s.getTransitionType() == SpaceInfo.TransitionType.None)
                      .flatMap(s -> Arrays.stream(s.getLabs()))
                      .forEach(Lab::resetForwardingAddresses);
            }
        }
    }

    public static void validateAllPointers(DetailedHeap heap) throws TraceException {
        class ThreadLocalPointerValidator implements ObjectVisitor {
            class InvalidPointerPair {
                public long from;
                public long to;

                public InvalidPointerPair(long from, long to) {
                    this.from = from;
                    this.to = to;
                }
            }

            public List<InvalidPointerPair> invalidPointerPairs = new ArrayList<>();

            @Override
            public void visit(long address,
                              ObjectInfo objectInfo,
                              SpaceInfo space,
                              AllocatedType type,
                              int size,
                              boolean isArray,
                              int arrayLength,
                              AllocationSite allocationSite,
                              long[] pointedFrom,
                              long[] pointsTo,
                              EventType eventType,
                              List<RootPtr> rootPtrs) {
                if (pointsTo != null) {
                    for (int ptrIdx = 0; ptrIdx < pointsTo.length; ptrIdx++) {
                        long ptrAddr = pointsTo[ptrIdx];
                        if (ptrAddr >= 0) {
                            boolean pointerIsFine;
                            try {
                                Space toSpace = heap.getSpace(ptrAddr);
                                ObjectInfo front = toSpace.getLabInFront(ptrAddr).getObject(ptrAddr);
                                pointerIsFine = front != null;
                            } catch (TraceException ex) {
                                pointerIsFine = false;
                            }

                            if (!pointerIsFine) {
                                // Pointer to an invalid position!!
                                // This is only allowed for G1 objects that are already dead (not root-referenced anymore) and will be collected somewhen in the
                                // future
                                boolean isRootPointed = rootPtrs.size() > 0;
                                if (!isRootPointed || heap.addressesThatMayHaveInvalidPointersDueToNullingOnNonDirtyObjects.contains(address)) {
                                    // Okay, special case: This is a dead object that will be collected in the future, and it has wrong pointer information...
                                    // Set pointer to NULL or something similar
                                    pointsTo[ptrIdx] = Lab.OBJECT_PTR_DELETED;
                                } else {
                                    // Damn, there is something wrong with our pointers
                                    invalidPointerPairs.add(new InvalidPointerPair(address, ptrAddr));
                                }
                            }
                        }
                    }
                }
                // Clearing the list
                heap.addressesThatMayHaveInvalidPointersDueToNullingOnNonDirtyObjects.remove(address);
            }
        }

        List<ThreadLocalPointerValidator.InvalidPointerPair> invalidPairs = heap.toObjectStream()
                                                                                .forEachParallel(() -> new ThreadLocalPointerValidator(),
                                                                                                 new ObjectVisitor.Settings(false, true, true))
                                                                                .stream()
                                                                                .flatMap(visitor -> visitor.invalidPointerPairs.stream())
                                                                                .collect(Collectors.toList());
        assert heap.addressesThatMayHaveInvalidPointersDueToNullingOnNonDirtyObjects.isEmpty() : "After validating pointers no 'dirty info' may be left";

        if (!invalidPairs.isEmpty()) {
            throw new TraceException(
                    "Something is wrong with the pointers!\n" +
                            invalidPairs
                                    .stream()
                                    .map(pair -> String.format("-> The object at address %,d points to position %,d, but no object is located at the to-position",
                                                               pair.from,
                                                               pair.to))
                                    .collect(Collectors.joining("\n")));
        }
    }

    private static List<RootPtr> rootPointer(DetailedHeap heap, long address) {
        List<RootPtr> ret = new ArrayList<>();

        Set<Long> visited = new HashSet<>();
        Set<Long> currentlyToVisit = new HashSet<>();
        Set<Long> nextVisitRound = new HashSet<>();

        nextVisitRound.add(address);
        visited.add(address);

        while (!nextVisitRound.isEmpty()) {
            currentlyToVisit.clear();
            currentlyToVisit.addAll(nextVisitRound);
            nextVisitRound.clear();
            for (Long pointeeAddress : currentlyToVisit) {

                // is there a corresponding root ptr?
                List<RootPtr> matchingRoots = heap.rootPtrs.get(pointeeAddress);
                if (matchingRoots != null && matchingRoots.size() != 0) {
                    ret.addAll(matchingRoots);
                }

                // ... and keep searching
                long[] fromPointers;
                try {
                    fromPointers = heap.getObjectFromPointers(pointeeAddress);
                    if (fromPointers != null) {
                        for (int i = 0; i < fromPointers.length; i++) {
                            if (!visited.contains(fromPointers[i])) {
                                visited.add(fromPointers[i]);
                                // extend path by next step
                                nextVisitRound.add(fromPointers[i]);
                            }
                        }
                    }
                } catch (TraceException e) {
                    e.printStackTrace();
                }

            }
        }

        return ret;
    }

    private static boolean mayBeNonDirtyObject(DetailedHeap heap, long[] pointers, long fromAddr) throws TraceException {
        return Arrays.stream(pointers).allMatch(addr -> addr == -1) && !hasCorrectPointerCount(heap, heap.getLabInBack(fromAddr), fromAddr, pointers);
    }
}
