package at.jku.mevss.anttracks.parser.printing;

import at.jku.mevss.anttracks.heap.DetailedHeap;
import at.jku.mevss.anttracks.heap.objects.ObjectInfo;
import at.jku.mevss.anttracks.heap.roots.RootPtr;
import at.jku.mevss.anttracks.heap.space.Space;
import at.jku.mevss.anttracks.heap.space.SpaceMode;
import at.jku.mevss.anttracks.heap.space.SpaceRedefinition;
import at.jku.mevss.anttracks.heap.space.SpaceType;
import at.jku.mevss.anttracks.heap.symbols.AllocatedType;
import at.jku.mevss.anttracks.heap.symbols.AllocationSite;
import at.jku.mevss.anttracks.parser.EventType;
import at.jku.mevss.anttracks.parser.ParserGCInfo;
import at.jku.mevss.anttracks.parser.ThreadLocalHeap;
import at.jku.mevss.anttracks.parser.TraceParsingEventHandler;
import at.jku.mevss.anttracks.util.TraceException;

import java.util.List;
import java.util.logging.Logger;

public class AdditionalPrintingEventHandler implements TraceParsingEventHandler {

    private static boolean PRINT_CLEANUP = false;

    private static boolean PRINT_SPACE_REDEFINE = true;
    private static boolean PRINT_SPACE_DESTROY = true;
    private static boolean PRINT_SPACE_RELEASE = true;
    private static boolean PRINT_SPACE_ALLOC = true;
    private static boolean PRINT_SPACE_CREATE = true;

    private static final boolean PRINT_GC_MOVE = false;

    private static final boolean PRINT_TAG = true;

    private static boolean PRINT_GC_START = true;
    private static boolean PRINT_GC_END = true;

    private final DetailedHeap heap;
    private Logger LOGGER = Logger.getLogger(AdditionalPrintingEventHandler.class.getSimpleName());

    public AdditionalPrintingEventHandler(DetailedHeap heap) {
        this.heap = heap;
    }

    @Override
    public void doCleanUp(ThreadLocalHeap threadLocalHeap) {
        if (PRINT_CLEANUP) {
            LOGGER.info("Cleanup");
        }
    }

    @Override
    public void doParseSpaceRedefine(int index, long startAddr, long size, boolean postponed, int source) {
        if (PRINT_SPACE_REDEFINE) {
            SpaceRedefinition redef = new SpaceRedefinition(heap, index, startAddr, size, true);
            LOGGER.info(String.format("Redefine space %d by %s (postponed: %s, source: %d)\nCurrent heap: %s\nRedefinition: %s",
                                      index,
                                      Thread.currentThread().getName(),
                                      postponed,
                                      source,
                                      heap.toString(),
                                      redef.toString()));
        }
    }

    @Override
    public void doParseSpaceDestroy(int firstIndex, long nRegions) {
        if (PRINT_SPACE_DESTROY) {
            LOGGER.info("Destroy space " + firstIndex + " to " + (firstIndex + nRegions - 1) + " by " + Thread.currentThread().getName());
        }
    }

    @Override
    public void doParseSpaceRelease(int index) {
        if (PRINT_SPACE_RELEASE) {
            LOGGER.info("Release space " + index + " by " + Thread.currentThread().getName());
        }
    }

    @Override
    public void doParseSpaceAlloc(int index, SpaceMode spaceMode, SpaceType spaceType) {
        if (PRINT_SPACE_ALLOC) {
            LOGGER.info("Alloc space " + index + " by " + Thread.currentThread().getName());
        }
    }

    @Override
    public void doParseSpaceCreate(int index, long startAddr, long size) {
        if (PRINT_SPACE_CREATE) {
            LOGGER.info("Create space " + index + " by " + Thread.currentThread().getName());
        }
    }

    @Override
    public void doParseGCInfo(int index, int id) {

    }

    @Override
    public void doParseGCFailed(int index) {

    }

    @Override
    public void doParseGCStart(ParserGCInfo gcInfo, long start, long end, boolean check) {
        if (PRINT_GC_START) {
            LOGGER.info(String.format("Start GC %d (%s, %s)", gcInfo.id, gcInfo.type, gcInfo.cause));
        }
    }

    @Override
    public void doParseGCEnd(ParserGCInfo info, ThreadLocalHeap threadLocalheap, long start, long end, boolean failed, boolean check) {
        if (PRINT_GC_END) {
            LOGGER.info("End GC " + info.id);
        }
    }

    @Override
    public void doParseGCInterrupt(int id, long address) {

    }

    @Override
    public void doParseGCContinue(int id, long address) {

    }

    @Override
    public long doGCMove(EventType type, ThreadLocalHeap threadLocalHeap, ObjectInfo obj, long fromAddr, long toAddr, SpaceType toSpaceType) {
        if (PRINT_GC_MOVE) {
            LOGGER.info(String.format("Object %s moved from %,d to %,d", obj.toString(), fromAddr, toAddr));
        }
        return 0;
    }

    @Override
    public ObjectInfo getObjectIsFiller(Space space, long addr) {
        return null;
    }

    @Override
    public void doParseGCClassLoaderRootPtr(long ptr, String loaderName) {

    }

    @Override
    public void doParseGCClassRootPtr(long ptr, int classId) {

    }

    @Override
    public void doParseGCStaticFieldRootPtr(long ptr, int classId, int offset) {

    }

    @Override
    public void doParseGCJNILocalRootPtr(long ptr, long threadId) {

    }

    @Override
    public void doParseGCLocalVariableRootPtr(long ptr, long threadId, int classId, int methodId, int slot) {

    }

    @Override
    public void doParseGCVMInternalThreadDataRootPtr(long ptr, long threadId) {

    }

    @Override
    public void doParseGCCodeBlobRootPtr(long ptr, int classId, int methodId) {

    }

    @Override
    public void doParseGCJNIGlobalRootPtr(long ptr, boolean weak) {

    }

    @Override
    public void doParseGCOtherRootPtr(long ptr, RootPtr.RootType rootType) {

    }

    @Override
    public void doParseGCDebugRootPtr(long ptr, String vmCall) {

    }

    @Override
    public List<ObjectInfo> doParseGCMoveRegion(EventType type, long fromAddr, long toAddr, SpaceType toSpaceType, int numOfObjects, ThreadLocalHeap threadLocalHeap) {
        return null;
    }

    @Override
    public Space getSpace(long addr) {
        return new Space("Dummy") {
            @Override
            public ObjectInfo getNonFillerObjectInfo(long addr) throws TraceException {
                return null;
            }
        };
    }

    @Override
    public void doParseThreadAlive(int header, long id, String name) {

    }

    @Override
    public void doParseThreadDeath(long id) {

    }

    @Override
    public void doParseSyncObj(EventType type,
                               int allocationSiteId,
                               int allocatedTypeId,
                               AllocatedType allocatedType,
                               long fromAddr,
                               long toAddr,
                               SpaceType toSpaceType,
                               int length,
                               int size,
                               ThreadLocalHeap threadLocalHeap) {

    }

    @Override
    public void doParseObjAllocFastCi(EventType type,
                                      int allocationSiteId,
                                      AllocationSite allocationSite,
                                      AllocatedType allocatedType,
                                      boolean isArray,
                                      int arrayLength,
                                      ThreadLocalHeap threadLocalHeap) {

    }

    @Override
    public void doParseObjAllocFastC2DeviantType(EventType type,
                                                 int header,
                                                 int allocationSiteId,
                                                 AllocatedType allocatedType,
                                                 boolean isArray,
                                                 int arrayLength,
                                                 ThreadLocalHeap threadLocalHeap) {

    }

    @Override
    public void doParseObjAllocFastIr(EventType type, int allocationSiteId, AllocationSite allocationSite, ThreadLocalHeap threadLocalHeap) {

    }

    @Override
    public void doParseObjAllocNormalCi(EventType type,
                                        int allocationSiteId,
                                        AllocationSite allocationSite,
                                        AllocatedType allocatedType,
                                        long addr,
                                        boolean isArray,
                                        int arrayLength,
                                        ThreadLocalHeap threadLocalHeap) {

    }

    @Override
    public void doParseObjAllocNormalIr(EventType type, int allocationSiteId, AllocationSite allocationSite, long addr, ThreadLocalHeap threadLocalHeap) {

    }

    @Override
    public void doParseObjAllocSlowCiIr_Deviant(EventType type,
                                                int allocationSiteId,
                                                AllocationSite allocationSite,
                                                long addr,
                                                int arrayLength,
                                                int realAllocatedTypeId,
                                                ThreadLocalHeap threadLocalHeap) {

    }

    @Override
    public void doParseObjAllocSlow(EventType type,
                                    int allocationSiteId,
                                    AllocationSite allocationSite,
                                    long addr,
                                    boolean isArray,
                                    int arrayLength,
                                    int size,
                                    boolean mayBeFiller,
                                    ThreadLocalHeap threadLocalHeap) {

    }

    @Override
    public void doParseTlabAlloc(int size, long addr, ThreadLocalHeap threadLocalHeap) {

    }

    @Override
    public void doParsePlabAlloc(int size, long addr, ThreadLocalHeap threadLocalHeap) {

    }

    @Override
    public long getMoveTarget(long fromAddr, long toAddr, SpaceType toSpaceType, ThreadLocalHeap threadLocalHeap) {
        return 0;
    }

    @Override
    public void doPtrEvent(EventType type, ThreadLocalHeap threadLocalHeap, long fromAddr, long toAddr, long[] ptrs, boolean isFiller) {

    }

    @Override
    public void doParseGCTag(String tagText) {
        if (PRINT_TAG) {
            LOGGER.info(String.format("Tag received: %s", tagText));
        }
    }
}
