package at.jku.mevss.anttracks.util;

import at.jku.mevss.anttracks.heap.datastructures.DataStructure;
import at.jku.mevss.anttracks.heap.datastructures.DataStructureDefinition;
import at.jku.mevss.anttracks.heap.datastructures.DataStructureDefinitions;
import at.jku.mevss.anttracks.heap.datastructures.cocor.DataStructurePart;
import at.jku.mevss.anttracks.heap.datastructures.cocor.Parser;
import at.jku.mevss.anttracks.heap.datastructures.cocor.Scanner;
import at.jku.mevss.anttracks.heap.symbols.AllocatedType;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.nio.file.*;
import java.util.*;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class DataStructureUtil {

    private static final Logger LOGGER = Logger.getLogger(DataStructureUtil.class.getSimpleName());

    /**
     * Artificial DataStructurePart that matches everything
     */
    public static final DataStructurePart starDummyDataStructurePart = new DataStructurePart("*", false);

    /**
     * Artificial AllocatedType that contains a DataStructureDefinition that matches all objects but is not to be followed
     */
    public static final AllocatedType flatStarDummyAllocatedType = new AllocatedType(Integer.MIN_VALUE + 1, -1, "Lanttracks/helper/FlatStarDummy;", -1);
    /**
     * Artificial AllocatedType that contains a DataStructureDefinition that matches all objects and is to be followed
     */
    public static final AllocatedType deepStarDummyAllocatedType = new AllocatedType(Integer.MIN_VALUE + 2, -1, "Lanttracks/helper/DeepStarDummy;", -1);

    {
        flatStarDummyAllocatedType.dataStructureDefinition = new DataStructureDefinition(deepStarDummyAllocatedType) {
            @Override
            public boolean containsFlat(AllocatedType flatty) {
                return true;
            }

            @Override
            public boolean hasFollower(AllocatedType follower) {
                return false;
            }
        };
        deepStarDummyAllocatedType.dataStructureDefinition = new DataStructureDefinition(deepStarDummyAllocatedType) {
            @Override
            public boolean containsFlat(AllocatedType flatty) {
                return false;
            }

            @Override
            public boolean hasFollower(AllocatedType follower) {
                return true;
            }
        };
    }

    private static final String DEFAULT_DATA_STRUCTURE_DEFINITIONS_RESOURCE_FOLDER = "/datastructures";

    /**
     * Parses default data structure definition files
     *
     * @return the {@link DataStructureDefinitions} defined in the default .ds files or empty {@link DataStructureDefinitions} if no files or definitions present
     */
    public static DataStructureDefinitions parseDataStructureDefinitionFiles() {
        return parseDataStructureDefinitionFiles(getDefaultDataStructureDefinitionFiles());
    }

    /**
     * Parses the given data structure definition files
     *
     * @param dataStructureDefinitionFiles URI of data structure definition files to be parsed
     * @return the {@link DataStructureDefinitions} defined in the given .ds files or empty {@link DataStructureDefinitions} if no files or definitions present
     */
    public static DataStructureDefinitions parseDataStructureDefinitionFiles(Collection<URI> dataStructureDefinitionFiles) {
        // parse data structure definitions out of all selected data structure definition files
        if (dataStructureDefinitionFiles == null || dataStructureDefinitionFiles.isEmpty()) {
            return new DataStructureDefinitions();
        }
        LOGGER.info(String.format("Merging data structure definition files: %s",
                                  String.join(", ", dataStructureDefinitionFiles.stream().map(URI::toString).toArray(String[]::new))));
        Set<FileSystem> fileSystems = new HashSet<>();
        dataStructureDefinitionFiles.stream()
                                    .filter(uri -> uri.getScheme().equals("jar"))
                                    .forEach(fileUri -> {
                                                 try {
                                                     fileSystems.add(FileSystems.getFileSystem(fileUri));
                                                 } catch (FileSystemNotFoundException e) {
                                                     try {
                                                         fileSystems.add(FileSystems.newFileSystem(fileUri, Collections.<String, String>emptyMap()));
                                                     } catch (IOException e1) {
                                                         e1.printStackTrace();
                                                     }
                                                 }
                                             }
                                    );

        Optional<String> mergedFiles = dataStructureDefinitionFiles.stream()
                                                                   .flatMap(uri -> {
                                                                       try {
                                                                           return Files.readAllLines(Paths.get(uri)).stream();
                                                                       } catch (IOException e) {
                                                                           e.printStackTrace();
                                                                           return Stream.empty();
                                                                       }
                                                                   })
                                                                   .reduce((line1, line2) -> line1 + line2);

        fileSystems.forEach(fs -> {
            try {
                fs.close();
            } catch (IOException e) {
                e.printStackTrace();
            }
        });

        if (mergedFiles.isPresent()) {
            Parser parser = new Parser(new Scanner(new ByteArrayInputStream(mergedFiles.get().getBytes(StandardCharsets.UTF_8))));
            parser.Parse();
            parser.getDataStructureDefinitions().parts.forEach(dsPart -> LOGGER.info(String.format("Parsed data structure part: %s", dsPart.getType())));
            return parser.getDataStructureDefinitions();
        }

        return new DataStructureDefinitions();
    }

    /**
     * Retrieves the default data structure definition files
     *
     * @return a {@link Set} of URIs that represent the default data structure definition files
     */
    public static Set<URI> getDefaultDataStructureDefinitionFiles() {
        URL defaultDataStructureDefinitionsFolderUrl = DataStructure.class.getResource(DEFAULT_DATA_STRUCTURE_DEFINITIONS_RESOURCE_FOLDER);
        URI defaultDataStructureDefinitionsFolderUri;
        Path defaultDataStructureDefinitionsFolderPath;
        FileSystem fs = null;
        try {
            defaultDataStructureDefinitionsFolderUri = defaultDataStructureDefinitionsFolderUrl.toURI();
            // JAR files are zip files and need some special handling in Java nio
            if (defaultDataStructureDefinitionsFolderUri.getScheme().equals("jar")) {
                fs = FileSystems.newFileSystem(defaultDataStructureDefinitionsFolderUri, Collections.<String, String>emptyMap());
            }

            Set<URI> fileUris = Files.list(Paths.get(defaultDataStructureDefinitionsFolderUri))
                                     .filter(path -> path.toString().endsWith(".ds"))
                                     .map(Path::toUri)
                                     .collect(Collectors.toSet());
            if (fs != null) {
                fs.close();
            }
            return fileUris;
        } catch (IOException | URISyntaxException e) {
            e.printStackTrace();
        }

        return new HashSet<>();
    }

}
