package at.jku.mevss.anttracks.util;

import at.jku.mevss.anttracks.diff.HeapDiff;
import at.jku.mevss.anttracks.diff.HeapDiffTraceParser;
import at.jku.mevss.anttracks.heap.*;
import at.jku.mevss.anttracks.heap.io.MetaDataReaderConfig;
import at.jku.mevss.anttracks.heap.io.MetaDataWriterConfig;
import at.jku.mevss.anttracks.heap.symbols.Symbols;
import at.jku.mevss.anttracks.parser.EventType;
import at.jku.mevss.anttracks.parser.TraceFile;
import at.jku.mevss.anttracks.parser.classdefinitions.ClassDefinitionsFile;
import at.jku.mevss.anttracks.parser.heap.HeapTraceParser;
import at.jku.mevss.anttracks.parser.symbols.SymbolsFile;
import at.jku.mevss.anttracks.parser.symbols.SymbolsParser;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.file.*;
import java.nio.file.attribute.BasicFileAttributes;
import java.util.Base64;

public class TestUtility {

    private static final String TRACE_STORAGE = "http://antonius.ssw.jku.at:8000/job/AntTracks/job/default/job/Trace-Storage/lastSuccessfulBuild/artifact/";
    private static final String LOGIN = "eg:test1234";

    public static String getExampleTrace(String name, String path) throws IOException, URISyntaxException {
        // downloads the trace from jenkins (path = save where?)

        if (!path.endsWith(File.separator)) {
            path += File.separator;
        }

        File outDir = new File(path);

        FileUtil.deleteRecursively(outDir);

        if (!outDir.mkdirs() && !outDir.exists()) {
            throw new IllegalStateException("Given path could not be created!");
        }

        // do: curl --basic -u eg:test1234 http://antonius.ssw.jku.at:8080/job/Trace-Storage/lastSuccessfulBuild/artifact/ClosureSize/trace/*zip*
        String encoding = Base64.getEncoder().encodeToString(LOGIN.getBytes("UTF-8"));

        HttpURLConnection connection = (HttpURLConnection) new URL(TRACE_STORAGE + name + "/*zip*").openConnection();
        connection.setRequestMethod("GET");
        connection.setDoOutput(true);
        connection.setRequestProperty("Authorization", "Basic " + encoding);
        InputStream content = connection.getInputStream();
        Path zipPath = Paths.get(path, "trace.zip");
        Files.copy(content, zipPath, StandardCopyOption.REPLACE_EXISTING);
        content.close();

        ZipFileUtil.unzip(new File(path), new File(zipPath.toAbsolutePath().toString()));
        FileUtil.deleteRecursively(zipPath.toAbsolutePath().toFile());

        Path pathPath = Paths.get(path);
        Files.walkFileTree(pathPath, new FileVisitor<Path>() {
            @Override
            public FileVisitResult preVisitDirectory(Path dir, BasicFileAttributes attrs) throws IOException {
                return FileVisitResult.CONTINUE;
            }

            @Override
            public FileVisitResult visitFile(Path file, BasicFileAttributes attrs) throws IOException {
                boolean successful = false;
                for (int i = 0; i < 10 && !successful; i++) {
                    successful = true;
                    try {
                        Path from = file;
                        Path to = Paths.get(file.getParent().getParent().toAbsolutePath().toString(), (i > 0 ? (i + 1) : "") + from.getFileName().toString());
                        System.out.println(String.format("Trying to move file from %s to %s", from, to));
                        Files.move(from, to);
                    } catch (Exception ex) {
                        System.out.println("Move unsuccessful");
                        System.err.println("Move unsuccessful");
                        if (i + 1 < 10) {
                            System.out.println("Try again by renaming file");
                            System.err.println("Try again by renaming file");
                        } else {
                            System.out.println("Abort, file move was unsuccessful!");
                            System.err.println("Abort, file move was unsuccessful!");
                            throw ex;
                        }
                        successful = false;
                    }
                }
                return FileVisitResult.CONTINUE;
            }

            @Override
            public FileVisitResult visitFileFailed(Path file, IOException exc) throws IOException {
                return FileVisitResult.CONTINUE;
            }

            @Override
            public FileVisitResult postVisitDirectory(Path dir, IOException exc) throws IOException {
                return FileVisitResult.CONTINUE;
            }
        });

        // TODO use some (more high-level) api ?
        // https://github.com/jenkinsci/java-client-api doesn't provide artifact download option!
        // JenkinsServer server = new JenkinsServer(new URI("http://brutus.ssw.jku.at:8080/jenkins"), "eg", "test1234");
        // System.out.println(server.getJob(name).getLastSuccessfulBuild().getUrl());

        return Paths.get(path).toAbsolutePath().toString();
    }

    public static PreprocessedHeap preprocess(String path, int dumps, HeapListener listener, boolean forcePreProcess) throws Exception {
        // parse through complete trace file such that metadata is written
        File trace = new File(path);
        File symbolsFile = SymbolsFile.findSymbolsFileToTrace(trace);
        File classDefinitionFile = ClassDefinitionsFile.findClassDefinitionsFileToTrace(trace);
        File metaDir = new File(symbolsFile.getParent() + File.separator + Consts.ANT_META_DIRECTORY);
        SymbolsParser symbolsParser = new SymbolsParser(symbolsFile, classDefinitionFile, null, Symbols.CALLCONTEXT_NONE);
        Symbols symbols = symbolsParser.parse();

        boolean hasCorrectMeta = false;
        if (metaDir.exists()) {
            String metaHeaderFilePath = metaDir.getAbsolutePath() + File.separator + Consts.HEADERS_META_FILE;
            if (TraceFile.hasValidMetaData(trace, metaHeaderFilePath)) {
                hasCorrectMeta = true;
            } else {
                FileUtil.deleteRecursively(metaDir);
            }
        }
        if (forcePreProcess || !hasCorrectMeta) {
            HeapTraceParser metaDataParser = new HeapTraceParser(symbols, new MetaDataWriterConfig(metaDir.getAbsolutePath(), dumps, false));
            if (listener != null) {
                metaDataParser.addHeapListener(listener);
            }
            metaDataParser.parse();
        }

        return new PreprocessedHeap(trace, symbolsFile, classDefinitionFile, metaDir, symbols);
    }

    public static PreprocessedHeap preprocess(String path, boolean forcePreProcess) throws Exception {
        return preprocess(path, 100, null, forcePreProcess);
    }

    public static DetailedHeap getHeap(PreprocessedHeap preprocessedHeap, GarbageCollectionLookup lookup) throws Exception {
        HeapTraceParser metaDataParser = new HeapTraceParser(preprocessedHeap.symbols, new MetaDataReaderConfig(preprocessedHeap.metaDir.getAbsolutePath()), lookup);
        return metaDataParser.parse();
    }

    public static DetailedHeap getHeap(String path, GarbageCollectionLookup lookup) throws Exception {
        // returns the heap at GCEND/START) after nth garbage collection (=gcIndex)
        File trace = new File(path);
        SymbolsParser symbolsParser = new SymbolsParser(SymbolsFile.findSymbolsFileToTrace(trace),
                                                        ClassDefinitionsFile.findClassDefinitionsFileToTrace(trace),
                                                        null,
                                                        Symbols.CALLCONTEXT_NONE);
        Symbols symbols = symbolsParser.parse();
        HeapTraceParser metaDataParser = new HeapTraceParser(symbols, new MetaDataReaderConfig(path + File.separator + Consts.ANT_META_DIRECTORY), lookup);

        return metaDataParser.parse();
    }

    public static HeapDiff getHeapDiff(PreprocessedHeap preprocessedHeap, GarbageCollectionLookup startLookup, GarbageCollectionLookup endLookup) throws Exception {
        HeapDiffTraceParser diffParser = new HeapDiffTraceParser(preprocessedHeap.symbols,
                                                                 new MetaDataReaderConfig(preprocessedHeap.metaDir.getAbsolutePath()),
                                                                 startLookup,
                                                                 endLookup);
        return diffParser.parse();
    }

    public static HeapDiff getHeapDiffEndToEnd(PreprocessedHeap preprocessedHeap, int startFullGCId, int endFullGCId) throws Exception {
        return getHeapDiff(preprocessedHeap,
                           new GarbageCollectionLookup(EventType.GC_END,
                                                       GarbageCollectionType.MAJOR,
                                                       preprocessedHeap.symbols.causes.getByName("System.gc()"),
                                                       startFullGCId),
                           new GarbageCollectionLookup(EventType.GC_END,
                                                       GarbageCollectionType.MAJOR,
                                                       preprocessedHeap.symbols.causes.getByName("System.gc()"),
                                                       endFullGCId));
    }

    public static HeapDiff getHeapDiffStartToEnd(PreprocessedHeap preprocessedHeap, int startFullGCId, int endFullGCId) throws Exception {
        return getHeapDiff(preprocessedHeap,
                           new GarbageCollectionLookup(EventType.GC_START,
                                                       GarbageCollectionType.MAJOR,
                                                       preprocessedHeap.symbols.causes.getByName("System.gc()"),
                                                       startFullGCId),
                           new GarbageCollectionLookup(EventType.GC_END,
                                                       GarbageCollectionType.MAJOR,
                                                       preprocessedHeap.symbols.causes.getByName("System.gc()"),
                                                       endFullGCId));
    }

    public static HeapDiff getHeapDiffStartToStart(PreprocessedHeap preprocessedHeap, int startFullGCId, int endFullGCId) throws Exception {
        return getHeapDiff(preprocessedHeap,
                           new GarbageCollectionLookup(EventType.GC_START,
                                                       GarbageCollectionType.MAJOR,
                                                       preprocessedHeap.symbols.causes.getByName("System.gc()"),
                                                       startFullGCId),
                           new GarbageCollectionLookup(EventType.GC_START,
                                                       GarbageCollectionType.MAJOR,
                                                       preprocessedHeap.symbols.causes.getByName("System.gc()"),
                                                       endFullGCId));
    }

    public static DetailedHeap getHeap(String path, int gcIndex, EventType event) throws Exception {
        return getHeap(path, new GarbageCollectionLookup(event, null, null, gcIndex));
    }

    public static void main(String args[]) throws Exception {
        // test
        getExampleTrace("PointerTraces/ClosureSize/trace", "./ClosureSize");
        preprocess("./ClosureSize/trace.zip", true);
        DetailedHeap heap = getHeap("/home/eg/ClosureSize/trace.zip", 5, EventType.GC_END);
        System.out.println(heap);
        System.out.println(heap.rootPtrs.size());
    }

    // TODO should downloaded tracefiles (used in preprocess() and getHeap()) be deleted at some point?
}
