package at.jku.mevss.anttracks.heap.datastructures

import at.jku.mevss.anttracks.heap.DetailedHeap
import at.jku.mevss.anttracks.heap.objects.ObjectInfo
import at.jku.mevss.anttracks.heap.space.Space
import at.jku.mevss.anttracks.heap.space.SpaceType
import at.jku.mevss.anttracks.heap.symbols.Symbols
import at.jku.mevss.anttracks.parser.*
import at.jku.mevss.anttracks.parser.heap.HeapEventHandler
import at.jku.mevss.anttracks.util.TraceException

class DataStructureDevelopmentEventHandler(initialHeap: DetailedHeap,
                                           val symbols: Symbols,
                                           val parsingInfo: ParsingInfo,
                                           val checkMovesFun: () -> Boolean,
                                           val reportGCFun: (Long) -> Unit,
                                           val dataStructures: () -> Array<MovingObject>?) : HeapEventHandler(initialHeap, symbols, parsingInfo) {
    override fun doParseGCStart(gcInfo: ParserGCInfo?, start: Long, end: Long, check: Boolean) {
        super.doParseGCStart(gcInfo, start, end, check)
        reportGCFun(gcInfo!!.time)
        // Full GCs always clean the whole heap
        // Minor GCs get following GCInfo events that mark spaces for cleaning
        if (gcInfo.type.full) {
            heap.spacesUncloned.forEach {
                markObjectsInSpaceForCollection(it)
            }
        }
    }

    override fun doParseGCEnd(gcInfo: ParserGCInfo?, threadlocalHeap: ThreadLocalHeap?, start: Long, end: Long, failed: Boolean, check: Boolean) {
        super.doParseGCEnd(gcInfo, threadlocalHeap, start, end, failed, check)
        reportGCFun(gcInfo!!.time)

        // If data structures exist, check for non-moved objects that were expected to be moved
        dataStructures()?.filter { it.died() }?.forEach { it.stopTracking() }
        dataStructures()?.forEach { it.clearMoveInfo() }

        if (TraceParser.CONSISTENCY_CHECK) {
            dataStructures()?.filter { it.currentAddress >= 0 }?.forEach {
                try {
                    if (heap.getObjectInFront(it.currentAddress) == null) {
                        throw TraceException("Object expected at address %,d but not found!".format(it.currentAddress))
                    }
                } catch (_: Exception) {
                    throw TraceException("Object expected at address %,d but not found!".format(it.currentAddress))
                }
            }
        }
    }

    override fun doGCMove(type: EventType?,
                          threadLocalHeap: ThreadLocalHeap?,
                          obj: ObjectInfo?,
                          fromAddr: Long,
                          toAddr: Long,
                          toSpaceType: SpaceType?): Long = super.doGCMove(type, threadLocalHeap, obj, fromAddr, toAddr, toSpaceType).also { updateAddress(fromAddr, it) }

    override fun doParseGCInfo(index: Int, gcId: Int) {
        super.doParseGCInfo(index, gcId)
        markObjectsInSpaceForCollection(heap.getSpace(index))
    }

    override fun doParseGCMoveRegion(type: EventType?,
                                     fromAddr: Long,
                                     toAddr: Long,
                                     toSpaceType: SpaceType?,
                                     numOfObjects: Int,
                                     threadLocalHeap: ThreadLocalHeap?): MutableList<ObjectInfo> =
            super.doParseGCMoveRegion(type, fromAddr, toAddr, toSpaceType, numOfObjects, threadLocalHeap).apply {
                val fromToDiff = toAddr - fromAddr
                fold(fromAddr) { from: Long, obj: ObjectInfo ->
                    updateAddress(from, from + fromToDiff)
                    from + obj.size
                }
            }

    private fun markObjectsInSpaceForCollection(space: Space) {
        fun spaceContainFunction(ds: MovingObject): Boolean = ds.currentAddress >= space.info.address && ds.currentAddress < space.info.end
        dataStructures()?.filter(::spaceContainFunction)?.forEach(MovingObject::markInCollectedSpace)
    }

    private fun updateAddress(from: Long, to: Long) {
        if (checkMovesFun()) {
            dataStructures()?.find({ it.currentAddress == from })?.move(to)
        }
    }
}