/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.trace.gui.chart.base;

import at.jku.mevss.anttracks.util.ApplicationStatistics;
import at.jku.mevss.trace.gui.model.ChartSelection;

import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;

public class ChartSynchronizer {

    private final Object lock;
    private final List<ZoomListener> zoomListeners = new CopyOnWriteArrayList<ZoomListener>();
    private final List<SelectionListener> selectionListeners = new CopyOnWriteArrayList<SelectionListener>();

    private double zoomXUpperBound = Double.MAX_VALUE;
    private double zoomXLowerBound = Double.MIN_VALUE;
    private int zoom;
    private int domain;
    private SelectedItemsList selections;

    // Last chart ID is always set when .sync() is called, no matter if a new
    // selection was added, an existing removed, etc
    private String lastChartId = "";
    private boolean inCompareMode;

    // lastSelection is set during .sync() if a new selection has been added. If
    // an existing selection has been removed, this will be null. If the
    // selection state has not been changed (nothing removed, nothing added)
    // lastSelection does not change
    private ChartSelection lastSelection;

    {
        lock = new Object();
        zoom = 0;
        domain = 0;
        selections = new SelectedItemsList();
        inCompareMode = false;
    }

    public ChartSelection getLastChartSelection() {
        return lastSelection;
    }

    public double getZoomXLowerBound() {
        return zoomXLowerBound;
    }

    public double getZoomXUpperBound() {
        return zoomXUpperBound;
    }

    public int getZoom() {
        synchronized (lock) {
            return zoom;
        }
    }

    public int getDomain() {
        synchronized (lock) {
            return domain;
        }
    }

    public SelectedItemsList getSelection() {
        synchronized (lock) {
            return selections;
        }
    }

    public boolean isInCompareMode() {
        synchronized (lock) {
            return inCompareMode;
        }
    }

    public void setZoomXBounds(double lowerBound, double upperBound) {
        ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance().createMeasurement("Chart Sync: Sync Zoom");
        synchronized (lock) {
            zoomXLowerBound = lowerBound;
            zoomXUpperBound = upperBound;
        }
        zoomListeners.forEach(l -> l.zoomChanged(this));
        m.end();
    }

    public void setZoom(int zoom) {
        synchronized (lock) {
            this.zoom = zoom;
        }
        zoomListeners.forEach(l -> l.zoomChanged(this));
    }

    public void setDomain(int domain) {
        synchronized (lock) {
            this.domain = domain;
        }
        zoomListeners.forEach(l -> l.domainChanged(this));
    }

    public void sync(ChartSelection selection) {
        synchronized (lock) {
            ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance().createMeasurement("Chart Sync: Sync Selection");
            // Clear all selections if different charts have been clicked
            if (!selection.chartId.equals(lastChartId)) {
                selections.clear();
            }

            if (selections.containsItemInAnySeries(selection)) {
                // Clicked on already existing selection
                selections.remove(selection);
                lastSelection = null;
            } else {
                if (isInCompareMode()) {
                    // Compare mode. New selection only allowed if less than 2
                    // points are selected currently
                    if (selections.size() != 2) {
                        selections.add(selection);
                        lastSelection = selection;
                    }
                } else {
                    // Single-selection mode. Clear current selection and then
                    // add new selection
                    selections.clear();
                    selections.add(selection);
                    lastSelection = selection;
                }
            }
            selections.sort();

            lastChartId = selection.chartId;
            selectionListeners.forEach(l -> l.selectionChanged(this));
            m.end();
        }
    }

    public void setInCompareMode(boolean inCompareMode) {
        synchronized (lock) {
            this.inCompareMode = inCompareMode;
        }

        if (!inCompareMode) {
            selections.clear();
            lastSelection = null;
            lastChartId = "";
        }
        selectionListeners.forEach(l -> {
            l.inCompareModeChanged(this);
            l.selectionChanged(this);
        });
    }

    public String getLastChartId() {
        return lastChartId;
    }

    public void addZoomListener(ZoomListener zoomListener) {
        if (zoomListener != null) {
            zoomListeners.add(zoomListener);
            zoomListener.zoomChanged(this);
            zoomListener.domainChanged(this);
        }
    }

    public void addSelectionListener(SelectionListener selectionListener) {
        if (selectionListener != null) {
            selectionListeners.add(selectionListener);
            selectionListener.selectionChanged(this);
        }
    }

    public void removeZoomListener(ZoomListener zoomListener) {
        if (zoomListener != null) {
            zoomListeners.remove(zoomListener);
        }
    }

    public void removeSelectionListener(SelectionListener selectionListener) {
        if (selectionListener != null) {
            zoomListeners.remove(selectionListener);
        }
    }

    public void clear() {
        synchronized (lock) {
            zoom = 0;
            domain = 0;
            selections.clear();
            inCompareMode = false;
            lastChartId = "";
            lastSelection = null;
            zoomListeners.forEach(l -> l.zoomChanged(this));
            zoomListeners.forEach(l -> l.domainChanged(this));
            selectionListeners.forEach(l -> l.selectionChanged(this));
        }
    }

    public static interface ZoomListener {
        public abstract void zoomChanged(ChartSynchronizer sender);

        public abstract void domainChanged(ChartSynchronizer sender);
    }

    public static interface SelectionListener {
        public abstract void inCompareModeChanged(ChartSynchronizer sender);

        public abstract void selectionChanged(ChartSynchronizer sender);
    }
}
