/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.trace.gui.chart.fxchart.xy.linepane;

import at.jku.mevss.anttracks.heap.GarbageCollectionCause;
import at.jku.mevss.anttracks.heap.StatisticGCInfo;
import at.jku.mevss.trace.gui.chart.base.AxisScaler;
import at.jku.mevss.trace.gui.chart.base.ChartSynchronizer;
import at.jku.mevss.trace.gui.chart.fxchart.xy.linepane.base.LineJavaFXChartPane;
import at.jku.mevss.trace.gui.model.AppInfo;
import at.jku.mevss.trace.gui.utils.FXMLUtil;
import at.jku.mevss.trace.gui.utils.WindowUtil;
import javafx.application.Platform;
import javafx.collections.ObservableList;
import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.scene.chart.XYChart;
import javafx.scene.chart.XYChart.Data;
import javafx.scene.text.Text;

import java.util.*;
import java.util.stream.Collectors;

public class GCTypesLineChartPane extends LineJavaFXChartPane<AppInfo> implements ShowGCCause {

    private static final class Info {
        public final GarbageCollectionCause cause;
        public final boolean failed;

        public Info(GarbageCollectionCause cause, boolean failed) {
            this.cause = cause;
            this.failed = failed;
        }
    }

    private static boolean useFirstDerivation = false;

    final Set<Integer> displayedCauses = new HashSet<>();
    private Map<Number, Info> infos;

    public GCTypesLineChartPane() {
        FXMLUtil.load(this, GCTypesLineChartPane.class);
    }

    @Override
    public void init(ChartSynchronizer chartSynchronizer) {
        super.init(chartSynchronizer);
        chart.setTitle("GC Pauses");
    }

    @Override
    public ObservableList<XYChart.Series<Number, Number>> updateDataSet(AppInfo appInfo) {
        double maxX = getMaxXValue(chart.getData());

        infos = new HashMap<>();

        if (useFirstDerivation) {
            double minor = 0;
            double major = 0;

            for (int i = 0; i + 1 < appInfo.getStatistics().size(); i += 2) {
                StatisticGCInfo startInfo = appInfo.getStatistics().get(i).info;
                StatisticGCInfo endInfo = appInfo.getStatistics().get(i + 1).info;

                double time = endInfo.time;
                double start = startInfo.time;
                double end = endInfo.time;
                double duration = end - start;

                if (startInfo.type.isFull()) {
                    major += duration;
                } else {
                    minor += duration;
                }

                if (time > maxX) {
                    minorSeries.getData().add(new Data<>(time, minor));
                    majorSeries.getData().add(new Data<>(time, major));
                    infos.put(time, new Info(startInfo.cause, endInfo.failed));
                }
            }
        } else {
            for (int i = 0; i + 1 < appInfo.getStatistics().size(); i += 2) {
                StatisticGCInfo startInfo = appInfo.getStatistics().get(i).info;
                StatisticGCInfo endInfo = appInfo.getStatistics().get(i + 1).info;
                long start = startInfo.time;
                long end = endInfo.time;
                long duration = end - start;

                if (start > maxX) {
                    XYChart.Series<Number, Number> series = startInfo.type.isFull() ?
                                                            (startInfo.concurrent ? majorConcurrentSeries : majorSeries) :
                                                            (startInfo.concurrent ? minorConcurrentSeries : minorSeries);

                    XYChart.Data<Number, Number> firstPoint = new Data<>(1.0 * start, 0);
                    XYChart.Data<Number, Number> secondPoint = new Data<>(1.0 * start, duration);
                    XYChart.Data<Number, Number> thirdPoint = new Data<>(1.0 * end, duration);
                    XYChart.Data<Number, Number> fourthPoint = new Data<>(1.0 * end, 0);
                    Platform.runLater(() -> {
                        series.getData().add(firstPoint);
                        series.getData().add(secondPoint);
                        series.getData().add(thirdPoint);
                        series.getData().add(fourthPoint);
                    });
                    infos.put(1.0 * start, new Info(startInfo.cause, endInfo.failed));
                }
            }
        }

        List<XYChart.Series<Number, Number>> series = new ArrayList<>();
        if (!minorSeries.getData().isEmpty()) {
            series.add(minorSeries);
        }
        if (!majorSeries.getData().isEmpty()) {
            series.add(majorSeries);
        }
        if (!minorConcurrentSeries.getData().isEmpty()) {
            series.add(minorConcurrentSeries);
        }
        if (!majorConcurrentSeries.getData().isEmpty()) {
            series.add(majorConcurrentSeries);
        }
        if (!unknownSeries.getData().isEmpty()) {
            series.add(unknownSeries);
        }

        displayedCauses.addAll(Arrays.stream(appInfo.getSymbols().causes.getAll()).filter(c -> !c.common).map(c -> c.id).collect(Collectors.toList()));

        return chart.getData();
    }

    XYChart.Series<Number, Number> minorSeries;
    XYChart.Series<Number, Number> majorSeries;
    XYChart.Series<Number, Number> minorConcurrentSeries;
    XYChart.Series<Number, Number> majorConcurrentSeries;
    XYChart.Series<Number, Number> unknownSeries;

    @Override
    protected void initializeChartSeries() {
        minorSeries = new XYChart.Series<Number, Number>();
        minorSeries.setName("Minor GC");
        majorSeries = new XYChart.Series<Number, Number>();
        majorSeries.setName("Major GC");
        minorConcurrentSeries = new XYChart.Series<Number, Number>();
        minorConcurrentSeries.setName("Minor GC (concurrent)");
        majorConcurrentSeries = new XYChart.Series<Number, Number>();
        majorConcurrentSeries.setName("Major GC (concurrent)");
        unknownSeries = new XYChart.Series<Number, Number>();
        unknownSeries.setName("Unknown GC");
        chart.getData().add(minorSeries);
        chart.getData().add(majorSeries);
        chart.getData().add(minorConcurrentSeries);
        chart.getData().add(majorConcurrentSeries);
        chart.getData().add(unknownSeries);
    }

    @Override
    protected void updateDataLabelContent() {
        dataLabels.forEach((dataPoint, dataLabel) -> getChildren().remove(dataLabel));
        dataLabels.clear();

        chart.getData().forEach(series -> series.getData().forEach(dataPoint -> {
            if (dataPoint.getYValue().doubleValue() > 0) {
                Number value = dataPoint.getXValue();
                Info info = infos.get(value);
                if (info != null && (displayedCauses.contains(info.cause.id) || info.failed)) {
                    Text dataLabel = new Text(info.cause.name + (info.failed ? "(failed)" : ""));
                    dataLabel.getStyleClass().add("data-label");
                    dataLabel.setManaged(false);
                    updateDataLabelLayout(dataPoint, dataLabel);
                    dataLabels.put(dataPoint, dataLabel);
                    getChildren().add(dataLabel);
                }
            }
        }));
    }

    @Override
    protected EventHandler<ActionEvent> getOnConfigureAction() {
        return ae -> {
            GCTypesChartConfigurationDialog dialog = new GCTypesChartConfigurationDialog();
            dialog.init(this);
            WindowUtil.centerInMainFrame(dialog);
            dialog.showAndWait();
        };
    }

    @Override
    public AxisScaler<ObservableList<XYChart.Series<Number, Number>>> getScaler() {
        return new AxisScaler.JavaFXAxisScaler.GCScaler(this);
    }

    @Override
    public Set<Integer> getDisplayedCauses() {
        return displayedCauses;
    }

    @Override
    public void updateDisplayCause(boolean show, GarbageCollectionCause... causes) {
        if (show) {
            displayedCauses.addAll(Arrays.stream(causes).map(c -> c.id).collect(Collectors.toList()));
        } else {
            displayedCauses.removeAll(Arrays.stream(causes).map(c -> c.id).collect(Collectors.toList()));
        }
        updateDataLabelContent();
    }
}
