package at.jku.mevss.trace.gui.chart.jfreechart.xy.stackedarea.memory;

import at.jku.mevss.anttracks.heap.Tag;
import at.jku.mevss.anttracks.heap.statistics.Statistics;
import at.jku.mevss.trace.gui.chart.base.ApplicationChartFactory;
import at.jku.mevss.trace.gui.chart.base.AxisScaler;
import at.jku.mevss.trace.gui.chart.base.ChartSynchronizer;
import at.jku.mevss.trace.gui.chart.jfreechart.base.JFreeChartFactory;
import at.jku.mevss.trace.gui.chart.jfreechart.xy.stackedarea.base.StackedAreaJFreeChartPane;
import at.jku.mevss.trace.gui.chart.jfreechart.xy.util.AddALotXYSeries;
import at.jku.mevss.trace.gui.model.AppInfo;
import at.jku.mevss.trace.gui.utils.FXMLUtil;
import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.scene.control.Alert;
import javafx.scene.control.ButtonBar;
import javafx.scene.control.ButtonType;
import javafx.stage.StageStyle;
import org.jfree.chart.labels.StandardXYToolTipGenerator;
import org.jfree.data.xy.AbstractIntervalXYDataset;
import org.jfree.data.xy.DefaultTableXYDataset;
import org.jfree.data.xy.XYDataItem;

import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

public class MemoryStackedAreaJFreeChart extends StackedAreaJFreeChartPane<AppInfo> {

    private ApplicationChartFactory.MemoryConsumptionUnit memoryUnit;
    private boolean showTags = true;
    private List<Tag> tags;

    public MemoryStackedAreaJFreeChart() {
        FXMLUtil.load(this, MemoryStackedAreaJFreeChart.class);
    }

    public void init(ChartSynchronizer chartSynchronizer, ApplicationChartFactory.MemoryConsumptionUnit memoryUnit) {
        this.memoryUnit = memoryUnit;
        super.init(chartSynchronizer);
    }

    public DefaultTableXYDataset createDataSet(AppInfo appInfo) {
        tags = appInfo.getTags();

        AddALotXYSeries edenSeries = new AddALotXYSeries("Eden", false);
        List<XYDataItem> edenData = new ArrayList<>();
        AddALotXYSeries survivorSeries = new AddALotXYSeries("Survivor", false);
        List<XYDataItem> survivorData = new ArrayList<>();
        AddALotXYSeries oldSeries = new AddALotXYSeries("Old", false);
        List<XYDataItem> oldData = new ArrayList<>();

        for (int i = 0; i < appInfo.getStatistics().size(); i++) {
            Statistics stat = appInfo.getStatistics().get(i);
            double time = stat.info.time;

            // Prevent duplicate x values
            if (i == 0 || time != appInfo.getStatistics().get(i - 1).info.time) {
                switch (memoryUnit) {
                    case BYTES:
                        edenData.add(new XYDataItem(time, stat.eden.memoryConsumption.getBytes()));
                        survivorData.add(new XYDataItem(time, stat.survivor.memoryConsumption.getBytes()));
                        oldData.add(new XYDataItem(time, stat.old.memoryConsumption.getBytes()));
                        break;
                    case OBJECTS:
                        edenData.add(new XYDataItem(time, stat.eden.memoryConsumption.getObjects()));
                        survivorData.add(new XYDataItem(time, stat.survivor.memoryConsumption.getObjects()));
                        oldData.add(new XYDataItem(time, stat.old.memoryConsumption.getObjects()));
                        break;
                    default:
                        throw new IllegalStateException();
                }
            }
        }

        edenSeries.setData(edenData, false);
        survivorSeries.setData(survivorData, false);
        oldSeries.setData(oldData, false);

        DefaultTableXYDataset dataset = new DefaultTableXYDataset();
        dataset.addSeries(oldSeries);
        dataset.addSeries(survivorSeries);
        dataset.addSeries(edenSeries);

        return dataset;
    }

    @Override
    protected void initializeChart() {
        chartViewer.setChart(JFreeChartFactory.createStackedXYAreaChart(this,
                                                                        memoryUnit.getLabel(),
                                                                        null,
                                                                        "x",
                                                                        "y",
                                                                        new DefaultTableXYDataset(),
                                                                        (dataset, series, item) -> {
                                                                            if (showTags) {
                                                                                try {
                                                                                    long x = dataset.getX(series, item).longValue();
                                                                                    return tags.stream()
                                                                                               .filter(tag -> tag.gcInfo.time == x)
                                                                                               .map(tag -> tag.text)
                                                                                               .collect(Collectors.joining("; "));
                                                                                } catch (Exception ex) {
                                                                                    return "";
                                                                                }
                                                                            } else {
                                                                                return "";
                                                                            }
                                                                        },
                                                                        new StandardXYToolTipGenerator(),
                                                                        null,
                                                                        chartSynchronizer));
    }

    @Override
    protected EventHandler<ActionEvent> getOnConfigureAction() {
        return new EventHandler<ActionEvent>() {
            @Override
            public void handle(ActionEvent actionEvent) {
                Alert dialog = new Alert(Alert.AlertType.INFORMATION);
                dialog.initStyle(StageStyle.UTILITY);
                dialog.setTitle("Datastructure development analysis");
                dialog.setHeaderText(dialog.getTitle());
                dialog.setContentText("Do you want to display tags?");
                //dialog.getDialogPane().setExpandableContent(additionalContent);

                ButtonType buttonTypeYes = new ButtonType("Yes", ButtonBar.ButtonData.YES);
                ButtonType buttonTypeNo = new ButtonType("No", ButtonBar.ButtonData.NO);
                ButtonType buttonTypeCancel = new ButtonType("Cancel", ButtonBar.ButtonData.CANCEL_CLOSE);

                dialog.getButtonTypes().setAll(buttonTypeYes, buttonTypeNo, buttonTypeCancel);

                ButtonType response = dialog.showAndWait().orElse(null);
                if (response == buttonTypeYes) {
                    showTags = true;
                    chartViewer.getChart().fireChartChanged();
                } else if (response == buttonTypeNo) {
                    showTags = false;
                    chartViewer.getChart().fireChartChanged();
                }
            }
        };
    }

    @Override
    public AxisScaler<AbstractIntervalXYDataset> getScaler() {
        switch (memoryUnit) {
            case OBJECTS:
                return new AxisScaler.JFreeChartAxisScaler.ObjectsScaler(this);
            case BYTES:
                return new AxisScaler.JFreeChartAxisScaler.BytesScaler(this);
            default:
                LOGGER.warning("No scaler defined for " + memoryUnit);
                return new AxisScaler.JFreeChartAxisScaler.NoAxisScaler(this);
        }
    }
}
