/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.trace.gui.classification.filter.component.selectionpane;

import at.jku.mevss.anttracks.classification.Filter;
import at.jku.mevss.anttracks.classification.FilterFactory;
import at.jku.mevss.anttracks.util.ImagePack;
import at.jku.mevss.trace.gui.classification.component.selectionpane.ClassificationSelectionPane;
import at.jku.mevss.trace.gui.classification.dialog.properties.ClassificationPropertiesDialog;
import at.jku.mevss.trace.gui.classification.filter.dialog.create.CreateFilterDialog;
import at.jku.mevss.trace.gui.classification.filter.dialog.edit.EditFilterDialog;
import at.jku.mevss.trace.gui.component.clickablepane.ClickablePane;
import at.jku.mevss.trace.gui.model.AvailableClassifierInfo;
import at.jku.mevss.trace.gui.utils.FXMLUtil;
import at.jku.mevss.trace.gui.utils.WindowUtil;
import javafx.scene.control.Dialog;
import javafx.scene.layout.Pane;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public final class ObjectFilterSelectionPane extends ClassificationSelectionPane<Filter, FilterFactory> {

    List<Pane> availableFilterPaneList;

    public ObjectFilterSelectionPane() {
        super();
        FXMLUtil.load(this, ObjectFilterSelectionPane.class);
    }

    @Override
    public void init(AvailableClassifierInfo availableClassifierInfo, int maxSelect, boolean prohibitDuplicates) {
        availableFilterPaneList = new ArrayList<>();
        super.init(availableClassifierInfo, maxSelect, prohibitDuplicates);
        Filter onlyDataStructureHeadsFilter = availableClassifierInfo.getAvailableFilter().get("Only data structure heads");
        if (onlyDataStructureHeadsFilter != null) {
            addExcludedFromAvailable(onlyDataStructureHeadsFilter);
        }
        refresh();
    }

    // SELECTED

    @Override
    protected ImagePack getIcon(Filter filter) {
        return null;
    }

    @Override
    protected String getName(Filter filter) {
        return filter.getName();
    }

    @Override
    protected String[] getSelectedDescriptions(Filter filter) {
        return Stream.of(filter.configurableProperties()).filter(prop -> prop.getPresentationName() != null && prop.get() != null).map(prop -> {
            return prop.getPresentationName() + ": " + prop.get();
        }).toArray(String[]::new);
    }

    @Override
    protected String getSelectedToolTip(Filter filter) {
        return filter.getDesc() + (filter.configurableProperties().length > 0 ? "\nRight click to configure" : "");
    }

    @Override
    protected boolean selectedRightClick(Filter filter) {
        if (filter == null) {
            return false;
        }
        if (filter.configurableProperties().length == 0) {
            return false;
        }

        ClassificationPropertiesDialog dialog = new ClassificationPropertiesDialog(filter, availableClassifierInfo, this);
        WindowUtil.centerInMainFrame(dialog);
        Optional<Boolean> dialogResult = dialog.showAndWait();
        if (dialogResult.isPresent() && dialogResult.get()) {
            // Refresh UI if some filter property has been changed
            refreshSelected();
            return true;
        }
        return false;
    }

    // AVAILABLE

    @Override
    protected ImagePack getAvailableIcon(FilterFactory fac) {
        return null;
    }

    @Override
    protected String getAvailableName(FilterFactory fac) {
        return fac.getName();
    }

    @Override
    protected String[] getAvailableDescriptions(FilterFactory fac) {
        return new String[]{};
    }

    @Override
    protected String getAvailableToolTip(FilterFactory fac) {
        return fac.getDesc() + (fac instanceof FilterFactory && fac.isOnTheFlyCompilable() ? "\nRight click to edit" : "");
    }

    @Override
    protected void addAvailablePane(FilterFactory fac, ClickablePane pane) {
        availableFilterPaneList.add(pane);
        availablePane.getChildren().add(pane);
    }

    @Override
    protected boolean availableRightClickAvailable(FilterFactory fac) {
        return fac != null && fac instanceof FilterFactory && fac.isOnTheFlyCompilable();
    }

    @Override
    public List<FilterFactory> getAvailableList() {
        return availableClassifierInfo.getAvailableFilter().stream()
                                      .filter(fac -> !isExcludedFromAvailable(fac.create()))
                                      .collect(Collectors.toList());
    }

    @Override
    protected CreateFilterDialog getCreateDialog() throws IOException {
        return new CreateFilterDialog(availableClassifierInfo.getSymbolsSupplier(),
                                      availableClassifierInfo.getFastHeapSupplier(),
                                      availableClassifierInfo.getAvailableFilter());
    }

    @Override
    protected Dialog<? extends FilterFactory> getEditDialog(FilterFactory fac) throws IOException {
        if (fac instanceof FilterFactory) {
            return new EditFilterDialog(availableClassifierInfo.getSymbolsSupplier(),
                                        availableClassifierInfo.getFastHeapSupplier(),
                                        fac,
                                        availableClassifierInfo.getAvailableFilter());
        } else {
            return null;
        }
    }

    @Override
    protected Comparator<FilterFactory> comperator() {
        return (o1, o2) -> o1.getName().compareTo(o2.getName());
    }

    @Override
    protected void clearAvailablePanes() {
        availablePane.getChildren().removeAll(availableFilterPaneList);
        availableFilterPaneList.clear();
    }

    @Override
    protected boolean removeSelected(FilterFactory y) {
        List<Filter> remainingSelected = getSelected().stream().filter(selected -> !selected.getName().equals(y.getName())).collect(Collectors.toList());
        if (getSelected().size() != remainingSelected.size()) {
            resetSelected(remainingSelected);
            return true;
        }
        return false;
    }
}
