/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.trace.gui.component.clickablepane;

import at.jku.mevss.anttracks.util.ImagePack;
import at.jku.mevss.trace.gui.utils.FXMLUtil;
import javafx.application.Platform;
import javafx.beans.property.*;
import javafx.fxml.FXML;
import javafx.scene.control.Button;
import javafx.scene.control.Label;
import javafx.scene.control.Separator;
import javafx.scene.control.Tooltip;
import javafx.scene.image.ImageView;
import javafx.scene.layout.HBox;
import javafx.scene.text.Text;
import javafx.scene.text.TextFlow;

import java.util.Arrays;
import java.util.List;
import java.util.concurrent.CompletableFuture;
import java.util.stream.Collectors;

public class ClickablePane extends HBox {

    @FXML
    private HBox container;
    @FXML
    private Label titleLabel;
    @FXML
    private TextFlow descriptionTextFlow;
    @FXML
    private Separator separator;
    @FXML
    private Button button;

    private final BooleanProperty clickableProperty = new SimpleBooleanProperty();
    private final StringProperty titleProperty = new SimpleStringProperty();
    private final SimpleObjectProperty<ImageView> iconProperty = new SimpleObjectProperty<ImageView>();
    private final ObjectProperty<List<String>> descriptionsProperty = new SimpleObjectProperty<List<String>>();
    private final StringProperty buttonTextProperty = new SimpleStringProperty();
    private final ObjectProperty<Tooltip> tooltipProperty = new SimpleObjectProperty<Tooltip>();
    private final ObjectProperty<Runnable> onRightClickProperty = new SimpleObjectProperty<Runnable>();
    private final ObjectProperty<Runnable> onLeftClickProperty = new SimpleObjectProperty<Runnable>();
    private final ObjectProperty<Runnable> onDoubleLeftClickProperty = new SimpleObjectProperty<Runnable>();

    public ClickablePane() {
        FXMLUtil.load(this, ClickablePane.class);
    }

    public void initialize(ImagePack icon,
                           String title,
                           String[] descriptions,
                           String buttonText,
                           String tooltip,
                           Runnable onDoubleLeftClick,
                           Runnable onRightClick,
                           boolean clickable) {
        iconProperty.set(icon != null ? icon.getAsNewNode() : null);
        iconProperty.addListener((observable, oldValue, newValue) -> {
            titleLabel.setGraphic(newValue);
        });

        titleProperty.set(title);
        titleProperty.addListener((observable, oldValue, newValue) -> {
            titleLabel.setText(newValue);
        });

        descriptionsProperty.addListener((observable, oldValue, newValue) -> {
            descriptionTextFlow.getChildren().setAll(createTextFlowElements(newValue));
                    /*
            descriptionTextFlow.getChildren().forEach(child -> {
                child.setVisible(descriptionTextFlow.isVisible());
                child.setManaged(descriptionTextFlow.isManaged());
            });*/
        });

        descriptions = Arrays.copyOf(descriptions, descriptions.length + (descriptions.length > 0 ? 2 : 1));
        for (int i = descriptions.length - 1; i > 1; i--) {
            descriptions[i] = descriptions[i - 2];
        }
        descriptions[0] = tooltip;
        if (descriptions.length > 1) {
            descriptions[1] = "";
        }

        descriptionsProperty.set(Arrays.asList(descriptions));

        buttonTextProperty.set(buttonText);
        buttonTextProperty.addListener((observable, oldValue, newValue) -> {
            button.setText(newValue);
        });

        tooltipProperty.set(new Tooltip(tooltip));
        tooltipProperty.addListener((observable, oldValue, newValue) -> {
            Tooltip.install(container, newValue);
        });

        onLeftClickProperty.set(() -> {
            boolean newVisible = descriptionTextFlow.getChildren().size() != 0 && !descriptionTextFlow.isVisible();

            /*
            descriptionTextFlow.getChildren().forEach(child -> {
                child.setVisible(newVisible);
                child.setManaged(newVisible);
            });*/

            descriptionTextFlow.setVisible(newVisible);
            descriptionTextFlow.setManaged(newVisible);
        });
        onDoubleLeftClickProperty.set(onDoubleLeftClick);
        onRightClickProperty.set(onRightClick);
        clickableProperty.set(clickable);

        initializeContainer();
        initializeTitleLabel();
        initializeButton();
    }

    private static final int SINGLE_CLICK_DELAY = 250;
    private ClickRunner latestClickRunner = null;

    private class ClickRunner implements Runnable {

        private final Runnable onSingleClick;
        private boolean aborted = false;

        public ClickRunner(Runnable onSingleClick) {
            this.onSingleClick = onSingleClick;
        }

        public void abort() {
            this.aborted = true;
        }

        @Override
        public void run() {
            try {
                Thread.sleep(SINGLE_CLICK_DELAY);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
            if (!aborted) {
                // System.out.println("Execute Single Click");
                Platform.runLater(() -> onSingleClick.run());
            }
        }
    }

    private void initializeContainer() {
        Tooltip.install(container, tooltipProperty.get());
        container.setOnMouseClicked(me -> {
            switch (me.getButton()) {
                case PRIMARY:
                    if (me.getClickCount() == 1 && onLeftClickProperty.get() != null) {
                        // System.out.println("Single Click");
                        latestClickRunner = new ClickRunner(onLeftClickProperty.get());
                        CompletableFuture.runAsync(latestClickRunner);
                    }
                    if (me.getClickCount() == 2 && onDoubleLeftClickProperty.get() != null) {
                        // System.out.println("Double Click");
                        if (latestClickRunner != null) {
                            // System.out.println("-> Abort Single Click");
                            latestClickRunner.abort();
                        }
                        onDoubleLeftClickProperty.get().run();
                    }
                    break;
                case SECONDARY:
                    if (onRightClickProperty.get() != null) {
                        onRightClickProperty.get().run();
                    }
                    break;
                default:
                    break;
            }
        });
    }

    private void initializeTitleLabel() {
        titleLabel.setGraphic(iconProperty.get());
        titleLabel.setText(titleProperty.get());
    }

    private void initializeButton() {
        button.setText(buttonTextProperty.get());
        button.setOnAction(ae -> {
            if (onDoubleLeftClickProperty.get() != null) {
                onDoubleLeftClickProperty.get().run();
            }
        });
        button.visibleProperty().bind(clickableProperty);
        button.managedProperty().bind(clickableProperty);
        separator.visibleProperty().bind(clickableProperty);
        separator.managedProperty().bind(clickableProperty);
    }

    private List<Text> createTextFlowElements(List<String> descriptions) {
        return descriptions.stream()
                           .map(description -> new Text(description + (descriptions.indexOf(description) != descriptions.size() - 1 ? "\n" : "")))
                           .collect(Collectors.toList());
    }

    public BooleanProperty clickableProperty() {
        return clickableProperty;
    }

    public boolean getClickable() {
        return clickableProperty.get();
    }

    public void setClickable(boolean clickable) {
        clickableProperty.set(clickable);
    }

    public StringProperty getTitleProperty() {
        return titleProperty;
    }

    public StringProperty getButtonTextProperty() {
        return buttonTextProperty;
    }

    public ObjectProperty<Tooltip> getTooltipProperty() {
        return tooltipProperty;
    }

    public ObjectProperty<Runnable> getOnRightClickProperty() {
        return onRightClickProperty;
    }

    public ObjectProperty<Runnable> getOnDoubleLeftClickProperty() {
        return onDoubleLeftClickProperty;
    }
}
