package at.jku.mevss.trace.gui.component.treetableview;

import at.jku.mevss.trace.gui.component.treetableview.cell.AntCell;
import at.jku.mevss.trace.gui.component.treetableview.cell.apporximatedoublecell.ApproximateDoubleCellContent;
import at.jku.mevss.trace.gui.component.treetableview.cell.memorycell.MemoryCellContent;
import at.jku.mevss.trace.gui.component.treetableview.cell.percentagecell.PercentageCellContent;
import at.jku.mevss.trace.gui.component.treetableview.cell.plotablediffingdata.PlotableDiffingDataCellContent;
import at.jku.mevss.trace.gui.component.treetableview.cell.valuewithreference.ValueWithReferenceCellContent;
import at.jku.mevss.trace.gui.frame.main.tab.heapdiffing.model.PlotableDiffingData;
import at.jku.mevss.trace.gui.model.ApproximateDouble;
import at.jku.mevss.trace.gui.model.Percentage;
import at.jku.mevss.trace.gui.model.ValueWithReference;
import at.jku.mevss.trace.gui.utils.FXMLUtil;
import javafx.collections.ListChangeListener;
import javafx.collections.ObservableList;
import javafx.css.PseudoClass;
import javafx.scene.control.*;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.logging.Logger;

public class AntTreeTableView<T> extends TreeTableView<T> {
    protected static Logger LOGGER = Logger.getLogger(AntTreeTableView.class.getSimpleName());

    /**
     * Used to remember expansions e.g., can be used to re-expand previously expanded rows after re-classification
     */
    private final List<T> expanded = new ArrayList<T>();

    protected static PseudoClass ancestorOfSelection = PseudoClass.getPseudoClass("ancestor-of-selection");

    public AntTreeTableView() {
        FXMLUtil.load(this, AntTreeTableView.class);
        this.getSelectionModel().setSelectionMode(SelectionMode.MULTIPLE);

        // Default cell factories and default sorting
        rootProperty().addListener((observableValue, oldVal, newValue) -> {
            if (newValue != null) {
                setDefaultCellFactories(getColumns());
                setDefaultSorting(getColumns());
            }
        });

        // selected row highlighting
        setRowFactory(ttv -> createTreeTableRow());

        setSortMode(TreeSortMode.ALL_DESCENDANTS);
    }

    /**
     * Creates table rows that also highlight parent rows when selected
     *
     * @return a parent-highlighting table row
     */
    protected TreeTableRow<T> createTreeTableRow() {
        return new TreeTableRow<T>() {
            {
                getSelectionModel().getSelectedItems().addListener((ListChangeListener<TreeItem<T>>) change -> updateStyleClass());
            }

            @Override
            protected void updateItem(T item, boolean empty) {
                super.updateItem(item, empty);
                updateStyleClass();
            }

            private void updateStyleClass() {
                if (getPseudoClassStates().contains(ancestorOfSelection)) {
                    pseudoClassStateChanged(ancestorOfSelection, false);
                }
                for (TreeItem<T> treeItem : getSelectionModel().getSelectedItems()) {
                    if (treeItem != null) {
                        for (TreeItem<T> parent = treeItem.getParent(); parent != null; parent = parent.getParent()) {
                            if (parent == getTreeItem()) {
                                if (!getPseudoClassStates().contains(ancestorOfSelection)) {
                                    pseudoClassStateChanged(ancestorOfSelection, true);
                                }
                                break;
                            }
                        }
                    }
                }
            }
        };
    }

    /**
     * Sets AntTracks's default cell visualizations (e.g., bars for percentages)
     *
     * @param columns the table's columns
     */
    @SuppressWarnings("unchecked")
    protected void setDefaultCellFactories(ObservableList<TreeTableColumn<T, ?>> columns) {
        for (TreeTableColumn<T, ?> column : columns) {
            // Check if cell factory is still default type and has not been changed by dev
            if (column.getCellFactory().getClass().getName().equals("javafx.scene.control.TreeTableColumn$1")) {
                // Exclude grouping columns
                if (column.getCellData(0) != null) {
                    switch (column.getCellData(0).getClass().getSimpleName()) {
                        case "PlotableDiffingData":
                            ((TreeTableColumn<T, PlotableDiffingData>) column).setCellFactory(param -> new AntCell<>(PlotableDiffingDataCellContent.class));
                            break;
                        case "ValueWithReference":
                            ((TreeTableColumn<T, ValueWithReference>) column).setCellFactory(param -> new AntCell<>(ValueWithReferenceCellContent.class));
                            break;
                        case "ApproximateDouble":
                            ((TreeTableColumn<T, ApproximateDouble>) column).setCellFactory(param -> new AntCell<>(ApproximateDoubleCellContent.class));
                            break;
                        case "Percentage":
                            ((TreeTableColumn<T, Percentage>) column).setCellFactory(param -> new AntCell<>(PercentageCellContent.class));
                            break;
                        case "Double":
                        case "Integer":
                        case "Long":
                        case "Number":
                            ((TreeTableColumn<T, Number>) column).setCellFactory(param -> new AntCell<>(MemoryCellContent.class));
                    }
                } else {
                    LOGGER.warning("Could not determine column type for column " + column.getText());
                }
            }

            setDefaultCellFactories(column.getColumns());
        }
    }

    @SuppressWarnings("unchecked")
    protected void setDefaultSorting(ObservableList<TreeTableColumn<T, ?>> columns) {
        for (TreeTableColumn<T, ?> column : columns) {
            // Check if comperator is still default type and has not been changed by dev
            if (column.getComparator().getClass().getName().startsWith("javafx.scene.control.TableColumnBase")) {
                // Exclude grouping columns
                if (column.getCellData(0) != null) {
                    switch (column.getCellData(0).getClass().getSimpleName()) {
                        case "PlotableDiffingData":
                            ((TreeTableColumn<T, PlotableDiffingData>) column).setComparator(Comparator.comparingDouble(value -> value.plotStyle == PlotableDiffingData
                                    .PlotStyle
                                    .PermDiedBorn ?
                                                                                                                                 value.after :
                                                                                                                                 value.temp));
                            break;
                    }
                }
            }

            setDefaultSorting(column.getColumns());
        }
    }

    protected void rememberExpansion() {
        expanded.clear();
        rememberExpansion(getRoot());
    }

    private void rememberExpansion(TreeItem<T> item) {
        if (item != null) {
            if (item.isExpanded()) {
                expanded.add(item.getValue());
            }
            item.getChildren().forEach(this::rememberExpansion);
        }
    }

    protected void restoreExpansion() {
        restoreExpansion(getRoot());
    }

    private void restoreExpansion(TreeItem<T> item) {
        if (item != null) {
            item.setExpanded(expanded.contains(item.getValue()));
            item.getChildren().forEach(this::restoreExpansion);
        }
    }
}
