/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.trace.gui.dialog.openfile;

import at.jku.mevss.trace.gui.model.AppInfo;
import at.jku.mevss.trace.gui.model.ClientInfo;
import at.jku.mevss.trace.gui.utils.Consts;
import at.jku.mevss.trace.gui.utils.FXMLUtil;
import at.jku.mevss.trace.gui.utils.ImageUtil;
import javafx.fxml.FXML;
import javafx.fxml.Initializable;
import javafx.scene.Node;
import javafx.scene.control.Button;
import javafx.scene.control.ButtonType;
import javafx.scene.control.DialogPane;
import javafx.scene.control.TextField;
import javafx.stage.FileChooser;

import java.io.File;
import java.net.URL;
import java.util.ResourceBundle;

final class OpenFileDialogPane extends DialogPane implements Initializable {

    @FXML
    TextField applicationNameTextField;
    @FXML
    TextField traceTextField;
    @FXML
    Button traceChooseButton;
    @FXML
    Button traceClearButton;
    @FXML
    TextField featuresTextField;
    @FXML
    Button featuresChooseButton;
    @FXML
    Button featuresClearButton;

    private final FileChooser traceFileChooser = new FileChooser();
    private final FileChooser featuresChooser = new FileChooser();

    public OpenFileDialogPane() {
        traceFileChooser.setInitialDirectory(new File(ClientInfo.traceDirectory));
        // traceFileChooser.getExtensionFilters().add(new ExtensionFilter("Loadable files only", ".zip", "*.*"));

        featuresChooser.setInitialDirectory(new File(ClientInfo.featureDirectory));
        // featuresChooser.getExtensionFilters().add(new ExtensionFilter("Loadable files only (*.features)", "*.features"));

        FXMLUtil.load(this, OpenFileDialogPane.class);
    }

    public OpenFileDialogPane(AppInfo appInfo) {
        this();
        applicationNameTextField.setText(appInfo.getAppName());
        traceTextField.setText(appInfo.getSelectedTraceFile() == null ? "" : appInfo.getSelectedTraceFile().getAbsolutePath());
        featuresTextField.setText(appInfo.getSelectedFeaturesFile() == null ? "" : appInfo.getSelectedFeaturesFile().getAbsolutePath());

        if (appInfo.getSelectedTraceFile() != null) {
            traceFileChooser.setInitialDirectory(new File(appInfo.getSelectedTraceFile().getParent()));
        }
        if (appInfo.getSelectedFeaturesFile() != null) {
            featuresChooser.setInitialDirectory(new File(appInfo.getSelectedFeaturesFile().getParent()));
        }

        checkAcceptButtonEnabling();
    }

    @Override
    public void initialize(URL location, ResourceBundle resources) {
        checkAcceptButtonEnabling();

        traceChooseButton.setOnAction(ae -> {
            traceFileChooser.setInitialDirectory(searchExistingDirectory(traceFileChooser.getInitialDirectory()));
            File traceFile = traceFileChooser.showOpenDialog(this.getScene().getWindow());
            if (traceFile != null) {
                if (applicationNameTextField.getText() == null || applicationNameTextField.getText().isEmpty()) {
                    applicationNameTextField.setText(traceFile.getName());
                }
                traceTextField.setText(traceFile.getAbsolutePath());
            }
        });

        featuresChooseButton.setOnAction(ae -> {
            featuresChooser.setInitialDirectory(searchExistingDirectory(featuresChooser.getInitialDirectory()));
            File featuresFile = featuresChooser.showOpenDialog(this.getScene().getWindow());
            if (featuresFile != null) {
                featuresTextField.setText(featuresFile.getAbsolutePath());
            }
        });

        traceClearButton.setGraphic(ImageUtil.getIconNode(Consts.DELETE_ICON));
        traceClearButton.setOnAction(ae -> {
            traceTextField.clear();
        });

        featuresClearButton.setGraphic(ImageUtil.getIconNode(Consts.DELETE_ICON));
        featuresClearButton.setOnAction(ae -> {
            featuresTextField.clear();
            ;
        });

        traceTextField.textProperty().addListener((observable, oldValue, newValue) -> {
            checkAcceptButtonEnabling();
        });

        featuresTextField.textProperty().addListener((observable, oldValue, newValue) -> {
            checkAcceptButtonEnabling();
        });
    }

    private File searchExistingDirectory(File directory) {
        while (directory != null && !directory.isDirectory()) {
            directory = directory.getParentFile();
        }
        if (directory != null) {
            return directory;
        }
        directory = new File(System.getProperty("user.home"));
        if (directory.isDirectory()) {
            return directory;
        }
        return null;
    }

    String getApplicationName() {
        return applicationNameTextField.getText();
    }

    File getTraceFile() {
        return (traceTextField.getText() == null || traceTextField.getText().isEmpty()) ? null : new File(traceTextField.getText());
    }

    File getFeaturesFile() {
        return (featuresTextField.getText() == null || featuresTextField.getText().isEmpty()) ? null : new File(featuresTextField.getText());
    }

    private void checkAcceptButtonEnabling() {
        boolean basicAppInfos = getTraceFile() != null && getApplicationName() != null;
        getApplyButton().setDisable(!basicAppInfos);
    }

    Button getApplyButton() {
        Node node = lookupButton(ButtonType.APPLY);
        if (node instanceof Button) {
            return (Button) node;
        } else {
            throw new IllegalStateException("Apply Button not found - DialogPane may not be initialized correctly");
        }
    }

}
