/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.trace.gui.dialog.preference;

import at.jku.mevss.anttracks.heap.io.MetaDataWriterConfig;
import at.jku.mevss.anttracks.heap.labs.Lab;
import at.jku.mevss.trace.gui.model.ClientInfo;
import at.jku.mevss.trace.gui.utils.Consts;

import java.io.*;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.function.Consumer;
import java.util.function.Supplier;
import java.util.logging.Level;
import java.util.logging.Logger;

import static at.jku.mevss.trace.gui.frame.main.GUI.CLIENT_INFO;

public class Settings {

    private static final Logger LOGGER = Logger.getLogger(Settings.class.getSimpleName());

    private static final Settings INSTANCE = new Settings();

    public static Settings getInstance() {
        return INSTANCE;
    }

    private Property<?>[] settings;

    public void setup() {
        settings = availableSettings();

        try (BufferedReader in = new BufferedReader(new InputStreamReader(new FileInputStream(Consts.SETTINGS_FILE)))) {
            for (String line = in.readLine(); line != null; line = in.readLine()) {
                int splitter = line.indexOf('=');
                String key = line.substring(0, splitter);
                String value = line.substring(splitter + 1, line.length());
                trySet(key, value);
            }
        } catch (FileNotFoundException e) {
            // nothing to do
        } catch (IOException e) {
            LOGGER.log(Level.WARNING, "error occured", e);
        }

        Runtime.getRuntime().addShutdownHook(new Thread(this::dump, "Settings Dumper"));
    }

    public static final class Property<T> {
        public final String name;
        public final String description;
        public final boolean configurable;
        private final Supplier<T> getter;
        private final Consumer<T> setter;

        private Property(String name, String description, boolean configurable, Supplier<T> getter, Consumer<T> setter) {
            this.name = name;
            this.description = description;
            this.configurable = configurable;
            this.getter = getter;
            this.setter = setter;
        }

        @SuppressWarnings("unchecked")
        public Class<T> getValueClass() {
            return (Class<T>) get().getClass();
        }

        public T get() {
            return getter.get();
        }

        @SuppressWarnings("unchecked")
        public void set(Object value) throws Exception {
            setter.accept((T) value);
        }
    }

    private Property<?>[] availableSettings() {
        return new Property<?>[]{
                new Property<>("lab_addr_granularity",
                               "Granularity for heap indexing maps",
                               true,
                               () -> Lab.ADDR_GRANULARITY,
                               value -> Lab.ADDR_GRANULARITY = value),
                new Property<>("meta_data_dump_count",
                               "# of heap dumps to be created during parsing for later fast acces",
                               true,
                               () -> MetaDataWriterConfig.DEFAULT_DUMP_COUNT,
                               value -> MetaDataWriterConfig.DEFAULT_DUMP_COUNT = value),
                new Property<>("meta_data_major_gcs_only",
                               "Whether to create dumps only at major GCs",
                               true,
                               () -> MetaDataWriterConfig.DEFAULT_AT_MAJOR_GCS_ONLY,
                               value -> MetaDataWriterConfig.DEFAULT_AT_MAJOR_GCS_ONLY = value),
                new Property<>("traceDirectory",
                               "Last accesses directory",
                               false,
                               () -> {
                                   if (!new File(ClientInfo.traceDirectory).exists()) {
                                       ClientInfo.featureDirectory = System.getProperty("user.home");
                                   }
                                   return ClientInfo.traceDirectory;
                               },
                               value -> ClientInfo.traceDirectory = new File(value).exists() ? value : System.getProperty("user.home")),
                new Property<>("featureDirectory",
                               "Last accesses directory",
                               false,
                               () -> {
                                   if (!new File(ClientInfo.featureDirectory).exists()) {
                                       ClientInfo.featureDirectory = System.getProperty("user.home");
                                   }
                                   return ClientInfo.featureDirectory;
                               },
                               value -> ClientInfo.featureDirectory = new File(value).exists() ? value : System.getProperty("user.home")),
                new Property<>("opening_x_location",
                               "Last X location",
                               false,
                               () -> CLIENT_INFO.getStage().getX(),
                               value -> CLIENT_INFO.getStage().setX(value))
        };
    }

    public int getSize() {
        return settings.length;
    }

    public Property<?>[] getAll() {
        return settings.clone();
    }

    public void set(String key, String value) {
        boolean success = trySet(key, value);
        if (success) {
            dump();
        }
    }

    private synchronized void dump() {
        new File(Consts.SETTINGS_FILE).getParentFile().mkdirs();
        try (BufferedWriter out = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(Consts.SETTINGS_FILE)))) {
            for (Property<?> property : settings) {
                out.write(property.name + "=" + property.get().toString() + "\n");
            }
        } catch (IOException e) {
            LOGGER.log(Level.WARNING, "error occured writing settings file", e);
        }
    }

    private boolean trySet(String key, String value) {
        Property<?> property = get(key);
        if (property != null) {
            try {
                Class<?> clazz = property.getValueClass();
                Constructor<?> ctor = clazz.getConstructor(String.class);
                Object obj = ctor.newInstance(value);
                property.set(obj);
                return true;
            } catch (NoSuchMethodException e) {
                LOGGER.log(Level.WARNING, "no suitable constructor", e);
            } catch (SecurityException e) {
                LOGGER.log(Level.WARNING, "could not access constructor", e);
            } catch (InstantiationException e) {
                LOGGER.log(Level.WARNING, "could not create object", e);
            } catch (IllegalAccessException e) {
                LOGGER.log(Level.WARNING, "internal error occured", e);
            } catch (IllegalArgumentException e) {
                LOGGER.log(Level.WARNING, "internal error occured", e);
            } catch (InvocationTargetException e) {
                LOGGER.log(Level.WARNING, "internal error occured", e);
            } catch (Exception e) {
                LOGGER.log(Level.WARNING, "internal error occured setting property '" + key + "'", e);
            }
        } else {
            LOGGER.log(Level.WARNING, "could not find (and set) property '" + key + "'");
        }
        return false;
    }

    private Property<?> get(String key) {
        for (Property<?> property : settings) {
            if (property.name.equals(key)) {
                return property;
            }
        }
        return null;
    }

}
