/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.trace.gui.frame.main.tab.heapdiffing.component.heapstatisticstreetableview;

import at.jku.mevss.anttracks.diff.DiffGrouping;
import at.jku.mevss.anttracks.util.ImagePack;
import at.jku.mevss.trace.gui.chart.base.ApplicationChartFactory;
import at.jku.mevss.trace.gui.component.treetableview.AntTreeTableView;
import at.jku.mevss.trace.gui.component.treetableview.AutomatedTreeItem;
import at.jku.mevss.trace.gui.component.treetableview.cell.AntCell;
import at.jku.mevss.trace.gui.component.treetableview.cell.keycell.KeyCellContent;
import at.jku.mevss.trace.gui.frame.main.tab.heapdiffing.model.DiffingData;
import at.jku.mevss.trace.gui.frame.main.tab.heapdiffing.model.PlotableDiffingData;
import at.jku.mevss.trace.gui.model.Percentage;
import at.jku.mevss.trace.gui.model.ValueWithReference;
import at.jku.mevss.trace.gui.utils.Consts;
import at.jku.mevss.trace.gui.utils.FXMLUtil;
import javafx.beans.property.ReadOnlyObjectWrapper;
import javafx.beans.property.ReadOnlyStringWrapper;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.scene.Node;
import javafx.scene.control.ContextMenu;
import javafx.scene.control.MenuItem;
import javafx.scene.control.TreeTableColumn;
import javafx.scene.image.ImageView;
import javafx.scene.layout.HBox;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.function.Function;

@SuppressWarnings("unchecked")
public class DiffingTreeTableView extends AntTreeTableView<DiffingData> {
    public enum DisplayMode {
        Plotting,
        Metric
    }

    public static class DiffingPlotInfo {
        public double maxDiedObjects = 0;
        public double maxDiedBytes = 0;
        public double maxAfterObjects = 0;
        public double maxAfterBytes = 0;
        public double maxTempObjects = 0;
        public double maxTempBytes = 0;
    }

    DiffingPlotInfo plotInfo = new DiffingPlotInfo();
    DisplayMode displayMode = DisplayMode.Plotting;

    private static final Function<DiffGrouping, String> titleFunction = c -> c.getKey().toString();
    private final Function<DiffGrouping, DiffingData> dataFunction = c -> {
        DiffingData data = new DiffingData(c.getKey().toString(),
                                           c.getPerm() != null ? c.getPerm().getObjectCount() : 0,
                                           c.getDied() != null ? c.getDied().getObjectCount() : 0,
                                           c.getBorn() != null ? c.getBorn().getObjectCount() : 0,
                                           c.getTemp() != null ? c.getTemp().getObjectCount() : 0,
                                           c.getPerm() != null ? c.getPerm().getByteCount() : 0,
                                           c.getDied() != null ? c.getDied().getByteCount() : 0,
                                           c.getBorn() != null ? c.getBorn().getByteCount() : 0,
                                           c.getTemp() != null ? c.getTemp().getByteCount() : 0,
                                           0);

        plotInfo.maxDiedObjects = Math.max(data.died.objects, plotInfo.maxDiedObjects);
        plotInfo.maxDiedBytes = Math.max(data.died.bytes, plotInfo.maxDiedBytes);
        plotInfo.maxAfterObjects = Math.max(data.after.objects, plotInfo.maxAfterObjects);
        plotInfo.maxAfterBytes = Math.max(data.after.bytes, plotInfo.maxAfterBytes);
        plotInfo.maxTempObjects = Math.max(data.temp.objects, plotInfo.maxTempObjects);
        plotInfo.maxTempBytes = Math.max(data.temp.bytes, plotInfo.maxTempBytes);

        return data;
    };
    private static final Function<DiffGrouping, Collection<? extends DiffGrouping>> childFunction = c -> c.getChildren();
    private static final Function<DiffGrouping, Integer> subTreeLevelFunction = c -> c.getSubTreeLevel();
    // Root has no classifier, every other Grouping has a classifier
    private static final Function<DiffGrouping, Node> iconFunction = c -> {
        List<ImagePack> icons = new ArrayList<>();
        if (c.getClassifier() != null) {
            ImagePack icon = c.getClassifier().getIcon(c.getKey());
            if (icon != null) {
                icons.add(icon);
            }
            if (c.getSubTreeLevel() > 0) {
                DiffGrouping current = c.getParent();

                while (current.getParent() != null) {
                    if (current.getParent().getSubTreeLevel() != current.getSubTreeLevel()) {
                        icon = current.getClassifier().getIcon(current.getKey());
                        if (icon != null) {
                            icons.add(0, icon);
                        }
                    }
                    current = current.getParent();
                }
            }
        }

        HBox hBox = new HBox(2, icons.stream().map(i -> i.getAsNewNode()).toArray(ImageView[]::new));
        return hBox;
    };

    TreeTableColumn<DiffingData, String> nameColumn = new TreeTableColumn<>("Name");

    TreeTableColumn<DiffingData, ValueWithReference> objectsColumn = new TreeTableColumn<>("Objects");
    TreeTableColumn<DiffingData, ValueWithReference> beforeObjectsCountColumn = new TreeTableColumn<>("Before");
    TreeTableColumn<DiffingData, ValueWithReference> permObjectsCountColumn = new TreeTableColumn<>("Perm");
    TreeTableColumn<DiffingData, ValueWithReference> bornObjectsCountColumn = new TreeTableColumn<>("Born");
    TreeTableColumn<DiffingData, ValueWithReference> diedObjectsCountColumn = new TreeTableColumn<>("Died");
    TreeTableColumn<DiffingData, ValueWithReference> tempObjectsCountColumn = new TreeTableColumn<>("Temp");
    TreeTableColumn<DiffingData, ValueWithReference> afterObjectsCountColumn = new TreeTableColumn<>("After");

    TreeTableColumn<DiffingData, ValueWithReference> bytesColumn = new TreeTableColumn<>("Bytes");
    TreeTableColumn<DiffingData, ValueWithReference> beforeBytesCountColumn = new TreeTableColumn<>("Before");
    TreeTableColumn<DiffingData, ValueWithReference> permBytesCountColumn = new TreeTableColumn<>("Perm");
    TreeTableColumn<DiffingData, ValueWithReference> bornBytesCountColumn = new TreeTableColumn<>("Born");
    TreeTableColumn<DiffingData, ValueWithReference> diedBytesCountColumn = new TreeTableColumn<>("Died");
    TreeTableColumn<DiffingData, ValueWithReference> tempBytesCountColumn = new TreeTableColumn<>("Temp");
    TreeTableColumn<DiffingData, ValueWithReference> afterBytesCountColumn = new TreeTableColumn<>("After");

    TreeTableColumn<DiffingData, PlotableDiffingData> diedPermBornObjectsPlottingColumn = new TreeTableColumn<>("Died / Perm / Born");
    TreeTableColumn<DiffingData, PlotableDiffingData> tempObjectsPlottingColumn = new TreeTableColumn<>("Temp");
    TreeTableColumn<DiffingData, PlotableDiffingData> diedPermBornBytesPlottingColumn = new TreeTableColumn<>("Died / Perm / Born");
    TreeTableColumn<DiffingData, PlotableDiffingData> tempBytesPlottingColumn = new TreeTableColumn<>("Temp");

    public DiffingTreeTableView() {
        FXMLUtil.load(this, DiffingTreeTableView.class);
    }

    public void init() {
        this.setPrefHeight(0);  // fixes cutoff in configuration pane - tree table shrinks in favor of classifier selection

        MenuItem entry = new MenuItem("Switch display mode", Consts.EXCHANGE_PACK.getAsNewNode());
        entry.setOnAction(ae -> {
            switch (displayMode) {
                case Plotting:
                    displayMode = DisplayMode.Metric;
                    initColumns();
                    break;
                case Metric:
                    displayMode = DisplayMode.Plotting;
                    initColumns();
                    break;
                default:
                    assert false;
            }
        });
        setContextMenu(new ContextMenu(entry));

        initColumns();
    }

    private void initColumns() {
        switch (displayMode) {
            case Plotting:
                initPlottingView();
                break;
            case Metric:
                initMetricView();
                break;
            default:
                assert false;
        }

        // Ensure that cell factories are correct after column change
        ObservableList<TreeTableColumn<DiffingData, ?>> columns = FXCollections.observableArrayList(nameColumn,
                                                                                                    objectsColumn,
                                                                                                    beforeObjectsCountColumn,
                                                                                                    permObjectsCountColumn,
                                                                                                    bornObjectsCountColumn,
                                                                                                    diedObjectsCountColumn,
                                                                                                    tempObjectsCountColumn,
                                                                                                    afterObjectsCountColumn,
                                                                                                    bytesColumn,
                                                                                                    beforeBytesCountColumn,
                                                                                                    permBytesCountColumn,
                                                                                                    bornBytesCountColumn,
                                                                                                    diedBytesCountColumn,
                                                                                                    tempBytesCountColumn,
                                                                                                    afterBytesCountColumn,
                                                                                                    diedPermBornObjectsPlottingColumn,
                                                                                                    tempObjectsPlottingColumn,
                                                                                                    diedPermBornBytesPlottingColumn,
                                                                                                    tempBytesCountColumn);
        setDefaultCellFactories(columns);
        setDefaultSorting(columns);
    }

    public void initPlottingView() {
        // Clear bindings
        nameColumn.prefWidthProperty().unbind();

        objectsColumn.prefWidthProperty().unbind();
        diedPermBornObjectsPlottingColumn.prefWidthProperty().unbind();
        tempObjectsPlottingColumn.prefWidthProperty().unbind();

        bytesColumn.prefWidthProperty().unbind();
        diedPermBornBytesPlottingColumn.prefWidthProperty().unbind();
        tempBytesPlottingColumn.prefWidthProperty().unbind();

        getColumns().clear();
        // Define columns
        nameColumn.setCellValueFactory(param -> new ReadOnlyStringWrapper(param.getValue().getValue().id));
        nameColumn.setCellFactory(param -> new AntCell<>(KeyCellContent.class));

        diedPermBornObjectsPlottingColumn.setCellValueFactory(param -> new ReadOnlyObjectWrapper<>(new PlotableDiffingData(param.getValue().getValue(),
                                                                                                                           plotInfo,
                                                                                                                           ApplicationChartFactory
                                                                                                                                   .MemoryConsumptionUnit.OBJECTS,
                                                                                                                           PlotableDiffingData.PlotStyle.PermDiedBorn)));
        tempObjectsPlottingColumn.setCellValueFactory(param -> new ReadOnlyObjectWrapper<>(new PlotableDiffingData(param.getValue().getValue(),
                                                                                                                   plotInfo,
                                                                                                                   ApplicationChartFactory.MemoryConsumptionUnit.OBJECTS,
                                                                                                                   PlotableDiffingData.PlotStyle.Temp)));
        diedPermBornBytesPlottingColumn.setCellValueFactory(param -> new ReadOnlyObjectWrapper<>(new PlotableDiffingData(param.getValue().getValue(),
                                                                                                                         plotInfo,
                                                                                                                         ApplicationChartFactory.MemoryConsumptionUnit
                                                                                                                                 .BYTES,
                                                                                                                         PlotableDiffingData.PlotStyle.PermDiedBorn)));
        tempBytesPlottingColumn.setCellValueFactory(param -> new ReadOnlyObjectWrapper<>(new PlotableDiffingData(param.getValue().getValue(),
                                                                                                                 plotInfo,
                                                                                                                 ApplicationChartFactory.MemoryConsumptionUnit.BYTES,
                                                                                                                 PlotableDiffingData.PlotStyle.Temp)));

        objectsColumn.getColumns().clear();
        objectsColumn.getColumns().add(diedPermBornObjectsPlottingColumn);
        objectsColumn.getColumns().add(tempObjectsPlottingColumn);

        bytesColumn.getColumns().clear();
        bytesColumn.getColumns().add(diedPermBornBytesPlottingColumn);
        bytesColumn.getColumns().add(tempBytesPlottingColumn);

        getColumns().add(nameColumn);
        getColumns().add(objectsColumn);
        getColumns().add(bytesColumn);

        nameColumn.prefWidthProperty().bind(widthProperty().multiply(4).divide(12) /* aaaand subtract scroll bar width, otherwise binding properties to width does not
                                       work at all */.subtract(20)); // = 11/15

        objectsColumn.prefWidthProperty().bind(widthProperty().multiply(4).divide(12));
        diedPermBornObjectsPlottingColumn.prefWidthProperty().bind(objectsColumn.widthProperty().multiply(3).divide(4));
        tempObjectsPlottingColumn.prefWidthProperty().bind(objectsColumn.widthProperty().multiply(1).divide(4));

        bytesColumn.prefWidthProperty().bind(widthProperty().multiply(4).divide(12));
        diedPermBornBytesPlottingColumn.prefWidthProperty().bind(objectsColumn.widthProperty().multiply(3).divide(4));
        tempBytesPlottingColumn.prefWidthProperty().bind(objectsColumn.widthProperty().multiply(1).divide(4));
    }

    public void initMetricView() {
        // Clear bindings
        nameColumn.prefWidthProperty().unbind();

        objectsColumn.prefWidthProperty().unbind();
        beforeObjectsCountColumn.prefWidthProperty().unbind();
        permObjectsCountColumn.prefWidthProperty().unbind();
        bornObjectsCountColumn.prefWidthProperty().unbind();
        diedObjectsCountColumn.prefWidthProperty().unbind();
        tempObjectsCountColumn.prefWidthProperty().unbind();
        afterObjectsCountColumn.prefWidthProperty().unbind();

        bytesColumn.prefWidthProperty().unbind();
        beforeBytesCountColumn.prefWidthProperty().unbind();
        permBytesCountColumn.prefWidthProperty().unbind();
        bornBytesCountColumn.prefWidthProperty().unbind();
        diedBytesCountColumn.prefWidthProperty().unbind();
        tempBytesCountColumn.prefWidthProperty().unbind();
        afterBytesCountColumn.prefWidthProperty().unbind();

        getColumns().clear();
        // Define columns
        nameColumn.setCellValueFactory(param -> new ReadOnlyStringWrapper(param.getValue().getValue().id));
        nameColumn.setCellFactory(param -> new AntCell<>(KeyCellContent.class));

        beforeObjectsCountColumn.setCellValueFactory(param -> new ReadOnlyObjectWrapper<>(new ValueWithReference(param.getValue().getValue().before.objects,
                                                                                                                 getParentDiffGrouping((AutomatedTreeItem<DiffGrouping,
                                                                                                                         DiffingData>) param
                                                                                                                         .getValue()).before.objects)));
        permObjectsCountColumn.setCellValueFactory(param -> new ReadOnlyObjectWrapper<>(new ValueWithReference(param.getValue().getValue().perm.objects,
                                                                                                               getParentDiffGrouping((AutomatedTreeItem<DiffGrouping,
                                                                                                                       DiffingData>) param
                                                                                                                       .getValue()).perm.objects)));
        bornObjectsCountColumn.setCellValueFactory(param -> new ReadOnlyObjectWrapper<>(new ValueWithReference(param.getValue().getValue().born.objects,
                                                                                                               getParentDiffGrouping((AutomatedTreeItem<DiffGrouping,
                                                                                                                       DiffingData>) param
                                                                                                                       .getValue()).born.objects)));
        diedObjectsCountColumn.setCellValueFactory(param -> new ReadOnlyObjectWrapper<>(new ValueWithReference(param.getValue().getValue().died.objects,
                                                                                                               getParentDiffGrouping((AutomatedTreeItem<DiffGrouping,
                                                                                                                       DiffingData>) param
                                                                                                                       .getValue()).died.objects)));
        tempObjectsCountColumn.setCellValueFactory(param -> new ReadOnlyObjectWrapper<>(new ValueWithReference(param.getValue().getValue().temp.objects,
                                                                                                               getParentDiffGrouping((AutomatedTreeItem<DiffGrouping,
                                                                                                                       DiffingData>) param
                                                                                                                       .getValue()).temp.objects)));
        afterObjectsCountColumn.setCellValueFactory(param -> new ReadOnlyObjectWrapper<>(new ValueWithReference(param.getValue().getValue().after.objects,
                                                                                                                getParentDiffGrouping((AutomatedTreeItem<DiffGrouping,
                                                                                                                        DiffingData>) param
                                                                                                                        .getValue()).after.objects)));

        beforeBytesCountColumn.setCellValueFactory(param -> new ReadOnlyObjectWrapper<>(new ValueWithReference(param.getValue().getValue().before.bytes,
                                                                                                               getParentDiffGrouping((AutomatedTreeItem<DiffGrouping,
                                                                                                                       DiffingData>) param
                                                                                                                       .getValue()).before.bytes)));
        permBytesCountColumn.setCellValueFactory(param -> new ReadOnlyObjectWrapper<>(new ValueWithReference(param.getValue().getValue().perm.bytes,
                                                                                                             getParentDiffGrouping((AutomatedTreeItem<DiffGrouping,
                                                                                                                     DiffingData>)
                                                                                                                                           param
                                                                                                                                                   .getValue()).perm
                                                                                                                     .bytes)));
        bornBytesCountColumn.setCellValueFactory(param -> new ReadOnlyObjectWrapper<>(new ValueWithReference(param.getValue().getValue().born.bytes,
                                                                                                             getParentDiffGrouping((AutomatedTreeItem<DiffGrouping,
                                                                                                                     DiffingData>)
                                                                                                                                           param
                                                                                                                                                   .getValue()).born
                                                                                                                     .bytes)));
        diedBytesCountColumn.setCellValueFactory(param -> new ReadOnlyObjectWrapper<>(new ValueWithReference(param.getValue().getValue().died.bytes,
                                                                                                             getParentDiffGrouping((AutomatedTreeItem<DiffGrouping,
                                                                                                                     DiffingData>)
                                                                                                                                           param
                                                                                                                                                   .getValue()).died
                                                                                                                     .bytes)));
        tempBytesCountColumn.setCellValueFactory(param -> new ReadOnlyObjectWrapper<>(new ValueWithReference(param.getValue().getValue().temp.bytes,
                                                                                                             getParentDiffGrouping((AutomatedTreeItem<DiffGrouping,
                                                                                                                     DiffingData>)
                                                                                                                                           param
                                                                                                                                                   .getValue()).temp
                                                                                                                     .bytes)));
        afterBytesCountColumn.setCellValueFactory(param -> new ReadOnlyObjectWrapper<>(new ValueWithReference(param.getValue().getValue().after.bytes,
                                                                                                              getParentDiffGrouping((AutomatedTreeItem<DiffGrouping,
                                                                                                                      DiffingData>) param
                                                                                                                      .getValue()).after.bytes)));

        objectsColumn.getColumns().clear();
        objectsColumn.getColumns().add(beforeObjectsCountColumn);
        objectsColumn.getColumns().add(permObjectsCountColumn);
        objectsColumn.getColumns().add(bornObjectsCountColumn);
        objectsColumn.getColumns().add(diedObjectsCountColumn);
        objectsColumn.getColumns().add(tempObjectsCountColumn);
        objectsColumn.getColumns().add(afterObjectsCountColumn);

        bytesColumn.getColumns().clear();
        bytesColumn.getColumns().add(beforeBytesCountColumn);
        bytesColumn.getColumns().add(permBytesCountColumn);
        bytesColumn.getColumns().add(bornBytesCountColumn);
        bytesColumn.getColumns().add(diedBytesCountColumn);
        bytesColumn.getColumns().add(tempBytesCountColumn);
        bytesColumn.getColumns().add(afterBytesCountColumn);

        getColumns().add(nameColumn);
        getColumns().add(objectsColumn);
        getColumns().add(bytesColumn);

        // Define column widths
        nameColumn.prefWidthProperty().bind(widthProperty().multiply(4).divide(12) /* aaaand subtract scroll bar width, otherwise binding properties to width does not
                                       work at all */.subtract(20)); // = 11/15

        objectsColumn.prefWidthProperty().bind(widthProperty().multiply(4).divide(12));
        beforeObjectsCountColumn.prefWidthProperty().bind(objectsColumn.widthProperty().divide(6));
        permObjectsCountColumn.prefWidthProperty().bind(objectsColumn.widthProperty().divide(6));
        bornObjectsCountColumn.prefWidthProperty().bind(objectsColumn.widthProperty().divide(6));
        diedObjectsCountColumn.prefWidthProperty().bind(objectsColumn.widthProperty().divide(6));
        tempObjectsCountColumn.prefWidthProperty().bind(objectsColumn.widthProperty().divide(6));
        afterObjectsCountColumn.prefWidthProperty().bind(objectsColumn.widthProperty().divide(6));

        bytesColumn.prefWidthProperty().bind(widthProperty().multiply(4).divide(12));
        beforeBytesCountColumn.prefWidthProperty().bind(bytesColumn.widthProperty().divide(6));
        permBytesCountColumn.prefWidthProperty().bind(bytesColumn.widthProperty().divide(6));
        bornBytesCountColumn.prefWidthProperty().bind(bytesColumn.widthProperty().divide(6));
        diedBytesCountColumn.prefWidthProperty().bind(bytesColumn.widthProperty().divide(6));
        tempBytesCountColumn.prefWidthProperty().bind(bytesColumn.widthProperty().divide(6));
        afterBytesCountColumn.prefWidthProperty().bind(bytesColumn.widthProperty().divide(6));
    }

    public Percentage getPercentage(double total, double x) {
        if (total == 0 || x == 0) {
            return new Percentage(0.0);
        }
        return new Percentage(new BigDecimal(x / total * 100).setScale(1, BigDecimal.ROUND_HALF_UP).doubleValue());
    }

    public void setLoadingMode(boolean isLoading) {
        setDisabled(isLoading);
    }

    private DiffingData getParentDiffGrouping(AutomatedTreeItem<DiffGrouping, DiffingData> item) {
        int subTreeLevel = item.getSubTreeLevel();
        AutomatedTreeItem<DiffGrouping, DiffingData> lastMatching = null;
        while (item != null) {
            if (item.getSubTreeLevel() == subTreeLevel) {
                lastMatching = item;
            }
            item = (AutomatedTreeItem<DiffGrouping, DiffingData>) item.getParent();
        }
        return lastMatching.getValue();
    }

    public void setRoot(DiffGrouping grouping) {
        // Reset plotting info
        plotInfo = new DiffingPlotInfo();
        // Remember expansion
        rememberExpansion();
        setRoot(new AutomatedTreeItem<>(grouping, titleFunction, dataFunction, childFunction, subTreeLevelFunction, iconFunction));
        restoreExpansion();
        setShowRoot(true);
        getRoot().setExpanded(true);

        // Sorting
        getSortOrder().clear();
        switch (displayMode) {
            case Plotting:
                diedPermBornObjectsPlottingColumn.setSortType(TreeTableColumn.SortType.DESCENDING);
                getSortOrder().add(diedPermBornObjectsPlottingColumn);
                break;
            case Metric:
                afterObjectsCountColumn.setSortType(TreeTableColumn.SortType.DESCENDING);
                getSortOrder().add(afterObjectsCountColumn);
                break;
            default:
                assert false;
        }
    }
}
