/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.trace.gui.frame.main.tab.heapstate;

import at.jku.mevss.anttracks.classification.ClassifierChain;
import at.jku.mevss.anttracks.classification.ConditionalClassifierList;
import at.jku.mevss.anttracks.classification.Filter;
import at.jku.mevss.anttracks.util.ParallelizationUtil;
import at.jku.mevss.trace.gui.frame.main.component.applicationbase.ApplicationBaseTab;
import at.jku.mevss.trace.gui.frame.main.tab.heapstate.tab.statistics.HeapStatisticsTab;
import at.jku.mevss.trace.gui.model.AppInfo;
import at.jku.mevss.trace.gui.model.AppInfo.ChangeType;
import at.jku.mevss.trace.gui.model.FastHeapInfo;
import at.jku.mevss.trace.gui.model.HeapStateStatisticsInfo;
import at.jku.mevss.trace.gui.utils.AntTask;
import at.jku.mevss.trace.gui.utils.Consts;
import at.jku.mevss.trace.gui.utils.FXMLUtil;
import at.jku.mevss.trace.gui.utils.ImageUtil;
import javafx.concurrent.Task;
import javafx.fxml.FXML;
import javafx.scene.control.Button;
import javafx.scene.control.Label;
import javafx.scene.control.TabPane;
import javafx.scene.layout.HBox;

import java.util.ArrayList;
import java.util.List;

public class HeapStateTab extends ApplicationBaseTab {

    public interface HeapStateSubTab {
        void heapStateChanged();
    }

    @FXML
    private HBox outdatedHeapStateHBox;
    @FXML
    private Label outdatedHeapStateLabel;
    @FXML
    private Button outdatedHeapStateButton;

    @FXML
    private TabPane tabPane;

    @FXML
    private HeapStatisticsTab statisticsTab;

    private FastHeapInfo detailsInfo;

    public HeapStateTab() {
        FXMLUtil.load(this, HeapStateTab.class);
    }

    public void init(FastHeapInfo detailsInfo, HeapStateTask task) {
        init(detailsInfo, task, null,null,null);
    }

    public void init(FastHeapInfo detailsInfo, HeapStateTask task,
                     ClassifierChain selectedClassifier,
                     ConditionalClassifierList conditionalClassifier,
                     List<Filter> selectedFilters) {
        super.init(detailsInfo.getAppInfo());
        this.detailsInfo = detailsInfo;

        initializeTab();
        initializeStatisticsTab(selectedClassifier, conditionalClassifier, selectedFilters);

        tasks.add(task);
    }

    private void initializeTab() {
        setGraphic(ImageUtil.getIconNode(Consts.DETAILS_IMAGE));
        updateTabTitle();
        setHeaderText("Detailed information for " + getAppInfo().getAppName() + " (" + getAppInfo().getSelectedTraceFile() + ")");
    }

    private void updateTabTitle() {
        setText("Heap state - " + getAppInfo().getAppName() + " @ " + detailsInfo.getTime() + "ms");
    }

    private void initializeStatisticsTab(ClassifierChain selectedClassifier,
                                         ConditionalClassifierList conditionalClassifier,
                                         List<Filter> selectedFilters) {
        statisticsTab.init(new HeapStateStatisticsInfo(detailsInfo, selectedClassifier, conditionalClassifier, selectedFilters));
    }

    public FastHeapInfo getDetailsInfo() {
        return detailsInfo;
    }

    public void setDefaultClassification() {
        statisticsTab.setDefaultClassification();
    }

    public HeapStatisticsTab getStatistics() {
        return statisticsTab;
    }

    @Override
    protected void appInfoChangeAction(ChangeType type) {
        if (type == ChangeType.NAME) {
            updateTabTitle();
        }

        if (type == AppInfo.ChangeType.DATA_STRUCTURE) {
            outdatedHeapStateLabel.setText("Data structure definitions changed! This heap state might be outdated.");
            outdatedHeapStateHBox.setVisible(true);
            outdatedHeapStateHBox.setManaged(true);

            outdatedHeapStateButton.setOnAction(evt -> {
                Task<Void> updateDataStructures = new AntTask<Void>() {
                    @Override
                    protected Void backgroundWork() throws Exception {
                        // update data structure information in the heap
                        detailsInfo.getFastHeapSupplier().get().initDataStructures();
                        return null;
                    }

                    @Override
                    protected void suceeded() {
                        // refresh the classifications/metrics shown in the sub tabs of this HeapStateTab
                        // copy tab list in case new tabs are opened by the handlers
                        new ArrayList<>(tabPane.getTabs()).forEach(tab -> {
                            if (tab instanceof HeapStateSubTab) {
                                ((HeapStateSubTab) tab).heapStateChanged();
                            }
                        });
                        outdatedHeapStateHBox.setVisible(false);
                        outdatedHeapStateHBox.setManaged(false);
                    }
                };

                tasks.add(updateDataStructures);
                ParallelizationUtil.submitTask(updateDataStructures);
            });
        }
    }

    @Override
    protected void cleanupOnClose() {
        // Remove all child tabs
        if (statisticsTab.getOnClosed() != null) {
            statisticsTab.getOnClosed().handle(null);
        }
        statisticsTab.getTabPane().getTabs().remove(statisticsTab);

        detailsInfo.getFastHeapSupplier().get().clear();
    }
}
