/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.trace.gui.frame.main.tab.heapstate.tab.statistics;

import at.jku.mevss.anttracks.classification.Filter;
import at.jku.mevss.anttracks.classification.trees.ClassificationTree;
import at.jku.mevss.anttracks.heap.ObjectStream;
import at.jku.mevss.anttracks.heap.statistics.Statistics;
import at.jku.mevss.anttracks.util.Counter;
import at.jku.mevss.anttracks.util.ThreadUtil;
import at.jku.mevss.trace.gui.model.HeapStateStatisticsInfo;
import at.jku.mevss.trace.gui.utils.AntTask;
import at.jku.mevss.trace.gui.utils.WindowUtil;
import javafx.application.Platform;
import javafx.concurrent.Task;
import javafx.scene.control.Alert;
import javafx.scene.control.Alert.AlertType;
import javafx.scene.control.ButtonType;
import javafx.scene.layout.Region;

import java.io.IOException;
import java.util.Optional;
import java.util.concurrent.FutureTask;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;

import static at.jku.mevss.trace.gui.frame.main.GUI.CLIENT_INFO;

public class GroupingTask extends AntTask<ClassificationTree> {

    private String taskName;

    private final HeapStatisticsTab heapStatisticsTab;
    private final HeapStateStatisticsInfo statisticsInfo;
    private Task<Void> closureInitTask;

    @SuppressWarnings("unused")
    private static boolean VERBOSE = false;

    public GroupingTask(HeapStateStatisticsInfo statisticsInfo, HeapStatisticsTab heapStatisticsTab) {
        this.statisticsInfo = statisticsInfo;
        this.heapStatisticsTab = heapStatisticsTab;
        this.heapStatisticsTab.setLoadingMode(true);

        // new grouping task -> cancel all currently running tasks in this tab because they concern current grouping
        heapStatisticsTab.tasks.forEach(Task::cancel);
        heapStatisticsTab.tasks.clear();
        heapStatisticsTab.tasks.add(this);

        try {
            taskName = GroupingTask.class.getSimpleName() + " - " + statisticsInfo.getDetailsInfo()
                                                                                  .getAppInfo()
                                                                                  .getSymbolsFile()
                                                                                  .getCanonicalPath() + " - " + statisticsInfo.getDetailsInfo()
                                                                                                                              .getTime() + " - " + statisticsInfo
                    .getSelectedClassifiers()
                    .toString();
        } catch (IOException e) {
            LOGGER.warning("Error on retrieving parser worker name:\n" + e);
            taskName = this.getClass().getSimpleName();
        }
    }

    @Override
    protected void cancelled() {
        super.cancelled();

        if (closureInitTask != null) {
            closureInitTask.cancel();
        }
        heapStatisticsTab.setLoadingMode(false);
        // cancel() suppresses succeeded() i.e. the treetable will not be updated
        // thus the last working classifier/filter configuration must be restored
        heapStatisticsTab.rollbackConfiguration();
    }

    private ClassificationTree groupHeapObjects() {
        updateTitle("Heap State: ");
        updateMessage(String.format("Classify heap objects for time %,.3fs using classifiers %s",
                                    statisticsInfo.getDetailsInfo().getTime() / 1000.0f,
                                    statisticsInfo.getSelectedClassifiers()));
        long objectCount = statisticsInfo.getDetailsInfo().getFastHeapSupplier().get().getObjectCount();

        ClassificationTree grouping = null;
        Counter objectsProcessed = new Counter();
        final long t = System.nanoTime();
        ObjectStream.IterationListener iterationListener = (oc) -> {
            objectsProcessed.add(oc);
            updateProgress(objectsProcessed.get(), objectCount);
            updateClassificationMetrics(objectsProcessed.get(), t);
        };

        //switch (CLASSIFICATION_MODE) {
        //    case LIST:
        //        grouping = ListClassificationTree.readTree(statisticsInfo.getDetailsInfo().getAppInfo().getMetaDataPath(),
        //                                                   statisticsInfo.getSelectedFilters().toArray(new Filter[0]),
        //                                                   statisticsInfo.getSelectedClassifiers(),
        //                                                   statisticsInfo.getSelectedConditionalClassifier(),
        //                                                   String.valueOf(statisticsInfo.getDetailsInfo().getTime()),
        //                                                   statisticsInfo.getDetailsInfo());
        if (grouping == null) {
            grouping = statisticsInfo.getDetailsInfo()
                                     .getFastHeapSupplier()
                                     .get()
                                     .groupListParallel(statisticsInfo.getSelectedFilters().toArray(new Filter[0]),
                                                        statisticsInfo.getSelectedClassifiers(),
                                                        statisticsInfo.getSelectedConditionalClassifier(),
                                                        iterationListener,
                                                        cancelProperty);

            CLIENT_INFO.meterRegistry.timer("classification." + statisticsInfo.getSelectedClassifiers().toString() + ".per_object")
                                     .record((long) ((1.0 * System.nanoTime() - t) / grouping.getRoot().getObjectCount() * 1_000_000), TimeUnit.NANOSECONDS);

            //        break;
            //}

            if (!isCancelled()) {
                updateClassificationMetrics(objectsProcessed.get(), t);

                updateMessage(String.format("Calculate object & byte counts and overall closure for time %,.3fs (This may take some seconds, please wait)",
                                            statisticsInfo.getDetailsInfo().getTime() / 1000.0f));
                long t2 = System.nanoTime();
                grouping.init();
                closureInitTask = grouping.initClosureTask();
                heapStatisticsTab.getHeapMetricsTable().groupingInitTime.valueProperty.setValue((System.nanoTime() - t2) / 1_000_000_000.0);

                long t3 = System.nanoTime();
                grouping.sampleTopDown();
                heapStatisticsTab.getHeapMetricsTable().samplingTime.valueProperty.setValue((System.nanoTime() - t3) / 1_000_000_000.0);
                LOGGER.log(Level.INFO, "Finished classification");
            }
        }

        if (grouping.getRoot().getObjectCount() != statisticsInfo.getDetailsInfo().getFastHeapSupplier().get().getObjectCount()) {
            LOGGER.info(String.format("Classification Tree Overall Object Count (%,d) does not match Fast Heap Overall Object Count (%,d)",
                                      grouping.getRoot().getObjectCount(),
                                      statisticsInfo.getDetailsInfo().getFastHeapSupplier().get().getObjectCount()));
        }

        if (grouping.getRoot().getByteCount() != statisticsInfo.getDetailsInfo().getFastHeapSupplier().get().getByteCount()) {
            LOGGER.info(String.format("Classification Tree Overall Byte Count (%,d) does not match Fast Heap Overall Byte Count (%,d)",
                                      grouping.getRoot().getByteCount(),
                                      statisticsInfo.getDetailsInfo().getFastHeapSupplier().get().getByteCount()));
        }

        Optional<Statistics> stat = statisticsInfo.getDetailsInfo()
                                                  .getAppInfo()
                                                  .getStatistics()
                                                  .stream()
                                                  .filter(statistics -> statistics.info.time == statisticsInfo.getDetailsInfo().getTime())
                                                  .findFirst();
        assert stat.isPresent() : "Statistics must be present for heap state time";
        long statObjectCount = stat.get().eden.memoryConsumption.getObjects() + stat.get().survivor.memoryConsumption.getObjects() + stat.get().old.memoryConsumption
                .getObjects();
        long statMemoryConsumption = stat.get().eden.memoryConsumption.getBytes() + stat.get().survivor.memoryConsumption.getBytes() + stat.get().old.memoryConsumption
                .getBytes();

        if (grouping.getRoot().getObjectCount() != statObjectCount) {
            LOGGER.info(String.format("Classification Tree Overall Object Count (%,d) does not match Statistics Object Count (%,d)",
                                      grouping.getRoot().getObjectCount(),
                                      statObjectCount));
        }

        if (grouping.getRoot().getByteCount() != statMemoryConsumption) {
            LOGGER.info(String.format("Classification Tree Overall Byte Count (%,d) does not match Statistics Byte Count (%,d)",
                                      grouping.getRoot().getByteCount(),
                                      statMemoryConsumption));
        }

        return grouping;
    }

    private void updateClassificationMetrics(long objectsProcessed, long startTime) {
        double seconds = (System.nanoTime() - startTime) / 1_000_000_000.0;
        heapStatisticsTab.getHeapMetricsTable().classificationTime.valueProperty.setValue(seconds);
        long objPerSec = (int) (objectsProcessed / seconds);
        heapStatisticsTab.getHeapMetricsTable().classificationThroughput.valueProperty.setValue(objPerSec);
        heapStatisticsTab.getHeapMetricsTable().refresh();
    }

    private void checkGrouping(ClassificationTree grouping) {
        /*
         * if (VERBOSE) { long nNodes = 1; long nNodesWithAtLeastOneLeaf = Math.min(1, grouping.getLeafs().size()); long nHashMapEntries =
         * grouping.getLeafs().size(); List<? extends HeapObjectGrouping> stillToCheck = grouping.getChildren(); while
         * (!stillToCheck.isEmpty()) { Grouping g = stillToCheck.remove(0); nNodes += 1 + g.getChildren().stream().filter(child ->
         * child.getChildren().size() == 0).count(); nHashMapEntries += g.getLeafs().size() + g.getChildren().stream().filter(child ->
         * child.getChildren().size() == 0) .mapToLong(child -> child.getLeafs().size()).sum(); nNodesWithAtLeastOneLeaf += Math.min(1,
         * g.getLeafs().size()) + g.getChildren().stream().filter(child -> child.getChildren().size() == 0) .mapToLong(child -> Math.min(1,
         * child.getLeafs().size())).sum(); if (g.getChildren().size() == 0 && g.getLeafs().size() == 0) { logger.log(Level.WARNING,
         * "Somethings wrong with the grouping tree. Leaf node does not have data stored." ); }
         * stillToCheck.addAll(g.getChildren().stream().filter(child -> child.getChildren().size() > 0) .collect(Collectors.toList())); }
         *
         * logger.log(Level.FINE, grouping.toString()); logger.log(Level.FINE, "Number of nodes: " + nNodes); logger.log(Level.FINE,
         * "Number of nodes with at least one leaf: " + nNodesWithAtLeastOneLeaf); logger.log(Level.FINE, "Number of leaf HashMap
         * entries: "
         * + nHashMapEntries); logger.log(Level.FINE, "Number of objects: " + grouping.getSummedObjects());
         *
         * float avgHashMapEntries = 1.0f * nHashMapEntries / nNodes; double avgObjects = 1.0f * grouping.getSummedObjects() / nNodes;
         *
         * logger.log(Level.FINE, "Avg. number of leaf entries per node: " + avgHashMapEntries); logger.log(Level.FINE,
         * "Avg. number of objects per node: " + avgObjects); logger.log(Level.FINE, "avg(n_E) * x = avg(n_O) ... x = " + (avgObjects /
         * avgHashMapEntries)); }
         */

        if (grouping.getExceptions().size() > 0) {
            LOGGER.log(Level.INFO, String.format("There were %d exceptions during grouping.", grouping.getExceptions().size()));
            for (int i = 0; i < grouping.getExceptions().size(); i++) {
                LOGGER.log(Level.INFO, grouping.getExceptions().get(i).getMessage());
                grouping.getExceptions().get(i).getCause().printStackTrace();
            }
        }
    }

    @Override
    protected ClassificationTree backgroundWork() throws Exception {
        try {
            ClassificationTree grouping = groupHeapObjects();
            checkGrouping(grouping);
            return grouping;
        } catch (Exception e) {
            e.printStackTrace(System.err);
            final FutureTask<Boolean> alertTask = new FutureTask<>(() -> {
                Alert alert = new Alert(AlertType.ERROR, "An internal error occured while classifying, do you want to retry?", ButtonType.YES, ButtonType.NO);
                alert.setTitle("Error");
                alert.getDialogPane().setMinHeight(Region.USE_PREF_SIZE);
                WindowUtil.centerInMainFrame(alert);
                Optional<ButtonType> retryChoice = alert.showAndWait();
                return retryChoice.isPresent() && retryChoice.get() == ButtonType.YES;
            });
            Platform.runLater(alertTask);
            if (alertTask.get()) {
                return call();
            } else {
                throw new Exception(e);
            }
        }
    }

    @Override
    protected void suceeded() {
        ClassificationTree grouping = getValue();
        heapStatisticsTab.updateGrouping(grouping);
        heapStatisticsTab.setLoadingMode(false);
        heapStatisticsTab.setConfigurationSavepoint();

        // start closure calculation after grouping
        if (closureInitTask != null) {
            ThreadUtil.runDeferred(() -> {
                Platform.runLater(() -> {
                    CLIENT_INFO.operationManager.addNewOperation(closureInitTask);
                    //                closureInitTask.setOnSucceeded(workerStateEvent -> {
                    //                    try {
                    //                        Task<Void> writeTask = grouping.writeTreeTask(statisticsInfo.getDetailsInfo().getAppInfo().getMetaDataPath(),
                    //                                                                      String.valueOf(statisticsInfo.getDetailsInfo().getTime()));
                    //
                    //                        Platform.runLater(() -> {
                    //                            getClientInfo().operationManager.addNewOperation(writeTask);
                    //                            ThreadUtil.startTask(writeTask);
                    //                            heapStatisticsTab.tasks.add(writeTask);
                    //                        });
                    //                    } catch (OperationNotSupportedException e) {
                    //                        e.printStackTrace();
                    //                    }
                    //                });
                    ThreadUtil.startTask(closureInitTask);
                });
                heapStatisticsTab.tasks.add(closureInitTask);
            }, ThreadUtil.DeferredPeriod.LONG);

        }
    }
}
