/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.trace.gui.frame.main.tab.lifetimevisualization.visualization;

import at.jku.mevss.anttracks.classification.Classifier;
import at.jku.mevss.anttracks.classification.ClassifierChain;
import at.jku.mevss.anttracks.classification.ClassifierFactory;
import at.jku.mevss.anttracks.classification.Filter;
import at.jku.mevss.anttracks.classification.nodes.DataGroupingNode;
import at.jku.mevss.anttracks.classification.nodes.GroupingNode;
import at.jku.mevss.anttracks.classification.nodes.ListGroupingNode;
import at.jku.mevss.anttracks.classification.nodes.MapGroupingNode;
import at.jku.mevss.anttracks.classification.trees.ListClassificationTree;
import at.jku.mevss.anttracks.heap.*;
import at.jku.mevss.anttracks.heap.io.MetaDataReaderConfig;
import at.jku.mevss.anttracks.parser.EventType;
import at.jku.mevss.anttracks.parser.heap.HeapTraceParser;
import at.jku.mevss.anttracks.parser.io.BaseFile;
import at.jku.mevss.anttracks.parser.symbols.SymbolsFile;
import at.jku.mevss.anttracks.util.ThreadUtil;
import at.jku.mevss.trace.gui.frame.main.tab.lifetimevisualization.LifetimeVisualizationInfo;
import at.jku.mevss.trace.gui.frame.main.tab.lifetimevisualization.LifetimeVisualizationTab;
import at.jku.mevss.trace.gui.model.AppInfo;
import at.jku.mevss.trace.gui.utils.AntTask;
import at.jku.mevss.trace.gui.utils.Consts;
import at.jku.mevss.trace.gui.utils.PlatformUtil;
import javafx.concurrent.Task;

import java.io.*;
import java.nio.file.Files;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ExecutionException;
import java.util.logging.Level;
import java.util.stream.Collectors;

import static at.jku.mevss.trace.gui.frame.main.GUI.CLIENT_INFO;

public class LifetimeVisualizationDiagramTask extends AntTask<Map<Long, GroupingNode>> {

    public static final boolean USE_INDEX_BASED_HEAP = true;
    public static final boolean PARSE_AT_GC_START = false;
    public static final boolean PARSE_AT_GC_END = true;

    private final AppInfo appInfo;
    private ClassifierChain chain;
    private Filter[] filters;
    private LifetimeVisualizationInfo lifetimeVisualizationInfo;
    private LifetimeVisualizationDiagramInfo diagramInfo;
    private LifetimeVisualizationTab lifetimeVisualizationTab;
    private LifetimeVisualizationPlottingTask currentChartUpdatingTask = null;
    private long lastPlotTime = 0;
    private static final int PLOT_THRESHOLD = 1_000;
    private String fullMetaDataPath;
    private String mapTreeFilePathWithoutTime;
    private long startMissingTime = -1;
    private long endMissingTime = -1;
    private Map<Long, GroupingNode> groupings = new HashMap<>();
    private HashMap<String, Classifier<?>> availableClassifier;
    private boolean parseXthHeap;
    private boolean parseHeapAllXSeconds;
    private int parseVal;
    private long lastParsedTime = -10_000; //such that the first tree always gets selected
    private int countSinceLastParse = -1;

    public LifetimeVisualizationDiagramTask(AppInfo appInfo,
                                            ClassifierChain chain,
                                            Filter[] filters,
                                            LifetimeVisualizationInfo lifetimeVisualizationInfo,
                                            LifetimeVisualizationDiagramInfo diagramInfo,
                                            LifetimeVisualizationTab lifetimeVisualizationTab,
                                            boolean parseXthHeap,
                                            boolean parseHeapAllXSeconds,
                                            int val) {
        this.parseXthHeap = parseXthHeap;
        this.parseHeapAllXSeconds = parseHeapAllXSeconds;
        parseVal = val;
        this.chain = chain;
        this.filters = filters;
        this.appInfo = appInfo;
        this.lifetimeVisualizationInfo = lifetimeVisualizationInfo;
        this.diagramInfo = diagramInfo;
        this.lifetimeVisualizationTab = lifetimeVisualizationTab;
        availableClassifier = new HashMap<>();
        for (ClassifierFactory classifierFactory : diagramInfo.getLifetimeVisualizationInfo().getAvailableClassifier()) {
            availableClassifier.put(classifierFactory.getName(), classifierFactory.create());
        }
        try {
            fullMetaDataPath = (BaseFile.isPlainFile(SymbolsFile.SYMBOL_FILE_ID, appInfo.getSymbolsFile()) ?
                                appInfo.getSymbolsFile().getParent() :
                                appInfo.getSymbolsFile()).toString() + File.separator + Consts.ANT_META_DIRECTORY;
            StringBuilder mapTreeFilePathBuilder = new StringBuilder(fullMetaDataPath);
            mapTreeFilePathBuilder.append(File.separator);
            for (int i = 0; i < diagramInfo.getClassifierChain().length(); i++) {
                mapTreeFilePathBuilder.append(diagramInfo.getClassifierChain().get(i).getName());
                mapTreeFilePathBuilder.append("_");
            }
            Arrays.stream(diagramInfo.getFilters()).map(Filter::getName).sorted().forEach(filterName -> {   // sort filters because their order does not matter
                mapTreeFilePathBuilder.append(filterName);
                mapTreeFilePathBuilder.append("_");
            });
            mapTreeFilePathWithoutTime = mapTreeFilePathBuilder.toString();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    @Override
    protected Map<Long, GroupingNode> backgroundWork() throws Exception {
        return parseTrace();
    }

    private Map<Long, GroupingNode> parseTrace() throws Exception {
        LOGGER.log(Level.INFO, "Start parsing trace");
        updateTitle("Calculating heap");

        List<Long> allTimes = appInfo.getStatistics()
                                     .stream()
                                     .filter(x -> (x.info.meta == EventType.GC_START && PARSE_AT_GC_START) || (x.info.meta == EventType.GC_END && PARSE_AT_GC_END))
                                     .map(x -> x.info.time)
                                     .filter(x -> x >= lifetimeVisualizationInfo.getStartTime() && x <= lifetimeVisualizationInfo.getEndTime())
                                     .distinct()
                                     .sorted(Long::compareTo)
                                     .collect(Collectors.toList());

        boolean reparseWholeTrace = false;
        for (int i = 0; i < allTimes.size(); i++) {
            Long curTime = allTimes.get(i);
            File curInputMapTreeFile = new File((mapTreeFilePathWithoutTime + curTime + (USE_INDEX_BASED_HEAP ? ".listTree" : ".mapTree")));
            if (!curInputMapTreeFile.exists()) {
                if (startMissingTime != -1) {
                    if (endMissingTime != -1) {
                        //the segment of missing times already ended previously, so this is the second missing segment
                        LOGGER.log(Level.INFO,
                                   "Going to reparse whole trace - startmissingtime: " + startMissingTime + " endmissingtime: " + endMissingTime + " cur " + "time: " +
                                           curTime);
                        reparseWholeTrace = true;
                        break;
                    }
                } else {
                    //this is the start of the first missing segment
                    startMissingTime = curTime;
                }
            } else {
                if (startMissingTime != -1 && endMissingTime == -1) {
                    //if this is the first data after a missing segment => the previous time was the last missing one
                    endMissingTime = allTimes.get(i - 1);
                }
            }
        }

        if (reparseWholeTrace) {
            groupings.clear();
            startMissingTime = lifetimeVisualizationInfo.getStartTime();
            endMissingTime = lifetimeVisualizationInfo.getEndTime();
        } else {
            if (startMissingTime != -1 && endMissingTime == -1) {
                //if there was no data after the missing part, we have to parse until the end
                endMissingTime = lifetimeVisualizationInfo.getEndTime();
                LOGGER.log(Level.INFO, "Reparsing partially between: " + startMissingTime + " and " + endMissingTime);
            }
            for (int i = 0; i < allTimes.size(); i++) {
                Long curTime = allTimes.get(i);
                if (curTime < startMissingTime || curTime > endMissingTime) {
                    File curInputMapTreeFile = new File((mapTreeFilePathWithoutTime + curTime + (USE_INDEX_BASED_HEAP ? ".listTree" : ".mapTree")));
                    LOGGER.log(Level.INFO, "Loading map grouping file: " + curInputMapTreeFile);
                    byte[] loadedMapTreeFile = Files.readAllBytes(curInputMapTreeFile.toPath());
                    DataInputStream curInput = new DataInputStream(new ByteArrayInputStream(loadedMapTreeFile));
                    groupings.put(curTime, USE_INDEX_BASED_HEAP ?
                                           DataGroupingNode.readTree(curInput, null, availableClassifier) :
                                           MapGroupingNode.readTree(curInput, null, availableClassifier));
                    curInput.close();
                }
                tryRunUpdatingThread();
            }
            if (startMissingTime == -1) {
                //no files are missing => we don't have to reparse
                return groupings;
            }
        }

        LOGGER.log(Level.INFO, "Parsing the heaps from " + startMissingTime + " until " + endMissingTime);
        //setup heap traceparser
        HeapTraceParser parser = new HeapTraceParser(appInfo.getSymbols(),
                                                     new MetaDataReaderConfig(appInfo.getSymbols().root + File.separator + Consts.ANT_META_DIRECTORY),
                                                     startMissingTime,
                                                     endMissingTime);
        parser.addTraceParserListener((from, to, position) -> updateProgress(position - from, to - from));

        parser.addHeapListener(new HeapListener() {
            @Override
            public void phaseChanging(DetailedHeap heap,
                                      GarbageCollectionType from,
                                      GarbageCollectionType to,
                                      GarbageCollectionCause cause,
                                      boolean concurrent,
                                      boolean failed,
                                      long time,
                                      long position,
                                      long length,
                                      boolean inParserTimeWindow) {
                if (PARSE_AT_GC_START && inParserTimeWindow && to != GarbageCollectionType.MUTATOR) {
                    handleHeap(heap, time);
                }
            }

            @Override
            public void phaseChanged(DetailedHeap heap,
                                     GarbageCollectionType from,
                                     GarbageCollectionType to,
                                     GarbageCollectionCause cause,
                                     boolean concurrent,
                                     boolean failed,
                                     long time,
                                     long position,
                                     long length,
                                     boolean inParserTimeWindow) {
                if (PARSE_AT_GC_END && inParserTimeWindow && to == GarbageCollectionType.MUTATOR) {
                    handleHeap(heap, time);
                }
            }

            @Override
            public void close(DetailedHeap heap) {

            }

        });
        DetailedHeap finalHeap = parser.parse(cancelProperty);
        if (startMissingTime == endMissingTime) {
            handleHeap(finalHeap, startMissingTime);
        }

        LOGGER.log(Level.INFO, "Finished parsing trace, heaps constructed");
        return groupings;
    }

    private boolean shouldParse(long time) {
        //TODO: decide whether to always parse last tree
        if (parseHeapAllXSeconds) {
            if (time - lastParsedTime >= parseVal) {
                lastParsedTime = time;
                return true;
            } else {
                return false;
            }
        }
        if (parseXthHeap) {
            System.out.println("count since last: " + countSinceLastParse);
            countSinceLastParse++;
            countSinceLastParse %= parseVal;

            if (countSinceLastParse == 0) {
                return true;
            } else {
                return false;
            }

        }
        return true;
    }

    private void handleHeap(DetailedHeap heap, long time) {
        if (time < startMissingTime || time > endMissingTime) {
            return;
        }
        if (!shouldParse(time)) {
            return;
        }
        LOGGER.log(Level.INFO, "Parsing heap at time: " + time);

        GroupingNode groupingNode = null;
        if (USE_INDEX_BASED_HEAP) {
            IndexBasedHeap indexBasedHeap = heap.toIndexBasedHeap();
            chain.forEach(classifier -> classifier.setup(indexBasedHeap::getSymbols, () -> indexBasedHeap));
            Arrays.stream(filters).forEach(filter -> filter.setup(indexBasedHeap::getSymbols, () -> indexBasedHeap));

            //calculate tree
            ListClassificationTree listClassificationTree = indexBasedHeap.groupListParallel(filters, chain, null, null, null);
            listClassificationTree.init();

            //calculate closures
            Task<Void> closureInitTask = listClassificationTree.initClosureTask();
            CLIENT_INFO.operationManager.addNewOperation(closureInitTask);
            lifetimeVisualizationTab.tasks.add(closureInitTask);
            closureInitTask.run();

            groupingNode = new DataGroupingNode((ListGroupingNode) listClassificationTree.getRoot());
        } else {
            // NOTE: can be removed if not needed anymore
            groupingNode = heap.toObjectStream().filter(filters).groupMapParallel(chain, null);
        }

        groupings.put(time, groupingNode);

        //if no file exits for this grouping, we write it
        File groupingFile = new File((mapTreeFilePathWithoutTime + time + (USE_INDEX_BASED_HEAP ? ".listTree" : ".mapTree")));
        if (!groupingFile.exists()) {
            DataOutputStream curOutput = null;
            try {
                curOutput = new DataOutputStream(new FileOutputStream(groupingFile));
                if (USE_INDEX_BASED_HEAP) {
                    DataGroupingNode.writeTree(curOutput, (DataGroupingNode) groupingNode, null, 0, null);
                } else {
                    MapGroupingNode.writeTree(curOutput, (MapGroupingNode) groupingNode);
                }

            } catch (Exception e) {
                e.printStackTrace();
            } finally {
                if (curOutput != null) {
                    try {
                        curOutput.close();
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                }
            }
        }
        tryRunUpdatingThread();

    }

    private void tryRunUpdatingThread() {
        if (currentChartUpdatingTask == null && System.currentTimeMillis() > lastPlotTime + PLOT_THRESHOLD) {
            PlatformUtil.runAndWait(() -> {
                currentChartUpdatingTask = new LifetimeVisualizationPlottingTask(lifetimeVisualizationTab, appInfo, groupings);
                currentChartUpdatingTask.setOnSucceeded(workerStateEvent -> {
                    currentChartUpdatingTask = null;
                    lastPlotTime = System.currentTimeMillis();
                });
                ThreadUtil.startTask(currentChartUpdatingTask);
            });
        }
    }

    @Override
    protected void suceeded() {
        try {
            Map<Long, GroupingNode> mapGroupings = get();
            lifetimeVisualizationTab.updateChartData(mapGroupings);
        } catch (InterruptedException | ExecutionException e) {
            e.printStackTrace();
        }
    }
}
