/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.trace.gui.model;

import at.jku.mevss.anttracks.heap.Tag;
import at.jku.mevss.anttracks.heap.statistics.Statistics;
import at.jku.mevss.anttracks.heap.symbols.Symbols;
import at.jku.mevss.anttracks.parser.classdefinitions.ClassDefinitionsFile;
import at.jku.mevss.anttracks.parser.symbols.SymbolsFile;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArrayList;

public class AppInfo {

    public interface AppInfoListener {
        void appInfoChanged(ChangeType type);
    }

    public enum AliveDeadPanelType {
        N_OBJECTS,
        BYTE,
        RELATIVE
    }

    public enum ChangeType {
        NAME,
        FILE,
        TRACE,
        SYMBOLS,
        FEATURE,
        SHOW_FEATURES,
        DATA_STRUCTURE,
        ALIVE_DEAD_PANEL_TYPE
    }

    private String appName;
    private Symbols symbols;
    private File selectedTraceFile;
    private File selectedFeaturesFile;
    private Set<URI> selectedDataStructureDefinitionFiles;
    private List<Statistics> statistics;

    private List<Tag> tags;
    private String metaDir;
    // By default feature charts are shown (but only if feature file is loaded, see isShowFeatures())
    private boolean isShowFeatures = true;

    private final CopyOnWriteArrayList<AppInfoListener> listener = new CopyOnWriteArrayList<>();
    private AliveDeadPanelType aliveDeadType;
    private boolean parsingCompleted;

    public AppInfo(String appName, File selectedTraceFile, File selectedFeaturesFile) {
        this(appName, null, selectedTraceFile, selectedFeaturesFile);
    }

    public AppInfo(String appName, Symbols symbols, File selectedTraceFile, File selectedFeaturesFile) {
        this.appName = appName;
        this.symbols = symbols;

        this.selectedTraceFile = selectedTraceFile;
        this.selectedFeaturesFile = selectedFeaturesFile;
        this.selectedDataStructureDefinitionFiles = new HashSet<>();
        addListener(changeType -> {
            if ((changeType == ChangeType.SYMBOLS || changeType == ChangeType.DATA_STRUCTURE) && this.symbols != null) {
                // push all selected data structure definition files to symbols
                this.symbols.addDataStructureDefinitionFiles(selectedDataStructureDefinitionFiles);
            }
        });

        statistics = new ArrayList<>();
        isShowFeatures = selectedFeaturesFile != null;
        aliveDeadType = AliveDeadPanelType.RELATIVE;
        parsingCompleted = false;
        tags = new ArrayList<>();
    }

    public File getSelectedTraceFile() {
        return selectedTraceFile;
    }

    public File getSymbolsFile() throws IOException {
        if (selectedTraceFile == null) {
            return null;
        }
        return SymbolsFile.findSymbolsFileToTrace(selectedTraceFile);
    }

    public File getClassDefinitionsFile() throws IOException {
        if (selectedTraceFile == null) {
            return null;
        }
        return ClassDefinitionsFile.findClassDefinitionsFileToTrace(selectedTraceFile);
    }

    public Symbols getSymbols() {
        return symbols;
    }

    public List<Statistics> getStatistics() {
        return statistics;
    }

    public String getAppName() {
        return appName;
    }

    public File getSelectedFeaturesFile() {
        return selectedFeaturesFile;
    }

    public void setShowFeatures(boolean b) {
        if (isShowFeatures != b) {
            isShowFeatures = b;
            fireListener(ChangeType.SHOW_FEATURES);
        }
    }

    public boolean isShowFeatures() {
        return isShowFeatures && selectedFeaturesFile != null;
    }

    public void setSymbols(Symbols symbols) {
        if (this.symbols != symbols) {
            this.symbols = symbols;
            fireListener(ChangeType.SYMBOLS);
        }
    }

    public void setAppName(String appName) {
        if (!this.appName.equals(appName)) {
            this.appName = appName;
            fireListener(ChangeType.NAME);
        }
    }

    public void setSelectedFeaturesFile(File selectedFeaturesFile) {
        this.selectedFeaturesFile = selectedFeaturesFile;
    }

    public void setSelectedTraceFile(File selectedTraceFile) { this.selectedTraceFile = selectedTraceFile; }

    public void selectTraceFile(File traceFile) {
        // sets trace file and fires event on change
        if (fileChanged(traceFile, selectedTraceFile)) {
            setSelectedTraceFile(traceFile);
            fireListener(ChangeType.TRACE);
        }
    }

    public void selectFeaturesFile(File featuresFile) {
        // sets feature file and fires event on change
        if (fileChanged(featuresFile, selectedFeaturesFile)) {
            setSelectedFeaturesFile(featuresFile);
            fireListener(ChangeType.FEATURE);
        }
    }

    public void selectFeaturesAndTraceFile(File traceFile, File featuresFile) {
        // sets feature and trace file and fires SINGLE event on change (avoid multiple parser tasks on redefine)
        if (fileChanged(traceFile, selectedTraceFile) || fileChanged(featuresFile, selectedFeaturesFile)) {
            setSelectedTraceFile(traceFile);
            setSelectedFeaturesFile(featuresFile);
            fireListener(ChangeType.FILE);
        }
    }

    public void selectDataStructureDefinitionFile(URI selectedDataStructureDefinitionFile) {
        if (selectedDataStructureDefinitionFiles.add(selectedDataStructureDefinitionFile)) {
            fireListener(ChangeType.DATA_STRUCTURE);
        }
    }

    public void selectDataStructureDefinitionFiles(List<URI> selectedDataStructureDefinitionFiles) {
        if (this.selectedDataStructureDefinitionFiles.addAll(selectedDataStructureDefinitionFiles)) {
            fireListener(ChangeType.DATA_STRUCTURE);
        }
    }

    private boolean fileChanged(File before, File after) {
        try {
            return before == null && after != null || before != null && after == null || !before.getCanonicalPath().equals(after.getCanonicalPath());
        } catch (IOException e) {
            e.printStackTrace();
            return true;
        }
    }

    private void fireListener(ChangeType t) {
        listener.stream().forEach(x -> x.appInfoChanged(t));
    }

    public void addListener(AppInfoListener l) {
        listener.add(l);
    }

    public void removeListener(AppInfoListener l) {
        listener.remove(l);
    }

    public void setAliveDeadPanelType(AliveDeadPanelType type) {
        aliveDeadType = type;
        fireListener(ChangeType.ALIVE_DEAD_PANEL_TYPE);
    }

    public AliveDeadPanelType getAliveDeadPanelType() {
        return aliveDeadType;
    }

    public void setParsingCompleted(boolean b) {
        this.parsingCompleted = b;
    }

    public boolean isParsingCompleted() {
        return parsingCompleted;
    }

    public void setMetaDataPath(String metaDir) {
        this.metaDir = metaDir;
    }

    public String getMetaDataPath() {
        return metaDir;
    }

    public List<Tag> getTags() {
        return tags;
    }

    public void setTags(List<Tag> tags) {
        this.tags = tags;
    }
}
