/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.trace.gui.task;

import at.jku.mevss.anttracks.heap.Heap;
import at.jku.mevss.anttracks.heap.io.MetaDataReaderConfig;
import at.jku.mevss.anttracks.parser.heap.HeapTraceParser;
import at.jku.mevss.anttracks.util.TraceException;
import at.jku.mevss.trace.gui.model.AppInfo;
import at.jku.mevss.trace.gui.utils.AntTask;
import at.jku.mevss.trace.gui.utils.Consts;
import at.jku.mevss.trace.gui.utils.WindowUtil;
import javafx.application.Platform;
import javafx.scene.control.Alert;
import javafx.scene.control.Alert.AlertType;
import javafx.scene.control.ButtonType;
import javafx.scene.layout.Region;

import java.io.File;
import java.io.IOException;
import java.util.Optional;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.FutureTask;
import java.util.logging.Level;

public abstract class HeapTask<T extends Heap> extends AntTask<T> {
    protected final AppInfo appInfo;
    protected final long time;

    public HeapTask(AppInfo appInfo, long time) {
        this.appInfo = appInfo;
        this.time = time;
    }

    @Override
    protected T backgroundWork() throws Exception {
        try {
            T heap = parseTrace();
            additionalBackgroundWork(heap);
            return heap;
        } catch (InterruptedException e) {
            // task about to end after being cancelled
            handleTaskCancelled();
            return null;
        } catch (Exception e) {
            final FutureTask<Boolean> alertTask = new FutureTask<>(() -> {
                Alert alert = new Alert(AlertType.ERROR, "An internal error occured while parsing, do you want to retry?", ButtonType.YES, ButtonType.NO);
                alert.setTitle("Error");
                alert.getDialogPane().setMinHeight(Region.USE_PREF_SIZE);
                WindowUtil.centerInMainFrame(alert);
                Optional<ButtonType> retryChoice = alert.showAndWait();
                return retryChoice.isPresent() && retryChoice.get() == ButtonType.YES;
            });
            Platform.runLater(alertTask);
            if (alertTask.get()) {
                return backgroundWork();
            } else {
                throw new Exception(e);
            }
        }
    }

    protected abstract void handleTaskCancelled();

    protected abstract void additionalBackgroundWork(T heap);

    private T parseTrace() throws Exception {
        LOGGER.log(Level.INFO, "Start parsing trace");
        updateTitle("Heap State");
        updateMessage("Calculate heap state for time " + (time / 1000.0) + "s");
        HeapTraceParser parser;
        parser = new HeapTraceParser(appInfo.getSymbols(), new MetaDataReaderConfig(appInfo.getSymbols().root + File.separator + Consts.ANT_META_DIRECTORY), time);
        parser.addTraceParserListener((from, to, position) -> updateProgress(position - from, to - from));
        T heap = parse(parser);
        LOGGER.log(Level.INFO, "Finished parsing trace, heap constructed");
        return heap;
    }

    protected abstract T parse(HeapTraceParser parser) throws InterruptedException, IOException, TraceException, ExecutionException;
}
