package at.jku.mevss.trace.gui.utils;

import at.jku.mevss.anttracks.heap.DetailedHeap;
import at.jku.mevss.anttracks.heap.GarbageCollectionCause;
import at.jku.mevss.anttracks.heap.GarbageCollectionType;
import at.jku.mevss.anttracks.heap.HeapListener;
import at.jku.mevss.anttracks.heap.io.MetaDataWriterConfig;
import at.jku.mevss.anttracks.parser.heap.HeapTraceParser;
import at.jku.mevss.anttracks.parser.printing.AdditionalPrintingEventHandler;
import at.jku.mevss.anttracks.util.ThreadUtil;
import at.jku.mevss.trace.gui.frame.main.tab.application.ApplicationTab;
import at.jku.mevss.trace.gui.frame.main.tab.application.ApplicationTabPlottingTask;
import at.jku.mevss.trace.gui.model.AppInfo;
import javafx.application.Platform;
import javafx.scene.control.Alert;
import javafx.scene.control.ButtonType;
import javafx.scene.layout.Region;

import java.io.File;
import java.util.Optional;
import java.util.concurrent.FutureTask;
import java.util.logging.Level;

public class MetaParserTask extends AntTask<Void> {
    private final AppInfo appInfo;
    private ApplicationTabPlottingTask currentChartUpdatingTask = null;
    private final String metaHeaderFilePath;
    private final ApplicationTab tab;
    private long lastPlotTime = 0;
    private static final int PLOT_THRESHOLD = 1_000;

    public MetaParserTask(AppInfo appInfo, ApplicationTab tab, String metaHeaderFilePath) {
        this.appInfo = appInfo;
        this.tab = tab;
        tab.setTask(this);
        this.metaHeaderFilePath = metaHeaderFilePath;
    }

    @Override
    protected Void backgroundWork() throws Exception {
        updateTitle("Construct meta-data for " + appInfo.getAppName() + " (File: " + appInfo.getSelectedTraceFile() + ")");
        updateMessage("Parsing trace");

        appInfo.getStatistics().clear();

        LOGGER.log(Level.INFO, "parsing");
        File file = new File(appInfo.getMetaDataPath());
        if (file.isDirectory()) {
            LOGGER.log(Level.INFO, "cleaning meta data directory");
            for (File child : file.listFiles()) {
                child.delete();
            }
        }

        try {
            HeapTraceParser heapParser = new HeapTraceParser(appInfo.getSymbols(), new MetaDataWriterConfig(appInfo.getMetaDataPath())) {
                @Override
                protected void doWorkspaceCompletion(DetailedHeap workspace) {
                    // Do not call super since we do not want to complete the heap (e.g., from-pointer setting), since this is not needed for meta-parsing
                }
            };
            heapParser.addTraceParserListener((from, to, position) -> updateProgress(position - from, to - from));
            heapParser.addHeapListener(new HeapListener() {
                @Override
                public void phaseChanging(DetailedHeap heap,
                                          GarbageCollectionType from,
                                          GarbageCollectionType to,
                                          GarbageCollectionCause cause,
                                          boolean concurrent,
                                          boolean failed,
                                          long time,
                                          long position,
                                          long length,
                                          boolean inSelectedTimeWindow) {

                }

                @Override
                public void phaseChanged(DetailedHeap heap,
                                         GarbageCollectionType from,
                                         GarbageCollectionType to,
                                         GarbageCollectionCause cause,
                                         boolean concurrent,
                                         boolean failed,
                                         long time,
                                         long position,
                                         long length,
                                         boolean inSelectedTimeWindow) {
                    synchronized (appInfo.getStatistics()) {
                        appInfo.getStatistics().add(heap.statistics);
                    }
                    synchronized (appInfo.getTags()) {
                        appInfo.setTags(heap.getTags());
                    }

                    if (currentChartUpdatingTask == null && System.currentTimeMillis() > lastPlotTime + PLOT_THRESHOLD) {
                        PlatformUtil.runAndWait(() -> {
                            currentChartUpdatingTask = new ApplicationTabPlottingTask(appInfo, tab);
                            currentChartUpdatingTask.setOnSucceeded(workerStateEvent -> {
                                currentChartUpdatingTask = null;
                                lastPlotTime = System.currentTimeMillis();
                            });
                            ThreadUtil.startTask(currentChartUpdatingTask);
                        });
                    }
                }

                @Override
                public void close(DetailedHeap heap) {
                }
            });

            //if (getClientInfo().assertionsEnabled) {
            heapParser.addEventHandler((heap, info) -> new AdditionalPrintingEventHandler(heap));
            //}
            heapParser.parse(cancelProperty);

            // The following parser is only for debug purpose
            /*
            PrintingTraceParser printTraceParser = new PrintingTraceParser(appInfo.getSymbols());
            printTraceParser.parse();
            */
        } catch (InterruptedException e) {
            // cancelled parsing stopped
            // invalidate metadata
            File headerFile = new File(metaHeaderFilePath);
            assert headerFile.exists();
            headerFile.delete();

        } catch (Exception e) {
            // Delete header file if parsing did not complete
            File headerFile = new File(metaHeaderFilePath);
            assert headerFile.exists();
            headerFile.delete();

            final FutureTask<Boolean> alertTask = new FutureTask<>(() -> {
                Alert alert = new Alert(Alert.AlertType.ERROR, "An internal error occured while parsing, do you want to retry?", ButtonType.YES, ButtonType.NO);
                alert.setTitle("Error");
                alert.getDialogPane().setMinHeight(Region.USE_PREF_SIZE);
                WindowUtil.centerInMainFrame(alert);
                Optional<ButtonType> retryChoice = alert.showAndWait();
                return retryChoice.isPresent() && retryChoice.get() == ButtonType.YES;
            });
            Platform.runLater(alertTask);
            if (alertTask.get()) {
                return backgroundWork();
            } else {
                throw new Exception(e);
            }
        }
        return null;
    }

    @Override
    protected void suceeded() {
    }
}
