package at.jku.mevss.trace.gui.model

import at.jku.mevss.anttracks.classification.ClassifierFactoryList
import at.jku.mevss.anttracks.classification.FilterFactoryList
import at.jku.mevss.anttracks.classification.enumerations.ClassifierSourceCollection
import at.jku.mevss.anttracks.classification.nodes.DataGroupingNode
import at.jku.mevss.anttracks.classification.nodes.GroupingNode
import at.jku.mevss.anttracks.classification.trees.ListClassificationTree
import at.jku.mevss.anttracks.heap.IndexBasedHeap
import at.jku.mevss.anttracks.heap.symbols.Symbols
import at.jku.mevss.trace.gui.frame.main.tab.datastructuredevelopment.parser.DataStructureDevelopmentParserTaskResult
import at.jku.mevss.trace.gui.utils.Consts
import java.util.function.Supplier

class DataStructureDevelopmentInfo(val appInfo: AppInfo,
                                   val startTime: Long,
                                   val endTime: Long) : AvailableClassifierInfo() {
    inner class DataStructureDevelopment(val startAddress: Long,
                                         val startIndex: Int,
                                         val endAddress: Long,
                                         val endIndex: Int) {
        fun survived(): Boolean {
            return endAddress >= 0
        }

        fun died(): Boolean {
            return !survived()
        }

        fun label(endHeap: IndexBasedHeap): String {
            return String.format("%s\n@ %,d", endHeap.getType(endIndex).getSimpleName(true), endHeap.getAddress(endIndex))
        }

        private fun extractFromAddressNode(tree: ListClassificationTree?, address: Long, extractor: (GroupingNode?) -> Long): Long =
                extractor(tree?.root?.children?.find { node -> node.key.toString().contains(String.format("%,d", address)) })

        fun startTransitiveSize(): Long = extractFromAddressNode(startClassificationTree, endAddress, { it?.transitiveClosureSizeProperty()?.longValue() ?: 0 })

        fun endTransitiveSize(): Long = extractFromAddressNode(endClassificationTree, endAddress, { it?.transitiveClosureSizeProperty()?.longValue() ?: 0 })

        fun absoluteTransitiveGrowthRate(): Long = endTransitiveSize() - startTransitiveSize()

        fun relativeTransitiveGrowthRate(): Float = 100.0f * absoluteTransitiveGrowthRate() / startTransitiveSize()

        fun transitiveHeapGrowthPortion() = absoluteTransitiveGrowthRate() * 100.0f / absoluteHeapGrowth()

        fun startRetainedSize(): Long = extractFromAddressNode(startClassificationTree, endAddress, { it?.retainedSizeProperty()?.longValue() ?: 0 })

        fun endRetainedSize(): Long = extractFromAddressNode(endClassificationTree, endAddress, { it?.retainedSizeProperty()?.longValue() ?: 0 })

        fun absoluteRetainedGrowthRate(): Long = endRetainedSize() - startRetainedSize()

        fun relativeRetainedGrowthRate(): Float = 100.0f * absoluteRetainedGrowthRate() / startRetainedSize()

        fun retainedHeapGrowthPortion() = absoluteRetainedGrowthRate() * 100.0f / absoluteHeapGrowth();

        fun startDataStructureSize(): Long = extractFromAddressNode(startClassificationTree, endAddress, { it?.dataStructureSizeProperty()?.longValue() ?: 0 })

        fun endDataStructureSize(): Long = extractFromAddressNode(endClassificationTree, endAddress, { it?.dataStructureSizeProperty()?.longValue() ?: 0 })

        fun absoluteDataStructureGrowthRate(): Long = endDataStructureSize() - startDataStructureSize()

        fun relativeDataStructureGrowthRate(): Float = 100.0f * absoluteDataStructureGrowthRate() / startDataStructureSize()

        fun dataStructureHeapGrowthPortion() = absoluteDataStructureGrowthRate() * 100.0f / absoluteHeapGrowth();

        fun startDeepDataStructure(): Long = extractFromAddressNode(startClassificationTree, endAddress, { it?.deepDataStructureSizeProperty()?.longValue() ?: 0 })

        fun endDeepDataStructureSize(): Long = extractFromAddressNode(endClassificationTree, endAddress, { it?.deepDataStructureSizeProperty()?.longValue() ?: 0 })

        fun absoluteDeepDataStructureGrowthRate(): Long = endDeepDataStructureSize() - startDeepDataStructure()

        fun relativeDeepDataStructureGrowthRate(): Float = 100.0f * absoluteDeepDataStructureGrowthRate() / startDeepDataStructure()

        fun deepDataStructureHeapGrowthPortion() = absoluteDeepDataStructureGrowthRate() * 100.0f / absoluteHeapGrowth();
    }

    val selectedClassifierInfo: SelectedClassifierInfo = SelectedClassifierInfo(null, null, null)
    val developments: MutableList<DataStructureDevelopment> = ArrayList()

    var startHeap: IndexBasedHeap? = null
    var endHeap: IndexBasedHeap? = null

    var startClassificationTree: ListClassificationTree? = null
    var endClassificationTree: ListClassificationTree? = null
    var classificationTreeDifference: DataGroupingNode? = null

    init {
        availableClassifier = ClassifierFactoryList(symbolsSupplier, fastHeapSupplier, Consts.CLASSIFIERS_DIRECTORY, ClassifierSourceCollection.FASTHEAP)
        availableFilter = FilterFactoryList(symbolsSupplier, fastHeapSupplier, Consts.FILTERS_DIRECTORY, ClassifierSourceCollection.FASTHEAP)
    }

    fun setDataStructureDevelopmentParserTaskResult(result: DataStructureDevelopmentParserTaskResult) {
        startHeap = result.beforeHeap
        endHeap = result.afterHeap

        developments.clear()
        for (i in 0 until result.dataStructures.size) {
            if(result.dataStructures[i].currentAddress >= 0) {
                val dsd = DataStructureDevelopment(
                        result.dataStructures[i].startAddress,
                        result.beforeHeap.toIndex(result.dataStructures[i].startAddress),
                        result.dataStructures[i].currentAddress,
                        result.afterHeap.toIndex(result.dataStructures[i].currentAddress))

                assert(startHeap!!.getType(dsd.startIndex) === endHeap!!.getType(dsd.endIndex)) { "Our object may not change its type!" }
                assert(startHeap!!.isDataStructureHead(dsd.startIndex)) { "Object must be a DS head" }
                assert(endHeap!!.isDataStructureHead(dsd.endIndex)) { "Object must be a DS head" }
                assert(startHeap!!.getType(dsd.startIndex).dataStructureDefinition.isHead) { "Object must be a DS head" }
                assert(endHeap!!.getType(dsd.endIndex).dataStructureDefinition.isHead) { "Object must be a DS head" }

                developments.add(dsd)
            }
        }
    }

    fun absoluteHeapGrowth(): Float = 1.0f * (endHeap?.getByteCount() ?: 0) - (startHeap?.getByteCount() ?: 0)

    fun getStartFastHeapSupplier(): Supplier<IndexBasedHeap?> = Supplier { startHeap }
    override fun getSymbolsSupplier(): Supplier<Symbols?> = Supplier { endHeap?.symbols }
    override fun getFastHeapSupplier(): Supplier<IndexBasedHeap?> = Supplier { endHeap }
}
